/*
 * AlignProbData.hpp
 *
 */
#ifndef ALIGN_PROB_DATA_HPP
#define ALIGN_PROB_DATA_HPP

#include "Util.hpp"
#include "Matrix.hpp"
#include "SparseProb.hpp"

class ScoreParam;
class SeqIds;
class SeqData;
class Weight;
class MSA;
class Aligner;
class PairProbData;
class Scheduler;

class AlignProbData {
public:
  //typedef NMatrix<float> ValueT;
  typedef SparseProb ValueT;
  typedef Array<ValueT> Data;
  //typedef Matrix<ValueT> Data; // internal error occurs for gcc-4.1.1 

  AlignProbData(SeqData& seqData);
  ~AlignProbData();
  bool empty() const {return data.empty();}
  int size() const {return data.size();}
  ValueT& ref(int seqId1, int seqId2) {
    //return data.ref(seqIds.getIndex(seqId1), seqIds.getIndex(seqId2));
    return data[seqIds.getIndex(seqId1) * stride + seqIds.getIndex(seqId2)];
  }
  const ValueT& ref(int seqId1, int seqId2) const {
    //return data.ref(seqIds.getIndex(seqId1), seqIds.getIndex(seqId2));
    return data[seqIds.getIndex(seqId1) * stride + seqIds.getIndex(seqId2)];
  }
  void run(Scheduler& scheduler);
  void bcastAlignProbs(Scheduler& scheduler);
  void smoothData();
  void smoothData(const SeqIds& seqIds1);
  //void smoothData(Scheduler& scheduler);
  //void smoothAlignProb(NMatrix<float>& mtx, const SeqIds& seqIds1, 
  //		       const SeqId& seqId1, const SeqId& seqId2) const;
  void smoothPairProbData(PairProbData& pairProbData) const;
  void smoothPairProbData(PairProbData& pairProbData,
  			  const SeqIds& seqIds1) const;
  //Matrix< SparseProb > getSparseProbs() const;
  void getAlignProb(NMatrix<float>& alignProb, const Weight& weights1,
		    const MSA& msaX, const MSA& msaY) const;
private:
  SeqData& seqData;
  SeqIds& seqIds;
  GuideTree& guideTree;
  Weight& weights;
  Data data;
  int stride;
  float cutoff;
 
  AlignProbData();
  AlignProbData(const AlignProbData& alignProbData);
  AlignProbData& operator=(const AlignProbData& alignProbData);
};

#endif
