/*
 * BasePairData.hpp
 *
 */
#ifndef BASE_PAIR_DATA_HPP
#define BASE_PAIR_DATA_HPP

#include "Util.hpp"
#include "SparseProb.hpp"

class ScoreParam;
class SeqIds;
class SeqData;
class McCaskill;
class GuideTree;
class Weight;
class Aligner;
class Scheduler;

class 
BasePairData 
{
  typedef SparseProb ValueT ;
  typedef Array<ValueT> Data;
public:
  BasePairData(SeqData& seqData);
  ~BasePairData();

  bool empty() const {return data.empty();}
  int size() const {return data.size();}
  ValueT& ref(int id) {return data[seqIds.getIndex(id)];}
  const ValueT& ref(int id) const {return data[seqIds.getIndex(id)];}
  void run(Scheduler& scheduler);
  void bcastPairProbs(Scheduler& scheduler);
  void getPairProb(UpTriMtx<float>& pairProb, 
		   const Weight& weights1, const MSA& msa) const;

private:
  SeqData& seqData;
  SeqIds& seqIds;
  GuideTree& guideTree;
  Weight& weights;
  Data data;
  float cutoff;
 
  BasePairData();
  BasePairData(const BasePairData& basePairData);
  BasePairData& operator=(const BasePairData& basePairData);
};

#endif
