/*
 * DPPath.hpp
 *
 */
#ifndef DP_PATH_HPP
#define DP_PATH_HPP
#include "Rect.hpp"
#include "EScript.hpp"

class DPRegion;
class Profile;
class MSA;

class DPPath {
public:
  PosMap posMapX;
  PosMap posMapY;
  CT ct;
  EScript esX;
  EScript esY;

  DPPath() {}
  DPPath(const DPPath& dpPath)
    : posMapX(dpPath.posMapX), posMapY(dpPath.posMapY),
      ct(dpPath.ct), esX(dpPath.esX), esY(dpPath.esY) {}
  ~DPPath() {};
  DPPath& operator=(const DPPath& dpPath) {
    if (this == &dpPath) return *this;

    posMapX = dpPath.posMapX;
    posMapY = dpPath.posMapY;
    ct = dpPath.ct;
    esX = dpPath.esX;
    esY = dpPath.esY;
    return *this;
  }
  bool operator==(const DPPath& dpPath) const {
    return (posMapX == dpPath.posMapX
	    && posMapY == dpPath.posMapY
	    && ct == dpPath.ct
	    && esX == dpPath.esX
	    && esY == dpPath.esY);
  }

  void clear();
  bool empty() const {return posMapX.empty() && posMapY.empty();}
  int getAlignLen() const {return (int) posMapX.size();}
  void pushMState(int i, int j);
  void pushIState(int i);
  void pushDState(int i);
  int getLettLenX() const {return posMapX.getSourceLength();}
  int getLettLenY() const {return posMapY.getSourceLength();} 
  void setEditScript();
  void setDefaultCT();
  void setEditScript(EScript& es, const PosMap& posMap);
  void expand(int chunkSize, int nTailX, int nTailY);
  void expandPosMaps(int chunkSize, int nTailX, int nTailY);
  void setDiagonalPath(int lengthX, int lengthY);
  void setDiagonalPath(const MSA& msaX, const MSA& msaY);
  void setFromMSAs(const MSA& msaX, const MSA& msaY, const CT& ct0 = CT());
  void setCTX(const CT& ctX);
  void setCTY(const CT& ctY);
  void unpairGapCols();
  void multiply(const DPPath& path);
  void flipXY();
  float computeMEPD(const DPPath& dpPath) const;
  float computeMaxStripWidth(const DPPath& dpPath) const;
  int computeDistSumMax(bool computeSum, const DPPath& dpPath) const;
  void draw(ostream& fo) const;
  void draw() const;
  string toString() const;
  void print() const;
  
  void getStripDPRegion(int stripWidth, DPRegion& dpRegion) const;
  void getStripDPRegionBBox(int stripWidth, const Rect& bbox, DPRegion& dpRegion) const;
  void getYRange(Vector<int>& y_begs, Vector<int>& y_lasts) const;
  void setFromPackedData(const Vector<int>& v);
  Vector<int> getPackedData() const;
  int getPackedDataSize() const;
  Rect getRect0() const;
  int getAlnPos(const Point& pt) const;
  void update(const DPPath& dpPath);

  // tentative.
  // Change CT coordinate system from alignPos to Sequence column
  // and have ctX and ctY
  bool isStemX(int x) const;
  bool isStemY(int y) const;
  bool isLoopX(int x) const;
  bool isLoopY(int y) const;
  bool pairsX(int x1, int x2) const;
  bool pairsY(int y1, int y2) const;

  struct Iterator {
    const DPPath& path;
    enum State {
      M_st = 0, I_st = 1, D_st = 2, X_st = 3
    };
    int i;
    int s;
    union {
      struct {int x, y;};
      Point pt;
    };
    Iterator(const DPPath& path)
      : path(path), i(-1), s(M_st), x(-1), y(-1) {}
    void reset() {i = -1, s = M_st, x = -1, y = -1;}
    int getState(int i) const;
    bool getNext();
    bool emitsX() const;
    bool emitsY() const;
  };
  void testConnectivity() const;
  void testConsistency() const;
};

#endif
