/*
 * LoadStat.cpp
 *
 */
#include "Util.hpp"
#include "LoadStat.hpp"

template <typename Vect>
static float
mean(const Vect& u)
{
  float sum = 0.0;
  for (int i = 0; i < (int) u.size(); i++) {
    sum += u[i];
  }
  return sum / (float) u.size();
}

float LoadStat::maxWorkspaceSizeInMB = 0.0;
float LoadStat::totalDPIterationCountInMT = 0.0;
int LoadStat::dpCount = 0;
bool LoadStat::accountForWidth = true;
Vector<int> LoadStat::lengths;
Vector<int> LoadStat::stripWidths;
Vector<float> LoadStat::vScore;
ostringstream LoadStat::os;

void
LoadStat::
addStripWidth(int stripWidth)
{
  if (accountForWidth) {
    stripWidths.push_back(stripWidth / 2);
  }
}
void
LoadStat::
addLengths(int lengthX, int lengthY)
{
  if (accountForWidth) {
    lengths.push_back((lengthX + lengthY) / 2);
  }
}
void
LoadStat::
addLengths(int lengthX, int lengthY, int stripWidth)
{
  if (accountForWidth) {
    lengths.push_back((lengthX + lengthY) / 2);
    stripWidths.push_back(stripWidth / 2);
  }
}
void
LoadStat::
setAccountForWidth(bool flag)
{
  accountForWidth = flag;
}
void
LoadStat::
incrementDPCount()
{
  dpCount++;
}
void
LoadStat::
addWorkspaceSize(float workspaceSizeInMB)
{
  if (maxWorkspaceSizeInMB < workspaceSizeInMB) {
    maxWorkspaceSizeInMB = workspaceSizeInMB;
  }
}
void
LoadStat::
addDPIterationCount(float dpIterationCountInMT)
{
  totalDPIterationCountInMT += dpIterationCountInMT;
}
// int
// LoadStat::
// getMemUse()
// {
//   static int pageSize = 0;
//   static char statm[64];
//   static FILE *fp = NULL;
//   
//   if (fp == NULL) {
//     if (pageSize == 0) {
//       pageSize = sysconf(_SC_PAGESIZE);
//       sprintf(statm, "/proc/%d/statm", getpid());
//     }
//     fp = fopen(statm, "r");
//     if (fp == NULL) {
//       //Warn("Cannot open %s\\n", statm);
//       return (-1);
//     }
//   }
// 
//   rewind(fp);
//   char buf[64];
//   if(fgets(buf, sizeof(buf), fp) == NULL){
//     //Warn("Cannot read %s", statm);
//     return (-1);
//   }
//   const int nPages = atoi(buf);
//   return (nPages * pageSize);
// }
// float
// LoadStat::
// getMemUseInMB()
// {
//   const float oneMegaBytes = (float) (1 << 20);
//   return (float) getMemUse() / oneMegaBytes;
// }
// void
// LoadStat::
// printMemUse(ostream& fo, int indentLevel)
// {
//   const float oneMB = (float) (1 << 20);
//   string indent(indentLevel * 2, ' ');
// 
//   int memUse = -1;
//   for (int i = 0; i < 5; i++) {
//     memUse = getMemUse();
//     if (memUse >= 0) break;
// 
//     sleep(1);
//   }
//       
//   fo << indent << "used_memory: ";
//   fo << setprecision(2);
//   if (memUse >= 0) {
//     fo << (memUse / oneMB) << "MB";
//   }
//   fo << endl;
// }
int
LoadStat::
getElapsedTimeInSec()
{
  return (int) (clock() / CLOCKS_PER_SEC);
}
void
LoadStat::
printElapsedTime(ostream& fo, int indentLevel)
{
  string indent(indentLevel * 2, ' ');
  const int tot_sec = getElapsedTimeInSec();
  fo << indent << "elapsed_time: { sec: " << tot_sec << ", ";
  fo << "string: " << sec2str(tot_sec) << " }" << endl;
}
void
LoadStat::
printReport(ostream& fo)
{
  fo << setprecision(3);
  fo << "---\n";
  fo << "mean_length: " << mean(lengths) << "\n";
  fo << "mean_strip_width: " << mean(stripWidths) << "\n";

  Assert(lengths.size() == stripWidths.size());
  Vector<float> ratios(lengths.size());
  for (int i = 0; i < (int) ratios.size(); i++) {
    ratios[i] = stripWidths[i] / (float) lengths[i];
  }
  // fo << "mean_lw_ratio: " << mean(ratios) << "\n";
  //fo << "dp_count: " << dpCount << "\n";
  //fo << "max_workspace_size: " << maxWorkspaceSizeInMB << "MB\n";
  //fo << "total_dp_iteration_count: " 
  //    << totalDPIterationCountInMT << "MT\n";
  //printMemUse(fo);
  printElapsedTime(fo);
  fo << os.str();
  fo << flush;
}
string
LoadStat::
toYAMLString()
{
  ostringstream oss;
  oss << "---\n"
      << "load_stat:\n"
      << "  max_workspace_size_in_mb: "
      << maxWorkspaceSizeInMB << "\n"
      << "  total_dp_iteration_count_in_mt: "
      << totalDPIterationCountInMT << "\n"
      << "  dp_count: "
      << dpCount << "\n"
      << "  account_for_width: "
      << accountForWidth << "\n"
      << "  lengths: [ ";
  for (int i = 0; i < (int) lengths.size(); i++) {
    oss << lengths[i] 
	<< (i == (int) lengths.size() - 1 ? "" : ", ");
  }
  oss << " ]\n";
  oss << "  strip_widths: [ ";
  for (int i = 0; i < (int) stripWidths.size(); i++) {
    oss << stripWidths[i] 
	<< (i == (int) stripWidths.size() - 1 ? "" : ", ");
  }
  oss << " ]\n";
  oss << "  vscore: [ ";
  for (int i = 0; i < (int) vScore.size(); i++) {
    oss << vScore[i] 
	<< (i == (int) vScore.size() - 1 ? "" : ", ");
  }
  oss << " ]\n";
  return oss.str();
}
void
LoadStat::
mergeYAMLString(const string& str)
{
  istringstream iss(str);
  string buf;
  while (getline(iss, buf)) {
    StringTokenizer st(buf);
    while (st.hasMoreTokens()) {
      const string& tok0 = st.nextToken();
      const string& tok1 = st.nextToken();
      if (tok0 == "---" || tok0 == "load_stat:") continue;
      if (tok0 == "max_workspace_size_in_mb:") {
	maxWorkspaceSizeInMB = 
	  max(maxWorkspaceSizeInMB, (float) atof(tok1.c_str()));
      } else if (tok0 == "total_dp_iteration_count_in_mt:") {
	totalDPIterationCountInMT += atof(tok1.c_str());
      } else if (tok0 == "dp_count:") {
	dpCount += atoi(tok1.c_str());
      } else if (tok0 == "account_for_width:") {
	;
      } else if (tok0 == "lengths:") {
	Assert(tok1 == "[");
	for (string tok2 = st.nextToken(); tok2 != "" && tok2 != "]";
	     tok2 = st.nextToken()) {
	  lengths.push_back(atoi(tok2.c_str()));
	}
      } else if (tok0 == "strip_widths:") {
	Assert(tok1 == "[");
	for (string tok2 = st.nextToken(); tok2 != "" && tok2 != "]";
	     tok2 = st.nextToken()) {
	  stripWidths.push_back(atoi(tok2.c_str()));
	}
      } else if (tok0 == "vscore:") {
	Assert(tok1 == "[");
	for (string tok2 = st.nextToken(); tok2 != "" && tok2 != "]";
	     tok2 = st.nextToken()) {
	  stripWidths.push_back(atoi(tok2.c_str()));
	}
      } else {
	Die("unrecognized field %s", tok0.c_str());
      }
    }
  }
}
string
LoadStat::
toString()
{
  ostringstream oss;
  oss << setprecision(3);
  oss << "---\n"
      << "max_workspace_size_in_mb: " 
      << maxWorkspaceSizeInMB << "\n"
      << "total_dp_iteration_count_in_mt: "
      << totalDPIterationCountInMT << "\n";
  oss << "dp_count: " << dpCount << "\n";
  oss << "account_for_width: " << accountForWidth << "\n";
  oss << "lengths: [ ";
  for (int i = 0; i < (int) lengths.size(); i++) {
    oss << lengths[i];
    oss << (i == (int) lengths.size() - 1 ? "" : ", ");
  }
  oss << " ]\n";
  oss << "strip_widths: [ ";
  for (int i = 0; i < (int) stripWidths.size(); i++) {
    oss << stripWidths[i];
    oss << (i == (int) stripWidths.size() - 1 ? "" : ", ");
  }
  oss << " ]\n";
  oss << "vscore: [ ";
  for (int i = 0; i < (int) vScore.size(); i++) {
    oss << vScore[i];
    oss << (i == (int) vScore.size() - 1 ? "" : ", ");
  }
  oss << " ]\n";
  return oss.str();
}
void
LoadStat::
print()
{
  cout << toString() << flush;
}


