/*
 * ProfileFreqs.hpp
 *
 */
#ifndef PROFILE_FREQS_HPP
#define PROFILE_FREQS_HPP

#include "Alpha.hpp"
class ScoreParam;

struct 
Profile::
FCol
{
  typedef Profile::FCol Self;
  enum {SIZE = Alpha::NBASE};
  CArray<float,SIZE> data;
  CArray<float,SIZE> data_0;

  typedef float T;
  typedef T              value_type;
  typedef T*             iterator;
  typedef const T*       const_iterator;
  typedef T&             reference;
  typedef const T&       const_reference;
  typedef std::size_t    size_type;
  typedef std::ptrdiff_t difference_type;
    
  //iterator support
  iterator begin() {return data.begin();}
  const_iterator begin() const {return data.begin();}
  iterator end() {return data.end();}
  const_iterator end() const {return data.end();}

  int size() const {return SIZE;}
  float& operator[](int i) {return data[i];}
  const float& operator[](int i) const {return data[i];}

  void clear();
  void clean();
  void normalize(float weight);
  void add(const Self& f);
  bool equals(const Self& f) const;
  string toString() const;
  void addCode(const char* s, float weight);
  void print() const;

};

struct 
Profile::
FPairFull 
{
  typedef FPairFull Self;
  static const int SIZE = Alpha::NPBASE;
  float data[SIZE];
  float data_0[SIZE];
    
  unsigned int size() const {return SIZE;}
  float& operator[](int i) {return data[i];}
  const float& operator[](int i) const {return data[i];}

  void clear();
  void clean();
  void normalize(float weight);
  void add(const Self& f);
  bool equals(const Self& f) const;
  void print() const;

  void addPairCode(const char* si, const char* sj, float weight);
  void addBasePairProb(const float& prob, const float& weight);
  string toString() const;
};

struct 
Profile::
FPairReduced 
{
  typedef FPairReduced Self;
  static const int SIZE = Alpha::REDUCED_NPBASE;
  float data[SIZE];
  float data_0[SIZE];
    
  unsigned int size() const {return SIZE;}
  float& operator[](int i) {return data[i];}
  const float& operator[](int i) const {return data[i];}

  void clear();
  void clean();
  void normalize(float weight);
  void add(const Self& f);
  bool equals(const Self& f) const;
  void addPairCode(const char* si, const char* sj, float weight);
  void addBasePairProb(const float& prob, const float& weight);
  string toString() const;
  void print() const;
};
struct 
Profile::
FDimer  
{
  typedef FDimer Self;
  static const int SIZE = 4;
  enum {LL = 0, LG = 1, GL = 2, GG = 3};
  float data[SIZE];

  float& getLL() {return data[LL];}
  float& getLG() {return data[LG];}
  float& getGL() {return data[GL];}
  float& getGG() {return data[GG];}
  const float& getLL() const {return data[LL];}
  const float& getLG() const {return data[LG];}
  const float& getGL() const {return data[GL];}
  const float& getGG() const {return data[GG];}
  int size() const {return SIZE;}
  float& operator[](int i) {return data[i];}
  const float& operator[](int i) const {return data[i];}
  void clear();
  void clean();
  void normalize(float weight);
  void add(const Self& f);
  void addPairCode(const char* si, const char* sj, float weight);
  float getBaseWeight() const {return data[LL] + data[GL];}
  float getGapWeight() const {return data[LG] + data[GG];}
  float getWeight() const {return getBaseWeight() + getGapWeight();}
  bool equals(const Self& f) const;
  string toString() const;
  void print() const;
};


struct 
Profile::
FStack 
{
  typedef FStack Self;
  enum {
    LENGTH = Alpha::REDUCED_NPBASE,
    SIZE = LENGTH * LENGTH,
    LENGTH1 = Alpha::REDUCED_NPBASEN
  };
  typedef CMatrix<float,LENGTH,LENGTH> Data;
  typedef CArray<float,LENGTH1> FPair;
  Data data;
  Data data_0;
  float basePairProb;
  FPair fPair;
  FPair fPair_0;
  float lodScore;

  typedef float T;
  typedef T              value_type;
  typedef T*             iterator;
  typedef const T*       const_iterator;
  typedef T&             reference;
  typedef const T&       const_reference;
  typedef std::size_t    size_type;
  typedef std::ptrdiff_t difference_type;
    
  //iterator support
  iterator begin() {return data.begin();}
  const_iterator begin() const {return data.begin();}
  iterator end() {return data.end();}
  const_iterator end() const {return data.end();}

  int size() const {return SIZE;}
  int length() const {return LENGTH;}
  float& ref(int i, int j) {return data.ref(i, j);}
  float& ref_0(int i, int j) {return data_0.ref(i, j);}
  const float& ref(int i, int j) const {return data.ref(i, j);}
  const float& ref_0(int i, int j) const {return data_0.ref(i, j);}
  void clear();
  void clean();
  void normalize(float weight);
  void add(const Self& f);
  bool equals(const Self& f) const;
  void addPairCode(const char* si, const char* sj, float weight);
  void addBasePairProb(const float& prob, const float& weight);
  void getFPair(float* fPair, float* fPair_0 = NULL) const;
  float getBasePairProb() const;
  void setFPair();
  void setLODScore(const FCol& fColL, const FCol& fColR);

  float getCanFrac() const;
  string toString() const;
  void print() const;
};

#endif
