/*
 * ProfileVectors.hpp
 *
 */
#ifndef PROFILE_VECTORS_HPP
#define PROFILE_VECTORS_HPP

class PosMap;
class EScript;
class Seq;

// TODO use private inheritance
struct 
Profile::VFCol 
  : public Vector<FCol> 
{
  typedef Vector<FCol> Super;
  int length() const {return Super::size();}
  void addSeq(const Vector<char>& seq, float seqWgt);
  void edit(const EScript& es);
  void merge(const VFCol& vFCol);
  float computeInterProfileScore(const VFCol& vFCol) const;
  void chunk(int chunkSize);
  void normalize();
  void normalize(float weight);
  void clean();
  string toString() const;
   void print() const;
};

struct Profile::VFPair 
  : public Vector<FPair> 
{
  typedef Vector<FPair> Super;
  static int computeNPair(int length);
  static int computeIndex(int length, int i, int j);

  void setLength(int length);
  void addSeq(const Vector<char>& seq, float seqWgt);
  int length() const;
  int getAreaAfter(int i , int j) const;
  int getIndex(int i, int j) const;
  const FPair& ref(int i, int j) const {return get(getIndex(i, j));}
  FPair& ref(int i, int j) {return get(getIndex(i, j));}
  void edit(const EScript& es);
  void merge(const VFPair& vfPair);
  float computeInterProfileScore(const VFPair& vFPair) const;
  void chunk(int chunkSize);
  void normalize();
  void normalize(float weight);
  void clean();
  string toString() const;
   void print() const;
};

struct 
Profile::VFDimer 
  : public Vector<FDimer> 
{
  typedef Vector<FDimer> Super;
  void addSeq(const Vector<char>& seq, float seqWgt);
  void edit(const EScript& es);
  void merge(const VFDimer& vFDimer);
  void chunk(int chunkSize);
  float getBaseWeightAt(int i) const;
  float getGapWeightAt(int i) const;
  float getWeight() const;
  float getLGOccupancy(int i) const;
  float getGLOccupancy(int i) const;
  void clean();
  void normalize(float weight);
  string toString() const;
   void print() const;
};

#endif
