/*
 * ProgressiveAligner.cpp
 *
 */
#include "Util.hpp"
#include "Alpha.hpp"
#include "Vector.hpp"
#include "Rect.hpp"
#include "SeqData.hpp"
#include "Profile.hpp"
#include "MSA.hpp"
#include "ScoreParam.hpp"
#include "EScript.hpp"
#include "Dmtx.hpp"
#include "Weight.hpp"
#include "GuideTree.hpp"
#include "Cluster.hpp"
#include "Trapezoid.hpp"
#include "DPPath.hpp"
#include "DPRegion.hpp"
#include "DPDeck.hpp"
#include "Aligner.hpp"
#include "Job.hpp"
// #include "EvalAlign.hpp"
#include "Scheduler.hpp"
#include "SeqFamAligner.hpp"
#include "ProgressiveAligner.hpp"

//pthread_mutex_t ProgressiveAligner::mutex = PTHREAD_MUTEX_INITIALIZER;

void*
ProgressiveAligner_Client_run(void* pProgressiveAligner)
{
  ProgressiveAligner& pa = *(ProgressiveAligner*) pProgressiveAligner;
  ProgressiveAligner::Client client(pa);
  client.run();
  return NULL;
}
ProgressiveAligner::Client::
Client(ProgressiveAligner& progressiveAligner)
  : progressiveAligner(progressiveAligner),
    clientID(-1),
    msaX(progressiveAligner.seqData),
    msaY(progressiveAligner.seqData),
    dpPath(),
    score(IMPOSSIBLE),
    params()
{
}
ProgressiveAligner::Client::
~Client()
{
  if (clientID >= 0) {
    progressiveAligner.scheduler.finish(clientID);
    progressiveAligner.scheduler.removeClient(clientID);
  }
}
void
ProgressiveAligner::Client::
run()
{
  ProgressiveAligner& pa = progressiveAligner;
  Scheduler& scheduler = pa.scheduler;
  const GuideTree& guideTree = pa.guideTree;
  const Weight& weights = pa.weights;
  Vector<Status>& vStatus = pa.vStatus;
  Vector<MSA>& vMSA = pa.vMSA;
  if (clientID < 0) {  
    clientID = scheduler.createClient();
  }

  //#ifdef DEBUG
  cout << "progressive alignment" << endl;
  //#endif

  int align_id = 0;
  ostringstream oss;
  const int size = vStatus.size();
  int begin = 0;
  //pthread_mutex_lock(&progressiveAligner.mutex);
  int counter = 0;
  int max_count = (guideTree.getNLeaf() - 1);
  while (true) {
    int nd = -1;
    int end = begin + size;
    for (int i = begin; i < end; i++) {
      int ndi = i % size;
      Status& status = vStatus[ndi];
      if (status != TOBEDONE) continue;

      const GuideTree::CLink& link = guideTree.getLink(ndi);
      if (link->isLeaf()) {
	nd = ndi;
	status = OCCUPIED;
	break;
      }
      const Status& status1 = vStatus[link->getChild(0)->getId()];
      const Status& status2 = vStatus[link->getChild(1)->getId()];
      if (status1 == FINISHED && status2 == FINISHED) {
	nd = ndi;
	status = OCCUPIED;
	break;
      }
    }
    if (nd == -1) break; // no work available to this thread.

    //pthread_mutex_unlock(&progressiveAligner.mutex);


    dpPath.clear();
    score = IMPOSSIBLE;
    params.clear();

    Status& status = vStatus[nd];
    MSA& msa0 = vMSA[nd];
    const GuideTree::CLink& link = guideTree.getLink(nd);
    if (link->isLeaf()) {
      msa0.setSeq(link->getSeqId());
    } else {
      // no thread except the present one touches these nodes
      MSA& msaX0 = vMSA[link->getChild(0)->getId()];
      MSA& msaY0 = vMSA[link->getChild(1)->getId()];

      bool print_seq_weight = true;
      if (print_seq_weight) {
	ostringstream oss1;
	oss1 << "\n---\n"
	     << "ac: " << msaX0.getAC() << "_a" << align_id << "\n"
	     << "id: " << msaX0.getID() << "_a" << align_id << "\n";
	align_id++;
	oss1 << "msaX: \n";
	for (int i = 0; i < (int)msaX0.size(); i++) {
	  const string& name = msaX0.getSeq(i).getName();
	  oss1 << "  - " << name << "\n";
	}
	oss1 << "msaY: \n";
	for (int i = 0; i < (int)msaY0.size(); i++) {
	  const string& name = msaY0.getSeq(i).getName();
	  oss1 << "  - " << name << "\n";
	}
	oss1 << "wgtX: \n";
	for (int i = 0; i < (int)msaX0.size(); i++) {
	  const SeqId& id = msaX0.getSeqId(i);
	  const float& weight = weights.ref(id);
	  oss1 << "  - " << weight << "\n";
	}
	oss1 << "wgtY: \n";
	for (int i = 0; i < (int)msaY0.size(); i++) {
	  const SeqId& id = msaY0.getSeqId(i);
	  const float& weight = weights.ref(id);
	  oss1 << "  - " << weight << "\n";
	}

	const string str = oss1.str();
	//pthread_mutex_lock(&progressiveAligner.mutex);
	(*progressiveAligner.pSLog) += str;
	//pthread_mutex_unlock(&progressiveAligner.mutex);
      }   

      cout << "--- \n";
      cout << ++counter << "/" << max_count << endl;

      Job job;
      job.set_Align(guideTree, weights, msaX0, msaY0);
      //job.set_Align1(guideTree, weights, msaX0, msaY0);
      //job.set_Align2(guideTree, weights, msaX0, msaY0);
      //job.set_Align3(guideTree, weights, msaX0, msaY0); // only pscfg case
      scheduler.addJob(clientID, job);
      scheduler.getResult(clientID, job);
      string message;
      job.get_AlignResult(msaX, msaY, dpPath, score, params, message);
      //cout << message << endl;
      (*progressiveAligner.pSLog) += message;
      const float score0 = score;

      job.set_Merge(msaX, msaY, dpPath);
      scheduler.addJob(clientID, job);
      msaX.merge(msaY, dpPath);

      dpPath.clear();
      params.clear();
      score = IMPOSSIBLE;
      job.set_Fold(guideTree, weights, msaX, dpPath);
      //job.set_Fold2Step(guideTree, weights, msaX, dpPath);
      scheduler.addJob(clientID, job);
      scheduler.getResult(clientID, job);
      job.get_AlignResult(msaX, msaX, dpPath, score, params);

      oss.str("");
      if (pa.outfile_format == "NCRNA_WEB_SERVER") {
	msaX.sortSeqId();
	oss << pa.get_ncrna_web_server_format_string(guideTree, msaX, dpPath.ct, score0);
      } else {
	msaX.sortSeqId();
	msaX.writeStockholm(oss, dpPath.ct.getSSCons(), score0);
      }

      cout << "done.\n";

      //pthread_mutex_lock(&progressiveAligner.mutex);
      string(oss.str()).swap(*(progressiveAligner.pSOut));

      float treeDistance = guideTree.getTreeDistance(link->getChild(0)->getId(),
						     link->getChild(1)->getId());

      msaX0.resetPosMap();
      msaY0.resetPosMap();
      float seqSim = msaX0.computeSeqIdentity(msaY0); // cannot compute in advance
      oss << "tree_dist: " << treeDistance << "\n";
      oss << "seq_ident: " << seqSim << "\n";
      //cout << oss.str();
      (*progressiveAligner.pSLog) += oss.str();
    
      msa0 = msaX;
      msaX0.clear();
      msaY0.clear();

#ifdef DEBUG
      cout << "---\n";
      cout << setprecision(3);
      cout << "score: " << score << "\n";
      msa0.print();
      //dpPath().print();
#endif
    }
  
    status = FINISHED;
    if (link->isRoot()) {
      break;
    } else {
      begin = link->getParent()->getId();
    }
  }
  //pthread_mutex_unlock(&progressiveAligner.mutex);

  //#ifdef DEBUG
  //cout << "progressiveAligner client" << clientID << ": exiting" << endl;
  //#endif
  scheduler.finish(clientID);
}

ProgressiveAligner::
ProgressiveAligner(SeqData& seqData,
		   GuideTree& guideTree,
		   Weight& weights,
		   Scheduler& scheduler,
		   int clientSize,
		   string outfile_format)
  : Super(seqData, scheduler),
    //clientSize(clientSize),
    guideTree(guideTree),
    weights(weights),
    vStatus(),
    vMSA()//,
    //vThread()
    ,outfile_format(outfile_format)
{
}

ProgressiveAligner::
ProgressiveAligner(const ProgressiveAligner& progressiveAligner)
  : Super(progressiveAligner),
    //clientSize(progressiveAligner.clientSize),
    guideTree(progressiveAligner.guideTree),
    weights(progressiveAligner.weights),
    vStatus(progressiveAligner.vStatus),
    vMSA(progressiveAligner.vMSA)//,
  //vThread(progressiveAligner.vThread)
    ,outfile_format(progressiveAligner.outfile_format)
{
}
ProgressiveAligner::
~ProgressiveAligner()
{
}
ProgressiveAligner&
ProgressiveAligner::
dup() const
{
  return *new ProgressiveAligner(*this);
}
void
ProgressiveAligner::
run()
{
  Job job;

  cout << "computing guide tree..." << endl;
  //guideTree.setFromTupleCount(seqData, seqIds);
  guideTree.setFromAlignDistance(seqData, seqIds, scheduler);
  weights.setFromGuideTree(guideTree);
  cout << "done." << endl;

  ostringstream oss;
  oss << guideTree.toString() << endl;
  oss << weights.toString() << endl;
  *pSLog += oss.str();

  int nNode = guideTree.getNNode();
  vStatus.assign(nNode, TOBEDONE);
  vMSA.assign(nNode, MSA(seqData));

  //int nThread = min(clientSize, (int)seqIds.size());
  ////TODO check thread size limitation using sysconf and MAX_THREADS
  //vThread.resize(nThread);
  //for (int i = 0; i < nThread; i++) {
  //  int err = pthread_create(&vThread[i], NULL,
  //			     ProgressiveAligner_Client_run, (void*)this);
  //  if (err) {
  //    vThread.resize(i);
  //    break;
  //  }
  //}
  //if (vThread.empty()) Die("cannot create client threads");

#ifdef DEBUG
  //cout << vThread.size() << " clients started." << endl;
#endif

  //for_each(vThread.begin(), vThread.end(),
  //	   bind2nd(ptr_fun(pthread_join), (void**)0));

  ProgressiveAligner_Client_run((void*)this);

  // for debug. this should be Assert
  Check(seqData.nColsAreEqual(seqIds));

  // add secondary structure prediction
  //MSA msaTotal(seqData, seqIds);
  //ostringstream oss;
  //msaTotal.writeStockholm(oss);
  //string(oss.str()).swap(*pSOut);
}
string
ProgressiveAligner::
get_ncrna_web_server_format_string(const GuideTree& guideTree,
				   const MSA& msa, const CT& ct, const float& score0) const
{
  ostringstream oss;
  oss << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
      << "<result>\n"
      << "<tree>\n";
  string nw = guideTree.getNewickString();
  nw.resize(nw.size() - 1);//remove semicolon from newick format
  oss << nw << "\n"
      << "</tree>\n";
  for (int i = 0; i < (int)msa.size(); i++) {
    const Seq& seq = msa[i];
    oss << "<php ref=\"" << seq.getId() << "\">\n"
	<< "<name>\n"
	<< seq.getName() << "\n"
	<< "</name>\n"
	<< "<sequence>\n";
    const string& str = seq.getSequenceString();
    const int maxSeqWidth = 70;
    for (int j = 0; j < (int)str.size(); j += maxSeqWidth) {
      int seqWidth = min(maxSeqWidth, (int)str.size() - j);
      for (int k = 0; k < seqWidth; k++) {
	oss << str[j + k];
      }
      oss << "\n";
    }
    oss << "</sequence>\n"
	<< "</php>\n";
  }
  oss << "<structure>\n";
  
  string sscons(ct.getSSCons());
  for (int i = 0; i < (int)sscons.size(); i++) {
    switch (sscons[i]) {
    case '<': sscons[i] = '('; break;
    case '>': sscons[i] = ')'; break;
    }
  }
  // const int maxSeqWidth = 70;
  // for (int j = 0; j < (int)str.size(); j += maxSeqWidth) {
  //   int seqWidth = min(maxSeqWidth, (int)str.size() - j);
  //   for (int k = 0; k < seqWidth; k++) {
  //     oss << str[j + k];
  //   }
  //   oss << "\n";
  // }
  oss << sscons
      << "</structure>\n"
      << "<coordinate>\n";
  int k = 0;
  for (int i = 0; i < (int)ct.size(); i++) {
    if (ct.isLeftBase(i)) {
      k++;
      oss << "<segment position=\"" << k << "\" "
	  << "starts=\"" << (i + 1) << "\" "
	  << "ends=\"" << (ct[i] + 1) << "\"/>\n";
    }
  }
  oss << "</coordinate>\n"
      << "<stockholm>\n";

  msa.writeStockholm(oss, sscons, score0);
  oss << "</stockholm>\n"
      << "<aln>\n";
  msa.writeClustal(oss);
  oss << "</aln>\n";
  oss << "<mfa>\n";

  ostringstream oss1;
  msa.writeFasta(oss1, sscons);
  istringstream iss1(oss1.str());
  string buf;
  while (getline(iss1, buf)) {
    if (buf[0] == '>') {
      buf[0] = ')';
    }
    oss << buf << "\n";
  }
  oss << "</mfa>\n"
      << "</result>\n";
  return oss.str();
}
