/*
 * Scheduler.hpp
 *
 */
#ifndef SCHEDULER_HPP
#define SCHEDULER_HPP

#include "Job.hpp"
#include "JobRunner.hpp"
class BasePairData;
// #include "BaliScore.hpp"
//#include <pthread.h>

class Scheduler {
public:
  Scheduler(const JobRunner& jobRunner);
  ~Scheduler();

  bool isRoot() const;
  int getNServant() const;
  int createClient();
  void removeClient(int clientID);
  int addJob(const int& clientID, const Job& job);
  int getResult(const int& cliendID, Job& job);
  void finish(int clientID);
  int getRemaining(int clientID) const;
  string toString() const;
  void print() const;

private:
  struct TaggedJob;
  struct ClientState;
  struct ServantState;
  class Servant;
  class Channel;
  struct TaggedJob {
    enum {ANY_SERVANT = -1};
    enum {CHANNEL_CLIENT = -1};
    Job job;
    int id;
    int clientID;
    int servantID;
    TaggedJob(const Job& job, int id, int clientID, int servantID);
    TaggedJob(const TaggedJob& taggedJob);
    ~TaggedJob();
  private:
    TaggedJob();
  };
  struct ClientState {
    bool isWaiting;
    list<TaggedJob> results;
    int remaining;
    // pthread_cond_t cv;

    ClientState();
    ~ClientState();
    string toString() const;
  };
  struct ServantState {
    bool isBusy;
    list<TaggedJob> jobQ;

    ServantState();
    ~ServantState();
    string toString() const;
  };
  class Channel {
  public:
    Channel(Scheduler& scheduler);
    ~Channel();
    void run();
    string toString() const;
  private:
    Scheduler& scheduler;
    bool shutdownRequested;
    // pthread_t thread;
    //pthread_cond_t cv;

    int collectResults();
    int throwJobs();
    bool servantsAreAllBusy() const;
    bool clientsAreAllWaiting() const;
    void waitAnyServant();
    void waitAnyClient();
    void timedWaitAnyClient();
    void sendRequest(int servantID, const Job& job);
    bool resultIsArrived(int servantID) const;
    Job recvResult(int servantID);
    friend class Servant;
    friend class Scheduler;
  };
  class Servant {
  public:
    Servant(Scheduler& scheduler, const JobRunner& jobRunner);
    ~Servant();
    void run();
    void recvRequest();
    void sendResult();
    void executeCommand();
    string toString() const;
    void setJob(const Job& job);
    Job getJob() const;
  private:
    Scheduler& scheduler;
    JobRunner jobRunner;
    // pthread_t thread;
    // pthread_cond_t cv;
    bool jobIsCompleted;
    friend class Channel;
    friend class Scheduler;
  };

  bool isMultiProcess;
  list<TaggedJob> jobQ;
  int jobCounter;
  int remaining;
  // static pthread_mutex_t mutex;

  typedef map<int, ClientState> ClientStates;
  typedef Vector<ServantState> ServantStates;
  ClientStates clientStates;
  ServantStates servantStates;
  Channel channel;
  Servant servant;

  Scheduler();
  Scheduler(const Scheduler& scheduler);
  Scheduler& operator=(const Scheduler& scheduler);

  void terminate();
  friend class Channel;
  friend class Servant;
  friend void* Scheduler_Channel_run(void* pScheduler);
  friend void* Scheduler_Servant_run(void* pScheduler);
};

#endif
