/*
 * StripRegion.hpp
 *
 */
#ifndef STRIP_REGION_HPP
#define STRIP_REGION_HPP

#include "Util.hpp"
#include "Rect.hpp"
#include "Trapezoid.hpp"

class StripRegion : private Vector<Trapezoid> {
  typedef Vector<Trapezoid> Super;
public:
  union {
    Rect rect0;
    struct {Point b0, e0;};
    struct {int bx0, by0, ex0, ey0;};
  };
  StripRegion();
  StripRegion(const StripRegion& sr);
  virtual ~StripRegion();
  StripRegion& operator=(const StripRegion& sr);
  bool operator==(const StripRegion& sr) const;

  using Super::empty;
  using Super::size;
  void clear();
  const Trapezoid& operator[](int i) const {return Super::operator[](i);}
  string toString() const;
  void print() const;
  void draw(ostream& fo) const;
  void draw() const {draw(cout);}

  void push(const Trapezoid& t);

  bool includes(const Point& pt) const;
  bool includes(const Rect& rect) const;
  int getBX(int y) const;
  int getEX(int y) const;
  int getXTIndex(int x) const;
  int getBY(int x) const {return Super::get(getXTIndex(x)).getBY(x);}
  int getEY(int x) const {return Super::get(getXTIndex(x)).getEY(x);}
  int getBXTIndex(int y) const;
  int getEXTIndex(int y) const;
  int getBXEY() const;
  int getEXBY() const;
  int getArea() const;
  int getMaxDiag() const;
  int getMinDiag() const;
  int getDiagWidth() const;
  int getAreaBefore(const Point& pt) const;
  int getAreaBeforeT(const Point& pt) const;
  int getAreaLeft(int y) const;
  int getAreaTopLeft(const Point& pt) const;
  int getAreaBottomRight(const Point& pt) const;
  //int hasLockedPath() const;
  void setFromYRange(const Rect& rect00,
		     const Vector<int>& y_begs,
		     const Vector<int>& y_lasts);
  void getYRange(Vector<int>& y_begs, Vector<int>& y_lasts) const;
  StripRegion getBoundedRegion(const Rect& bbox) const;
  StripRegion getBoundingRegion(const StripRegion& stripRegion) const;

  // margin methods
  int getAreaMargin() const;
  int getAreaMarginBefore(const Point& pt) const;
  int getTopLeftMargin(const int& i) const;
  int getBottomRightMargin(const int& i) const;

  void testConsistency() const;
  template <bool fwd> class RowIterator;
  template <bool fwd> class Iterator;

private:
  void resetRect0();
  const Trapezoid& get(int i) const {return Super::get(i);}

  //int getAreaInside(const Rect& rect) const;
  //int getAreaBetween(const Rect& rect) const;
  //int getEXEY(int y) const;
  //int getLX() const {return (ex0 - bx0 + 1);}
  //int getLY() const {return (ey0 - by0 + 1);}
  //int getLY(int x) const;
  //bool getCoordFromArea(Point& pt, const int& area) const;
  //
  //int getLeftMargin(const int& i, const int& x) const;
  //int getRightMargin(const int& i, const int& x) const;
  //int getLYMargin(const int& x) const;
  //int getLYMargin(const int& i, const int& x) const;
  //int getAreaMarginBetween(const Rect& rect) const;
  //void testGetAreaBottomRight() const;
  //void testGetAreaTopLeft() const;
};

template <bool fwd>
class StripRegion::RowIterator {
public:
  int x;
  int ly;
  int y_beg;
  int y_last;

  RowIterator(const StripRegion& sr);
  RowIterator(const StripRegion& sr, const Rect& bbox);
  ~RowIterator();
  void reset(const Rect& rect0);
  void reset();
  bool getNext();
private:
  const StripRegion& sr;
  Rect bbox;
  int idx_lim;
  int idx;
  int x_lim;
  int next_x;
};

template <bool fwd>
class StripRegion::Iterator {
public:
  union {
    Point pt;
    struct {int x, y;};
  };
  Iterator(const StripRegion& sr);
  Iterator(const StripRegion& sr, const Rect& bbox);
  ~Iterator();
  bool getNext();
  void reset(const Rect& rect0);
  void reset();
private:
  RowIterator<fwd> rit;
  int y_lim;
  int next_y;
  void reset0();
};

#include "StripRegionIterators.cpp"
#endif
