/*
 * Weight.hpp
 *
 */
#ifndef WEIGHT_HPP
#define WEIGHT_HPP

#include "Util.hpp"
#include "SeqData.hpp"

class GuideTree;
class MSA;

class Weight {
public:
  Weight();
  Weight(const Weight& weight);
  ~Weight();
  Weight& operator=(const Weight& weight);
  string toString() const;
  void print() const;

  //float& operator[](int i) {return data[i];}
  //const float& operator[](int i) const {return data[i];}

  bool empty() const;
  int size() const;
  void clear();
  void fill(const float& val);
  const float& ref(const SeqId& seqId) const;
  float& ref(const SeqId& seqId);

  void setType(const string& weightType);
  void setSeqIds(const SeqIds& seqIds1);
  const SeqIds& getSeqIds() const {return seqIds;}
  void assign(const SeqIds& seqIds1, const Vector<float>& values1);
  void appendPairs(const SeqIds& seqIds1,
		   const Vector<float>& values1);
  void append(const Weight& weights1);
  Vector<char> getPackedData() const;
  void setFromPackedData(const Vector<char>& packedData);

  void setFromTupleCount(const SeqData& seqData,
			 const SeqIds& seqIds1);
  void setFromSeqIdentity(const MSA& msa);
  void setFromGuideTree(const GuideTree& guideTree);
  void setFlatWeight(const SeqIds& seqIds1);
  
 
  //iterator support
  //iterator begin() {return data.begin();}
  //const_iterator begin() const {return data.begin();}
  //iterator end() {return data.end();}
  //const_iterator end() const {return data.end();}

protected:
  typedef Vector<float> Data;
  static const float MIN_VALUE;
  enum Type {
    FLAT,
    CLUSTALW
  };
  Type type;
  SeqIds seqIds;
  Data data;

  void setFlatWeight(const GuideTree& guideTree); 
  void setClustalWWeight(const GuideTree& guideTree);
  void testConsistency() const;
  float getMinWeight() const;
};

#endif
