/*
 * DPDeckPSCFG.hpp
 *
 */
#ifndef DP_DECK_PSCFG_HPP
#define DP_DECK_PSCFG_HPP

#include "Util.hpp"
#include "DPDeck.hpp"
#include "DPScore.hpp"
#include "PSCFG_Base.hpp"
#include "PSCFG_Energy.hpp"
#include "PSCFGEnergyScoreManager.hpp"
//#include "PSCFGBaseScoreManager.hpp"

class ScoreParam;
class MSA;
class DPRegion;
class DPPath;
class SkipSet;
class PairProbData;
class AlignProbData;

class 
DPDeckPSCFG 
  : public DPDeck 
{
public:
  typedef PSCFG_Energy  Model;
  typedef Model::State  State;
  typedef Model::TBNode TBNode;
  typedef Model::ScoreManager ScoreManager;
  typedef Model::EmitModel EmitModel;// should be enum EmitType
  typedef Model::EmitState EmitState;// should be enum EmitType
  typedef DPScore::DPScoreT DPScoreT;
  static const DPScoreT NegInfDPScore;
  static const DPScoreT ZeroDPScore;
  inline static bool IsFinite(const DPScoreT& z) {
    return DPScore::IsFiniteDPScore(z);
  }
  class DPRegion;
  class DPTask;
  class TBTree;
  typedef UpTriMtx<float> BasePairProb;

  DPDeckPSCFG();
  ~DPDeckPSCFG();
  string toString() const;
  void print() const;
  string getLogString() const {return oss_log.str();}

  void setScoreParam(ScoreParam& scoreParam);
  void setScoreParamList(Vector<ScoreParam>& scoreParamList);
  void setGuideTree(GuideTree& guideTree);
  void setWeight(Weight& weights);
  void setAlignProb(const NMatrix<float>& alignProb);
  void setPairProbData(const PairProbData& pairProbData);
  void setAlignProbData(const AlignProbData& alignProbData);
  void setSkipSize(int skipSize0);
  int getSkipSize() const;
  void setRefinementStrategy(const string& strategy);
  void setMSAs(const MSA& msaX, const MSA& msaY);
  void setMSAs(const MSA& msaX, const MSA& msaY,
	       const BasePairProb& basePairProbX1,
	       const BasePairProb& basePairProbY1) {
    scoreManager.setMSAs(msaX, msaY, basePairProbX1, basePairProbY1);
  }
  void setSequenceSimilarity(const float& z) {scoreManager.setSequenceSimilarity(z);}
  void setDPRegion(const ::DPRegion& dpRegion);
  void setDPPath(const DPPath& dpPath0);
  const DPPath& getDPPath() const;
  float getScore() const;
  void run();

  Vector<float> computeFeatureVector() const;
  Vector<float> getParamVector() const;
  int getParamVectorSize() const;
  const char* getParamNameFromOffset(int offset) const;
  void setParamVector(const Vector<float>& pv);
  string featureVect2String(const Vector<float>& fv, const int& indentLevel) const;
  Vector<float> getParamVectorFromCountVector(const Vector<float>& countVector) const;
  float computeWorkspaceSizeInMB(); 
  float computeDPIterationCountInUnitsOfMegaTimes() const;
  map<string, float> computeDPIterationCount() const;
  float getTimeEstimate() const;

  ScoreManager& getScoreManager() {return scoreManager;}
  DPScoreT computeScoreFromTree(const TBTree& tree,
				const Vector<State>& states, bool log_progress = false) const;
private:
  enum BFType {
    BF_TYPE_0 = 0,
    BF_TYPE_1 = 1,
    BF_TYPE_2 = 2,
    INVALID_BF_TYPE = -1
  };
  enum SplitType {
    SPLIT_TYPE_B = 0,
    SPLIT_TYPE_M = 1,
    SPLIT_TYPE_E = 2,
    INVALID_SPLIT_TYPE = -1
  };
  struct TZNode;
  struct XNode;
  struct YNode;
  struct PairNode;
  struct BENode;
  struct TBBFNode;
  
  class VTZ;
  class VX;
  class VY;
  class VPairX;
  class VPairY;
  class VBE;
  class VSC;
  class VSkip;
  class VSTB;
  class VSTE;
  class VTB;
  class VTBBF;
  class VSFBWB;
  class VSFBWE;
  class VFBW;
  template <typename Task> class DPLocals;
  //class DPLocals;

  const Model& model;
  ScoreManager& scoreManager;
  SkipSet& skipSet;
  DPRegion& dpRegion;
  VTZ& vt;
  VX& vx;
  VY& vy;
  VPairX& vpairX;
  VPairY& vpairY;
  VBE& vbe;
  VSC& vsc;
  VSkip& vskip;
  VTB& vtb;
  VTBBF& vtbbf;
  VSFBWB& vsfbwB;
  VSFBWE& vsfbwE;
  VFBW& vfw;
  VFBW& vbw;
  TBTree& tbTree;
  DPPath& dpPath;
  DPScoreT dpScore;
  mutable ostringstream oss_log;

  TBTree cyk(const DPTask& dpTask0);
  TBTree cyk_small(const DPTask& dpTask0);
  DPScoreT inside_outside(const DPTask& dpTask0);
  DPScoreT likelihood(const DPTask& dpTask0);

  float computeCYKMemorySizeInMB(const DPTask& dpTask) const;
  float computeSmallCYKMemorySizeInMB(const DPTask& dpTask) const;

  class EscVec;
  class TscVec;
  template <typename FeatureCounter>
  void countFeatures(FeatureCounter& fc) const;

  DPDeckPSCFG(const DPDeckPSCFG& dpDeck);
};

#endif
