/* 
 * DPDeckPSCFGDPLocals.cpp
 *
 */
#ifndef DP_DECK_PSCFG_DP_LOCALS_CPP
#define DP_DECK_PSCFG_DP_LOCALS_CPP

#include "DPDeckPSCFGDPLocals.hpp"
#include "DPDeckPSCFGVectors.hpp"
#include "DPDeckPSCFGDPTask.hpp"
#include "DPDeckPSCFGDPTaskSubTask.cpp"
#include "DPDeckPSCFGDPLocalsTest.cpp"

#define TEMPLATE template<typename Task>
#define CLASS DPDeckPSCFG::DPLocals<Task>

#ifdef DEBUG
extern bool DEBUG_PSCFG_DP_LOCALS;
static int debug = true;
static int dpCounter = 0;
static int dpCounter1 = 0;
#endif

TEMPLATE
const DPDeckPSCFG::Model&
CLASS::
model = DPDeckPSCFG::Model::getModel();

TEMPLATE
inline
CLASS::
DPLocals(Task& task)
  : 
#ifdef DEBUG
  dpDeck(task.dpDeck),
  vt(dpDeck.vt),
  vx(dpDeck.vx),
  vy(dpDeck.vy),
  vpairX(dpDeck.vpairX),
  vpairY(dpDeck.vpairY),
  vbe(dpDeck.vbe),
  vsc(dpDeck.vsc),
  vskip(dpDeck.vskip),
  vtb(dpDeck.vtb),
  vtbbf(dpDeck.vtbbf),
  vsfbwB(dpDeck.vsfbwB),
  vsfbwE(dpDeck.vsfbwE),
  vfw(dpDeck.vfw),
  vbw(dpDeck.vbw),
  dpTask(task.dpTask),
  vstb(dpTask.vstb),
  vste(dpTask.vste),
  dpRegion(task.dpRegion),
  region(task.region),
  skipSet(region.getSkipSet()),
#endif
  scoreManager(task.dpDeck.getScoreManager()),
  task(task),
  escTable(scoreManager),
  tscTable(scoreManager)

#ifdef DEBUG
    ,
    dpi(region),
    bfi(region)
#endif
{
#ifdef DEBUG
  debug = DEBUG_PSCFG_DP_LOCALS;
#endif
}

TEMPLATE
inline
CLASS::
~DPLocals()
{
}
TEMPLATE
void
CLASS::
clear()
{
  stateFlags.clear();

  dpAreaMargin = 0;
  state0 = Model::X_st;
  state_lim = Model::X_st;

  scb_store = NULL;
  sce_store = NULL;

  tau = NULL;
  tau_BF = NULL;

  bt_lim = NULL;
  et0 = NULL;
  et_lim = NULL;
  bt = NULL;
  et = NULL;

  bx0 = NULL;    
  bx_lim = NULL; 
  by0 = NULL;    
  by_lim = NULL; 
  ex0 = NULL;    
  ex_lim = NULL; 
  ey0 = NULL;    
  ey_lim = NULL; 
  bx = NULL;     
  by = NULL;     
  ex = NULL;     
  ey = NULL;     

  l_margin = 0;

  pairX = NULL;
  pairY = NULL;
  prev_pairX = NULL;
  prev_pairY = NULL;

  bn = NULL;
  en0 = NULL;
  en = NULL;

  curr_sc = NULL;
  prevIL = NULL;
  prevDL = NULL;
  prevDR = NULL;

  prev_sc.setNull();
  diag_sc.setNull();

  bxn_beg = 0; 
  bxn_last = 0;
  exn_beg = 0; 
  exn_last = 0;
  mxn_beg = 0; 
  mxn_last = 0;

  bxr0 = 0;
  exr0 = 0;

  bxr = 0;
  exr = 0;
  bxn = 0;
  exn = 0;

  skipSize = 1;
  pSkipCountB0 = NULL;
  pSkipCountE00 = NULL;
  pSkipCountE0 = NULL;
  pSkipCountM0 = NULL;

  h_bxn = 0;
  h_exn = 0;
  h_byexn = 0;
  h_eybxn = 0;
  h_intsec = 0;
  pSkipCount0 = NULL;
  pSkipCountL = NULL;
  pSkipShiftL = NULL;
  pSkipCountR = NULL;
  pSkipShiftR = NULL;

  skipRowCount = 0;
  skipCount = 0;
  pSkipCount = NULL;
  scVSTRect1 = NULL;
  scVSTRect2 = NULL;

  best_child_index = (-1);
  best_score = NegInfDPScore;
  best_score1 = NegInfDPScore;
  best_split_ptr_plus_1 = NULL;
  best_state = Model::X_st;
  best_bfType = INVALID_BF_TYPE;
  best_split_offset = 0;
#ifdef SLR_STATE
  best_score_L = NegInfDPScore;
  best_score_R = NegInfDPScore;
#endif
  escTable.clear();
  tscTable.clear();
}
TEMPLATE
void
CLASS::
reset()
{
  clear();

  dpAreaMargin = task.dpAreaMargin; 
  state0 = task.state0;
  state_lim = task.state_lim;

  scb_store = task.scb_store;
  sce_store = task.sce_store;

  tau = task.tau;
  tau_BF = task.tau_BF;

  bt_lim = task.bt_lim;
  bt = task.bt;
  et0 = task.et0;
  et_lim = task.et_lim;

  bx = (Task::insideType ? bt->ex : bt->bx);

  bn = task.bn;
  en0 = task.en0;

  bxn_beg  = task.bxn_beg;
  bxn_last = task.bxn_last;
  exn_beg  = task.exn_beg;
  exn_last = task.exn_last ;
  mxn_beg  = task.mxn_beg ;
  mxn_last = task.mxn_last;

  bxr0 = task.bxr0;
  exr0 = task.exr0;

  bxr = bxr0;
  exr = exr0;
  bxn = (Task::insideType ? (bxn_last + 1) : (bxn_beg - 1));
  //exn

  skipSize = task.skipSize;
  pSkipCountB0 = task.pSkipCountB0;
  pSkipCountE00 = task.pSkipCountE0;
  pSkipCountE0 = task.pSkipCountE0;
  pSkipCountM0 = task.pSkipCountM0;

  escTable.reset();
  tscTable.reset();
}

TEMPLATE
inline void
CLASS::
run()
{
#ifdef DEBUG
  dpCounter1 = 0;
  if (debug) {
    testInitialState();
    //region.print();
    //region.draw();
  }
#endif
  reset();
  dpProcBT();
  computeFinalScoreState();
#ifdef DEBUG
  if (debug) {
    testFinalState();
  }
#endif
}
TEMPLATE
void
CLASS::
computeFinalScoreState()
{
  const ScoreManager::SCNode& sc = *curr_sc;
  best_score = NegInfDPScore;
  best_state = Model::X_st;
  if (Task::insideType) {
    if (state_lim == Model::BEG_st) {
      for (int i = 0; i < model.getNChild(Model::BEG_st); i++) {
	const State& s = model.getChild(Model::BEG_st, i);
	const DPScoreT& score1 = sc[s];
	if (Task::takesMaxScore()) {
	  if (best_score < score1) {
	    best_score = score1;
	    if (Task::isCYK()) {// cyk is inside. hence prev is child
	      //best_child_index = i;
	      best_state = s;
	    }
	  }
	} else {
	  scoreManager.logAdd(best_score, score1);
	}
      }
      if (Task::isCYK()) {
	Assert(IsFinite(best_score));
	Assert(Model::isNormalState(best_state));
      }
    } else if (model.isNormalState(state_lim)) {
      best_score = sc[state_lim];
      best_state = state_lim;
    } else {
      Assert(state_lim == Model::X_st);
      Model::Iterator it;
      while (it.getNext()) {
	const State& s = it.s;
	const DPScoreT& score1 = sc[s];
	if (Task::takesMaxScore()) {
	  if (best_score < score1) {
	    best_score = score1;
	    if (Task::isCYK()) {// cyk is inside. hence prev is child
	      //best_child_index = i;
	      best_state = s;
	    }
	  }
	} else {
	  scoreManager.logAdd(best_score, score1);
	}
      }
    }
  }
  finalScore = best_score;
  finalState = best_state;
}
TEMPLATE
DPDeckPSCFG::DPScoreT
CLASS::
getFinalScore() const
{
  return finalScore;
}
TEMPLATE
DPDeckPSCFG::State
CLASS::
getFinalState() const
{
  return finalState;
}
TEMPLATE
inline void
CLASS::
dpProcBT()
{
#define CHOOSE_DP_PROC_BX(btFlag0) do {\
  stateFlags.resetBT(BtFlag(btFlag0));\
  chooseDPProcBX();\
} while (0)

  if (bt == bt_lim) {
    //dpProcBTInit< BT<(BT_IS_BT0 | BT_IS_BT_LIM)> >();
    CHOOSE_DP_PROC_BX(BT_IS_BT0 | BT_IS_BT_LIM);
    return;
  } else {
    //dpProcBTInit< BT<BT_IS_BT0> >();
    CHOOSE_DP_PROC_BX(BT_IS_BT0);
    dpProcBTUpdate();
    //dpProcBTUpdate();
    while (true) {
      if (bt == bt_lim) {
	CHOOSE_DP_PROC_BX(BT_IS_BT_LIM);
	return;
      }
      CHOOSE_DP_PROC_BX(BT_IS_GENERIC);
      //dpProcBTUpdate< BT<BT_IS_GENERIC > >();
      dpProcBTUpdate();
    }
  }
#undef CHOOSE_DP_PROC_BX
}
//TEMPLATE
//template <typename BT0>
//inline void
//CLASS::
//dpProcBTInit()
//{
//}
TEMPLATE
inline void
CLASS::
dpProcBTUpdate()
{
  (Task::insideType ? bt-- : bt++);
  (Task::insideType ? bx-- : bx++);
  bn += (Task::insideType ? -bt->bottomRightMargin : bt->topLeftMargin);
}
TEMPLATE
inline void
CLASS::
chooseDPProcBX()
{
#define DP_PROC_BX(tzType0)  do {\
  stateFlags.Bt.reset(tzType0);\
  dpProcBX();\
} while (0)

  switch (bt->type) {
  case Trapezoid::LV_RV: DP_PROC_BX(Trapezoid::LV_RV); return;
  case Trapezoid::LV_RD: DP_PROC_BX(Trapezoid::LV_RD); return;
  case Trapezoid::LD_RV: DP_PROC_BX(Trapezoid::LD_RV); return;
  case Trapezoid::LD_RD: DP_PROC_BX(Trapezoid::LD_RD); return;
  default: Die("bad tzType");
  }

#undef DP_PROC_BX
}
TEMPLATE
inline void
CLASS::
dpProcBX()
{
#define DP_PROC_BX_SKIP(bxFlag0) do {\
  stateFlags.resetBX(BxFlag(bxFlag0));\
  dpProcBXSkip();\
} while (0)
#define DP_PROC_BX_UPDATE() dpProcBXUpdate()

  dpProcBXInit();
  if (bx == bx_lim) {
    DP_PROC_BX_SKIP(BX_IS_BX0 | BX_IS_BX_LIM);
    return;
  } else {
    DP_PROC_BX_SKIP(BX_IS_BX0);
    DP_PROC_BX_UPDATE();
    while (true) {
      if (bx == bx_lim) {
	DP_PROC_BX_SKIP(BX_IS_BX_LIM);
	return;
      }
      DP_PROC_BX_SKIP(BX_IS_GENERIC);
      DP_PROC_BX_UPDATE();
    }
  }
#undef DP_PROC_BX_SKIP
#undef DP_PROC_BX_UPDATE 
}
TEMPLATE
inline void
CLASS::
dpProcBXInit()
{
  bx_lim = (Task::insideType ? bt->bx : bt->ex);
  by0    = (Task::insideType ? bt->ey : bt->by);
  by_lim = (Task::insideType ? bt->exby : bt->bxey);
}
TEMPLATE
inline void
CLASS::
dpProcBXUpdate()
{
  const TZState& Bt = stateFlags.Bt;

  (Task::insideType ? bx-- : bx++);
  by0 += (Task::insideType ? -Bt.dR : Bt.dL);
  by_lim += (Task::insideType ? -Bt.dL : Bt.dR);
  bn += (Task::insideType ? -Bt.bottomRightMargin : Bt.topLeftMargin);
}
TEMPLATE
inline void
CLASS::
dpProcBXSkip()
{
#define DP_PROC_BY(bx_is_on_skip0) do {\
  stateFlags.resetBXK(bx_is_on_skip0);\
  dpProcBY();\
} while (0)

  if ((bxn_beg < 0) || (--bxr > 0)) {// bxn_beg < 0 should be in BX0
    DP_PROC_BY(BX_IS_NOT_ON_SKIP);
  } else {
    bxr = skipSize;
    (Task::insideType ? bxn-- : bxn++);
    (Task::insideType ? pSkipCountB0-- : pSkipCountB0++);

    Assert(skipSet.isSkipRow(vx.getPos(bx)));// bx is at a skipRow
    DP_PROC_BY(BX_IS_ON_SKIP);
  }
#undef DP_PROC_BY
}
TEMPLATE
inline void
CLASS::
dpProcBY()
{
  const StateFlags& T = stateFlags;
  const TZState& Bt =  stateFlags.Bt;

//#define DP_PROC_BY_SKIP_0(prevFlag0)
//  dpProcBYInit< BY<T, BY_IS_BY0, prevFlag0> >();
//  dpProcBYSkip< BY<T, BY_IS_BY0, prevFlag0> >()
#define DP_PROC_BY_SKIP_0(prevFlag0) do {\
  stateFlags.resetBY(BY_IS_BY0, PrevFlag(prevFlag0));\
  dpProcBYSkip();\
} while (0)
#define DP_PROC_BY_SKIP_1(prevFlag0) do {\
  stateFlags.resetBY(BY_IS_NOT_BY0, PrevFlag(prevFlag0));\
  dpProcBYSkip();\
} while (0)
#define DP_PROC_BY_UPDATE() dpProcBYUpdate()

  dpProcBYInit();
  if (T.bx_is_bx0) {
    Assert(Task::insideType ? (bx == bt->ex) : (bt->bx == bx));
    if (T.bt_is_bt0) {
      DP_PROC_BY_SKIP_0(FLAG_NULL);
      if (by == by_lim) return;

      DP_PROC_BY_UPDATE();
      while (true) {
	DP_PROC_BY_SKIP_1(FLAG_PREV_IL);
	if (by == by_lim) return;

	DP_PROC_BY_UPDATE();
      }
    } else {
      const int& havePrevML =
	(Task::insideType ? (by < (bt+1)->bxey) : ((bt-1)->exby) < by);
      const int& havePrevDL =
	(Task::insideType ? ((bt+1)->by <= by) : (by <= (bt-1)->ey));
      if (havePrevML) {// have lower right (upper left) cell
	if (havePrevDL) {// have bottom (top) cell
	  DP_PROC_BY_SKIP_0(FLAG_PREV_ML | FLAG_PREV_DL);
	} else {// only have lower right (upper left) cell
	  Assert(by == (Task::insideType
			? ((bt+1)->by - 1) : ((bt-1)->ey + 1)));
	  DP_PROC_BY_SKIP_0(FLAG_PREV_ML);
	}
      } else {// only have bottom (top) cell
	Assert(by == (Task::insideType ? (bt+1)->bxey : (bt-1)->exby));
	DP_PROC_BY_SKIP_0(FLAG_PREV_DL);
      }

      if (havePrevDL) {
	int count_by = 
	  (Task::insideType ? (by - (bt+1)->by) : ((bt-1)->ey - by));
	while (count_by-- > 0) {
	  DP_PROC_BY_UPDATE();
	  DP_PROC_BY_SKIP_1((FLAG_PREV_ML | FLAG_PREV_IL | FLAG_PREV_DL));
	}
	if (by == by_lim) return;

	DP_PROC_BY_UPDATE();
	DP_PROC_BY_SKIP_1((FLAG_PREV_ML | FLAG_PREV_IL));
      }
      if (by == by_lim) return;

      DP_PROC_BY_UPDATE();
      while (true) {
	DP_PROC_BY_SKIP_1(FLAG_PREV_IL);
	if (by == by_lim) return;

	DP_PROC_BY_UPDATE();
      }	
    }
  } else {// ! T.bx_is_bx0
    Assert(Task::insideType ? (bx < bt->ex) : (bt->bx < bx));

    if (by == by_lim) {// by_lim == by0
      const int& havePrevML = 
	(Task::insideType ? Bt.isRD : Bt.isLD);
      const int& havePrevDL =
	(Task::insideType ? (!Bt.isLD) : (!Bt.isRD));
      if (havePrevML) {
	if (havePrevDL)  {
	  DP_PROC_BY_SKIP_0((FLAG_PREV_ML | FLAG_PREV_DL));
	} else {
	  DP_PROC_BY_SKIP_0(FLAG_PREV_ML);
	}
      } else {// havePrevDL == true
	DP_PROC_BY_SKIP_0(FLAG_PREV_DL);
      }
    } else {// by != by_lim
      const int& havePrevML =
	(Task::insideType ? Bt.isRD : Bt.isLD);
      if (havePrevML) {
	DP_PROC_BY_SKIP_0((FLAG_PREV_ML | FLAG_PREV_DL));
      } else {
	DP_PROC_BY_SKIP_0(FLAG_PREV_DL);
      }
      //if (by == by_lim) return; <- not possible
      DP_PROC_BY_UPDATE();
      while (by != by_lim) {
	DP_PROC_BY_SKIP_1((FLAG_PREV_ML | FLAG_PREV_IL | FLAG_PREV_DL));
	DP_PROC_BY_UPDATE();
      }
      const int& havePrevDL =
	(Task::insideType ? (!Bt.isLD) : (!Bt.isRD));
      if (havePrevDL)  {
	DP_PROC_BY_SKIP_1((FLAG_PREV_ML | FLAG_PREV_IL | FLAG_PREV_DL));
      } else {
	DP_PROC_BY_SKIP_1((FLAG_PREV_ML | FLAG_PREV_IL));
      }
    }
  }
#undef DP_PROC_BY_SKIP
}
TEMPLATE
inline void
CLASS::
dpProcBYInit()
{
  by = by0;
}
TEMPLATE
inline void
CLASS::
dpProcBYUpdate()
{
  (Task::insideType ? by-- : by++);
  (Task::insideType ? bn-- : bn++);
}
TEMPLATE
inline void
CLASS::
dpProcBYSkip()
{
#define DP_PROC_ET(b_is_on_skip0) do {\
  stateFlags.resetBYK(b_is_on_skip0);\
  dpProcET();\
} while (0)

  const StateFlags& T = stateFlags;

  if (!T.bx_is_on_skip) {
    DP_PROC_ET(B_IS_NOT_ON_SKIP);
  } else {
    if (Task::insideType) {
      // ref(max(bx, ybx), y)
      if (by->bxn >= 0) {
	by->pSkipCountB--;
	by->pSkipShiftB--;
	by->pSkipCountE--;
	by->pSkipShiftE--;
      }
    } else {//outsideType
      if (by->bxn >= 0) {
	by->pSkipCountB++;
	by->pSkipShiftB++;
      }
    }

    // use more simple condition
    if (bn->st[0] || bn->st[1] || bn->st[2] || bn->st[3]) {
      DP_PROC_ET(B_IS_ON_SKIP);
    } else {
      DP_PROC_ET(B_IS_NOT_ON_SKIP);
    }
  }
  if (Task::storeE()) {
    dpProcBYSkipStoreE();
  }
#undef DP_PROC_ET
}
TEMPLATE
inline void
CLASS::
dpProcBYSkipStoreE()
{
  //const BY& T = BY0;

  if (Task::insideType) {
    ScoreManager::SCNode* sc1 = curr_sc - (et->ey - max(by, et->exby));

#ifdef DEBUG
    Rect rect = getCurrRect();
    int ey1 = max(region.sre.getBY(rect.ex), rect.by);
    {
      Assert(rect.ex == region.sre.ex0);
      Assert(rect.ey == region.sre.ey0);
      Assert(curr_sc == &task.vsc_ref(rect.by - rect.bx, rect.ex, rect.ey));
      Assert(sc1 == &task.vsc_ref(rect.by - rect.bx, rect.ex, ey1));
      Assert(sce_store == &task.vsfbwE_refI(rect.bx, rect.by, ey1));
    }
#endif

    while (sc1 <= curr_sc) {

#ifdef DEBUG
      Assert(ey1 <= rect.ey);
      Assert(sc1 == &task.vsc_ref(rect.by - rect.bx, rect.ex, ey1));
      Assert(sce_store == &task.vsfbwE_refI(rect.bx, rect.by, ey1));
      ey1++;
#endif

      *sce_store++ = *sc1++;
    }
  } else {// outsideType
    ScoreManager::SCNode* sc1 = curr_sc + (et->bxey - max(by, et->by));

#ifdef DEBUG
    Rect rect = getCurrRect();
    int ey1 = region.sre.getEY(rect.ex);
    {
      Assert(rect.ex == region.sre.bx0);
      Assert(rect.ey == max(region.sre.by0, rect.by));
      Assert(curr_sc == &task.vsc_ref(rect.by - rect.bx, rect.ex, rect.ey));
      Assert(sc1 == &task.vsc_ref(rect.by - rect.bx, rect.ex, ey1));
      Assert(sce_store == &task.vsfbwE_refO(rect.bx, rect.by, ey1));
    }
#endif

    while (sc1 >= curr_sc) {

#ifdef DEBUG
      Assert(rect.ey <= ey1);
      Assert(sc1 == &task.vsc_ref(rect.by - rect.bx, rect.ex, ey1));
      Assert(sce_store == &task.vsfbwE_refO(rect.bx, rect.by, ey1));
      ey1--;
#endif

      *sce_store-- = *sc1--;
    }
  }
}
TEMPLATE
inline void
CLASS::
dpProcET()
{
#define CHOOSE_DP_PROC_EX(et_is_et0, et_cmp_byext0) do {\
  stateFlags.resetET(et_is_et0, EtCmpByext(et_cmp_byext0));\
  chooseDPProcEX();\
} while (0)
#define DP_PROC_ET_INIT()   dpProcETInit()
#define DP_PROC_ET_UPDATE() dpProcETUpDate()

  TZNode*const& byext = by->ext;
  if (Task::isTriShape()
      || (Task::isRectIn() && (et0 <= byext))
      || (Task::isRectOut() && (et_lim <= byext))) {
    enum EtPhase {
      PHASE1 = (Task::insideType ? ET_LT_BYEXT : ET_GT_BYEXT),
      PHASE2 = (Task::insideType ? ET_GT_BYEXT : ET_LT_BYEXT)
    };
    TZNode*const& et0_1 = (Task::isTriIn() ? bt : et0);
    if (et0_1 == byext) {
      DP_PROC_ET_INIT();
      CHOOSE_DP_PROC_EX(ET_IS_ET0, ET_IS_BYEXT);
      if (et == et_lim) return;

      DP_PROC_ET_UPDATE();
    } else {
      DP_PROC_ET_INIT();
      CHOOSE_DP_PROC_EX(ET_IS_ET0, PHASE1);
      DP_PROC_ET_UPDATE();
      while (et != byext) {
	CHOOSE_DP_PROC_EX(ET_IS_NOT_ET0, PHASE1);
	DP_PROC_ET_UPDATE();
      }
      CHOOSE_DP_PROC_EX(ET_IS_NOT_ET0, ET_IS_BYEXT);
      if (et == et_lim) return;

      DP_PROC_ET_UPDATE();
    }
    while (true) {
      CHOOSE_DP_PROC_EX(ET_IS_NOT_ET0, PHASE2);
      if (et == et_lim) return;

      DP_PROC_ET_UPDATE();
    }
  } else {// byext < et0, et_lim
    DP_PROC_ET_INIT();
    CHOOSE_DP_PROC_EX(ET_IS_ET0, ET_GT_BYEXT);
    if (et == et_lim) return;
	
    DP_PROC_ET_UPDATE();
    while (true) {
      CHOOSE_DP_PROC_EX(ET_IS_NOT_ET0, ET_GT_BYEXT);
      if (et == et_lim) return;
	
      DP_PROC_ET_UPDATE();
    }
  }
#undef CHOOSE_DP_PROC_EX
#undef DP_PROC_ET_INIT
#undef DP_PROC_ET_UPDATE
}
TEMPLATE
inline void
CLASS::
dpProcETInit()
{
  const StateFlags& T = stateFlags;

  if (Task::isTriOut()) {
    et_lim = bt;
  }
  et = (Task::isTriIn() ? bt : et0);

  dpProcETInitSCMP();
  if (Task::isTriIn()) {
    ex = bx;
  } else {
    ex = (Task::insideType ? et->bx : et->ex);
  }
  pairX = bx->pn_beg;
  prev_pairX = ((bx+1)->pn_beg - (Task::isTriIn() ? 1 : 0));

  if (Task::isTriIn()) {
    ey0 = by;
    en = bn;
  } else if (Task::isRectIn()) {
    YNode* const& etby = et->by;
    if (etby < by) {
      ey0 = by;
      en = (en0 + (by - etby));
    } else {
      ey0 = etby;
      en = en0;
    }
  } else {
    ey0 = (Task::insideType ? et->by : et->ey);
    en = en0;
  }

  pairY = by->pn_beg;
  prev_pairY = ((by+1)->pn_beg 
		- ((Task::insideType && by == ey0) ? 1 : 0));

  curr_sc = bn->sc;

  prevDR = NULL;

  if (T.hasPrevIL) {
    prevIL = (curr_sc 
	      + (Task::insideType ? dpAreaMargin : -dpAreaMargin));
  } else {
    prevIL = NULL;
  }
  if (T.hasPrevDL) {
    prevDL = (curr_sc 
	      + (Task::insideType ? -dpAreaMargin : +dpAreaMargin));
  } else {
    prevDL = NULL;
  }
  prev_sc.setNull();

  if (T.bx_is_on_skip) {
    Assert(!Task::isTriIn() || exr0 == 1);
    exr = exr0;
    if (Task::isTriIn()) {
      exn = (bxn - 1);
      pSkipCountE0 = (pSkipCountB0 - 1);
    } else if (Task::isRectIn()) {
      exn = (exn_beg - 1);
      pSkipCountE0 -= (exn_beg >= 0 ? (exn_last - exn_beg + 1) : 0);
    } else {
      exn = (exn_last + 1);
      pSkipCountE0 = pSkipCountE00;
    }
    dpProcETInitSkipCountE();
  }
}
TEMPLATE
inline void
CLASS::
dpProcETUpDate()
{
  const StateFlags& T = stateFlags;

  (Task::insideType ? et++ : et--);
  if (Task::isTriOut() && (et == et_lim)) {
    ex_lim = bx;
  } else {
    ex_lim = (Task::insideType ? et->ex : et->bx);
  }
  (Task::insideType ? ex++ : ex--);
  (Task::insideType ? pairX++ : pairX--);
  (Task::insideType ? prev_pairX++ : prev_pairX--);

  ey0 = (Task::insideType ? max(by, et->by) : et->ey);

  const int& dey = (ey0 - ey_lim);
  pairY += dey;
  prev_pairY += dey;
  prevDR = curr_sc + dey;

  BENode* en_old = en;// instanciate value
  if (Task::insideType) {
    en += et->topLeftMargin;
    if (T.et_lt_byext) {// old et
      en += (by - et->by);
    }
  } else {
    if (T.et_le_byext) {
      en -= (by - (et+1)->by);
    }
    en -= et->bottomRightMargin;
  }
  const int& den = (en - en_old);
  curr_sc += den;
  if (T.hasPrevIL) {
    prevIL += den;
  }
  if (T.hasPrevDL) {
    prevDL += den;
  }
  if (Task::insideType
      && T.hasPrevML
      && T.et_ge_byext
      && (et-1)->exby < ey0) {
    prev_sc = (ey0-1)->scMP;
  } else if (Task::outsideType
	     && T.hasPrevML
	     && ey0 < (et+1)->bxey) {
    prev_sc = (ey0+1)->scMP;
  } else {
    prev_sc.setNull();
  }
}
TEMPLATE
inline void
CLASS::
dpProcETInitSCMP()
{
  YNode* const& ey1 = (Task::insideType ? et->by : et_lim->by);
  YNode* const& ey2 = (Task::insideType ? et_lim->ey : et->ey);
  ey = ey1;
  int count1 = (ey2 - ey1 + 1);
  while (count1-- > 0) {
    (ey++)->scMP.setNull();
  }
}
TEMPLATE
inline void
CLASS::
dpProcETInitSkipCountE()
{
  if (Task::insideType) {

    // set (by..et_lim->ey)->pSkipCountE
    int const& nB = (by->exn >= 0 ? (by->exn - bxn + 1) : 0);
    int*const& pCntB = by->pSkipCountB; //ref(bx, by)
    int*const& pCntB_end = (pCntB + nB);//ref(byexk+1,by)

#ifdef DEBUG
    Rect rect = getCurrRect();
    {
      int byexk = region.getEXK(rect.by);
      if (byexk >= 0) {
	Assert(pCntB == &task.vskip_skipCountB_ref(rect.bx, rect.by));
	Assert(pCntB_end == &task.vskip_skipCountB_ref(byexk, rect.by)+1);
      } else {
	Assert(pCntB == pCntB_end);
      }
    }
#endif
	   
    for (YNode* ey1 = max(by, et0->by); ey1 <= et_lim->ey; ey1++) {
      const int& nE = ((ey1->exn >= 0)
		       ? (ey1->exn - max(bxn, ey1->bxn) + 1)
		       : 0);
      int* pCntE1 = ey1->pSkipCountE;           //ref(max(bx,eybxk),ey)
      int* pSftE1 = ey1->pSkipShiftE;           //ref(max(bx,eybxk),ey)
                                                //ref(eyexk+1, ey)
      int*const& pCntE1_end = (pCntE1 + nE);
                        
      // ref(max(bx,eybxk),ey) includes by->exn < 0 or ey1->bxn < 0 case 
      // since bxn >= 0
      // simplify this expression using bybxn, byexn, ey1bxn, ey1exn etc.
      const int& nB1 = ((nB > 0 && nE > 0)
			? max(0, (by->exn - max(bxn, ey1->bxn) + 1))
			: 0);
      int* pCntB1 = (pCntB_end - nB1);

#ifdef DEBUG
      int dxk = skipSet.getSkipRowPeriod();
      int y1 = vy.getPos(ey1);
      int eybxk = region.getBXK(y1);
      int eyexk = region.getEXK(y1);
      int x1 = (eybxk >= 0 ? max(rect.bx, eybxk) : rect.bx);
      int bx_last = region.srb.ex0;
      int x1_end = (eyexk >= 0 ? (eyexk + dxk) : rect.bx);
      {
	if (eybxk >= 0) {
	  Assert(pCntE1 == &task.vskip_skipCountE_ref(x1, y1));
	  Assert(pSftE1 == &task.vskip_skipShiftE_ref(x1, y1));
	  Assert(pCntE1_end == &task.vskip_skipCountE_ref(eyexk, y1) + 1);
	  if (nB1 > 0) {
	    Assert(pCntB1 == &task.vskip_skipCountB_ref(x1, rect.by));
	  }
	} else {
	  Assert(pCntE1 == pCntE1_end);
	  Assert(pCntB1 == pCntB_end);
	}
      }
#endif

      while (pCntB1 < pCntB_end) {
#ifdef DEBUG
	{
	  Assert(pCntE1 == &task.vskip_skipCountE_ref(x1, y1));
	  Assert(pSftE1 == &task.vskip_skipShiftE_ref(x1, y1));
	  Assert(pCntB1 == &task.vskip_skipCountB_ref(x1, rect.by));

	  int exey = ((x1 <= bx_last) 
		      ? region.srb.getEY(x1)
		      : region.sre.getEY(x1));
	  int cntB = skipSet.getNSkipPointY(x1, rect.by, exey);
	  int sftE = skipSet.getNSkipPointY(x1, y1 + 1, exey);
	  int ly = skipSet.getNSkipPointY(x1, rect.by, y1);
	  Assert(*pCntB1 == cntB);
	  Assert(*pSftE1 == sftE);
	  Assert(ly == ((*pCntB1) - (*pSftE1)));
	  if (Task::isRectIn() && x1 <= bx_last) {
	    x1 += dxk;
	    if (bx_last < x1) {
	      int exk_beg = region.srke.getFirstSkipRow();
	      if (exk_beg >= 0) {
		x1 = exk_beg;
	      }
	    }
	  } else {
	    x1 += dxk;
	  }
	}
#endif
	*pCntE1++ = ((*pCntB1++) - (*pSftE1++)); //subtract margin region
      }

      int* pCnt01 = (pSkipCountB0 
		     + (ey1->exn - bxn + 1) 
		     - (pCntE1_end - pCntE1));

      while (pCntE1 < pCntE1_end) {
#ifdef DEBUG
	{
	  Assert(pCntE1 == &task.vskip_skipCountE_ref(x1, y1));
	  Assert(pSftE1 == &task.vskip_skipShiftE_ref(x1, y1));
	  Assert(pCnt01 == &task.vskip_skipCount0_ref(x1));

	  int exey = ((x1 <= bx_last) 
		      ? region.srb.getEY(x1)
		      : region.sre.getEY(x1));
	  int exby = ((x1 <= bx_last)
		      ? region.srb.getBY(x1)
		      : region.sre.getBY(x1));
	  int cntB = skipSet.getNSkipPointY(x1, exby, exey);
	  int sftE = skipSet.getNSkipPointY(x1, y1 + 1, exey);
	  int ly = skipSet.getNSkipPointY(x1, exby, y1);
	  Assert(cntB - sftE == ly);
	  Assert(*pCnt01 == cntB);
	  Assert(*pSftE1 == sftE);
	  Assert(ly == ((*pCnt01) - (*pSftE1)));
	  if (Task::isRectIn() && x1 <= bx_last) {
	    x1 += dxk;
	    if (bx_last < x1) {
	      int exk_beg = region.exk_beg;
	      if (exk_beg >= 0) {
		x1 = exk_beg;
	      }
	    }
	  } else {
	    x1 += dxk;
	  }
	}
#endif
	*pCntE1++ = ((*pCnt01++) - (*pSftE1++));
      }

#ifdef DEBUG
      Assert(x1 == x1_end);
#endif
    }

  } else {//outside

#ifdef DEBUG
    Rect rect = getCurrRect();
    int dxk = skipSet.getSkipRowPeriod();
    int bx_last = region.srb.ex0;
#endif

    for (YNode* ey1 = max(by, et_lim->by); ey1 <= et0->ey; ey1++) {
      if (ey1->exn < 0) continue;

      const int& n1 = (ey1->pSkipCountE0 - ey1->pSkipCountE);
      ey1->pSkipCountE += n1;
      ey1->pSkipShiftE += n1;

#ifdef DEBUG
      int y1 = vy.getPos(ey1);
      int eyexk = region.getEXK(y1);
      Assert(eyexk >= 0);
      if (region.exk_beg <= eyexk) {
	Assert(ey1->pSkipCountE == 
	       &task.vskip_skipCountE_ref(eyexk, y1) + 1);
	Assert(ey1->pSkipShiftE == 
	       &task.vskip_skipShiftE_ref(eyexk, y1) + 1);
      }
      int exk = region.convertXN2XK(exn - 1);
      if (exk >= 0) {
	Assert(pSkipCountE0 == &task.vskip_skipCount0_ref(exk) + 1);
      } else {
	Assert(exn_beg < 0);
	Assert(pSkipCountE0 == NULL);
      }
#endif

      const int& ey1exn = ey1->exn;
      const int& ey1bxn = ey1->bxn;
      const int& byexn = by->exn;
	
      int nE0 = 0;
      int nE = 0;
      int nB = 0;
      if (byexn >= 0) {
	if (byexn < ey1bxn) {
	  nE0 = (ey1exn - ey1bxn + 1);
	} else {
	  nE0 = (ey1exn - byexn);
	    
	  if (bxn <= ey1bxn) {
	    nE = (byexn - ey1bxn + 1);
	  } else {
	    nE = (byexn - bxn + 1);
	    nB = (bxn - ey1bxn);
	  }
	}
      }

      int* pCntE0 = (pSkipCountE0 + (ey1exn - exn));
      int* pCntE  = (ey1->pSkipCountE - 1);
      int* pSftE  = (ey1->pSkipShiftE - 1);
      int* pCntE1 = (by->pSkipCountE - 1);
      int* pCntB  = (by->pSkipCountB - 1);
      int* pCntB1 = (ey1->pSkipCountB - (ey1 == by ? 1 : 0));

#ifdef DEBUG
      int ey1exk = region.getEXK(y1);
      int ey1bxk = region.getBXK(y1);
      int byexk = region.getEXK(rect.by);
      int bxk = rect.bx;
      Assert(ey1bxk >= 0 && ey1bxk >= 0);
      int xk = ey1exk;
      if (nE0 > 0) {
	Assert(pCntE0 == &task.vskip_skipCount0_ref(ey1exk));
	Assert(pCntE == &task.vskip_skipCountE_ref(ey1exk, y1));
	Assert(pSftE == &task.vskip_skipShiftE_ref(ey1exk, y1));
      } else {

      }
#endif
      while (nE0-- > 0) {
#ifdef DEBUG
	Assert(pCntE0 == &task.vskip_skipCount0_ref(xk));
	Assert(pCntE == &task.vskip_skipCountE_ref(xk, y1));
	Assert(pSftE == &task.vskip_skipShiftE_ref(xk, y1));
	int xkey = ((xk <= bx_last) 
		    ? region.srb.getEY(xk)
		    : region.sre.getEY(xk));
	int xkby = max(rect.by,
		       ((xk <= bx_last) 
			? region.srb.getBY(xk)
			: region.sre.getBY(xk)));
	int ly = skipSet.getNSkipPointY(xk, xkby, xkey);
	int cntE = skipSet.getNSkipPointY(xk, y1, xkey);
	int sftE = skipSet.getNSkipPointY(xk, xkby, y1 - 1);
	Assert(*pCntE0 == ly);
	Assert(*pCntE == cntE);
	Assert(sftE == ly - cntE);

	if (Task::isRectShape() 
	    && (0 <= region.exk_beg && region.exk_beg <= xk)) {
	  xk -= dxk;
	  if (xk < region.exk_beg) {
	    xk = region.bxk_last;
	  }
	} else {
	  xk -= dxk;
	}
#endif
	(*pSftE--) = ((*pCntE0--) - (*pCntE--));
      }
#ifdef DEBUG
      if (nE > 0) {
	Assert(xk == byexk);
	Assert(pCntE1 == &task.vskip_skipCountE_ref(xk, rect.by));
	Assert(pCntE == &task.vskip_skipCountE_ref(xk, y1));
	Assert(pSftE == &task.vskip_skipShiftE_ref(xk, y1));
      }
#endif
      while (nE-- > 0) {
#ifdef DEBUG
	Assert(pCntE1 == &task.vskip_skipCountE_ref(xk, rect.by));
	Assert(pCntE == &task.vskip_skipCountE_ref(xk, y1));
	Assert(pSftE == &task.vskip_skipShiftE_ref(xk, y1));
	int xkey = ((xk <= bx_last) 
		    ? region.srb.getEY(xk)
		    : region.sre.getEY(xk));
	int xkby = max(rect.by,
		       ((xk <= bx_last) 
			? region.srb.getBY(xk)
			: region.sre.getBY(xk)));
	int ly = skipSet.getNSkipPointY(xk, xkby, xkey);
	int cntE = skipSet.getNSkipPointY(xk, y1, xkey);
	int sftE = skipSet.getNSkipPointY(xk, rect.by, y1 - 1);
	Assert(*pCntE1 == ly);
	Assert(*pCntE == cntE);
	Assert(sftE == ly - cntE);

	if (Task::isRectShape()
	    && (0 <= region.exk_beg && region.exk_beg <= xk)) {
	  xk -= dxk;
	  if (xk < region.exk_beg) {
	    xk = region.bxk_last;
	  }
	} else {
	  xk -= dxk;
	}
#endif
	(*pSftE--) = ((*pCntE1--) - (*pCntE--));
      }
#ifdef DEBUG
      if (nB > 0) {
	Assert(xk + dxk == bxk
	       || (xk == region.bxk_last && xk == region.exk_beg));
	Assert(pCntB == &task.vskip_skipCountB_ref(xk, rect.by));
	Assert(pCntB1 == &task.vskip_skipCountB_ref(xk, y1));
	Assert(pCntE == &task.vskip_skipCountE_ref(xk, y1));
	Assert(pSftE == &task.vskip_skipShiftE_ref(xk, y1));
      }
#endif
      while (nB-- > 0) {
#ifdef DEBUG
	Assert(pCntB == &task.vskip_skipCountB_ref(xk, rect.by));
	Assert(pCntB1 == &task.vskip_skipCountB_ref(xk, y1));
	//Assert(pCntE == &task.vskip_skipCountE_ref(xk, y1));
	Assert(pSftE == &task.vskip_skipShiftE_ref(xk, y1));

	int xkby = ((xk <= bx_last) 
		    ? region.srb.getBY(xk)
		    : region.sre.getBY(xk));
	int xkey = y1;
	int ly = skipSet.getNSkipPointY(xk, xkby, xkey);
	int cntB = skipSet.getNSkipPointY(xk, xkby, rect.by);
	int sftE = skipSet.getNSkipPointY(xk, rect.by + 1, y1);
	Assert(*pCntB1 == ly);
	Assert(*pCntB == cntB);
	Assert(sftE == ly - cntB);
	if (Task::isRectShape()
	    && (0 <= region.exk_beg && region.exk_beg <= xk)) {
	  xk -= dxk;
	  if (xk < region.exk_beg) {
	    xk = region.bxk_last;
	  }
	} else {
	  xk -= dxk;
	}
#endif
	(*pSftE--) = ((*pCntB1--) - (*pCntB--));
      }
#ifdef DEBUG
      Assert((xk + dxk == ey1bxk) 
	     || (xk == region.bxk_last && ey1bxk == region.exk_beg));
#endif
    }
  }
}
TEMPLATE
inline void
CLASS::
chooseDPProcEX()
{
#define DP_PROC_EX(tzType0) do {\
  stateFlags.Et.reset(tzType0);\
  dpProcEX();\
} while (0)

  switch (et->type) {
  case Trapezoid::LV_RV: DP_PROC_EX(Trapezoid::LV_RV); return;
  case Trapezoid::LV_RD: DP_PROC_EX(Trapezoid::LV_RD); return;
  case Trapezoid::LD_RV: DP_PROC_EX(Trapezoid::LD_RV); return;
  case Trapezoid::LD_RD: DP_PROC_EX(Trapezoid::LD_RD); return;
  default: Die("bad tzType"); return;
  }

#undef DP_PROC_EX
}
TEMPLATE
inline void
CLASS::
dpProcEX()
{
  // warning: dpProcEXInit uses current stateFlags.Et
  // Et should be set before this function
#define DP_PROC_EX_INIT() dpProcEXInit()
#define DP_PROC_EX_UPDATE() dpProcEXUpdate()
#define DP_PROC_EX_SKIP(ex_is_ex0, ex_cmp_byex0) do {\
  stateFlags.resetEX(ex_is_ex0, ExCmpByex(ex_cmp_byex0));\
  dpProcEXSkip();\
} while (0)

  const StateFlags& T = stateFlags;

  if (T.et_is_byext) {
    enum ExPhase {
      PHASE1 = (Task::insideType ? EX_LT_BYEX : EX_GT_BYEX),
      PHASE2 = (Task::insideType ? EX_GT_BYEX : EX_LT_BYEX)
    };
    XNode* const& byex = by->ex;
    if (ex == byex) {
      DP_PROC_EX_INIT();
      DP_PROC_EX_SKIP(EX_IS_EX0, EX_IS_BYEX);
      if (ex == ex_lim) return;

      DP_PROC_EX_UPDATE();
    } else {
      DP_PROC_EX_INIT();
      DP_PROC_EX_SKIP(EX_IS_EX0, PHASE1);
      DP_PROC_EX_UPDATE();
      while (ex != byex) {
	DP_PROC_EX_SKIP(EX_IS_NOT_EX0, PHASE1);
	DP_PROC_EX_UPDATE();
      }
      DP_PROC_EX_SKIP(EX_IS_NOT_EX0, EX_IS_BYEX);
      if (ex == ex_lim) return;

      DP_PROC_EX_UPDATE();
    }
    while (true) {
      DP_PROC_EX_SKIP(EX_IS_NOT_EX0, PHASE2);
      if (ex == ex_lim) return;

      DP_PROC_EX_UPDATE();
    }
  } else {
    const int& EX_CMP_BYEX = (T.et_lt_byext ? EX_LT_BYEX : EX_GT_BYEX);
    DP_PROC_EX_INIT();
    DP_PROC_EX_SKIP(EX_IS_EX0, EX_CMP_BYEX);
    if (ex == ex_lim) return;

    DP_PROC_EX_UPDATE();
    while (true) {
      DP_PROC_EX_SKIP(EX_IS_NOT_EX0, EX_CMP_BYEX);
      if (ex == ex_lim) return;

      DP_PROC_EX_UPDATE();
    }
  }
#undef DP_PROC_EX_SKIP
#undef DP_PROC_EX_UPDATE
#undef DP_PROC_EX_SKIP
}
TEMPLATE
inline void
CLASS::
dpProcEXInit()
{// use current Et
  const StateFlags& T = stateFlags;
  const TZState& Et = stateFlags.Et;

  if (Task::isTriOut() && (et == et_lim)) {
    ex_lim = bx;
  } else {
    ex_lim = (Task::insideType ? et->ex : et->bx);
  }
  if (Task::insideType) {
    ey_lim = et->bxey;
    if (Task::isTriIn() && T.et_is_et0 && Et.isRD) {
      ey_lim += (bx - et->bx);
    }
  } else {
    ey_lim = (T.et_le_byext ? max(by, et->exby) : et->exby);
  }
  if (T.et_le_byext) {
    if (Task::insideType) {
      if (Task::isTriShape() && T.et_is_et0) {
	l_margin = ((by - by_lim) - Et.dL + Et.topLeftMargin);
      } else {
	l_margin = ((by - et->by) - Et.dL + Et.topLeftMargin);
      }
    } else {
      l_margin = (max(0, (int)(by - et->exby)) + Et.bottomRightMargin);
    }
  }
}
TEMPLATE
inline void
CLASS::
dpProcEXUpdate()
{
  const StateFlags& T = stateFlags;
  const TZState& Et = stateFlags.Et;

  (Task::insideType ? ex++ : ex--);
  (Task::insideType ? pairX++ : pairX--);
  (Task::insideType ? prev_pairX++ : prev_pairX--);

  if (Task::insideType && Et.isLD && T.ex_ge_byex) {
    ey0++;
  } else if (Task::outsideType && Et.isRD) {
    ey0--;
  }

  const int& dey = (ey0 - ey_lim);
  pairY += dey;
  prev_pairY += dey;
  prevDR = curr_sc + dey;

  if (Task::insideType && Et.isRD) {
    ey_lim++;
  } else if (Task::outsideType && Et.isLD && T.ex_gt_byex) {
    ey_lim--;
  }
  const int& den = 
    (Task::insideType
     ? (T.ex_lt_byex ? l_margin : Et.topLeftMargin)
     : (-(T.ex_le_byex ? l_margin : Et.bottomRightMargin)));

  en += den;
  curr_sc += den;
  if (T.hasPrevIL) {
    prevIL += den;
  }
  if (T.hasPrevDL) {
    prevDL += den;
  }
  if (Task::insideType && T.hasPrevML && Et.isLD && T.ex_ge_byex) {
    prev_sc = (ey0-1)->scMP;
  } else if (Task::outsideType && T.hasPrevML && Et.isRD) {
    prev_sc = (ey0+1)->scMP;
  } else {
    prev_sc.setNull();
  }
  if (Task::insideType && Et.isLD && T.ex_lt_byex) {
    l_margin--;
  } else if (Task::outsideType && Et.isLD && T.ex_le_byex) {
    l_margin++;
  }
}
TEMPLATE
inline void
CLASS::
dpProcEXSkip()
{
#define DP_PROC_EY(ex_is_on_skip0) do {\
  stateFlags.resetEXK(ex_is_on_skip0);\
  dpProcEY();\
} while (0)

  const StateFlags& T = stateFlags;
  // (exn_beg should be in EX0 T.have_no_exn
  if ((exn_beg < 0) || !T.bx_is_on_skip || (--exr > 0)) {
    DP_PROC_EY(EX_IS_NOT_ON_SKIP);
  } else {
    exr = skipSize;
    (Task::insideType ? exn++ : exn--);
    (Task::insideType ? pSkipCountE0++ : pSkipCountE0--);
    Assert(skipSet.isSkipRow(vx.getPos(ex)));// bx is at a skipRow
    DP_PROC_EY(EX_IS_ON_SKIP);
  }
#undef DP_PROC_EY
}
TEMPLATE
inline void
CLASS::
dpProcEY()
{
#define DP_PROC_EY_SKIP(ey_is_ey0) do {\
  stateFlags.resetEY(ey_is_ey0);\
  dpProcEYSkip();\
} while (0)
#define DP_PROC_EY_UPDATE() dpProcEYUpdate()

  //dpProcEYInit< EY<EXK0, EY_IS_EY0> >();
  dpProcEYInit();
  DP_PROC_EY_SKIP(EY_IS_EY0);
  if (ey == ey_lim) return;

  DP_PROC_EY_UPDATE();
  while (true) {
    DP_PROC_EY_SKIP(EY_IS_NOT_EY0);
    if (ey == ey_lim) return;

    DP_PROC_EY_UPDATE();
  }
#undef DP_PROC_EY_SKIP
#undef DP_PROC_EY_UPDATE
}
TEMPLATE
inline void
CLASS::
dpProcEYInit()
{
  ey = ey0;
}
TEMPLATE
inline void
CLASS::
dpProcEYUpdate()
{
  const StateFlags& T = stateFlags;

  (Task::insideType ? ey++ : ey--);
  (Task::insideType ? pairY++ : pairY--);
  (Task::insideType ? prev_pairY++ : prev_pairY--);
  (Task::insideType ? en++ : en--);
  (Task::insideType ? curr_sc++ : curr_sc--);
  if (T.hasPrevIL) {
    (Task::insideType ? prevIL++ : prevIL--);
  }
  if (T.hasPrevDL) {
    (Task::insideType ? prevDL++ : prevDL--);
  }
  if (!T.ex_is_etex0) {
    (Task::insideType ? prevDR++ : prevDR--);
  }
}
TEMPLATE
inline void
CLASS::
dpProcEYSkip()
{
#define DP_PROC(rect_is_on_skip0) do {\
  stateFlags.resetEYK(rect_is_on_skip0);\
  dpProc();\
} while (0)

 const StateFlags& T = stateFlags; 

  if (!T.ex_is_on_skip) {
    DP_PROC(RECT_IS_NOT_ON_SKIP);
  } else {
    if (Task::outsideType) {
      if (ey->exn >= 0) {
	ey->pSkipCountE--;
	ey->pSkipShiftE--;
      }
    }
    if (T.b_is_on_skip
	&& (en->st[0] || en->st[1] || en->st[2] || en->st[3])) {
      // simplify this
      DP_PROC(RECT_IS_ON_SKIP);
    } else {
      DP_PROC(RECT_IS_NOT_ON_SKIP);
    }
  }
#undef DP_PROC
}
TEMPLATE
inline void
CLASS::
dpProc()
{
#define DP_PROC_STATE(s) dpProcState(s)
#define TEST_DP_PROC_INITIAL_STATE() do {\
  if (debug) {testDPProcInitialState(); Break::Here1();}\
} while (0)
#define TEST_DP_PROC_FINAL_STATE() do {\
  if (debug) {testDPProcFinalState();}\
} while (0)

//#define DP_PROC_EMISSION_STATES() do {
//  const int& i_end = model.getNEmitState();
//  for (int i = 0; i < i_end; i++) {
//    DP_PROC_STATE(model.getEmitState(i));
//  }
//} while (0)
#define DP_PROC_EMISSION_STATES() do {\
  int n = model.getNEmitState();\
  int s = model.getEmitState(0);\
  while (n-- > 0) {\
    DP_PROC_STATE(State(s++));\
  }\
} while (0)
  const StateFlags& T = stateFlags;

#ifdef SLR_STATE

#else
  // inside: bifurcation parents, outside: bifurcation children
  const State& s00 = 
    (Task::insideType ? Model::BF_st : Model::ST_st);
  // inside: no emission parents, outside: no emission children
  const State& s01 =
    (Task::insideType ? Model::ST_st : Model::BF_st);
#endif

  diag_sc = prev_sc;
  prev_sc = ey->scMP;
  ey->scMP = *curr_sc;
  curr_sc->setNull();

#ifdef DEBUG
  TEST_DP_PROC_INITIAL_STATE();
#endif

  if (Task::isCYK()) {
    tau->setNull();
    if (T.rect_is_on_skip) {
      tau_BF->setNull();
    }
  }
  if (T.rect_is_on_skip) {
    dpProcInitST();
    dpProcBifurcate();
  } else {
#ifdef SLR_STATE
    if (Task::insideType) {
      (*curr_sc)[Model::BF_st] = NegInfDPScore;
    } else {
      (*curr_sc)[Model::SL_st] = NegInfDPScore;
      (*curr_sc)[Model::SR_st] = NegInfDPScore;
    }
#else
    (*curr_sc)[s00] = NegInfDPScore;
#endif
  }

  switch (T.getType()) {
  case BE_IS_BE0:
    if (Task::isTriIn()) {
      // assuming there is no BF->END transition
      const int& i_end = model.getNParent(Model::END_st);
      for (int i = 0; i < i_end; i++) {
	const State& s = model.getParent(Model::END_st, i);
	//(*curr_sc)[s] = ZeroDPScore;
	(*curr_sc)[s] = tsc(s, Model::END_st);
      }
//#if defined(USE_PSCFG_NUSSINOV)
//#ifdef SLR_STATE
//      DP_PROC_STATE(Model::SL_st);
//      DP_PROC_STATE(Model::SR_st);
//#else
//      DP_PROC_STATE(s01);
//#endif
//#endif
    } else if (Task::isRectIn()) {
      if (model.isNormalState(state0)) {
	(*curr_sc)[state0] = ZeroDPScore;
      }
#ifdef SLR_STATE
      DP_PROC_STATE(Model::SL_st);
      DP_PROC_STATE(Model::SR_st);
#else
      DP_PROC_STATE(s01);
#endif
    } else {// outside type
      if (state0 == Model::BEG_st) {
	// first set ST, SR, SL states
	const int& i_end = model.getNChild(Model::BEG_st);
	for (int i = 0; i < i_end; i++) {
	  const State& s = model.getChild(Model::BEG_st, i);
	  if (!model.isEmitState(s) && !model.bifurcates(s)) {
	    // only non emission *parent* needs this ST, SR, SL
	    (*curr_sc)[s] = ZeroDPScore;
	  }
	}
	// needed if ST, SR, SL exists
	//// account for ST->Emission States
	DP_PROC_EMISSION_STATES();
	//// account for ST->BF
#ifdef SLR_STATE
	DP_PROC_STATE(Model::BF_st);
#else
	DP_PROC_STATE(s01);
#endif	
	// then set other states
	for (int i = 0; i < i_end; i++) {
	  const State& s = model.getChild(Model::BEG_st, i);
	  if (!model.isEmitState(s) && !model.bifurcates(s)) continue;

	  DPScoreT& score0 = (*curr_sc)[s];
	  const DPScoreT& score1 = ZeroDPScore;
	  if (Task::takesMaxScore()) {
	    if (score0 < score1) {			     
	      score0 = score1;
	    }
	  } else {	
	    scoreManager.logAdd(score0, score1);		     
	  }
	}						     
      } else if (model.isNormalState(state0)) {
	(*curr_sc)[state0] = ZeroDPScore;
	if (!model.isEmitState(state0) && (state0 != Model::BF_st)) {
	  // only non emission *parent* needs this ST, SR, SL
	  DP_PROC_EMISSION_STATES();
	}
#ifdef SLR_STATE
	DP_PROC_STATE(Model::BF_st);
#else
	DP_PROC_STATE(s01);
#endif
      } else {
	Assert(state0 == Model::X_st);
	DP_PROC_EMISSION_STATES();
#ifdef SLR_STATE
	DP_PROC_STATE(Model::BF_st);
#else
	DP_PROC_STATE(s01);
#endif
      }
    }
    break;
  case EXY_IS_EXY0:
    if (Task::isTriIn()) {
      const int& i_end = model.getNParent(Model::END_st);
      for (int i = 0; i < i_end; i++) {
	const State& s = model.getParent(Model::END_st, i);
	//(*curr_sc)[s] = ZeroDPScore;
	(*curr_sc)[s] = tsc(s, Model::END_st);
      }
    } else {
      DP_PROC_EMISSION_STATES();
#ifdef SLR_STATE
      if (Task::insideType) {
	DP_PROC_STATE(Model::SL_st);
	DP_PROC_STATE(Model::SR_st);
      } else {
	DP_PROC_STATE(Model::BF_st);
      }
#else
      DP_PROC_STATE(s01);
#endif
    }
    break;
  case EXY_IS_GENERIC:
    DP_PROC_EMISSION_STATES();
#ifdef SLR_STATE
    if (Task::insideType) {
      DP_PROC_STATE(Model::SL_st);
      DP_PROC_STATE(Model::SR_st);
    } else {
      DP_PROC_STATE(Model::BF_st);
    }
#else
    DP_PROC_STATE(s01);
#endif
    break;
  default:
    Die("bad type=%d", T.getType());
    break;
  }

  if (T.rect_is_on_skip) {
    dpProcUpdateST();
  }
  if (Task::isCYK()) {
    tau++;
    if (T.rect_is_on_skip) {
      tau_BF++;
    }
  } else if (Task::isInside()) {
    *sce_store++ = *curr_sc;
  } else if (Task::isOutside()) {
    *sce_store-- = *curr_sc;
  }

  if (Task::storeB() && T.bx_is_btbx_lim) {
#ifdef DEBUG
    if (debug) {
    const Rect& rect = getCurrRect();
    Assert(rect.bx == 
	   (Task::insideType ? region.srb.bx0 : region.srb.ex0));
    Assert(scb_store == 
	   (Task::insideType
	    ? &task.vsfbwB_refI(rect.by, rect.ex, rect.ey)
	    : &task.vsfbwB_refO(rect.by, rect.ex, rect.ey)));
    }
#endif
    *scb_store = *curr_sc;
    (Task::insideType ? scb_store++ : scb_store--);
  }

#ifdef DEBUG
  Break::Here2();
  if (debug) {
    TEST_DP_PROC_FINAL_STATE();
  }
  dpCounter++;
  dpCounter1++;
#endif
#undef DP_PROC_STATE
#undef TEST_DP_PROC_INITIAL_STATE
#undef TEST_DP_PROC_FINAL_STATE
}
TEMPLATE
inline void
CLASS::
dpProcState(const State& s)
{
#define N_PREV_STATE(s) \
  (Task::insideType ? model.getNChild(s) : model.getNParent(s))
#define PREV_STATE(s, i) \
  (Task::insideType ? model.getChild(s, i) : model.getParent(s, i))
#define PREV_SCORE(s, s1) getPreviousScore(s, s1)
#define TSC(s, s1) tsc(s, s1)
#define ESC(s, s1) esc(s, s1)
#define SET_VAL(s, s1) setVal(s, s1)
#define GET_EMIT_STATE(s) model.toPSCFGBaseState(s)

#if 0
  const StateFlags& T = stateFlags;
  if (Task::insideType) {
    switch (GET_EMIT_STATE(s)) {
    case EmitModel::MP_st: case EmitModel::ML_st:
      if (!T.hasPrevML) {
	(*curr_sc)[s] = NegInfDPScore;
	return;
      }
      break;
    case EmitModel::IL_st:
      if (!T.hasPrevIL) {
	(*curr_sc)[s] = NegInfDPScore;
	return;
      }
      break;
    case EmitModel::DL_st:
      if (!T.hasPrevDL) {
	(*curr_sc)[s] = NegInfDPScore;
	return;
      }
      break;
    default:
      break;
    }
  }
#endif
                                                        
  //const State& s0 = PREV_STATE(s, 0);			     
  //const DPScoreT& score1 =
  //  (PREV_SCORE(s, s0) + TSC(s, s0)
  //   + (Task::insideType ? ZeroDPScore : ESC(s, s0)));
  //best_score = score1;
  //if (Task::isCYK()) {				     
  //  best_child_index = 0;					     
  //}
  best_score = NegInfDPScore;
  if (Task::isCYK()) {
    best_child_index = 0;
  }
  const int& i_end = N_PREV_STATE(s);
  //for (int i = 1; i < i_end; i++) {
  for (int i = 0; i < i_end; i++) {
    const State& s1 = PREV_STATE(s, i);
    //const int& n0 = (N_PREV_STATE(s) - 1);
    //int n = n0;
    //const State* ps1 = &PREV_STATE(s, 1);
    //while (n-- > 0) {
    //  const State& s1 = (*ps1++);
    const DPScoreT& score1 = 
      (PREV_SCORE(s, s1) + TSC(s, s1)
       + (Task::insideType ? ZeroDPScore : ESC(s, s1)));  
    if (Task::takesMaxScore()) {
      if (best_score < score1) {			     
	best_score = score1;
	if (Task::isCYK()) {
	  best_child_index = i;
	  //best_child_index = (n0 - n);
	}
      }
    } else {	
      scoreManager.logAdd(best_score, score1);		     
    }						     
  }
  if (Task::insideType) {
    best_score += ESC(s, Model::X_st);
  }					     
  if (IsFinite(best_score)) {			     
    (*curr_sc)[s] = best_score;
    if (Task::isCYK()) {
      tau->SET_VAL(s, best_child_index);
    }			     
  } else {						     
    (*curr_sc)[s] = NegInfDPScore;			     
  }

#undef N_PREV_STATE
#undef PREV_STATE
#undef PREV_SCORE
#undef TSC
#undef ESC
#undef SET_VAL
#undef GET_EMIT_STATE
}
TEMPLATE
inline DPDeckPSCFG::DPScoreT
CLASS::
getPreviousScore(const State& s, const State& s1) const
{
#define GET_EMIT_STATE(s) model.toPSCFGBaseState(s)

  const StateFlags& T = stateFlags;
  enum {in = Task::insideType};
  const State& sp = (in ? s : s1);
  //const State& sc = (in ? s1 : s);
  const EmitState& sp_es = GET_EMIT_STATE(sp);
  //const EmitState& sc_es = GET_EMIT_STATE(sc);
  switch (sp_es) {
  case EmitModel::ST_st: 
    return (*curr_sc)[s1];
  case EmitModel::MP_st: 
    return (T.hasPrevML ? diag_sc[s1] : NegInfDPScore);
  case EmitModel::MR_st: 
    return (prevDR ? (*(prevDR+(in ? (-1) : (+1))))[s1] : NegInfDPScore);
  case EmitModel::IR_st: 
    return (*(curr_sc + (in ? (-1) : (+1))))[s1];
  case EmitModel::DR_st: 
    return (prevDR ? (*prevDR)[s1] : NegInfDPScore);
  case EmitModel::ML_st: 
    return (T.hasPrevML ? (ey->scMP)[s1] : NegInfDPScore);
  case EmitModel::IL_st: 
    return (T.hasPrevIL ? (*prevIL)[s1] : NegInfDPScore);
  case EmitModel::DL_st: 
    return (T.hasPrevDL ? (*prevDL)[s1] : NegInfDPScore);
  default:
    Assert(false, "bad type");
    return NegInfDPScore;
  }
#undef GET_EMIT_STATE
}
TEMPLATE
inline DPDeckPSCFG::DPScoreT
CLASS::
esc(const State& s, const State& s1) const
{
  enum {in = Task::insideType};
  const State& sp = (in ? s : s1);

#if 0
#define ESC(result,s,xColB,yColB,xColE,yColE,xPair,yPair,mColB,mColE,escTable) \
  PSCFG_ENERGY_SCORE_MANAGER_ESC(result,s,xColB,yColB,xColE,yColE,xPair,yPair,mColB,mColE,escTable)
#else
#define ESC(result,s,xColB,yColB,xColE,yColE,xPair,yPair,mColB,mColE,escTable) \
  (result = scoreManager.esc(s,xColB,yColB,xColE,yColE,xPair,yPair,mColB,mColE,escTable))
#endif
  //const Rect& d = model.getNEmit(sp);

  // only used by (d.bx > 0)
  //XNode*const& bx1 = (bx - (in ? 0 : d.bx));
  XNode*const& bx1 = (bx - (in ? 0 : 1));

  //only used by (d.ex > 0)
  //YNode*const& by1 = (by - (in ? 0 : d.by));
  YNode*const& by1 = (by - (in ? 0 : 1));

  //only used by (d.ex > 0)
  //XNode*const& ex1 = (ex - 1 + (in ? 0 : d.ex));
  XNode*const& ex1 = (ex - 1 + (in ? 0 : 1));

  //only used by (d.ey > 0)
  //YNode*const& ey1 = (ey - 1 + (in ? 0 : d.ey));
  YNode*const& ey1 = (ey - 1 + (in ? 0 : 1));

  //only used by ({d.bx, d.ex} = {1, 1})
  //PairNode*const& pairX1 = (in 
  //			    ? pairX 
  //			    : ((d.bx ? pairX : prev_pairX) - 1 + d.ex));

  //only used by ({d.by, d.ey} = {1, 1})
  //PairNode*const& pairY1 = (in 
  //			    ? pairY 
  //			    : ((d.by ? pairY : prev_pairY) - 1 + d.ey));


#ifdef DEBUG
  if (debug) {
  const Rect& d1 = model.getNEmit(sp);
  Rect rect = getCurrRect();
  Rect rectp(Rect::null());
  bool doTest = false;
  if (Task::insideType) {
    //Assert(pairX == &vpairX.ref(rect.bx, rect.ex - 1));
    //Assert(prev_pairX == &vpairX.ref(rect.bx + 1, rect.ex - 1);
    //Assert(pairY == &vpairY.ref(rect.by, rect.ey - 1);
    //Assert(prev_pairY == &vpairY.ref(rect.by + 1, rect.ey - 1);
    rectp = rect;
    doTest = true;
  } else {
    //Assert(pairX == &vpairX.ref(rect.bx - 1, rect.ex));
    //Assert(prev_pairX == &vpairX.ref(rect.bx, rect.ex);
    //Assert(pairY == &vpairY.ref(rect.by - 1, rect.ey));
    //Assert(prev_pairY == &vpairY.ref(rect.by, rect.ey));
    Rect rect1 = model.subtractNEmit(rect, s1);
    if (region.includes(rect1)) {
      rectp = rect1;
      doTest = true;
    }
  }
  if (doTest) {
    if (d1.bx > 0) {
      Assert(bx1 == &task.vx.refB(rectp.bx));
    }
    if (d1.by > 0) {
      Assert(by1 == &task.vy.refB(rectp.by));
    }
    if (d1.ex > 0) {
      Assert(ex1 == &task.vx.refE(rectp.ex - 1));
    }
    if (d1.ey > 0) {
      Assert(ey1 == &task.vy.refE(rectp.ey - 1));
    }
    if (d1.bx > 0 && d1.ex > 0) {
      Assert(pairX == &task.vpairX.ref(rectp.bx, rectp.ex - 1));
    }
    if (d1.by > 0 && d1.ey > 0) {
      Assert(pairY == &task.vpairY.ref(rectp.by, rectp.ey - 1));
    }
    //this is wrong
    //Assert(bn == &task.vbe_ref(rectp.bx, rectp.by));
    //Assert(en == &task.vbe_ref(rectp.ex - 1, rectp.ey - 1));
    //
    // correct statement is 
    // if (d.bx > 0 && d.by > 0) {
    //   Assert(bn->mColB == getMatchCol(rectp.bx, rectp.by);
    // }
    // if (d.ex > 0 && d.ey > 0) {
    //   Assert(en->mColE == getMatchCol(rectp.ex-1, rectp.ey - 1));
    // }
  }
  }
#endif

  DPScoreT result;
  ESC(result, sp, 
      bx1->xCol, by1->yCol, ex1->xCol, ey1->yCol,
      pairX->pairCol, pairY->pairCol,
      bn->mColB, en->mColE,
      escTable);
  return result;
#undef ESC
}
TEMPLATE
inline DPDeckPSCFG::DPScoreT
CLASS::
tsc(const State& s, const State& s1) const
{
#if 1
#define TSC(result,sp,sc,xColB,yColB,xColE,yColE,xPair,yPair,xPairR,yPairR,xPairL,yPairL,tscTable) \
  PSCFG_ENERGY_SCORE_MANAGER_TSC(result,sp,sc,xColB,yColB,xColE,yColE,xPair,yPair,xPairR,yPairR,xPairL,yPairL,tscTable)
#else
#define TSC(result,sp,sc,xColB,yColB,xColE,yColE,xPair,yPair,xPairR,yPairR,xPairL,yPairL,tscTable) \
  (result = scoreManager.tsc(result,sp,sc,xColB,yColB,xColE,yColE,xPair,yPair,xPairR,yPairR,xPairL,yPairL,tscTable))
#endif
 

#ifdef DEBUG
  enum {in = Task::insideType};
  const State& sp = (in ? s : s1);
  const State& sc = (in ? s1 : s);

  const Rect& d = model.getNEmit(sp);

  XNode*const& bx1 = (bx + (in ? d.bx : 0));
  YNode*const& by1 = (by + (in ? d.by : 0));
  XNode*const& ex1 = (ex - (in ? d.ex : 0));
  YNode*const& ey1 = (ey - (in ? d.ey : 0));

  PairNode*const& pairX1 = (in 
			    ? pairX 
			    : ((d.bx ? pairX : prev_pairX) - 1 + d.ex));
  PairNode*const& pairY1 = (in 
			    ? pairY 
			    : ((d.by ? pairY : prev_pairY) - 1 + d.ey));
#endif
#ifdef DEBUG
  //used only if ({d.bx,d.ex} = {0,1})
  PairNode*const& pairXR = ((in ? pairX : prev_pairX) - 1);
  //used only if ({d.bx,d.ex} = {1,0})
  PairNode*const& pairYR = ((in ? pairY : prev_pairY) - 1);
  //used only if ({d.by,d.ey} = {0,1}
  PairNode*const& pairXL = (prev_pairX - (in ? 0 : 1));
  //used only if ({d.by,d.ey} = {1,0})
  PairNode*const& pairYL = (prev_pairY - (in ? 0 : 1));
#endif

#ifdef DEBUG
  if (debug) {
  Rect rect = getCurrRect();
  Rect rectp(Rect::null());
  Rect rectc(Rect::null());
  bool doTest = false;
  if (Task::insideType) {
    //Assert(pairX == &vpairX.ref(rect.bx, rect.ex - 1));
    //Assert(prev_pairX == &vpairX.ref(rect.bx + 1, rect.ex - 1);
    //Assert(pairY == &vpairY.ref(rect.by, rect.ey - 1);
    //Assert(prev_pairY == &vpairY.ref(rect.by + 1, rect.ey - 1);
    Rect rect1 = model.addNEmit(rect, s);
    if (region.includes(rect1)) {
      rectp = rect;
      rectc = rect1;
      doTest = true;
    }
  } else {
    //Assert(pairX == &vpairX.ref(rect.bx - 1, rect.ex));
    //Assert(prev_pairX == &vpairX.ref(rect.bx, rect.ex);
    //Assert(pairY == &vpairY.ref(rect.by - 1, rect.ey));
    //Assert(prev_pairY == &vpairY.ref(rect.by, rect.ey));
    Rect rect1 = model.subtractNEmit(rect, s1);
    if (region.includes(rect1)) {
      rectp = rect1;
      rectc = rect;
      doTest = true;
    }
  }
  if (doTest) {
    Assert(bx1 == &task.vx.refB(rectc.bx));
    Assert(by1 == &task.vy.refB(rectc.by));
    Assert(ex1 == &task.vx.refE(rectc.ex));
    Assert(ey1 == &task.vy.refE(rectc.ey));
    Assert(pairX1 == &task.vpairX.ref(rectp.bx, rectp.ex - 1));
    Assert(pairY1 == &task.vpairY.ref(rectp.by, rectp.ey - 1));
    if (d.bx == 0 && d.ex > 0) {
      Assert(pairXR == &task.vpairX.ref(rectp.bx, rectp.ex - 2));
    }
    if (d.by == 0 && d.ey > 0) {
      Assert(pairYR == &task.vpairY.ref(rectp.by, rectp.ey - 2));
    }
    if (d.bx > 0 && d.ex == 0) {
      Assert(pairXL == &task.vpairX.ref(rectp.bx + 1, rectp.ex - 1));
    }
    if (d.by > 0 && d.ey == 0) {
      Assert(pairYL == &task.vpairY.ref(rectp.by + 1, rectp.ey - 1));
    }
  }
  }
#endif

  DPScoreT result;
  TSC(result, sp, sc,
      bx1->xCol, by1->yCol,
      ex1->xCol, ey1->yCol,
      pairX1->pairCol, pairY1->pairCol,
      pairXR->pairCol, pairYR->pairCol,
      pairXL->pairCol, pairYL->pairCol,
      tscTable);
  return result;
#undef TSC
}
TEMPLATE
inline void
CLASS::
dpProcInitST()
{
  *((Task::insideType ? en : bn)->st[0]) = NegInfDPScore;
  *((Task::insideType ? bn : en)->st[1]) = NegInfDPScore;
}
TEMPLATE
inline void
CLASS::
dpProcUpdateST()
{// inside: bifurcation children, outside: bifurcation parents
#ifdef SLR_STATE
  if (Task::insideType) {
    const DPScoreT& z0 = (*curr_sc)[Model::SR_st];
    const DPScoreT& z1 = (*curr_sc)[Model::SL_st];
    *(en->st[0])-- = z0;
    *(bn->st[1])++ = z1;
  } else {
     const DPScoreT& z = (*curr_sc)[Model::BF_st];
    *(bn->st[0])-- = z;
    *(en->st[1])++ = z;
  }
#else
  enum {s01 = (Task::insideType ? Model::ST_st : Model::BF_st)};
  const DPScoreT& z = (*curr_sc)[s01];
  *((Task::insideType ? en : bn)->st[0])-- = z;
  *((Task::insideType ? bn : en)->st[1])++ = z;
#endif
}
TEMPLATE
inline void
CLASS::
dpProcBifurcate()
{
#ifdef DEBUG
  if (debug) {
  testDPProcBifurcateInitialState();
  }
#endif

#ifdef SLR_STATE
  if (Task::insideType) {
    best_score = NegInfDPScore;
  } else {
    best_score_L = NegInfDPScore;
    best_score_R = NegInfDPScore;
  }
#else
  best_score = NegInfDPScore;
#endif

  if (Task::isCYK()) {
    best_split_offset = (-1);
    best_bfType = INVALID_BF_TYPE;
  }
  
  dpProcBifurcateBFType0();
  if (Task::hasBFType1()) {
    dpProcBifurcateBFType1();
  }
  if (Task::hasBFType2()) {
    dpProcBifurcateBFType2();
  }

#ifdef SLR_STATE
  if (Task::insideType) {
    best_score += esc(Model::BF_st, Model::SL_st);
    if (IsFinite(best_score)) {
      (*curr_sc)[Model::BF_st] = best_score;
      if (Task::isCYK()) {
	Assert(best_split_offset > 0);
	tau_BF->setVal(best_bfType, best_split_offset);
      }
    } else {
      (*curr_sc)[Model::BF_st] = NegInfDPScore;
    }
  } else {//outside type
    //SL_st
    best_score_L += esc(Model::BF_st, Model::SL_st);
    if (IsFinite(best_score_L)) {
      (*curr_sc)[Model::SL_st] = best_score_L;
    } else {
      (*curr_sc)[Model::SL_st] = NegInfDPScore;
    }
    //SR_st
    // esc should be equal to SL case
    best_score_R += esc(Model::BF_st, Model::SR_st);
    if (IsFinite(best_score_R)) {
      (*curr_sc)[Model::SR_st] = best_score_R;
    } else {
      (*curr_sc)[Model::SR_st] = NegInfDPScore;
    }
  }
#else
  enum {
    s00 = (Task::insideType ? Model::BF_st : Model::ST_st),
    s01 = (Task::insideType ? Model::ST_st : Model::BF_st)
  };
  best_score += esc(State(s00), State(s01));
  if (IsFinite(best_score)) {
    (*curr_sc)[s00] = best_score;
    if (Task::isCYK()) {
      Assert(best_split_offset > 0);
      tau_BF->setVal(best_bfType, best_split_offset);
    }
  } else {
    (*curr_sc)[s00] = NegInfDPScore;
  }
#endif
#ifdef DEBUG
  if (debug) {
  testDPProcBifurcateFinalState();
  }
#endif
}
TEMPLATE
inline void
CLASS::
dpProcBifurcateBFType0()
{
#ifdef DEBUG
  if (debug) {
    //testDPProcBifurcateLoopInitialState<BF_TYPE_0>();
    testDPProcBifurcateLoopInitialState(BF_TYPE_0);
  }
  Rect rect = getCurrRect();
  //int dxk = skipSet.getSkipRowPeriod();
#endif

  best_score1 = NegInfDPScore;
  if (Task::isCYK()) {
    best_split_ptr_plus_1 = NULL;
  }

  if (Task::insideType) {
    if (Task::isTriIn()) {
      scVSTRect1 = bn->st[2];
      scVSTRect2 = en->st[0];
    } else if (Task::isRectIn()) {
      scVSTRect1 = bn->st[0];
      scVSTRect2 = en->st[0];
    }
    scVSTRect20 = scVSTRect2;

    h_bxn = bxn;
    h_exn = min(exn, bxn_last);
    //h_byexn = max(min(by->exn, h_exn), h_bxn - 1);
    h_byexn = min(by->exn, h_exn);
    h_eybxn = min(max(ey->bxn, h_bxn), h_exn + 1);
    h_intsec = (h_byexn - h_eybxn + 1);
    pSkipCount0 = pSkipCountB0;
    pSkipCountL = by->pSkipCountB;
    pSkipShiftL = by->pSkipShiftB;
    pSkipCountR = (ey->pSkipCountE + (h_eybxn - max(ey->bxn, bxn)));
    pSkipShiftR = (ey->pSkipShiftE + (h_eybxn - max(ey->bxn, bxn)));

#ifdef DEBUG
    if (debug) {
      Assert(scVSTRect1 == &task.vst_rect1_ref(BF_TYPE_0,
					       rect.bx, rect.by,
					       rect.bx, rect.by));
      Assert(scVSTRect2 == &task.vst_rect2_ref(BF_TYPE_0,
					       rect.bx, rect.by,
					       rect.ex, rect.ey));
      int h_eybxk = region.convertXN2XK(h_eybxn);
      int h_bxk = region.convertXN2XK(h_bxn);
      Assert(pSkipCount0 == &task.vskip_skipCount0_ref(h_bxk));
      Assert(pSkipCountL == &task.vskip_skipCountB_ref(h_bxk, rect.by));
      Assert(pSkipShiftL == &task.vskip_skipShiftB_ref(h_bxk, rect.by));
      if (h_eybxn < h_exn + 1) {
	Assert(pSkipCountR == 
	       &task.vskip_skipCountE_ref(h_eybxk, rect.ey));
	Assert(pSkipShiftR == 
	       &task.vskip_skipShiftE_ref(h_eybxk, rect.ey));
      }
    }
#endif

    if (h_intsec >= 0) {
#if defined(USE_TEMPLATE_BF)
      dpProcBifurcateHexagon<FORWARD, L2_R1, HAS_INTSEC, BF_TYPE_0>();
#else
      dpProcBifurcateHexagon(FORWARD, L2_R1, HAS_INTSEC, BF_TYPE_0);
#endif
    } else {
#if defined(USE_TEMPLATE_BF)
      dpProcBifurcateHexagon<FORWARD, L2_R1, NO_INTSEC, BF_TYPE_0>();
#else
      dpProcBifurcateHexagon(FORWARD, L2_R1, NO_INTSEC, BF_TYPE_0);
#endif
    }
    if (Task::takesMaxScore()) {
      if (best_score < best_score1) {
	best_score = best_score1;
	if (Task::isCYK()) {
	  best_split_offset = (best_split_ptr_plus_1 - scVSTRect20);
	  best_bfType = BF_TYPE_0;
	}
      }
    } else {
      scoreManager.logAdd(best_score, best_score1);
    }
  } else {// outsideType
    if (Task::isTriOut()) {
      scVSTRect1 = bn->st[0];
      scVSTRect2 = en->st[2];
    } else if (Task::isRectOut()) {
      scVSTRect1  = bn->st[0];
      scVSTRect2  = en->st[0];
    }
    scVSTRect20 = scVSTRect2;

    h_bxn = exn;
    h_exn = exn_last;
    //h_byexn = max(min(ey->exn, h_exn), h_bxn - 1);
    h_byexn = min(ey->exn, h_exn);
    //h_eybxn = min(max((ey_last+1)->bxn, h_bxn), h_exn + 1);
    h_eybxn = (h_exn + 1);
    h_intsec = (h_byexn - h_eybxn + 1);
    pSkipCount0 = pSkipCountE0;
    pSkipCountL = ey->pSkipCountE;
    pSkipShiftL = ey->pSkipShiftE;
    //pSkipCountR = NULL;
    //pSkipShiftR = NULL;

#ifdef DEBUG
    if (debug) {
      Assert(scVSTRect1 == &task.vst_rect1_ref(BF_TYPE_0,
					       rect.bx, rect.by,
					       rect.ex, rect.ey));
      Assert(scVSTRect2 == &task.vst_rect2_ref(BF_TYPE_0,
					       rect.ex, rect.ey,
					       rect.ex, rect.ey));
      int h_bxk = region.convertXN2XK(h_bxn);
      //int h_exk = region.convertXN2XK(h_exn);
      //int h_byexk = region.convertXN2XK(h_byexn);
      //int h_eybxk = region.convertXN2XK(h_eybxn);
      Assert(pSkipCount0 == &task.vskip_skipCount0_ref(h_bxk));
      Assert(pSkipCountL == &task.vskip_skipCountE_ref(h_bxk, rect.ey));
      Assert(pSkipShiftL == &task.vskip_skipShiftE_ref(h_bxk, rect.ey));
      //Assert(pSkipCountR == &task.vskip_skipCountE_ref());
      //Assert(pSkipShiftR == &task.vskip_skipShiftE_ref());
    }
#endif

#if defined(USE_TEMPLATE_BF)
    dpProcBifurcateHexagon<FORWARD, L1_R2, NO_INTSEC, BF_TYPE_0>();
#else
    dpProcBifurcateHexagon(FORWARD, L1_R2, NO_INTSEC, BF_TYPE_0);
#endif

#ifdef SLR_STATE
    if (Task::takesMaxScore()) {
      if (best_score_L < best_score1) {
	best_score_L = best_score1;
      }
    } else {
      scoreManager.logAdd(best_score_L, best_score1);
    }
#else
    if (Task::takesMaxScore()) {
      if (best_score < best_score1) {
	best_score = best_score1;
      }
    } else {
      scoreManager.logAdd(best_score, best_score1);
    }
#endif
  } 

#ifdef DEBUG
  if (debug) {
    //testDPProcBifurcateLoopFinalState<BF_TYPE_0>();
    testDPProcBifurcateLoopFinalState(BF_TYPE_0);
  }
#endif
}
TEMPLATE
inline void
CLASS::
dpProcBifurcateBFType1()
{
#ifdef DEBUG
  if (debug) {
    //testDPProcBifurcateLoopInitialState<BF_TYPE_1>();
    testDPProcBifurcateLoopInitialState(BF_TYPE_1);
  }
  Rect rect = getCurrRect();
  //int dxk = skipSet.getSkipRowPeriod();
#endif

  best_score1 = NegInfDPScore;
  if (Task::isCYK()) {
    best_split_ptr_plus_1 = NULL;
  }

  if (Task::insideType) {
    if (Task::isRectIn()) {
      scVSTRect1 = bn->st[1];
      scVSTRect2 = en->st[1];
    }
    scVSTRect20 = scVSTRect2;

    h_bxn = max(bxn, exn_beg);
    h_exn = exn;
    h_byexn = max(min(by->exn, h_exn), h_bxn - 1);
    //h_eybxn = min(max(ey->bxn, h_bxn), h_exn + 1);
    h_eybxn = max(ey->bxn, h_bxn);
    h_intsec = (h_byexn - h_eybxn + 1);
    pSkipCount0 = (pSkipCountE0 + (h_exn - exn));
    pSkipCountL = (by->pSkipCountB + (h_byexn - bxn));
    pSkipShiftL = (by->pSkipShiftB + (h_byexn - bxn));
    pSkipCountR = (ey->pSkipCountE + (h_exn - max(ey->bxn, bxn)));
    pSkipShiftR = (ey->pSkipShiftE + (h_exn - max(ey->bxn, bxn)));

#ifdef DEBUG
    if (debug) {
      Assert(scVSTRect1 == &task.vst_rect1_ref(BF_TYPE_1,
					       rect.bx, rect.by,
					       rect.ex, rect.ey));
      Assert(scVSTRect2 == &task.vst_rect2_ref(BF_TYPE_1,
					       rect.ex, rect.ey,
					       rect.ex, rect.ey));
      int h_byexk = region.convertXN2XK(h_byexn);
      int h_exk = region.convertXN2XK(h_exn);
      Assert(pSkipCount0 == &task.vskip_skipCount0_ref(h_exk));
      if (h_bxn <= h_byexn) {
	Assert(pSkipCountL == 
	       &task.vskip_skipCountB_ref(h_byexk, rect.by));
	Assert(pSkipShiftL == 
	       &task.vskip_skipShiftB_ref(h_byexk, rect.by));
      }
      Assert(pSkipCountR == &task.vskip_skipCountE_ref(h_exk, rect.ey));
      Assert(pSkipShiftR == &task.vskip_skipShiftE_ref(h_exk, rect.ey));
    }
#endif

    if (h_intsec >= 0) {
#if defined(USE_TEMPLATE_BF)
      dpProcBifurcateHexagon<BACKWARD, L2_R1, HAS_INTSEC, BF_TYPE_1>();
#else
      dpProcBifurcateHexagon(BACKWARD, L2_R1, HAS_INTSEC, BF_TYPE_1);
#endif
    } else {
#if defined(USE_TEMPLATE_BF)
      dpProcBifurcateHexagon<BACKWARD, L2_R1, NO_INTSEC, BF_TYPE_1>();
#else
      dpProcBifurcateHexagon(BACKWARD, L2_R1, NO_INTSEC, BF_TYPE_1);
#endif
    }

    if (Task::takesMaxScore()) {
      if (best_score < best_score1) {
	best_score = best_score1;
	if (Task::isCYK()) {
	  Assert((scVSTRect20 - best_split_ptr_plus_1) > 0);
	  best_split_offset = (scVSTRect20 - best_split_ptr_plus_1);
	  best_bfType = BF_TYPE_1;
	}
      }
    } else {
      scoreManager.logAdd(best_score, best_score1);
    }
  } else {// outsideType

    if (Task::isTriOut()) {
      scVSTRect1 = en->st[1];
      scVSTRect2 = bn->st[3];
    } else if (Task::isRectOut()) {
      scVSTRect1 = en->st[1];
      scVSTRect2 = bn->st[1];
    }
    scVSTRect20 = scVSTRect2;

    h_bxn = (min(ey->bxn, bxn) + 1);
    h_exn = bxn;
    //h_byexn = max(min((by_beg-1)->exn, h_exn), h_bxn - 1);
    h_byexn = (h_bxn - 1);
    //h_eybxn = min(max(by->bxn, h_bxn), h_exn + 1);
    h_eybxn = h_bxn;
    h_intsec = (h_byexn - h_eybxn + 1);
    pSkipCount0 = pSkipCountB0;
    //pSkipCountL = NULL;
    //pSkipShiftL = NULL;
    pSkipCountR = by->pSkipCountB;
    pSkipShiftR = (ey->pSkipShiftE - (exn - bxn));

#ifdef DEBUG
    if (debug) {
      Assert(scVSTRect1 == &task.vst_rect1_ref(BF_TYPE_1,
					       rect.bx,rect.by,
					       rect.ex,rect.ey));
      Assert(scVSTRect2 == &task.vst_rect2_ref(BF_TYPE_1,
					       rect.bx,rect.by,
					       rect.bx,rect.by));
      int h_bxk = region.convertXN2XK(h_bxn);
      int h_exk = region.convertXN2XK(h_exn);
      //int h_byexk = region.convertXN2XK(h_byexn);
      //int h_eybxk = region.convertXN2XK(h_eybxn);
      if (0 <= h_bxk && h_bxk <= h_exk) {
	Assert(pSkipCount0 == &task.vskip_skipCount0_ref(h_exk));
	Assert(pSkipCountR == &task.vskip_skipCountB_ref(h_exk, rect.by));
	Assert(pSkipShiftR == &task.vskip_skipShiftE_ref(h_exk, rect.ey));
      }
    }
#endif
#if defined(USE_TEMPLATE_BF)
    dpProcBifurcateHexagon<BACKWARD, L2_R1, NO_INTSEC, BF_TYPE_1>();
#else
    dpProcBifurcateHexagon(BACKWARD, L2_R1, NO_INTSEC, BF_TYPE_1);
#endif

    h_bxn = bxn_beg;
    h_exn = min(ey->bxn, bxn);
    //h_byexn = max(min((by_beg-1)->exn, h_exn), h_bxn - 1);
    h_byexn = (h_bxn - 1);
    h_eybxn = min(max(by->bxn, h_bxn), h_exn + 1);
    h_intsec = (h_byexn - h_eybxn + 1);
    pSkipCount0 = (pSkipCountB0 - (bxn - h_exn));
    //pSkipCountL = NULL;
    //pSkipShiftL = NULL;
    pSkipCountR = (by->pSkipCountB - (bxn - h_exn));
    pSkipShiftR = (by->pSkipShiftB - (bxn - h_exn));

#ifdef DEBUG
    if (debug) {
      int h_bxk = region.convertXN2XK(h_bxn);
      int h_exk = region.convertXN2XK(h_exn);
      //int h_byexk = region.convertXN2XK(h_byexn);
      //int h_eybxk = region.convertXN2XK(h_eybxn);
      if (0 <= h_bxk && h_bxk <= h_exk) {
	Assert(pSkipCount0 == &task.vskip_skipCount0_ref(h_exk));
	Assert(pSkipCountR == &task.vskip_skipCountB_ref(h_exk, rect.by));
	Assert(pSkipShiftR == &task.vskip_skipShiftB_ref(h_exk, rect.by));
      }
    }
#endif

#if defined(USE_TEMPLATE_BF)
    dpProcBifurcateHexagon<BACKWARD, L2_R1, NO_INTSEC, BF_TYPE_1>();
#else
    dpProcBifurcateHexagon(BACKWARD, L2_R1, NO_INTSEC, BF_TYPE_1);
#endif
#ifdef SLR_STATE
    if (Task::takesMaxScore()) {
      if (best_score_R < best_score1) {
	best_score_R = best_score1;
      }
    } else {
      scoreManager.logAdd(best_score_R, best_score1);
    }
#else
    if (Task::takesMaxScore()) {
      if (best_score < best_score1) {
	best_score = best_score1;
      }
    } else {
      scoreManager.logAdd(best_score, best_score1);
    }
#endif
  }
#ifdef DEBUG
  if (debug) {
    //testDPProcBifurcateLoopFinalState<BF_TYPE_1>();
    testDPProcBifurcateLoopFinalState(BF_TYPE_1);
  }
#endif
}
TEMPLATE
inline void
CLASS::
dpProcBifurcateBFType2()
{
#ifdef DEBUG
  if (debug) {
    //testDPProcBifurcateLoopInitialState<BF_TYPE_2>();
    testDPProcBifurcateLoopInitialState(BF_TYPE_2);
  }
  Rect rectkm = region.srkm.getRectK();
  Rect rect = getCurrRect();
  //int dxk = skipSet.getSkipRowPeriod();
#endif

  if (mxn_beg >= 0) {
    best_score1 = NegInfDPScore;
    if (Task::isCYK()) {
      best_split_ptr_plus_1 = NULL;
    }

    if (Task::insideType) {
      Assert(Task::isRegionType(DPRegion::NOSE));
      scVSTRect1 = bn->st[2];
      scVSTRect2 = en->st[2];
      scVSTRect20 = scVSTRect2;

      h_bxn = mxn_beg;
      h_exn = mxn_last;
      //h_byexn = max(min(by->exn, h_exn), h_bxn - 1);
      h_byexn = (h_bxn - 1);
      //h_eybxn = min(max(ey->bxn, h_bxn), h_exn + 1);
      h_eybxn = (h_exn + 1);
      h_intsec = (h_byexn - h_eybxn + 1);
      pSkipCount0 = pSkipCountM0;
      //pSkipCountL = (by->pSkipCountB + (h_bxn - bxn));
      //pSkipShiftL = (by->pSkipShiftB + (h_bxn - bxn));
      //pSkipCountR = (ey->pSkipCountE + (h_eybxn - max(ey->bxn, bxn)));
      //pSkipCountR = (ey->pSkipShiftE + (h_eybxn - max(ey->bxn, bxn)));

#ifdef DEBUG
      if (debug) {
	if (rectkm.isNormal()) {
	  Assert(scVSTRect1 == &task.vst_rect1_ref(BF_TYPE_2,
						   rect.bx, rect.by,
						   rectkm.bx, rectkm.by));
	  Assert(scVSTRect2 == &task.vst_rect2_ref(BF_TYPE_2,
						   rectkm.bx, rectkm.by,
						   rect.ex, rect.ey));
	}
	if (mxn_beg >= 0) {
	  int h_bxk = region.convertXN2XK(h_bxn);
	  //int h_exk = region.convertXN2XK(h_exn);
	  //int h_byexk = region.convertXN2XK(h_byexn);
	  //int h_eybxk = region.convertXN2XK(h_eybxn);
	  Assert(pSkipCount0 == &task.vskip_skipCount0_ref(h_bxk));
	}
      }
#endif

      //if (h_intsec >= 0) {
      //  dpProcBifurcateHexagon<FORWARD, L2_R1, HAS_INTSEC>();
      //} else {
#if defined(USE_TEMPLATE_BF)
      dpProcBifurcateHexagon<FORWARD, L2_R1, NO_INTSEC, BF_TYPE_2>();
#else
      dpProcBifurcateHexagon(FORWARD, L2_R1, NO_INTSEC, BF_TYPE_2);
#endif
      //}

      if (Task::takesMaxScore()) {
	if (best_score < best_score1) {
	  best_score = best_score1;
	  if (Task::isCYK()) {
	    Assert((best_split_ptr_plus_1 - scVSTRect20) > 0);
	    best_split_offset = (best_split_ptr_plus_1 - scVSTRect20);
	    best_bfType = BF_TYPE_2;
	  }
	}
      } else {
	scoreManager.logAdd(best_score, best_score1);
      }
    } else {// outsideType
      if (Task::isRegionType(DPRegion::CARET_CUT_L)) {
	scVSTRect1 = bn->st[2];
	scVSTRect2 = en->st[2];
	scVSTRect20 = scVSTRect2;

	h_bxn = mxn_beg;
	h_exn = mxn_last;
	//h_byexn = max(min(ey->exn, h_exn), h_bxn - 1);
	h_byexn = (h_bxn - 1);
	//h_eybxn = min(max((my_last+1)->bxn, h_bxn), h_exn + 1);
	h_eybxn = (h_exn + 1);
	h_intsec = (h_byexn - h_eybxn + 1);
	pSkipCount0 = pSkipCountM0;
	//pSkipCountL = (ey->pSkipCountE + (h_bxn - exn));
	//pSkipShiftL = (ey->pSkipShiftE + (h_bxn - exn));
	//pSkipCountR = NULL;
	//pSkipShiftR = NULL;

#ifdef DEBUG
	if (debug) {
	  if (rectkm.isNormal()) {
	    Assert(scVSTRect1 == 
		   &task.vst_rect1_ref(BF_TYPE_2,
				       rect.bx, rect.by,
				       rectkm.bx, rectkm.by));
	    Assert(scVSTRect2 == 
		   &task.vst_rect2_ref(BF_TYPE_2,
				       rect.ex, rect.ey,
				       rectkm.bx, rectkm.by));
	  }
	  if (mxn_beg >= 0) {
	    int h_bxk = region.convertXN2XK(h_bxn);
	    //int h_exk = region.convertXN2XK(h_exn);
	    //int h_byexk = region.convertXN2XK(h_byexn);
	    //int h_eybxk = region.convertXN2XK(h_eybxn);
	    Assert(pSkipCount0 == &task.vskip_skipCount0_ref(h_bxk));
	  }
	}
#endif
#if defined(USE_TEMPLATE_BF)
	dpProcBifurcateHexagon<FORWARD, L1_R2, NO_INTSEC, BF_TYPE_2>();
#else
	dpProcBifurcateHexagon(FORWARD, L1_R2, NO_INTSEC, BF_TYPE_2);
#endif
#ifdef SLR_STATE
	if (Task::takesMaxScore()) {
	  if (best_score_L < best_score1) {
	    best_score_L = best_score1;
	  }
	} else {
	  scoreManager.logAdd(best_score_L, best_score1);
	}
#else
	if (Task::takesMaxScore()) {
	  if (best_score < best_score1) {
	    best_score = best_score1;
	  }
	} else {
	  scoreManager.logAdd(best_score, best_score1);
	}
#endif
      } else {//CARET_CUT_R

	Assert(Task::isRegionType(DPRegion::CARET_CUT_R));
	scVSTRect1 = en->st[2];
	scVSTRect2 = bn->st[2];
	scVSTRect20 = scVSTRect2;

	h_bxn = mxn_beg;
	h_exn = mxn_last;
	//h_byexn = max(min((my_beg-1)->exn, h_exn), h_bxn - 1);
	h_byexn = (h_bxn - 1);
	//h_eybxn = min(max(by->bxn, h_bxn), h_exn + 1);
	h_eybxn = (h_exn + 1);
	h_intsec = (h_byexn - h_eybxn + 1);
	pSkipCount0 = pSkipCountM0;
	//pSkipCountL = NULL;
	//pSkipShiftL = NULL;
	//pSkipCountR = (by->pSkipCountB + (h_exn - bxn));
	//pSkipShiftR = (by->pSkipShiftB + (h_exn - bxn));

#ifdef DEBUG
	if (debug) {
	  if (rectkm.isNormal()) {
	    Assert(scVSTRect1 == 
		   &task.vst_rect1_ref(BF_TYPE_2,
				       rectkm.bx, rectkm.by,
				       rect.ex, rect.ey));
	    Assert(scVSTRect2 == 
		   &task.vst_rect2_ref(BF_TYPE_2,
				       rectkm.bx, rectkm.by,
				       rect.bx, rect.by));
	  }
	  if (mxn_beg >= 0) {
	    int h_bxk = region.convertXN2XK(h_bxn);
	    //int h_exk = region.convertXN2XK(h_exn);
	    //int h_byexk = region.convertXN2XK(h_byexn);
	    //int h_eybxk = region.convertXN2XK(h_eybxn);
	    Assert(pSkipCount0 == &task.vskip_skipCount0_ref(h_bxk));
	  }
	}
#endif
#if defined(USE_TEMPLATE_BF)
	dpProcBifurcateHexagon<FORWARD, L1_R2, NO_INTSEC, BF_TYPE_2>();
#else
	dpProcBifurcateHexagon(FORWARD, L1_R2, NO_INTSEC, BF_TYPE_2);
#endif
#ifdef SLR_STATE
	if (Task::takesMaxScore()) {
	  if (best_score_R < best_score1) {
	    best_score_R = best_score1;
	  }
	} else {
	  scoreManager.logAdd(best_score_R, best_score1);
	}
#else
	if (Task::takesMaxScore()) {
	  if (best_score < best_score1) {
	    best_score = best_score1;
	  }
	} else {
	  scoreManager.logAdd(best_score, best_score1);
	}
#endif
      }
    }
  }

#ifdef DEBUG
  testDPProcBifurcateLoopFinalState(BF_TYPE_2);
#endif
}

TEMPLATE
#if defined(USE_TEMPLATE_BF)
template <typename CLASS::Forward forward,
	  typename CLASS::L1R2 l1_r2,
	  typename CLASS::HasIntsec hasIntSec,
	  DPDeckPSCFG::BFType bfType>
#endif
inline void
CLASS::
#if defined(USE_TEMPLATE_BF)
dpProcBifurcateHexagon()
#else
dpProcBifurcateHexagon(Forward forward, L1R2 l1_r2, 
		       HasIntsec hasIntSec, BFType bfType)
#endif
{
  // h_bxn, h_byexn, h_exn, h_eybxn, pSkip(Count/Shift)L,R, pSkipCount0
  const int& h_xn1 = (hasIntSec ? (h_eybxn - 1) : h_byexn);
  const int& h_xn2 = (hasIntSec ? h_byexn : (h_eybxn - 1));
  const int& skipRowCountL = (h_xn1 - h_bxn + 1);
  const int& skipRowCountR = (h_exn - h_xn2);

  pSkipCount = (forward ? pSkipCountL : pSkipCountR);
  skipRowCount = (forward ? skipRowCountL : skipRowCountR);
  (forward ? pSkipShiftL++ : pSkipShiftR--);

#if defined(USE_TEMPLATE_BF)
  dpProcBifurcateLoop
    <forward, l1_r2, bfType, (forward ? MARGIN_L : MARGIN_R)>();
#else
  dpProcBifurcateLoop(forward, l1_r2, bfType, (forward ? MARGIN_L : MARGIN_R));
#endif

  if (hasIntSec) {
    if (forward) {
      pSkipCount = pSkipCountR;
      //pSkipShiftR = pSkipShiftR;
    }

    skipRowCount = h_intsec;
#if defined(USE_TEMPLATE_BF)
    dpProcBifurcateLoop<forward, l1_r2, bfType, MARGIN_LR>();
#else
    dpProcBifurcateLoop(forward, l1_r2, bfType, MARGIN_LR);
#endif

    if (!forward) {
      pSkipCount = (pSkipCountL - h_intsec);
      //pSkipShiftL -= h_intsec; <- wrong pSkipShiftL is decreased
      // already by MARGIN_LR
    }
  } else {
    skipRowCount = (-h_intsec);
    pSkipCount = (pSkipCount0 
		  + (forward ? skipRowCountL : (-skipRowCountR)));
#if defined(USE_TEMPLATE_BF)
    dpProcBifurcateLoop<forward, l1_r2, bfType, NO_MARGIN>();
#else
    dpProcBifurcateLoop(forward, l1_r2, bfType, NO_MARGIN);
#endif

    if (forward) {
      pSkipCount = pSkipCountR;
      //pSkipShiftR = pSkipShiftR;
    }  else {
      pSkipCount = pSkipCountL;
      //pSkipShiftL = pSkipShiftL;
    }
  }
  skipRowCount = (forward ? skipRowCountR : skipRowCountL);
#if defined(USE_TEMPLATE_BF)
  dpProcBifurcateLoop
    <forward, l1_r2, bfType, (forward ? MARGIN_R : MARGIN_L)>();
#else
  dpProcBifurcateLoop(forward, l1_r2, bfType, (forward ? MARGIN_R : MARGIN_L));
#endif
}
TEMPLATE
#if defined(USE_TEMPLATE_BF)
template <typename CLASS::Forward forward,
	  typename CLASS::L1R2 l1_r2,
	  DPDeckPSCFG::BFType bfType,
	  typename CLASS::MarginType marginType>
#endif
inline void
CLASS::
#if defined(USE_TEMPLATE_BF)
dpProcBifurcateLoop()
#else
dpProcBifurcateLoop(Forward forward, L1R2 l1_r2, 
		    BFType bfType, MarginType marginType)
#endif
{
  while (skipRowCount-- > 0) {
    skipCount = (forward ? (*pSkipCount++) : (*pSkipCount--));
    while (skipCount-- > 0) {
#ifdef DEBUG
      if (debug) {
	//testDPProcBifurcateLoopTransientState
	//  <forward, l1_r2, bfType, marginType>();
	testDPProcBifurcateLoopTransientState(forward, l1_r2, bfType, marginType);
      }
#endif
      const DPScoreT& score1 = (forward
				? ((*scVSTRect1++) + (*scVSTRect2++))
				: ((*scVSTRect1--) + (*scVSTRect2--)));

      if (Task::takesMaxScore()) {
	if (best_score1 < score1) {
	  best_score1 = score1;
	  if (Task::isCYK()) {
	    best_split_ptr_plus_1 = scVSTRect2;// using next position
	  }
	}
      } else {
	scoreManager.logAdd(best_score1, score1);
      }
    }
    if (marginType & MARGIN_L) {
      // using next row (access ex0 + 1)
      // *(++pSkipShift1) should be zero when skipRowCount == 0
      const int& shiftL = (forward 
			   ? (*pSkipShiftL++) 
			   : (-(*pSkipShiftL--)));
      if (l1_r2) {
	scVSTRect1 += shiftL;
      } else {
	scVSTRect2 += shiftL;
      }
    }
    if (marginType & MARGIN_R) {
      const int& shiftR = (forward 
			   ? (*pSkipShiftR++) 
			   : (-(*pSkipShiftR--)));
      if (l1_r2) {
	scVSTRect2 += shiftR;
      } else {
	scVSTRect1 += shiftR;
      }
    }
  }
}

#undef TEMPLATE
#undef CLASS
#endif
