/*
 * DPDeckPSCFGDPRegionRegion.hpp
 *
 */
#ifndef DP_DECK_PSCFG_DP_REGION_REGION_HPP
#define DP_DECK_PSCFG_DP_REGION_REGION_HPP

#include "StripRegion.hpp"
#include "StripRegionSkip.hpp"
#include "DPDeckPSCFGDPRegion.hpp"
class DPDeckPSCFG::DPRegion::Region {
public:
  const DPRegion& dpRegion;
  const SkipSet& skipSet;
  RNum regionNum;
  RShape shape;
  StripRegion srb;
  StripRegion sre;
  StripRegion srm;
  StripRegionSkip srkb;
  StripRegionSkip srke;
  StripRegionSkip srkm;
  int bxk_beg;
  int bxk_last;
  int exk_beg;
  int exk_last;
  int mxk_beg;
  int mxk_last;
  int bxn_beg;
  int bxn_last;
  int exn_beg;
  int exn_last;
  int mxn_beg;
  int mxn_last;

  Region(const DPRegion& dpRegion);
  Region(const DPRegion& dpRegion, const Region& region);
  Region(const DPRegion& dpRegion, int regionNum, int shape,
	 const StripRegion& srb, const StripRegion& sre,
	 const StripRegion& srm = StripRegion());
  Region(const Region& region);
  ~Region();
  Region& operator=(const Region& region);
  string toString() const;
  void print() const;
  void draw(ostream& fo) const;
  void draw() const;

  bool isRegionNum(int regionNum0) const;
  bool isTriShape() const;
  bool isRectVertShape() const;
  bool isRectDiagShape() const;
  bool isRectShape() const;
  bool includes(const Rect& rect) const;
  bool includes(const Point& pt) const;
  bool includes(const Region& region) const;
  int getArea() const;
  int getB0Size() const;
  int getE0Size() const;
  int getM0Size() const;
  int getVEYRangeSize() const;
  int getVSTBVEYRangeSize() const;
  int getVSTEVBYRangeSize() const;
  int getVTBSize() const;
  int getVTBIndex(const Rect& rect) const;
  int getVSkipArea() const;
  int getVSkipIndex(const Point& pt) const;
  int getVSTSize() const;
  int getVSTBIndex(const Rect& rect) const;
  int getVSTEIndex(const Rect& rect) const;

  int getVSFBWBSize(const int& bx) const;
  int getVSFBWESize(const int& ex) const;
  int getVSFBWBSizeI() const;
  int getVSFBWESizeI() const;
  int getVSFBWBSizeO() const;
  int getVSFBWESizeO() const;
  int getVSFBWBIndex(const Rect& rect) const;
  int getVSFBWEIndex(const Rect& rect) const;
  int getVSFBWBIndexI(const int& by, const int& ex, const int& ey) const;
  int getVSFBWBIndexO(const int& by, const int& ex, const int& ey) const;
  int getVSFBWEIndexI(const int& bx, const int& by, const int& ey) const;
  int getVSFBWEIndexO(const int& bx, const int& by, const int& ey) const;

  int getVBESize() const;
  const SkipSet& getSkipSet() const {return skipSet;};
  Rect getRectFromVSTBArea(const int& areaSkip) const;
  int getBXK(int y) const;
  int getEXK(int y) const;
  int getBXN(int y) const;
  int getEXN(int y) const;
  int convertXK2XN(int xk) const;
  int convertXN2XK(int xn) const;

  template <bool inside0> class Iterator;
  template <bool inside0> class BFIterator;
  
private:
  Region();

  static int getVTBSizeBE(const Trapezoid& t);
  static int getVTBSizeProdSum(int trapezoidType, int width, 
			       int blx, int elx, int etAreaRight);
  int getVTBSize(const StripRegion& srb, const StripRegion& sre) const;
};

class DPDeckPSCFG::DPRegion::VRegion
  : private Vector<DPDeckPSCFG::DPRegion::Region>
{
  typedef Vector<DPDeckPSCFG::DPRegion::Region> Super;
public:
  VRegion() : Super() {}
  ~VRegion() {}
  using Super::empty;
  using Super::size;
  using Super::clear;
  using Super::push_back;
  using Super::operator[];
private:
  VRegion(const VRegion& vregion);
  VRegion& operator=(const VRegion& vregion);
};

template <bool inside0>
class DPDeckPSCFG::DPRegion::Region::Iterator {
public:
  Rect rect;

  Iterator(const Region& region);
  Iterator(const Iterator& it);
  ~Iterator();
  void reset();
  bool getNext();

private:
  const Region& region;
  StripRegion::Iterator<!inside0> itb;
  StripRegion::Iterator<inside0> ite;

  Iterator();
  Iterator& operator=(const Iterator& it);
};

template <bool inside0>
class DPDeckPSCFG::DPRegion::Region::BFIterator {
public:
  Rect rect1;
  Rect rect2;
  Point mpt;

  const DPRegion& dpRegion;
  const SkipSet& skipSet;
  const Region& region;
  StripRegionSkip::Iterator<true>* pitm;
  StripRegionSkip::Iterator<false>* pitm_b;
  int bfType;
  Rect rect;

  BFIterator(const Region& region);
  ~BFIterator();
  string toString() const;
  void print() const;
  void reset(const Rect& rect1, const int& bfType1);
  void reset();
  bool getNext();
private:

  BFIterator();
  BFIterator(const BFIterator& it);
  BFIterator& operator=(const BFIterator& it);
};
#include "DPDeckPSCFGDPRegionRegionIterators.cpp"

#endif
