/*
 * DPDeckPSCFGDPTask.cpp
 *
 */
#include "DPDeckPSCFGVectors.hpp"
#include "DPDeckPSCFGDPTask.hpp"
#include "DPDeckPSCFGDPTaskSubTask.hpp"
#include "DPDeckPSCFGDPTaskSubTask.cpp"
#include "DPDeckPSCFGDPLocals.hpp"
#include "DPDeckPSCFGDPLocals.cpp"
#ifdef DEBUG
#include "DPDeckPSCFGDPLocalsTest.cpp"
#endif

#define CHECK_RECT 1

const char* 
DPDeckPSCFG::DPTask::
type2str(int dpType0) {
  static const char names[][20] = {
    "CYK",
    "SMALL_INSIDE",
    "SMALL_OUTSIDE",
    "INSIDE",
    "OUTSIDE",
    "LIKELIHOOD",
    "SMALL_CYK",
    "INVALID_DP_TYPE"
  };
  return names[(0 <= dpType0 && dpType0 < INVALID_DP_TYPE)
	       ? dpType0 : INVALID_DP_TYPE];
}
DPDeckPSCFG::DPTask::
DPTask(DPDeckPSCFG& dpDeck,
       DPTask* parent,
       const DPRegion& dpRegion,
       const Vector<State>& states)
: model(dpDeck.model),
  dpDeck(dpDeck),
  scoreManager(dpDeck.scoreManager),
  vtb(dpDeck.vtb),
  vtbbf(dpDeck.vtbbf),
  vsfbwB(dpDeck.vsfbwB),
  vsfbwE(dpDeck.vsfbwE),
  vfw(dpDeck.vfw),
  vbw(dpDeck.vbw),
  parent(parent),
  num_unprocessed_children(0),
  dpType(INVALID_DP_TYPE),
  dpRegion(dpRegion),
  treeNode(NULL, NULL, NULL, dpRegion.rect0, Model::X_st),
  states(states),
  vstb(*new VSTB),
  vste(*new VSTE),
  finalScore(NegInfDPScore),
  finalState(Model::X_st),
  split_rect(Rect::null()),
  split_mpt(Point::null()),
  splitType(INVALID_SPLIT_TYPE)
{
}
DPDeckPSCFG::DPTask::
DPTask(const DPTask& dpTask)
: model(dpTask.model),
  dpDeck(dpTask.dpDeck),
  scoreManager(dpTask.scoreManager),
  vtb(dpTask.vtb),
  vtbbf(dpTask.vtbbf),
  vsfbwB(dpTask.vsfbwB),
  vsfbwE(dpTask.vsfbwE),
  vfw(dpTask.vfw),
  vbw(dpTask.vbw),
  parent(dpTask.parent),
  num_unprocessed_children(dpTask.num_unprocessed_children),
  dpType(dpTask.dpType),
  dpRegion(dpTask.dpRegion),
  treeNode(dpTask.treeNode),
  states(dpTask.states),
  vstb(*new VSTB),
  vste(*new VSTE),
  finalScore(dpTask.finalScore),
  finalState(dpTask.finalState),
  split_rect(dpTask.split_rect),
  split_mpt(dpTask.split_mpt),
  splitType(dpTask.splitType)
{
}
DPDeckPSCFG::DPTask::
~DPTask()
{
  delete &vstb;
  delete &vste;
}
string
DPDeckPSCFG::DPTask::
toString() const
{
  ostringstream oss;
  oss << "dpType: " << type2str(dpType) << "\n";
  oss << "parent: " << (void*)parent << "\n";
  oss << "left: " << (void*)left << "\n";
  oss << "right: " << (void*)right << "\n";
  oss << "treeNode: [ " << (void*)(treeNode.parent) << ", ";
  oss << (void*)(treeNode.leftChild) << ", ";
  oss << (void*)(treeNode.rightChild) << " ]\n";
  oss << "states: [ " << model.state2str(states[0]) << ", ";
  oss << model.state2str(states[1]) << ", ";
  oss << model.state2str(states[2]) << " ]\n";
  oss << "dpRegion: \n" << dpRegion.toString();
  return oss.str();
}
void
DPDeckPSCFG::DPTask::
print() const
{
  cout << toString() << flush;
}

bool
DPDeckPSCFG::DPTask::
isDPType(int dpType0) const
{
  return (dpType == dpType0);
}
Vector<DPDeckPSCFG::TBTree::Link>
DPDeckPSCFG::DPTask::
getTreeNodes()
{
  Vector<TBTree::Link> treeNodes(3, (TBTree::Link)NULL);
  treeNodes[0] = &treeNode;
  if (dpRegion.hasLeftChildRect()) {
    treeNodes[1] = &treeNode;
  }
  if (dpRegion.hasRightChildRect()) {
    treeNodes[2] = &treeNode;
  }
  return treeNodes;
}
Vector<DPDeckPSCFG::DPTask*>
DPDeckPSCFG::DPTask::
cyk_small()
{
#ifdef RELEASE0
  Die("cyk_small is not supported");
  return Vector<DPDeckPSCFG::DPTask*>();
#else
#ifdef DEBUG
  cyk_dp();
  const Vector<TBTree::Link>& treeNodes = 
    traceback(dpRegion.rect0, states[0]);
  treeNodes[0]->testConsistencyR();
  TBTree tbTree(treeNodes, getStates());
  const DPScoreT& dpScore1 =
    dpDeck.computeScoreFromTree(tbTree, states);
#if 0
  cout << dpRegion.type2str(dpRegion.getType()) << "\n"
       << "cyk score: " << dpScore1 << "\n"
       << "cyk tbTree: \n"
       << tbTree.toString();
#endif
#endif

  // this is necessary
  vstb.clear();
  vste.clear();

  dpType = SMALL_CYK;
  switch (dpRegion.getType()) {
  case TRIANGLE:
    {
      dpRegion.insertCut();
      {SubTask<SMALL_INSIDE, TRIANGLE_CUT, REGION_EE> task0(*this);}
      {SubTask<SMALL_INSIDE, TRIANGLE_CUT, REGION_BB> task1(*this);}
      {SubTask<SMALL_OUTSIDE, TRIANGLE_CUT, REGION_BE> task2(*this);}
    }
    break;
  case CARET:
    dpRegion.insertCut();
    if (RType(dpRegion.getType()) == CARET_CUT_L) {
      {SubTask<SMALL_INSIDE, CARET_CUT_L, REGION_EE> task0(*this);}
      {SubTask<SMALL_INSIDE, CARET_CUT_L, REGION_MM> task1(*this);}
      {SubTask<SMALL_INSIDE, CARET_CUT_L, REGION_ME> task2(*this);}
      {SubTask<SMALL_INSIDE, CARET_CUT_L, REGION_BB> task3(*this);}
      {SubTask<SMALL_OUTSIDE, CARET_CUT_L, REGION_BE> task4(*this);}
      {SubTask<SMALL_OUTSIDE, CARET_CUT_L, REGION_BM> task5(*this);}
    } else {
      {SubTask<SMALL_INSIDE, CARET_CUT_R, REGION_BB> task0(*this);}
      {SubTask<SMALL_INSIDE, CARET_CUT_R, REGION_EE> task1(*this);}
      {SubTask<SMALL_INSIDE, CARET_CUT_R, REGION_MM> task2(*this);}
      {SubTask<SMALL_INSIDE, CARET_CUT_R, REGION_BM> task3(*this);}
      {SubTask<SMALL_OUTSIDE, CARET_CUT_R, REGION_BE> task4(*this);}
      {SubTask<SMALL_OUTSIDE, CARET_CUT_R, REGION_ME> task5(*this);}
    }
    break;
  case NOSE:
    {
      {SubTask<SMALL_INSIDE, NOSE, REGION_EE> task0(*this);}
      {SubTask<SMALL_INSIDE, NOSE, REGION_BB> task1(*this);}
      {SubTask<SMALL_INSIDE, NOSE, REGION_MM> task2(*this);}
      {SubTask<SMALL_INSIDE, NOSE, REGION_ME> task3(*this);}
      {SubTask<SMALL_INSIDE, NOSE, REGION_BM> task4(*this);}
      {SubTask<SMALL_OUTSIDE, NOSE, REGION_BE> task5(*this);}
    }
    break;
  default:
    Die("bad dpRegion.type=%d", dpRegion.getType());
    break;
  }

  finalScore = 
    findBestSplitPoint(split_rect, finalState, split_mpt, splitType);

#ifdef DEBUG
  if (abs(dpScore1 - finalScore) > 0.01) {
    cout << "tbTree:\n" << tbTree.toString();
    Assert(false,
	   "dpScore1=%f,best_score=%f,rect=[%d,%d,%d,%d],state=%s",
    	   dpScore1, finalScore,
	   split_rect.bx, split_rect.by, 
	   split_rect.ex, split_rect.ey,
	   model.state2str(finalState));
  }
#endif

  if (splitType < 0 || !IsFinite(finalScore)) {
    cout << "error: \n"
	 << "  dpType: SMALL_CYK \n"
	 << "  splitType: " << splitType << "\n"
	 << "  finalScore: " << finalScore << "\n"
	 << "  dpRegion: \n"
	 << dpRegion.toString() << endl;
    Break::Here();
    return Vector<DPTask*>();
  }

  return split(split_rect, finalState, split_mpt, splitType);
#endif
}
Vector<DPDeckPSCFG::TBTree::Link>
DPDeckPSCFG::DPTask::
cyk()
{
  // be careful
  vstb.clear();
  vste.clear();

  dpType = CYK;
  cyk_dp();
  if (!IsFinite(finalScore)) {
    cout << "error: \n"
	 << "  dpType: CYK \n"
	 << "  finalScore: " << finalScore << "\n"
	 << "  dpRegion: \n"
	 << dpRegion.toString() << endl;
    Break::Here();
    return Vector<TBTree::Link>();
  }

  const Vector<TBTree::Link>& treeNodes =
    traceback(dpRegion.rect0, states[0]);

  if (treeNode.parent) {
    Assert(treeNodes[0]);
    treeNodes[0]->parent = treeNode.parent;
    treeNode.parent->replaceChild(&treeNode, treeNodes[0]);
  }
  if (treeNode.leftChild) {
    Assert(treeNodes[1]);
    treeNodes[1]->leftChild = treeNode.leftChild;
    treeNode.leftChild->parent = treeNodes[1];
  }
  if (treeNode.rightChild) {
    Assert(treeNodes[2]);
    treeNodes[2]->leftChild = treeNode.rightChild;
    treeNode.rightChild->parent = treeNodes[2];
  }

#ifdef DEBUG
  if (treeNode.parent) {
    treeNode.parent->testConsistencyR();
  }
  if (treeNode.leftChild) {
    treeNode.leftChild->testConsistencyR();
  }
  if (treeNode.rightChild) {
    treeNode.rightChild->testConsistencyR();
  }
#endif

  return treeNodes;
}
void
DPDeckPSCFG::DPTask::
cyk_dp()
{
  // this is necessary
  vstb.clear();
  vste.clear();

  dpType = CYK;
  switch (dpRegion.type) {
  case TRIANGLE:
    {
      SubTask<CYK, TRIANGLE, REGION_EE>
	task0(*this,finalScore,finalState);
    }
    break;
#ifdef RELEASE0

#else
  case CARET:
    {
      {SubTask<CYK, CARET, REGION_EE> task0(*this);}
      {SubTask<CYK, CARET, REGION_BB> task1(*this);}
      {SubTask<CYK, CARET, REGION_BE>
	 task2(*this,finalScore,finalState);}
    }
    break;
  case NOSE:
    {
      {SubTask<CYK, NOSE, REGION_EE> task0(*this);}
      {SubTask<CYK, NOSE, REGION_MM> task1(*this);}
      {SubTask<CYK, NOSE, REGION_ME> task2(*this);}
      {SubTask<CYK, NOSE, REGION_BB> task3(*this);}
      {SubTask<CYK, NOSE, REGION_BM> task4(*this);}
      {SubTask<CYK, NOSE, REGION_BE> 
	task5(*this,finalScore,finalState);}
    }
    break;
#endif
  default:
    Die("bad dpRegion.type=%d", dpRegion.type);
    break;
  }
}
DPDeckPSCFG::DPScoreT
DPDeckPSCFG::DPTask::
inside()
{
#ifdef RELEASE0
  Die("inside() is not supported");
  return NegInfDPScore;
#else
#ifdef DEBUG
  likelihood();
  const DPScoreT& score1 = finalScore;
#endif

  // be careful
  vstb.clear();
  vste.clear();
  dpType = INSIDE;
  Assert(dpRegion.isRegionType(TRIANGLE));
  SubTask<INSIDE, TRIANGLE, REGION_EE>
    task0(*this, finalScore, finalState);

#ifdef DEBUG
  Assert(abs(finalScore - score1) < 0.01);
#endif

  return finalScore;
#endif
}
void
DPDeckPSCFG::DPTask::
outside()
{
#ifdef RELEASE0
  Die("outside() is not supported");
#else
  // be careful
  vstb.clear();
  vste.clear();

  dpType = OUTSIDE;
  Assert(dpRegion.isRegionType(TRIANGLE));
  SubTask<OUTSIDE, TRIANGLE, REGION_EE> task0(*this);
#endif
}
DPDeckPSCFG::DPScoreT
DPDeckPSCFG::DPTask::
likelihood()
{
#ifdef RELEASE0
  Die("likelihood() is not supported");
  return NegInfDPScore;
#else
#ifdef DEBUG
  //cyk_dp();
  //const DPScoreT& score1 = finalScore;
#endif

  // be careful
  vstb.clear();
  vste.clear();

  dpType = LIKELIHOOD;
  Assert(dpRegion.isRegionType(TRIANGLE));
  SubTask<LIKELIHOOD, TRIANGLE, REGION_EE>
    task0(*this, finalScore, finalState);

#ifdef DEBUG
  //Assert((finalScore - score1) > -0.01);
#endif
  return finalScore;
#endif
}
DPDeckPSCFG::DPScoreT
DPDeckPSCFG::DPTask::
findBestSplitPoint(Rect& rect, State& state,
		   Point& ptm, SplitType& best_split_type) const
{
  enum {
    EMIT_BX = true,
    EMIT_EX = false
  };
  rect.setNull();
  state = Model::X_st;
  ptm.setNull();
  best_split_type = INVALID_SPLIT_TYPE;
  DPScoreT best_score = NegInfDPScore;

  switch (dpRegion.getType()) {
  case TRIANGLE_CUT:
    {
      findBestSplitPointRect(REGION_BE, REGION_BB, EMIT_EX, SPLIT_TYPE_B,
			     rect, best_split_type, best_score, state);
      findBestSplitPointRect(REGION_BE, REGION_EE, EMIT_BX, SPLIT_TYPE_E,
			     rect, best_split_type, best_score, state);
#ifdef DEBUG
      Assert(IsFinite(best_score));
      const Rect rect1 = model.addNEmit(rect, state);
      Assert(dpRegion.includes(rect1));
#endif
    }
    break;
  case CARET_CUT_L:
    {
      findBestSplitPointRect(REGION_BM, REGION_BB, EMIT_EX, SPLIT_TYPE_B,
			     rect, best_split_type, best_score, state);
      findBestSplitPointRect(REGION_BM, REGION_MM, EMIT_BX, SPLIT_TYPE_M,
			     rect, best_split_type, best_score, state);
      findBestSplitPointRect(REGION_BE, REGION_ME, EMIT_BX, SPLIT_TYPE_E,
			     rect, best_split_type, best_score, state);
#ifdef DEBUG
      Assert(IsFinite(best_score));
      const Rect& rect1 = model.addNEmit(rect, state);
      Assert(dpRegion.includes(rect1));
#endif
    }
    break;
  case CARET_CUT_R:
    {
      findBestSplitPointRect(REGION_BE, REGION_BM, EMIT_EX, SPLIT_TYPE_B,
			     rect, best_split_type, best_score, state);
      findBestSplitPointRect(REGION_ME, REGION_MM, EMIT_EX, SPLIT_TYPE_M,
			     rect, best_split_type, best_score, state);
      findBestSplitPointRect(REGION_ME, REGION_EE, EMIT_BX, SPLIT_TYPE_E,
			     rect, best_split_type, best_score, state);
#ifdef DEBUG
      Assert(IsFinite(best_score));
      const Rect rect1 = model.addNEmit(rect, state);
      Assert(dpRegion.includes(rect1));
#endif
    }
    break;
  case NOSE:
    {
      findBestBifurcatePoint(rect, ptm, best_score);
      if (IsFinite(best_score)) {
	best_split_type = SPLIT_TYPE_M;
	state = Model::BF_st;
      }
    }
    break;
  default:
    Die("bad dpRegion.type=%d", dpRegion.getType());
    break;
  }
  return best_score;
}
void
DPDeckPSCFG::DPTask::
findBestSplitPointRect(const int& regionO, const int& regionI,
			 const bool& emitBX, const SplitType& splitType1,
			 Rect& best_rect, SplitType& best_splitType,
			 DPScoreT& best_score, State& best_state) const
{
  const DPRegion::Region& regI = dpRegion.getSubRegion(regionI);
  const DPRegion::Region& regO = dpRegion.getSubRegion(regionO);
  Rect best_rect1;
  best_rect1.setNull();
  const StripRegion& srb = regO.srb;
  const StripRegion& sre = regO.sre;
  if (emitBX) {// cut point is emitted as left emission;determine right pt
    const Rect r0 = {
      {srb.ex0, srb.ey0},
      {sre.bx0, max(sre.by0, srb.ey0)}
    };
    const ScoreManager::SCNode* fwe = &vsfbwB.refI(regI,r0.by,r0.ex,r0.ey);
    const ScoreManager::SCNode* bwe = &vsfbwB.refO(regO,r0.by,r0.ex,r0.ey);

    StripRegion::Iterator<true> ite(sre);
    const int& by_beg = srb.getEXBY();
    const int& by_last = srb.ey0;
    for (int by = by_last; by >= by_beg; by--) {
      Rect bbox = {
	{sre.bx0, by},
	{sre.ex0, sre.ey0}
      };
      ite.reset(bbox);
      while (ite.getNext()) {
	const Rect r1 = {
	  {srb.ex0, by},
	  {ite.pt.x, ite.pt.y}
	};
	Assert(fwe == &vsfbwB.refI(regI, r1.by, r1.ex, r1.ey));
	Assert(bwe == &vsfbwB.refO(regO, r1.by, r1.ex, r1.ey));

	Model::Iterator its;
	while (its.getNext()) {
	  const State& s = its.s;
	  if (!model.emitBX(s)) continue;

	  const DPScoreT& score1 = (*fwe)[s] + (*bwe)[s];
	  if (IsFinite(score1) && (best_score < score1)) {
	    best_score = score1;
	    best_rect1 = r1;
	    best_state = s;
	  }
	}
	fwe++;
	bwe++;
      }
      if (regI.isTriShape()) {
	fwe += (by_last - (by - 1) + 1);
      }
    }
 } else {// emitEX: cutpoint is emitted as right emission; determine left pt
    const Rect r0 = {
      {srb.ex0, srb.ey0},
      {sre.bx0, max(sre.by0, srb.ey0)}
    };
    const ScoreManager::SCNode* fwb = &vsfbwE.refI(regI, r0.bx, r0.by, r0.ey);
    const ScoreManager::SCNode* bwb = &vsfbwE.refO(regO, r0.bx, r0.by, r0.ey);
 
    StripRegion::Iterator<false> itb(srb);
    while (itb.getNext()) {
      const int& ey_beg = max(sre.by0, itb.pt.y);
      const int& ey_last = sre.getBXEY();
      for (int ey = ey_beg; ey <= ey_last; ey++) {
	const Rect r1 = {
	  {itb.pt.x, itb.pt.y},
	  {sre.bx0, ey}
	};
	Assert(fwb == &vsfbwE.refI(regI, r1.bx, r1.by, r1.ey));
	Assert(bwb == &vsfbwE.refO(regO, r1.bx, r1.by, r1.ey));
  
	Model::Iterator its;
	while (its.getNext()) {
	  const State& s = its.s;
	  if (!model.emitEX(s)) continue;

	  const DPScoreT& score1 = (*fwb)[s] + (*bwb)[s];
	  if (IsFinite(score1) && best_score < score1) {
	    best_score = score1;
	    best_rect1 = r1;
	    best_state = s;
	  }
	}
	fwb++;
	bwb++;
      }
    }
  }
  if (!best_rect1.isNull()) {
    best_rect = best_rect1;
    best_splitType = splitType1;
  }
}
void
DPDeckPSCFG::DPTask::
findBestBifurcatePoint(Rect& best_rect, Point& best_ptm,
		       DPScoreT& best_score) const
{
  best_rect.setNull();
  best_ptm.setNull();
  best_score = NegInfDPScore;

  const Region& rbe = dpRegion.getSubRegion(REGION_BE);
  const Region& rbm = dpRegion.getSubRegion(REGION_BM);
  const Region& rme = dpRegion.getSubRegion(REGION_ME);

  const int& countB0 = rbe.srkb.getArea();
  const int& countE0 = rbe.srke.getArea();
  const int& countM0 = rbe.srkm.getArea();
#ifdef SLR_STATE
  const DPScoreT*const& st_p0 = &vstb.head(rbe, Model::BF_st);
  const DPScoreT*const& st_l0 = &vstb.head(rbm, Model::SL_st);
  const DPScoreT*const& st_r0 = &vste.head(rme, Model::SR_st);
#else
  const DPScoreT*const& st_p0 = &vstb.head(rbe);
  const DPScoreT*const& st_l0 = &vstb.head(rbm);
  const DPScoreT*const& st_r0 = &vste.head(rme);
#endif
  int countB, countE, countM;
  const DPScoreT* st_p = st_p0;
  const DPScoreT* st_l = st_l0;
  const DPScoreT* st_r = st_r0;
  const DPScoreT* best_st_p = NULL;
  const DPScoreT* best_st_l_plus_1 = NULL;

#ifdef DEBUG
  StripRegionSkip::Iterator<false> itb(rbe.srkb);
  StripRegionSkip::Iterator<true> ite(rbe.srke);
  StripRegionSkip::Iterator<true> itm(rbe.srkm);
#endif

  st_p = st_p0;
  countB = countB0;
  while (countB-- > 0) {

#ifdef DEBUG
    Assert(itb.getNext());
    ite.reset();
#endif
    countE = countE0;
    while (countE-- > 0) {
#ifdef DEBUG
      Assert(ite.getNext());
      itm.reset();
#endif
      countM = countM0;
      while (countM-- > 0) {
#ifdef DEBUG
	Assert(itm.getNext());
	Rect rect0 = {{itb.pt}, {ite.pt}};
	Rect rect1 = {{itb.pt}, {itm.pt}};
	Rect rect2 = {{itm.pt}, {ite.pt}};
#ifdef SLR_STATE
	Assert(st_p == &vstb.ref(rbe, rect0, Model::BF_st));
	Assert(st_l == &vstb.ref(rbm, rect1, Model::SL_st));
	Assert(st_r == &vste.ref(rme, rect2, Model::SR_st));
#else
	Assert(st_p == &vstb.ref(rbe, rect0));
	Assert(st_l == &vstb.ref(rbm, rect1));
	Assert(st_r == &vste.ref(rme, rect2));
#endif
#endif
	const DPScoreT& score1 = (*st_p) + (*st_l++) + (*st_r++);
	if (IsFinite(score1) && (best_score < score1)) {
	  best_score = score1;
	  best_st_p = st_p;
	  best_st_l_plus_1 = st_l;
	}
      }
      Assert(!itm.getNext());
      st_l -= countM0;
      st_p++;
    }
    Assert(!ite.getNext());
    st_r = st_r0;
    st_l += countM0;
  }
  Assert(!itb.getNext());
  Assert(best_st_p != NULL);
  Assert(best_st_l_plus_1 != NULL);
  best_rect = rbe.getRectFromVSTBArea(best_st_p - st_p0 + 1);
  Rect rect_l = rbm.getRectFromVSTBArea(best_st_l_plus_1 - st_l0);
  best_ptm = rect_l.e;
  best_score += scoreManager.esc(best_rect, Model::BF_st);

#ifdef DEBUG
  Assert(best_rect.b == rect_l.b);
#ifdef SLR_STATE
  Assert(best_st_p == &vstb.ref(rbe, best_rect, Model::BF_st));
  Assert(best_st_l_plus_1 - 1 == &vstb.ref(rbm, rect_l, Model::SL_st));
#else
  Assert(best_st_p == &vstb.ref(rbe, best_rect));
  Assert(best_st_l_plus_1 - 1 == &vstb.ref(rbm, rect_l));
#endif
  Assert(IsFinite(best_score));
  Assert(best_rect.includes(best_ptm));
  const Rect rect1 = {{best_rect.b}, {best_ptm}};
  const Rect rect2 = {{best_ptm}, {best_rect.e}};
  Assert(dpRegion.includes(rect1));
  Assert(dpRegion.includes(rect2));
#ifdef SLR_STATE
  Assert(abs(best_score 
	     - (scoreManager.esc(best_rect,Model::BF_st)
		+ vstb.ref(rbe, best_rect, Model::BF_st)
		+ vstb.ref(rbm, rect1, Model::SL_st)
		+ vste.ref(rme, rect2, Model::SR_st))) < 0.0001);
#else
  Assert(abs(best_score 
	     - (scoreManager.esc(best_rect,Model::BF_st)
		+ vstb.ref(rbe, best_rect)
		+ vstb.ref(rbm, rect1)
		+ vste.ref(rme, rect2))) < 0.0001);
#endif
#endif
}

Vector<DPDeckPSCFG::DPTask*>
DPDeckPSCFG::DPTask::
split(const Rect& rect, const State& state,
      const Point& ptm, const SplitType& splitType)
{
  typedef TBTree::Node Node;
#ifdef DEBUG
  if (treeNode.parent) {
    treeNode.parent->testConsistencyR();
  }
#endif
  Vector<DPTask*> vDPTask(3, (DPTask*)NULL);
  const Vector<DPRegion>& vdpr = dpRegion.split(rect, ptm, splitType);

  Assert(vdpr[0].regions[0].srkb.getSkipSize() > 0);
  Assert(vdpr[1].regions[0].srkb.getSkipSize() > 0);

  Vector<State> states0(3, Model::X_st);
  Vector<State> states1(3, Model::X_st);
  Vector<State> states2(3, Model::X_st);

  switch (dpRegion.getType()) {
  case TRIANGLE_CUT:
    {
      states0[0] = states[0];
      states0[1] = state;
      states1[0] = state;
      
      // constructer should take Node as an argument to
      // remove duplicated codes
      DPTask& dpTask0 = *new DPTask(dpDeck, this, vdpr[0], states0);
      DPTask& dpTask1 = *new DPTask(dpDeck, this, vdpr[1], states1);
   
      Node& node0 = dpTask0.treeNode;
      Node& node1 = dpTask1.treeNode;

      node0.leftChild = &node1;
      node1.parent = &node0;

      if (treeNode.parent) {
	node0.parent = treeNode.parent;
	treeNode.parent->replaceChild(&treeNode, &node0);
      }

#ifdef DEBUG
      if (treeNode.parent) {
	treeNode.parent->testConsistencyR();
      }
      if (treeNode.leftChild) {
	treeNode.leftChild->testConsistencyR();
      }
      if (treeNode.rightChild) {
	treeNode.rightChild->testConsistencyR();
      }
#endif

      vDPTask[0] = &dpTask0;
      vDPTask[1] = &dpTask1;
    }
    break;
  case CARET_CUT_L:
    {
      if (vdpr[0].isType(NOSE)) {// CL -> N + T
	states0[0] = states[0];
	states0[1] = state;
	states0[2] = states[1];
	states1[0] = state;

	DPTask& dpTask0 = *new DPTask(dpDeck, this, vdpr[0], states0);
	DPTask& dpTask1 = *new DPTask(dpDeck, this, vdpr[1], states1);

	Node& node0 = dpTask0.treeNode;
	Node& node1 = dpTask1.treeNode;

	node0.leftChild = &node1;
	node1.parent = &node0;

	if (treeNode.parent) {
	  node0.parent = treeNode.parent;
	  treeNode.parent->replaceChild(&treeNode, &node0);
	}
	if (treeNode.leftChild) {
	  node0.rightChild = treeNode.leftChild;
	  treeNode.leftChild->parent = &node0;
	}

#ifdef DEBUG
	if (treeNode.parent) {
	  treeNode.parent->testConsistencyR();
	}
	if (treeNode.leftChild) {
	  treeNode.leftChild->testConsistencyR();
	}
	if (treeNode.rightChild) {
	  treeNode.rightChild->testConsistencyR();
	}
#endif

	vDPTask[0] = &dpTask0;
	vDPTask[1] = &dpTask1;
      } else {// CL -> C + C
	states0[0] = states[0];
	states0[1] = state;
	states1[0] = state;
	states1[1] = states[1];

	DPTask& dpTask0 = *new DPTask(dpDeck, this, vdpr[0], states0);
	DPTask& dpTask1 = *new DPTask(dpDeck, this, vdpr[1], states1);

	Node& node0 = dpTask0.treeNode;
	Node& node1 = dpTask1.treeNode;

	node0.leftChild = &node1;
	node1.parent = &node0;

	if (treeNode.parent) {
	  node0.parent = treeNode.parent;
	  treeNode.parent->replaceChild(&treeNode, &node0);
	}
	if (treeNode.leftChild) {
	  node1.leftChild = treeNode.leftChild;
	  treeNode.leftChild->parent = &node1;
	}

#ifdef DEBUG
	if (treeNode.parent) {
	  treeNode.parent->testConsistencyR();
	}
	if (treeNode.leftChild) {
	  treeNode.leftChild->testConsistencyR();
	}
	if (treeNode.rightChild) {
	  treeNode.rightChild->testConsistencyR();
	}
#endif

	vDPTask[0] = &dpTask0;
	vDPTask[1] = &dpTask1;
      }
    }
    break;
  case CARET_CUT_R:
    {
      if (vdpr[0].isType(NOSE)) {// CR -> N + T
	states0[0] = states[0];
	states0[1] = states[1];
	states0[2] = state;
	states1[0] = state;

	DPTask& dpTask0 = *new DPTask(dpDeck, this, vdpr[0], states0);
	DPTask& dpTask1 = *new DPTask(dpDeck, this, vdpr[1], states1);
   
	Node& node0 = dpTask0.treeNode;
	Node& node1 = dpTask1.treeNode;

	node0.rightChild = &node1;
	node1.parent = &node0;

	if (treeNode.parent) {
	  node0.parent = treeNode.parent;
	  treeNode.parent->replaceChild(&treeNode, &node0);
	}
	if (treeNode.leftChild) {
	  node0.leftChild = treeNode.leftChild;
	  treeNode.leftChild->parent = &node0;
	}

#ifdef DEBUG
	if (treeNode.parent) {
	  treeNode.parent->testConsistencyR();
	}
	if (treeNode.leftChild) {
	  treeNode.leftChild->testConsistencyR();
	}
	if (treeNode.rightChild) {
	  treeNode.rightChild->testConsistencyR();
	}
#endif

	vDPTask[0] = &dpTask0;
	vDPTask[2] = &dpTask1;//be careful [2]
	
      } else {// CR -> C + C
	states0[0] = states[0];
	states0[1] = state;
	states1[0] = state;
	states1[1] = states[1];

	DPTask& dpTask0 = *new DPTask(dpDeck, this, vdpr[0], states0);
	DPTask& dpTask1 = *new DPTask(dpDeck, this, vdpr[1], states1);
   
	Node& node0 = dpTask0.treeNode;
	Node& node1 = dpTask1.treeNode;

	node0.leftChild = &node1;
	node1.parent = &node0;

	if (treeNode.parent) {
	  node0.parent = treeNode.parent;
	  treeNode.parent->replaceChild(&treeNode, &node0);
	}
	if (treeNode.leftChild) {
	  node1.leftChild = treeNode.leftChild;
	  treeNode.leftChild->parent = &node1;
	}

#ifdef DEBUG
	if (treeNode.parent) {
	  treeNode.parent->testConsistencyR();
	}
	if (treeNode.leftChild) {
	  treeNode.leftChild->testConsistencyR();
	}
	if (treeNode.rightChild) {
	  treeNode.rightChild->testConsistencyR();
	}
#endif

	vDPTask[0] = &dpTask0;
	vDPTask[1] = &dpTask1;// be careful about [1]
      }
    }
    break;
  case NOSE:
    {
      states0[0] = states[0];
      states0[1] = Model::BF_st;
#ifdef SLR_STATE
      states1[0] = Model::SL_st;
      states1[1] = states[1];
      states2[0] = Model::SR_st;
      states2[1] = states[2];
#else
      states1[0] = Model::ST_st;
      states1[1] = states[1];
      states2[0] = Model::ST_st;
      states2[1] = states[2];
#endif
      DPTask& dpTask0 = *new DPTask(dpDeck, this, vdpr[0], states0);
      DPTask& dpTask1 = *new DPTask(dpDeck, this, vdpr[1], states1);
      DPTask& dpTask2 = *new DPTask(dpDeck, this, vdpr[2], states2);

      Node& node0 = dpTask0.treeNode;
      Node& node1 = dpTask1.treeNode;
      Node& node2 = dpTask2.treeNode;

      Node& bfNode = 
	*new Node(&node0, &node1, &node2, rect, Model::BF_st);
  
      node0.leftChild = &bfNode;
      node1.parent = &bfNode;
      node2.parent = &bfNode;

      if (treeNode.parent) {
	node0.parent = treeNode.parent;
	treeNode.parent->replaceChild(&treeNode, &node0);
      }
      if (treeNode.leftChild) {
	node1.leftChild = treeNode.leftChild;
	treeNode.leftChild->parent = &node1;
      }
      if (treeNode.rightChild) {
	node2.leftChild = treeNode.rightChild;
	treeNode.rightChild->parent = &node2;
      }

#ifdef DEBUG
      if (treeNode.parent) {
	treeNode.parent->testConsistencyR();
      }
      if (treeNode.leftChild) {
	treeNode.leftChild->testConsistencyR();
      }
      if (treeNode.rightChild) {
	treeNode.rightChild->testConsistencyR();
      }
#endif

      vDPTask[0] = &dpTask0;
      vDPTask[1] = &dpTask1;
      vDPTask[2] = &dpTask2;
    }
    break;
  default:
    Die("bad dpRegion.type=%d", dpRegion.type);
  }
#ifdef DEBUG
  if (treeNode.parent) {
    treeNode.parent->testConsistencyR();
  }
  for (int i = 0; i < (int)vDPTask.size(); i++) {
    if (vDPTask[i]) {
      Assert(vDPTask[i]->dpRegion.regions[0].srkb.getSkipSize() > 0);
    }
  }
#endif
  return vDPTask;
}
Vector<DPDeckPSCFG::TBTree::Link>
DPDeckPSCFG::DPTask::
traceback(const Rect& rect0, const State& state0) const
{
  typedef TBTree::Node Node;
  typedef TBTree::Link Link;
  Vector<TBTree::Link> treeNodes(3, (TBTree::Link)NULL);
  Assert(rect0.isNormal());
  if ((rect0.b == rect0.e)
      || isLeftChild(rect0, state0)
      || isRightChild(rect0, state0)) {
    return treeNodes;
  }

  const Link& root = new Node(NULL, NULL, NULL, rect0, state0);
  treeNodes[0] = root;

  Link link = root;
  if (state0 == Model::BEG_st) {
    const Link& link1 = 
      new Node(NULL, NULL, NULL, rect0, finalState);
    root->leftChild = link1;
    link1->parent = root;
    link = link1;
  }

  Vector<Link> arr;
  arr.push_back(link);
  
  Rect rect = rect0;
  State state = link->state;
  while (!arr.empty()) {
    Link link = arr.back(); arr.pop_back();
    rect = link->rect;
    state = link->state;
    if (model.bifurcates(state)) {// BEG->BF && BF is left-rightEnd is not taken into account
      const Point& m = getBifurcationPoint(rect);
      Rect rect1 = {{rect.b}, {m}};
      Rect rect2 = {{m}, {rect.e}};
#if CHECK_RECT
      Check(rect1.isNormal());
      Check(rect2.isNormal());
#endif
#ifdef SLR_STATE
      const Link& link1 = new Node(link, NULL, NULL,rect1,Model::SL_st);
      const Link& link2 = new Node(link, NULL, NULL,rect2,Model::SR_st);
#else
      const Link& link1 = new Node(link, NULL, NULL,rect1,Model::ST_st);
      const Link& link2 = new Node(link, NULL, NULL,rect2,Model::ST_st);
#endif
      link->leftChild = link1;
      link->rightChild = link2;

      arr.push_back(link2);
      arr.push_back(link1);
      continue;
    }
      
    int child_status = getChild(rect, state);
    while (child_status == GENERIC_STATE) {
      const Link& link1 = new Node(link, NULL, NULL, rect, state);
      link->leftChild = link1;
      link = link1;
      child_status = getChild(rect, state);
    }
    if (child_status == LEFT_CHILD) {
      treeNodes[1] = link;
    } else if (child_status == RIGHT_CHILD) {
      treeNodes[2] = link;
    } else if (child_status == BIFURCATION) {
      // assuming bifurcation state is not left right child. Is this correct?
      const Link& link0 = new Node(link, NULL, NULL, rect, state);
      link->leftChild = link0;
      arr.push_back(link0);
    } else {
      Assert(child_status == END_STATE);
      const Link& link0 = new Node(link, NULL, NULL, rect, state);
      link->leftChild = link0;
    }
  }
#ifdef DEBUG
  treeNodes[0]->testConsistencyR();
#endif
  return treeNodes;
}
int
DPDeckPSCFG::DPTask::
getChild(Rect& rect, State& state) const
{
  Assert(model.isNormalState(state));
  Assert(rect.bx <= rect.ex
	 && rect.by <= rect.ey
	 && !(rect.b == rect.e));
  Assert(dpRegion.includes(rect));

#if CHECK_RECT
  Check(rect.isNormal(), "rect: [%d,%d,%d,%d]", rect.bx,rect.by,rect.ex,rect.ey);
#endif
  const DPRegion::Region& region = dpRegion.findRegion(rect);
  const TBNode& tbn = vtb.ref(region, rect);
  const Rect& rect1 = model.addNEmit(rect, state);

  Assert(dpRegion.includes(rect1)
	 && rect1.bx <= rect1.ex
	 && rect1.by <= rect1.ey);
  rect = rect1;
  if (rect.b == rect.e) {// leftChild and rightChild should be b < e

    state = Model::END_st;
    return END_STATE;
  } else {
    const State& state1 = tbn.getVal(state);
    state = state1;
    if (isLeftChild(rect, state)) {
      return LEFT_CHILD;
    } else if (isRightChild(rect, state)) {
      return RIGHT_CHILD;
    } else  if (model.bifurcates(state)) {
      return BIFURCATION;
    } else {
      return GENERIC_STATE;
    }
  }
}
bool
DPDeckPSCFG::DPTask::
isLeftChild(const Rect& rect, const State& state) const
{
  if (!dpRegion.hasLeftChildRect()) return false;
  if (state != states[1]) return false;
  if (!(rect == dpRegion.getLeftChildRect())) return false;

  return true;
}
bool
DPDeckPSCFG::DPTask::
isRightChild(const Rect& rect, const State& state) const
{
  if (!dpRegion.hasRightChildRect()) return false;
  if (state != states[2]) return false;
  if (!(rect == dpRegion.getRightChildRect())) return false;

  return true;
}
Point
DPDeckPSCFG::DPTask::
getBifurcationPoint(const Rect& rect) const
{
  Assert(dpRegion.isSkipRect(rect));
#if CHECK_RECT
  Check(rect.isNormal(), "rect: [%d,%d,%d,%d]", rect.bx,rect.by,rect.ex,rect.ey);
#endif
  const DPRegion::Region& region = dpRegion.findRegion(rect);
  const TBBFNode& nd = vtbbf.ref(region, rect);
  const int& bfType = nd.getBFType();
  const int& areaSkip = nd.getArea();
  Assert(areaSkip > 0);
  Rect rect1;
  Rect rect2;
  return getBifurcationPoint(region, rect, bfType, areaSkip,
			     rect1, rect2);
}
Point
DPDeckPSCFG::DPTask::
getBifurcationPoint(const DPRegion::Region& region,
		    const Rect& rect, const int& bfType,
		    const int& areaSkip,
		    Rect& rect1, Rect& rect2) const
{
  Assert(dpType == CYK);
  const bool inside0 = true;
  Assert(region.includes(rect));
  Assert(dpRegion.isSkipRect(rect));
  Point mpt = Point::null();
  if (inside0) {
    Rect bbox = {
      {region.srb.bx0, region.srb.by0}, 
      {region.sre.ex0, rect.ey}
    };
    switch (bfType) {
    case BF_TYPE_0:
      mpt = (region.srkb
	     .getBoundedRegion(bbox)
	     .advance(rect.b, (areaSkip - 1)));
      break;
    case BF_TYPE_1:
      mpt = (region.srke
	     .getBoundedRegion(bbox)
	     .advance(rect.e, (-(areaSkip - 1))));// minus since backward
      break;
    case BF_TYPE_2:
      mpt = (region.srkm
	     .advance(region.srkm.getRectK().b, (areaSkip - 1)));
      break;
    default:
      Die("bad bfType=%d", bfType);
      mpt = Point::null();
      break;
    }
    rect1.b = rect.b;
    rect1.e = mpt;
    rect2.b = mpt;
    rect2.e = rect.e;

    // used for debug
    //  } else { //outside
    //    Rect bbox = {{region.srb.b0}, {region.sre.e0}};
    //    switch (bfType) {
    //    case BF_TYPE_0:
    //      bbox.b = rect.e;
    //      mpt = rect.e;
    //      mpt = (region.srke
    //	     .getBoundedRegion(bbox)
    //	     .advance(mpt, areaSkip));
    //      rect1.b = rect.b;
    //      rect1.e = mpt;
    //      rect2.b = rect.e;
    //      rect2.e = mpt;
    //      break;
    //    case BF_TYPE_1:
    //      bbox.e = rect.b;
    //      mpt = region.srb.b0;
    //      mpt = region.srkb.getBoundedRegion(bbox).advance(mpt, areaSkip);
    //      rect1.b = mpt;
    //      rect1.e = rect.e;
    //      rect2.b = mpt;
    //      rect2.e = rect.b;
    //      break;
    //    case BF_TYPE_2:
    //      mpt = region.srm.b0;
    //      mpt = region.srkm.advance(mpt, areaSkip);
    //      if (dpRegion.isRegionType(CARET_CUT_L)) {
    //	rect1.b = rect.b;
    //	rect1.e = mpt;
    //	rect2.b = rect.e;
    //	rect2.e = mpt;
    //      } else {
    //	rect1.b = mpt;
    //	rect1.e = rect.e;
    //	rect2.b = mpt;
    //	rect2.e = rect.b;
    //      }
    //      break;
    //    default:
    //      Die("bad bfType=%d", bfType);
    //      break;
    //    }

  }
#ifdef DEBUG
  Assert(rect.includes(mpt),
	 "rect: [%d,%d,%d,%d], mpt: [%d,%d]",
	 rect.bx,rect.by,rect.ex,rect.ey, mpt.x,mpt.y);
  Assert(!(rect.b == mpt) && !(rect.e == mpt));
  Assert(dpRegion.isSkipPoint(mpt), "m: [%d,%d]", mpt.x, mpt.y);
#endif

  return mpt;
}
float
DPDeckPSCFG::DPTask::
computeMemorySizeInMB(const int& dpType) const
{
  const float& oneMB = (1 << 20);
  float n = 0;
  float m = 0;
  for (int i = 0; i < (int)dpRegion.getNRegion(); i++) {
    const DPRegion::Region& region = dpRegion.regions[i];
    float m1 = 0;
    m1 += VTZ::computeSizeInBytes(region);
    m1 += VX::computeSizeInBytes(region);
    m1 += VY::computeSizeInBytes(region);
    m1 += VBE::computeSizeInBytes(region);
    m1 += VPairX::computeSizeInBytes(region);
    m1 += VPairY::computeSizeInBytes(region);
    m1 += VSC::computeSizeInBytes(region);
    m1 += VSkip::computeSizeInBytes(region);
    m = max(m, m1);

    n += VSTB::computeSizeInBytes(region);
    n += VSTE::computeSizeInBytes(region);
    if (dpType == CYK) {
      n += VTB::computeSizeInBytes(region);
      n += VTBBF::computeSizeInBytes(region);
    } else if (dpType == SMALL_CYK
	       || dpType == SMALL_INSIDE
	       || dpType == SMALL_OUTSIDE) {
      n += VSFBWB::computeSizeInBytes(region);
      n += VSFBWE::computeSizeInBytes(region);
    } else if (dpType == INSIDE || dpType == OUTSIDE) {
      // always needed in pair
      n += VFBW::computeSizeInBytes(region);
      n += VFBW::computeSizeInBytes(region);
    }
  }
  return ((n + m) / oneMB);
}
float
DPDeckPSCFG::DPTask::
computeDPIterationCountVTB() const
{
  float n = 0;
  for (int i = 0; i < (int)dpRegion.getNRegion(); i++) {
    const DPRegion::Region& region = dpRegion.regions[i];
    n += region.getVTBSize();
  }
  return n;
}
float
DPDeckPSCFG::DPTask::
computeDPIterationCountVST() const
{
  float n = 0;
  for (int i = 0; i < (int)dpRegion.getNRegion(); i++) {
    const DPRegion::Region& region = dpRegion.regions[i];
    n += region.getVSTSize();
  }
  return n;
}
float
DPDeckPSCFG::DPTask::
computeCYKMemorySizeInMB() const
{
  return computeMemorySizeInMB(CYK);
}
float
DPDeckPSCFG::DPTask::
computeSmallCYKMemorySizeInMB() const
{
  return computeMemorySizeInMB(SMALL_CYK);
}
float
DPDeckPSCFG::DPTask::
computeInsideOutsideMemorySizeInMB() const
{
  return computeMemorySizeInMB(INSIDE);
}
float
DPDeckPSCFG::DPTask::
computeLikelihoodMemorySizeInMB() const
{
  return computeMemorySizeInMB(LIKELIHOOD);
}
#ifdef SLR_STATE
const DPDeckPSCFG::DPScoreT&
DPDeckPSCFG::DPTask::
vstb_ref(int regionNum, int bx, int by, int ex, int ey, State s) const
{
  Rect rect = {{bx, by}, {ex, ey}};
  Assert(dpRegion.isSkipRect(rect));
  const DPRegion::Region& r = dpRegion.getSubRegion(regionNum);
  Assert(r.includes(rect));
  return vstb.ref(r, rect, s);
}
const DPDeckPSCFG::DPScoreT&
DPDeckPSCFG::DPTask::
vste_ref(int regionNum, int bx, int by, int ex, int ey, State s) const
{
  Rect rect = {{bx, by}, {ex, ey}};
  Assert(dpRegion.isSkipRect(rect));
  const DPRegion::Region& r = dpRegion.getSubRegion(regionNum);
  Assert(r.includes(rect));
  return vste.ref(r, rect, s);
}
const DPDeckPSCFG::DPScoreT&
DPDeckPSCFG::DPTask::
vstb_ref(int bx, int by, int ex, int ey, State s) const
{
  Rect rect = {{bx, by}, {ex, ey}};
  Assert(dpRegion.isSkipRect(rect));
#if CHECK_RECT
  Check(rect.isNormal(), "rect: [%d,%d,%d,%d]", rect.bx,rect.by,rect.ex,rect.ey);
#endif
  const DPRegion::Region& r = dpRegion.findRegion(rect);
  return vstb.ref(r, rect, s);
}
const DPDeckPSCFG::DPScoreT&
DPDeckPSCFG::DPTask::
vste_ref(int bx, int by, int ex, int ey, State s) const
{
  Rect rect = {{bx, by}, {ex, ey}};
  Assert(dpRegion.isSkipRect(rect));
#if CHECK_RECT
  Check(rect.isNormal(), "rect: [%d,%d,%d,%d]", rect.bx,rect.by,rect.ex,rect.ey);
#endif
  const DPRegion::Region& r = dpRegion.findRegion(rect);
  return vste.ref(r, rect, s);
}
#else
const DPDeckPSCFG::DPScoreT&
DPDeckPSCFG::DPTask::
vstb_ref(int regionNum, int bx, int by, int ex, int ey) const
{
  Rect rect = {{bx, by}, {ex, ey}};
  Assert(dpRegion.isSkipRect(rect));
  const DPRegion::Region& r = dpRegion.getSubRegion(regionNum);
  Assert(r.includes(rect));
  return vstb.ref(r, rect);
}
const DPDeckPSCFG::DPScoreT&
DPDeckPSCFG::DPTask::
vste_ref(int regionNum, int bx, int by, int ex, int ey) const
{
  Rect rect = {{bx, by}, {ex, ey}};
  Assert(dpRegion.isSkipRect(rect));
  const DPRegion::Region& r = dpRegion.getSubRegion(regionNum);
  Assert(r.includes(rect));
  return vste.ref(r, rect);
}
const DPDeckPSCFG::DPScoreT&
DPDeckPSCFG::DPTask::
vstb_ref(int bx, int by, int ex, int ey) const
{
  Rect rect = {{bx, by}, {ex, ey}};
  Assert(dpRegion.isSkipRect(rect));
#if CHECK_RECT
  Check(rect.isNormal(), "rect: [%d,%d,%d,%d]", rect.bx,rect.by,rect.ex,rect.ey);
#endif
  const DPRegion::Region& r = dpRegion.findRegion(rect);
  return vstb.ref(r, rect);
}
const DPDeckPSCFG::DPScoreT&
DPDeckPSCFG::DPTask::
vste_ref(int bx, int by, int ex, int ey) const
{
  Rect rect = {{bx, by}, {ex, ey}};
  Assert(dpRegion.isSkipRect(rect));
#if CHECK_RECT
  Check(rect.isNormal(), "rect: [%d,%d,%d,%d]", rect.bx,rect.by,rect.ex,rect.ey);
#endif
  const DPRegion::Region& r = dpRegion.findRegion(rect);
  return vste.ref(r, rect);
}
#endif
const DPDeckPSCFG::TBNode&
DPDeckPSCFG::DPTask::
vtb_ref(int regionNum, int bx, int by, int ex, int ey) const
{
  Rect rect = {{bx, by}, {ex, ey}};
  const DPRegion::Region& r = dpRegion.getSubRegion(regionNum);
  Assert(r.includes(rect));
  return dpDeck.vtb.ref(r, rect);
}
const DPDeckPSCFG::TBBFNode&
DPDeckPSCFG::DPTask::
vtbbf_ref(int regionNum, int bx, int by, int ex, int ey) const
{
  Rect rect = {{bx, by}, {ex, ey}};
  Assert(dpRegion.isSkipRect(rect));
  const DPRegion::Region& r = dpRegion.getSubRegion(regionNum);
  Assert(r.includes(rect));
  return dpDeck.vtbbf.ref(r, rect);
}
const DPDeckPSCFG::ScoreManager::SCNode& 
DPDeckPSCFG::DPTask::
vsfbwB_refI(int regionNum, int by, int ex, int ey) const
{
  const DPRegion::Region& r = dpRegion.getSubRegion(regionNum);
  return vsfbwB.refI(r, by, ex, ey);
}
const DPDeckPSCFG::ScoreManager::SCNode&
DPDeckPSCFG::DPTask::
vsfbwB_refO(int regionNum, int by, int ex, int ey) const
{
  const DPRegion::Region& r = dpRegion.getSubRegion(regionNum);
  return vsfbwB.refO(r, by, ex, ey);
}
const DPDeckPSCFG::ScoreManager::SCNode&
DPDeckPSCFG::DPTask::
vsfbwE_refI(int regionNum, int bx, int by, int ey) const
{
  const DPRegion::Region& r = dpRegion.getSubRegion(regionNum);
  return vsfbwE.refI(r, bx, by, ey);
}
const DPDeckPSCFG::ScoreManager::SCNode&
DPDeckPSCFG::DPTask::
vsfbwE_refO(int regionNum, int bx, int by, int ey) const
{
  const DPRegion::Region& r = dpRegion.getSubRegion(regionNum);
  return vsfbwE.refO(r, bx, by, ey);
}
DPDeckPSCFG::DPScoreT
DPDeckPSCFG::DPTask::
getScore() const
{
  return finalScore;
}
DPDeckPSCFG::State
DPDeckPSCFG::DPTask::
getFinalState() const
{
  return finalState;
}

void
DPDeckPSCFG::DPTask::
testConsistency() const
{
#ifdef DEBUG
  Assert(model.isNormalState(states[0]));
  Assert(TRIANGLE <= dpRegion.getType() 
	 && dpRegion.getType() <= CARET_CUT_R);
#endif
}
#ifdef SLR_STATE
void
DPDeckPSCFG::DPTask::
loadVSTB(const DPRegion::Region& region, const State& s)
{
  if (!vstb.isEmpty(region, s)) return; //do nothing

  vstb.resize(region, s);
  if (vstb.isEmpty(region, s)) return;

  if (!vste.isEmpty(region, s)) {
    vstb.copy(region, vste, s);
    return;
  } 
  if (parent != NULL) {
    const DPTask& dpTask1 = *parent;
    const DPRegion& dpRegion1 = dpTask1.dpRegion;
    const Region& region1 = dpRegion1.findRegion(region);
    const VSTB& vstb1 = dpTask1.vstb;
    const VSTE& vste1 = dpTask1.vste;
    if (!vstb1.isEmpty(region1, s)) {
      vstb.copy(region, vstb1, region1, s);
    } else if (!vste1.isEmpty(region1, s)) {
      vstb.copy(region, vste1, region1, s);
    } else {
      Die("cannot find copy source of vstb in parent");
    }
    return;
  }
  Die("cannot find copy source of vstb");
}
void
DPDeckPSCFG::DPTask::
loadVSTE(const DPRegion::Region& region, const State& s)
{
  if (!vste.isEmpty(region, s)) return; //do nothing

  vste.resize(region, s);
  if (vste.isEmpty(region, s)) return;

  if (!vstb.isEmpty(region, s)) {
    vste.copy(region, vstb, s);
    return;
  } 
  if (parent != NULL) {
    const DPTask& dpTask1 = *parent;
    const DPRegion& dpRegion1 = dpTask1.dpRegion;
    const Region& region1 = dpRegion1.findRegion(region);
    const VSTB& vstb1 = dpTask1.vstb;
    const VSTE& vste1 = dpTask1.vste;
    if (!vste1.isEmpty(region1, s)) {
      vste.copy(region, vste1, region1, s);
    } else if (!vstb1.isEmpty(region1, s)) {
      vste.copy(region, vstb1, region1, s);
    } else {
      Die("cannot find copy source of vste in parent");
    }
    return;
  }
  Die("cannot find copy source of vste");
}
#else
void
DPDeckPSCFG::DPTask::
loadVSTB(const DPRegion::Region& region)
{
  if (!vstb.isEmpty(region)) return; //do nothing

  vstb.resize(region, Model::ST_st);
  if (vstb.isEmpty(region)) return;

  if (!vste.isEmpty(region)) {
    vstb.copy(region, vste);
    return;
  } 
  if (parent != NULL) {
    const DPTask& dpTask1 = *parent;
    const DPRegion& dpRegion1 = dpTask1.dpRegion;
    const Region& region1 = dpRegion1.findRegion(region);
    const VSTB& vstb1 = dpTask1.vstb;
    const VSTE& vste1 = dpTask1.vste;
    if (!vstb1.isEmpty(region1)) {
      vstb.copy(region, vstb1, region1);
    } else if (!vste1.isEmpty(region1)) {
      vstb.copy(region, vste1, region1);
    } else {
      Die("cannot find copy source of vstb in parent");
    }
    return;
  }
  Die("cannot find copy source of vstb");
}
void
DPDeckPSCFG::DPTask::
loadVSTE(const DPRegion::Region& region)
{
  if (!vste.isEmpty(region)) return; //do nothing

  vste.resize(region, Model::ST_st);
  if (vste.isEmpty(region)) return;

  if (!vstb.isEmpty(region)) {
    vste.copy(region, vstb);
    return;
  } 
  if (parent != NULL) {
    const DPTask& dpTask1 = *parent;
    const DPRegion& dpRegion1 = dpTask1.dpRegion;
    const Region& region1 = dpRegion1.findRegion(region);
    const VSTB& vstb1 = dpTask1.vstb;
    const VSTE& vste1 = dpTask1.vste;
    if (!vste1.isEmpty(region1)) {
      vste.copy(region, vste1, region1);
    } else if (!vstb1.isEmpty(region1)) {
      vste.copy(region, vstb1, region1);
    } else {
      Die("cannot find copy source of vste in parent");
    }
    return;
  }
  Die("cannot find copy source of vste");
}
#endif

