/*
 * DPDeckPSCFGDPTaskSubTask.hpp
 *
 */
#ifndef DP_DECK_PSCFG_DP_TASK_SUB_TASK_HPP
#define DP_DECK_PSCFG_DP_TASK_SUB_TASK_HPP

#include "DPDeckPSCFG.hpp"
#include "DPDeckPSCFGDPTask.hpp"
#include "DPDeckPSCFGVectors.hpp"

template <DPDeckPSCFG::DPTask::DPType dpType0,
	  DPDeckPSCFG::DPTask::RType regionType0,
	  DPDeckPSCFG::DPTask::RNum regionNum0>
class DPDeckPSCFG::DPTask::SubTask {
public:
  enum {
    dpType = dpType0,
    regionType = regionType0,
    regionNum = regionNum0,
    insideType = ((dpType0 != OUTSIDE)
		  && (dpType0 != SMALL_OUTSIDE)),
    outsideType = !insideType
  };
  static bool isDPType(int dpt1);
  static bool isRegionType(int rt1);
  static bool isRegionNum(int rn1);
  static bool isRegionTypeNum(int rt1, int rn1);
  static bool isTriShape();
  static bool isRectShape();
  static bool isRectVertShape();
  static bool isRectDiagShape();
  static bool storeB();
  static bool storeE();
  static bool isInsideType();
  static bool isOutsideType();
  static bool takesMaxScore();
  static bool isSmallCYK();
  static bool isCYK();
  static bool isInside();
  static bool isOutside();
  static bool isLikelihood();
  static bool isTriIn();
  static bool isTriOut();
  static bool isRectIn();
  static bool isRectVertIn();
  static bool isRectDiagIn();
  static bool isRectOut();
  static bool isRectVertOut();
  static bool isRectDiagOut();
  static bool hasBFType1();
  static bool hasBFType2();

  DPDeckPSCFG& dpDeck;
  ScoreManager& scoreManager;
  DPPath& dpPath;
  VTZ& vt;
  VX& vx;
  VY& vy;
  VPairX& vpairX;
  VPairY& vpairY;
  VBE& vbe;
  VSC& vsc;
  VSkip& vskip;
  VTB& vtb;
  VTBBF& vtbbf;
  VSFBWB& vsfbwB;
  VSFBWE& vsfbwE;
  VFBW& vfw;
  VFBW& vbw;
  DPTask& dpTask;
  VSTB& vstb;
  VSTE& vste;
  const DPRegion& dpRegion;
  const SkipSet& skipSet;
  const DPRegion::Region& region;
  const StripRegion& srb;
  const StripRegion& sre;
  const StripRegion& srm;
  const StripRegionSkip& srkb;
  const StripRegionSkip& srke;
  const StripRegionSkip& srkm;
  const Rect rectb;
  const Rect recte;
  const Rect rectm;
  Rect rectkb;
  Rect rectke;
  Rect rectkm;
  DPScoreT finalScore;
  State finalState;

  int dpAreaMargin;
  State state0;
  State state_lim;

  ScoreManager::SCNode* scb_store;
  ScoreManager::SCNode* sce_store;

  TBNode* tau;
  TBBFNode* tau_BF;

  TZNode* bt_lim;
  TZNode* bt;
  TZNode* et0;
  TZNode* et_lim;

  BENode* bn;
  BENode* en0;

  int bxn_beg;
  int bxn_last;
  int exn_beg;
  int exn_last;
  int mxn_beg;
  int mxn_last;

  int bxr0;
  int exr0;

  int skipSize;
  int* pSkipCountB0;
  int* pSkipCountE0;
  int* pSkipCountM0;

  SubTask(DPTask& dpTask);
  SubTask(DPTask& dpTask, DPScoreT& finalScore0, State& finalState0);
  ~SubTask();
  string toString() const;
  void print() const;
  void run();

  const DPRegion::Region& getRegionTriE() const;
  const DPRegion::Region& getRegionTriB() const;
  const DPRegion::Region& getRegionRectB() const;
  const DPRegion::Region& getRegionRectE() const;
  const DPRegion::Region& getRegionRectO() const;
  const DPRegion::Region& getRegionRectI() const;
  const DPRegion::Region& getRegion1(const int& bfType) const;
  const DPRegion::Region& getRegion2(const int& bfType) const;

  const PairNode& vpairX_ref(int bx, int ex) const {
    return vpairX.ref(bx, ex);
  }
  const PairNode& vpairY_ref(int by, int ey) const {
    return vpairY.ref(by, ey);
  }
  const BENode& vbe_ref(int x, int y) const;
  const ScoreManager::SCNode& vsc_ref(int diag, int ex, int ey) const;
  const ScoreManager::SCNode& vsfbwB_refI(int by, int ex, int ey) const;
  const ScoreManager::SCNode& vsfbwB_refO(int by, int ex, int ey) const;
  const ScoreManager::SCNode& vsfbwE_refI(int bx, int by, int ey) const;
  const ScoreManager::SCNode& vsfbwE_refO(int bx, int by, int ey) const;

  const DPScoreT& vst_rect1_ref(int bfType,
				int bx, int by, int ex, int ey) const;
  const DPScoreT& vst_rect2_ref(int bfType,
				int bx, int by, int ex, int ey) const;
#ifdef SLR_STATE
  const DPScoreT& vstb_ref(int bx, int by, int ex, int ey, State s) const;
  const DPScoreT& vste_ref(int bx, int by, int ex, int ey, State s) const;
#else
  const DPScoreT& vstb_ref(int bx, int by, int ex, int ey) const;
  const DPScoreT& vste_ref(int bx, int by, int ex, int ey) const;
#endif
  const int& vskip_skipCountB0_ref(int x) const;
  const int& vskip_skipCountE0_ref(int x) const;
  const int& vskip_skipCountM0_ref(int x) const;
  const int& vskip_skipCount0_ref(int x) const;
  const int& vskip_skipCountB_ref(int x, int y) const;
  const int& vskip_skipShiftB_ref(int x, int y) const;
  const int& vskip_skipCountE_ref(int x, int y) const;
  const int& vskip_skipShiftE_ref(int x, int y) const;
  const TBNode& vtb_ref(int bx, int by, int ex, int ey) const;
  const TBBFNode& vtbbf_ref(int bx, int by, int ex, int ey) const;
  const ScoreManager::SCNode& vfw_ref(int bx, int by,
				      int ex, int ey) const;
  const ScoreManager::SCNode& vbw_ref(int bx, int by,
				      int ex, int ey) const;
private:
  typedef DPRegion::Region Region;
  enum RType {
    TRIANGLE     = DPRegion::TRIANGLE,
    CARET        = DPRegion::CARET,
    NOSE         = DPRegion::NOSE,
    TRIANGLE_CUT = DPRegion::TRIANGLE_CUT,
    CARET_CUT_L  = DPRegion::CARET_CUT_L,
    CARET_CUT_R  = DPRegion::CARET_CUT_R,
    INVALID_R_TYPE = DPRegion::INVALID_R_TYPE
  };
  enum RShape {
    TRI_SHAPE       = DPRegion::TRI_SHAPE,
    RECT_VERT_SHAPE = DPRegion::RECT_VERT_SHAPE,
    RECT_DIAG_SHAPE = DPRegion::RECT_DIAG_SHAPE
  };
  enum RNum {
    REGION_EE    = DPRegion::REGION_EE,
    REGION_BB    = DPRegion::REGION_BB, 
    REGION_BE    = DPRegion::REGION_BE,
    REGION_ME    = DPRegion::REGION_ME, 
    REGION_BM    = DPRegion::REGION_BM, 
    REGION_MM    = DPRegion::REGION_MM
  };

  
  void initData();

  void resizeVectors();
  void initVTZ();
  void initVXY();
  void initVPair();
  void initVSC();
  void initVSkip();
  void initVBE();
  void initVBE_SC();
  void initVBE_ESC();
  void initVBE_TriIn();
  void initVBE_TriOut();
  void initVBE_RectIn(const Region& regionTriB,
		      const Region& regionTriE);
  void initVBE_RectIn_M(const Region& regionRectB,
			const Region& regionRectE);
  void initVBE_RectOut(const Region& regionTriB,
		       const Region& regionTriE);
  void initVBE_RectOut_L(const Region& regionRectO,
			 const Region& regionRectI);
  void initVBE_RectOut_R(const Region& regionRectO,
			 const Region& regionRectI);

  void clean();
  void initState0();
  void initStateLim();
  void initSCStore();
  void initTau();
  void initTZNodes();
  void initBENodes();
  void initSkip();
  int getSkipRowCountM0() const;

  bool onlyToLoadVST() const;
  bool toLoadVSTB() const;
  bool toLoadVSTE() const;
  void loadVST();

  SubTask();
  SubTask(const SubTask& subTask);
  SubTask& operator=(const SubTask& subTask);
};

#endif
