/*
 * DPDeckPSCFGNodes.hpp
 *
 */
#ifndef DP_DECK_PSCFG_NODES_HPP
#define DP_DECK_PSCFG_NODES_HPP

#include "DPDeckPSCFG.hpp"
#include "PSCFGEnergyScoreManager.hpp"
// #include "PSCFGBaseScoreManager.hpp"

struct DPDeckPSCFG::TZNode {
  int type;
  int topLeftMargin;
  int bottomRightMargin;
  XNode* bx;
  YNode* by;
  XNode* ex;
  YNode* ey;
  YNode* bxey;
  YNode* exby;

  string toString() const;
  static const TZNode& null() {
    static const TZNode nullNode = {
      0,
      0,
      0,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL,
      NULL
    };
    return nullNode;
  }
};

struct DPDeckPSCFG::XNode {
  ScoreManager::XCol xCol;
  PairNode* pn_beg;
  static const XNode& null() {
    static const XNode nullNode = {
      ScoreManager::XCol::null(),
      NULL
    };
    return nullNode;
  }
  string toString() const;
};

struct DPDeckPSCFG::YNode {
  ScoreManager::SCNode scMP;
  int* pSkipCountB;
  int* pSkipShiftB;
  int* pSkipCountE;
  int* pSkipShiftE;
  int* pSkipCountE0;

  int bxn;// nSkipRow from bxk_beg to the first skipRow for this y
  int exn;// nSkipRow form bxk_beg to the first skipRow for this y

  XNode* ex;
  TZNode* ext;

  ScoreManager::YCol yCol;
  PairNode* pn_beg;
  static const YNode& null() {
    static const YNode nullNode = {
      ScoreManager::SCNode::null(),

      NULL,
      NULL,
      NULL,
      NULL,
      NULL,

      0,
      0,

      NULL,
      NULL,

      ScoreManager::YCol::null(),

      NULL
    };
    return nullNode;
  }
  string toString() const;
};

struct DPDeckPSCFG::PairNode {
  ScoreManager::PairCol pairCol;
  static const PairNode& null() {
    static const PairNode nullNode = {
      ScoreManager::PairCol::null()
    };
    return nullNode;
  }
  string toString() const;
  bool equals(const PairNode& pn) const;
};

struct DPDeckPSCFG::BENode {
  ScoreManager::SCNode* sc;
  DPScoreT* st[4];
  ScoreManager::MatchCol mColB;
  ScoreManager::MatchCol mColE;
  static const BENode& null() {
    static const BENode nullNode = {
      NULL,
      {NULL, NULL, NULL},
      ScoreManager::MatchCol::null(),
      ScoreManager::MatchCol::null()
    };
    return nullNode;
  }

  string toString() const;
  bool equals(const BENode& bn) const;
};

struct DPDeckPSCFG::TBBFNode {
  unsigned int data;
  // area should not exceed 2**(sizeof(unsigned int) - 1) - 1
  // segType should be less than 4
  static const TBBFNode& null() {
    static const TBBFNode nullNode = {0};
    return nullNode;
  }
  static const unsigned int SHIFT = ((sizeof(unsigned int) * 8) - 2);
  static const unsigned int MASK = ~(~0 << SHIFT);
  inline void setNull() {data = 0;}
  void setVal(const unsigned int& bfType, const unsigned int& area) {
    Assert(area >> SHIFT == 0);
    Assert(bfType < 4);
    data = (area | (bfType << SHIFT));
  }
  unsigned int getBFType() const {
    return (data >> SHIFT);
  }
  unsigned int getArea() const {
    return (data & MASK);
  }
};

#endif
