/*
 * DPDeckPSCFGVectors.hpp
 *
 */
#ifndef DP_DECK_PSCFG_VECTORS_HPP
#define DP_DECK_PSCFG_VECTORS_HPP

#include "Util.hpp"
#include "DPDeckPSCFG.hpp"
#include "PSCFGEnergyScoreManager.hpp"
// #include "PSCFGBaseScoreManager.hpp"
#include "DPDeckPSCFGNodes.hpp"
#include "DPDeckPSCFGDPRegion.hpp"

class 
DPDeckPSCFG::
VTZ 
  : private Vector<TZNode> 
{
  typedef TZNode ValueT;
  typedef Vector<ValueT> Super;
public:
  VTZ();
  ~VTZ();
  string toString() const;
  void print() const;

  bool empty() const;
  int size() const;
  static int computeSizeInBytes(const DPRegion::Region& region);
  void resize(const DPRegion::Region& region);
  const ValueT& refB(int i) const {return Super::get(getIndexB(i));}
  ValueT& refB(int i) {return Super::get(getIndexB(i));}
  const ValueT& refE(int i) const {return Super::get(getIndexE(i));}
  ValueT& refE(int i) {return Super::get(getIndexE(i));}
  const ValueT& getHeadB() const {return Super::get(headB);}
  ValueT& getHeadB() {return Super::get(headB);}
  const ValueT& getHeadE() const {return Super::get(headE);}
  ValueT& getHeadE() {return Super::get(headE);}
  const ValueT& getTailB() const {return Super::get(tailB);}
  ValueT& getTailB() {return Super::get(tailB);}
  const ValueT& getTailE() const {return Super::get(tailE);}
  ValueT& getTailE() {return Super::get(tailE);}
private:
  static const int margin_l = 1;
  static const int margin_m = 1;
  static const int margin_r = 1;
  static const int marginSize = (margin_l + margin_m + margin_r);
  int headB;
  int tailB;
  int headE;
  int tailE;
  static int computeSize(const DPRegion::Region& region);
  int getIndexB(int i) const {return (headB + i);}
  int getIndexE(int i) const {return (headE + i);}
};
class 
DPDeckPSCFG::
VX 
: private Vector<XNode> 
{
  typedef Vector<XNode> Super;
  typedef XNode ValueT;
public:
  string toString() const;
  void print() const;

  bool empty() const;
  int size() const;
  static int computeSizeInBytes(const DPRegion::Region& region);
  void resize(const DPRegion::Region& region);
  const ValueT& refB(int y) const {return Super::get(getIndexB(y));}
  ValueT& refB(int y) {return Super::get(getIndexB(y));}
  const ValueT& refE(int y) const {return Super::get(getIndexE(y));}
  ValueT& refE(int y) {return Super::get(getIndexE(y));}
  int getPos(ValueT* const& p) const;
private:
  static const int margin_l = 1;
  static const int margin_r = 1;
  static const int margin_m = 3;
  static const int marginSize = (margin_l + margin_r + margin_m);
  int headB;
  int headE;
  int bx_beg;
  int bx_last;
  int ex_beg;
  int ex_last;

  int getIndexB(int x) const;
  int getIndexE(int x) const;
  static int computeSize(const DPRegion::Region& region);
};
class 
DPDeckPSCFG::
VY 
: private Vector<YNode> 
{
  typedef YNode ValueT;
  typedef Vector<YNode> Super;
public:
  string toString() const;
  void print() const;

  bool empty() const;
  int size() const;
  static int computeSizeInBytes(const DPRegion::Region& region);
  void resize(const DPRegion::Region& region);
  const ValueT& refB(int y) const {return Super::get(getIndexB(y));}
  ValueT& refB(int y) {return Super::get(getIndexB(y));}
  const ValueT& refE(int y) const {return Super::get(getIndexE(y));}
  ValueT& refE(int y) {return Super::get(getIndexE(y));}
  int getPos(ValueT*const& p) const;
private:
  static const int margin_l = 1;
  static const int margin_r = 1;
  static const int margin_m = 3;
  static const int marginSize = (margin_l + margin_r + margin_m);
  int headB;
  int headE;
  int by_beg;
  int by_last;
  int ey_beg;
  int ey_last;

  int getIndexB(int y) const;
  int getIndexE(int y) const;
  static int computeSize(const DPRegion::Region& region);
};
class 
DPDeckPSCFG::
VPairX 
: private Matrix<PairNode, Vector<PairNode> > 
{
  typedef PairNode ValueT;
  typedef Matrix<PairNode, Vector<PairNode> > Super;
public:
  string toString() const;
  void print() const;

  static int computeSizeInBytes(const DPRegion::Region& region);
  void resize(const DPRegion::Region& region);
  const ValueT& ref(int bx, int ex) const;
  ValueT& ref(int bx, int ex);
private:
  static const int rowMarginB = 1;
  static const int rowMarginE = 1;
  static const int colMarginB = 1;
  static const int colMarginE = 1;
  int bx_beg;
  int bx_last;
  int ex_beg;
  int ex_last;
  static void getLengths(const DPRegion::Region& region,
			 int& rowLength, int& colLength);
  static int computeSize(const DPRegion::Region& region);
};

class 
DPDeckPSCFG::
VPairY 
: private Matrix<PairNode, Vector<PairNode> > 
{
  typedef PairNode ValueT;
  typedef Matrix<PairNode, Vector<PairNode> > Super;
public:
  string toString() const;
  void print() const;

  static int computeSizeInBytes(const DPRegion::Region& region);
  void resize(const DPRegion::Region& region);
  const ValueT& ref(int by, int ey) const;
  ValueT& ref(int by, int ey);
private:
  static const int rowMarginB = 1;
  static const int rowMarginE = 1;
  static const int colMarginB = 1;
  static const int colMarginE = 1;
  int by_beg;
  int by_last;
  int ey_beg;
  int ey_last;
  static void getLengths(const DPRegion::Region& region,
			 int& rowLength, int& colLength);
  static int computeSize(const DPRegion::Region& region);
};

class 
DPDeckPSCFG::
VBE 
: private Vector<BENode> 
{
  typedef BENode ValueT;
  typedef Vector<ValueT> Super;
public:
  VBE();
  ~VBE();
  string toString() const;
  void print() const;

  static int computeSizeInBytes(const DPRegion::Region& region);
  void resize(const DPRegion::Region& region);
  const ValueT& ref(const DPRegion::Region& region,
		    const Point& pt) const {
    return Super::get(getIndex(region, pt));
  }
  ValueT& ref(const DPRegion::Region& region, const Point& pt) {
    return Super::get(getIndex(region, pt));
  }
  ValueT& getHeadB() {
    return Super::get(headB);
  }
  ValueT& getHeadE() {
    return Super::get(headE);
  }
  ValueT& getTailB() {
    return Super::get(tailB);
  }
  ValueT& getTailE() {
    return Super::get(tailE);
  }
private:
  int bSize;
  int eSize;
  int headB;
  int tailB;
  int headE;
  int tailE;
  int getIndex(const DPRegion::Region& region, const Point& pt) const {
    return (pt.x <= region.srb.ex0
	    ? getIndexB(region.srb, pt)
	    : getIndexE(region.sre, pt));
  }
  int getIndexB(const StripRegion& srb, const Point& pt) const;
  int getIndexE(const StripRegion& sre, const Point& pt) const;
  static int computeSize(const DPRegion::Region& region);
};

class 
DPDeckPSCFG::
VSC 
: private Vector<ScoreManager::SCNode> 
{
  typedef ScoreManager::SCNode ValueT;
  typedef Vector<ValueT> Super;
public:
  VSC();
  ~VSC();
  string toString() const;

  static int computeSizeInBytes(const DPRegion::Region& region);
  void resize(const DPRegion::Region& region);
  const ValueT& ref(const DPRegion::Region& region, 
	    const int& diagB, const Point& ptE) const {
    return Super::get(getIndex(region, diagB, ptE));
  }
  ValueT& ref(const DPRegion::Region& region,
      const int& diagB, const Point& ptE)  {
    return Super::get(getIndex(region, diagB, ptE));
  }

private:
  int getIndex(const DPRegion::Region& region,
	       const int& diag, const Point& pt) const;
  static int computeSize(const DPRegion::Region& region);
};
  
class 
DPDeckPSCFG::
VSkip 
{
public:
  VSkip();
  ~VSkip();
  string toString() const;
  void print() const;

  static int computeSizeInBytes(const DPRegion::Region& region);
  void resize(const DPRegion::Region& region);
  int& getHeadB0() {return vSkipCount0[headB0];}
  int& getTailB0() {return vSkipCount0[tailB0];}
  int& getHeadE0() {return vSkipCount0[headE0];}
  int& getTailE0() {return vSkipCount0[tailE0];}
  int& getHeadM0() {return vSkipCount0[headM0];}
  int& getTailM0() {return vSkipCount0[tailM0];}
  int& skipCountB0_ref(const DPRegion::Region& region, const int& bx);
  int& skipCountE0_ref(const DPRegion::Region& region, const int& ex);
  int& skipCountM0_ref(const DPRegion::Region& region, const int& mx);
  int& skipCount0_ref(const DPRegion::Region& region, const int& x);
  int& skipCountB_ref(const DPRegion::Region& region, const Point& pt);
  int& skipShiftB_ref(const DPRegion::Region& region, const Point& pt);
  int& skipCountE_ref(const DPRegion::Region& region, const Point& pt);
  int& skipShiftE_ref(const DPRegion::Region& region, const Point& pt);
private:
  int headB0;
  int tailB0;
  int headE0;
  int tailE0;
  int headM0;
  int tailM0;
  Vector<int> vSkipCount0;

  Vector<int> vSkipCountB;
  Vector<int> vSkipShiftB;

  Vector<int> vSkipCountE;
  Vector<int> vSkipShiftE;
};

class 
DPDeckPSCFG::
VSTB 
{
  typedef DPScoreT ValueT;
public:
  string toString() const;
  static int computeSize(const DPRegion::Region& region);
  static int computeSizeInBytes(const DPRegion::Region& region);
  void resize(const DPRegion::Region& region, State s);
  void clear() {
    for (int i = 0; i < (int)data.size(); i++) {
      data[i].clear();
    }
  }
#ifdef SLR_STATE
  const ValueT& ref(const DPRegion::Region& region, const Rect& rect, State s) const  {
    const Datum& datum = data[indexOf(region)];
    return datum.ref(region, rect, s);
  }
  ValueT& ref(const DPRegion::Region& region, const Rect& rect, State s)  {
    Datum& datum = data[indexOf(region)];
    return datum.ref(region, rect, s);
  }
  ValueT& head(const DPRegion::Region& region, State s) {
    return data[indexOf(region)].head(s);
  }
  ValueT& tail(const DPRegion::Region& region, State s) {
    return data[indexOf(region)].tail(s);
  }
  const ValueT& head(const DPRegion::Region& region, State s) const {
    return data[indexOf(region)].head(s);
  }
  const ValueT& tail(const DPRegion::Region& region, State s) const {
    return data[indexOf(region)].tail(s);
  }
  bool isEmpty(const DPRegion::Region& region, State s) const {
    const Datum& datum = data[indexOf(region)];
    return datum.isEmpty(s);
  }
  void copy(const DPRegion::Region& region,
	    const VSTE& vste, const State& s);
  void copy(const DPRegion::Region& region,
	    const VSTB& vstb1, const DPRegion::Region& region1, const State& s);
  void copy(const DPRegion::Region& region,
	    const VSTE& vste1, const DPRegion::Region& region1, const State& s);
#else 
  const ValueT& ref(const DPRegion::Region& region, const Rect& rect) const  {
    const Datum& datum = data[indexOf(region)];
    return datum.ref(region, rect);
  }
  ValueT& ref(const DPRegion::Region& region, const Rect& rect)  {
    Datum& datum = data[indexOf(region)];
    return datum.ref(region, rect);
  }
  ValueT& head(const DPRegion::Region& region) {
    return data[indexOf(region)].front();
  }
  ValueT& tail(const DPRegion::Region& region) {
    return data[indexOf(region)].back();
  }
  const ValueT& head(const DPRegion::Region& region) const {
    return data[indexOf(region)].front();
  }
  const ValueT& tail(const DPRegion::Region& region) const {
    return data[indexOf(region)].back();
  }
  void copy(const DPRegion::Region& region, const VSTE& vste);
  void copy(const DPRegion::Region& region,
	    const VSTB& vstb1, const DPRegion::Region& region1);
  void copy(const DPRegion::Region& region,
	    const VSTE& vste1, const DPRegion::Region& region1);
  bool isEmpty(const DPRegion::Region& region) const {
    const Datum& datum = data[indexOf(region)];
    return datum.isEmpty();
  }
#endif

  
private:
  struct Datum {
    struct EYRange {
      //ValueT* pVal;
      int offset;
      int ey_beg;
#ifdef DEBUG
      int ey_last;
#endif
      //inline ValueT& ref(const int& ey, const int& periodY) const {
      inline int ref(const int& ey, const int& periodY) const {
	Assert(ey_beg <= ey && ey <= ey_last);
	//return pVal[(ey - ey_beg) / periodY];
	return (offset + (ey - ey_beg) / periodY);
      }
    };
    struct EXRange {
      EYRange* pEYRange;
      int ex_beg;
#ifdef DEBUG
      int ex_last;
#endif
      inline EYRange& getEYRange(const int& ex, const int& periodX) const {
	Assert(ex_beg <= ex && ex <= ex_last);
	return pEYRange[(ex - ex_beg) / periodX];
      }
    };
    struct BYRange {
      EXRange* pEXRange;
#ifdef DEBUG
      int by_beg;
#endif
      int by_last;
      inline EXRange& getEXRange(const int& by, const int& periodY) const {
	Assert(by_beg <= by && by <= by_last);
	return pEXRange[(by - by_last) / periodY];
      }
    };
    struct BXRange {
      BYRange* pBYRange;
#ifdef DEBUG
      int bx_beg;
#endif
      int bx_last;
      inline BYRange& getBYRange(const int& bx, const int& periodX) const {
	Assert(bx_beg <= bx && bx <= bx_last);
	return pBYRange[(bx - bx_last) / periodX];
      }
    };

    int periodX;
    int periodY;
    BXRange bxRange;
    Vector<BYRange> vBYRange;
    Vector<EXRange> vEXRange;
    Vector<EYRange> vEYRange;
#ifdef SLR_STATE
    enum {
      NST_STATE = 3
    };
#else
    enum {
      NST_STATE = 1
    };
#endif
    CArray<Vector<ValueT>, NST_STATE> v;
    CArray<ValueT*, NST_STATE> v_ptrs;

    void resize(const DPRegion::Region& region, State s);
#ifdef SLR_STATE
    const ValueT& ref(const DPRegion::Region& region,
		      const Rect& rect, const State& s) const;
    ValueT& ref(const DPRegion::Region& region,
		const Rect& rect, const State& s);
    const ValueT& head(const State& s) const {
      return v[Model::getSTStateIndex(s)].front();
    }
    ValueT& head(const State& s) {
      return v[Model::getSTStateIndex(s)].front();
    }
    const ValueT& tail(const State& s) const {
      return v[Model::getSTStateIndex(s)].back();
    }
    ValueT& tail(const State& s) {
      return v[Model::getSTStateIndex(s)].back();
    }
    bool isEmpty(State s) const {
      return v[Model::getSTStateIndex(s)].empty();
    }
#else
    const ValueT& ref(const DPRegion::Region& region,
		      const Rect& rect) const;
    ValueT& ref(const DPRegion::Region& region,
		const Rect& rect);
    const ValueT& front() const {return v[0].front();}
    ValueT& front() {return v[0].front();}
    const ValueT& back() const {return v[0].back();}
    ValueT& back() {return v[0].back();}
    bool isEmpty() const {
      return v[0].empty();
    }
#endif
    void clear() {
      periodX = 0;
      periodY = 0;
      vBYRange.clear();
      vEXRange.clear();
      vEYRange.clear();
      for (int i = 0; i < (int)NST_STATE; i++) {
	v[i].clear();
      }
    }
  };
  //typedef CArray<Vector<ValueT>, DPRegion::RMAX> Data;
  typedef CArray<Datum, DPRegion::RMAX> Data;
  Data data;

  int indexOf(const DPRegion::Region& region) const {
    return DPRegion::regionNum2Index(region.regionNum);
  }
};
class 
DPDeckPSCFG::
VSTE
{
  typedef DPScoreT ValueT;
public:
  string toString() const;
  static int computeSize(const DPRegion::Region& region);
  static int computeSizeInBytes(const DPRegion::Region& region);
  void resize(const DPRegion::Region& region, State s);
  void clear() {
    for (int i = 0; i < (int)data.size(); i++) {
      data[i].clear();
    }
  }
#ifdef SLR_STATE
  const ValueT& ref(const DPRegion::Region& region, const Rect& rect, State s) const  {
    const Datum& datum = data[indexOf(region)];
    return datum.ref(region, rect, s);
  }
  ValueT& ref(const DPRegion::Region& region, const Rect& rect, State s)  {
    Datum& datum = data[indexOf(region)];
    return datum.ref(region, rect, s);
  }
  ValueT& head(const DPRegion::Region& region, State s) {
    return data[indexOf(region)].head(s);
  }
  ValueT& tail(const DPRegion::Region& region, State s) {
    return data[indexOf(region)].tail(s);
  }
  const ValueT& head(const DPRegion::Region& region, State s) const {
    return data[indexOf(region)].head(s);
  }
  const ValueT& tail(const DPRegion::Region& region, State s) const {
    return data[indexOf(region)].tail(s);
  }
  bool isEmpty(const DPRegion::Region& region, State s) const {
    const Datum& datum = data[indexOf(region)];
    return datum.isEmpty(s);
  }
  void copy(const DPRegion::Region& region,
	    const VSTB& vstb, const State& s);
  void copy(const DPRegion::Region& region,
	    const VSTE& vste1, const DPRegion::Region& region1, const State& s);
  void copy(const DPRegion::Region& region,
	    const VSTB& vstb1, const DPRegion::Region& region1, const State& s);
#else 
  const ValueT& ref(const DPRegion::Region& region, const Rect& rect) const  {
    const Datum& datum = data[indexOf(region)];
    return datum.ref(region, rect);
  }
  ValueT& ref(const DPRegion::Region& region, const Rect& rect)  {
    Datum& datum = data[indexOf(region)];
    return datum.ref(region, rect);
  }
  ValueT& head(const DPRegion::Region& region) {
    return data[indexOf(region)].front();
  }
  ValueT& tail(const DPRegion::Region& region) {
    return data[indexOf(region)].back();
  }
  const ValueT& head(const DPRegion::Region& region) const {
    return data[indexOf(region)].front();
  }
  const ValueT& tail(const DPRegion::Region& region) const {
    return data[indexOf(region)].back();
  }
  void copy(const DPRegion::Region& region, const VSTB& vstb);
  void copy(const DPRegion::Region& region,
	    const VSTE& vste1, const DPRegion::Region& region1);
  void copy(const DPRegion::Region& region,
	    const VSTB& vstb1, const DPRegion::Region& region1);
  bool isEmpty(const DPRegion::Region& region) const {
    const Datum& datum = data[indexOf(region)];
    return datum.isEmpty();
  }
#endif

private:
  struct Datum {
    struct BYRange {
      //ValueT* pVal;
      int offset;
      int by_beg;
#ifdef DEBUG
      int by_last;
#endif
      inline int ref(const int& by, const int& periodY) const {
	Assert(by_beg <= by && by <= by_last);
	//return pVal[(by - by_beg) / periodY];
	return (offset + (by - by_beg) / periodY);
      }
    };
    struct BXRange {
      BYRange* pBYRange;
      int bx_beg;
#ifdef DEBUG
      int bx_last;
#endif
      inline BYRange& getBYRange(const int& bx, const int& periodX) const {
	Assert(bx_beg <= bx && bx <= bx_last);
	return pBYRange[(bx - bx_beg) / periodX];
      }
    };
    struct EYRange {
      BXRange* pBXRange;
      int ey_beg;
#ifdef DEBUG
      int ey_last;
#endif
      inline BXRange& getBXRange(const int& ey, const int& periodY) const {
	Assert(ey_beg <= ey && ey <= ey_last);
	return pBXRange[(ey - ey_beg) / periodY];
      }
    };
    struct EXRange {
      EYRange* pEYRange;
      int ex_beg;
#ifdef DEBUG
      int ex_last;
#endif
      inline EYRange& getEYRange(const int& ex, const int& periodX) const {
	Assert(ex_beg <= ex && ex <= ex_last);
	return pEYRange[(ex - ex_beg) / periodX];
      }
    };

    int periodX;
    int periodY;
    EXRange exRange;
    Vector<EYRange> vEYRange;
    Vector<BXRange> vBXRange;
    Vector<BYRange> vBYRange;
#ifdef SLR_STATE
    enum {
      NST_STATE = 3
    };
#else
    enum {
      NST_STATE = 1
    };
#endif
    CArray<Vector<ValueT>, NST_STATE> v;
    CArray<ValueT*, NST_STATE> v_ptrs;

    void resize(const DPRegion::Region& region, State s);
#ifdef SLR_STATE
    const ValueT& ref(const DPRegion::Region& region,
		      const Rect& rect, const State& s) const;
    ValueT& ref(const DPRegion::Region& region,
		const Rect& rect, const State& s);
    const ValueT& head(const State& s) const {
      return v[Model::getSTStateIndex(s)].front();
    }
    ValueT& head(const State& s) {
      return v[Model::getSTStateIndex(s)].front();
    }
    const ValueT& tail(const State& s) const {
      return v[Model::getSTStateIndex(s)].back();
    }
    ValueT& tail(const State& s) {
      return v[Model::getSTStateIndex(s)].back();
    }
    bool isEmpty(State s) const {
      return v[Model::getSTStateIndex(s)].empty();
    }
#else
    const ValueT& ref(const DPRegion::Region& region,
		      const Rect& rect) const;
    ValueT& ref(const DPRegion::Region& region,
		const Rect& rect);
    const ValueT& front() const {return v[0].front();}
    ValueT& front() {return v[0].front();}
    const ValueT& back() const {return v[0].back();}
    ValueT& back() {return v[0].back();}
    bool isEmpty() const {
      return v[0].empty();
    }
#endif
    void clear() {
      periodX = 0;
      periodY = 0;
      vEYRange.clear();
      vBXRange.clear();
      vBYRange.clear();
      for (int i = 0; i < (int)NST_STATE; i++) {
	v[i].clear();
      }
    }
  };
  //typedef CArray<Vector<ValueT>, DPRegion::RMAX> Data;
  typedef CArray<Datum, DPRegion::RMAX> Data;
  Data data;
  int indexOf(const DPRegion::Region& region) const {
    return DPRegion::regionNum2Index(region.regionNum);
  }
};
class 
DPDeckPSCFG::
VTB 
{
  typedef TBNode ValueT;
public:
  string toString() const;
  void print() const;
  static int computeSizeInBytes(const DPRegion::Region& region);
  void resize(const DPRegion::Region& region);
  const ValueT& ref(const DPRegion::Region& region,
		    const Rect& rect) const  {
    const Datum& datum = data[indexOf(region)];
    return datum.ref(region, rect);
  }
  ValueT& ref(const DPRegion::Region& region, const Rect& rect)  {
    Datum& datum = data[indexOf(region)];
    return datum.ref(region, rect);
  }
  ValueT& head(const DPRegion::Region& region) {
    return data[indexOf(region)].front();
  }
  ValueT& tail(const DPRegion::Region& region) {
    return data[indexOf(region)].back();
  }
private:
  struct Datum {
    struct EYRange {
      ValueT* pVal;
      int ey_beg;
#ifdef DEBUG
      int ey_last;
#endif
      inline ValueT& ref(const int& ey) const {
	Assert(ey_beg <= ey && ey <= ey_last);
	return pVal[ey - ey_beg];
      }
    };
    struct EXRange {
      EYRange* pEYRange;
      int ex_beg;
#ifdef DEBUG
      int ex_last;
#endif
      inline EYRange& getEYRange(const int& ex) const {
	Assert(ex_beg <= ex && ex <= ex_last);
	return pEYRange[ex - ex_beg];
      }
    };
    struct BYRange {
      EXRange* pEXRange;
#ifdef DEBUG
      int by_beg;
#endif
      int by_last;
      inline EXRange& getEXRange(const int& by) const {
	Assert(by_beg <= by && by <= by_last);
	return pEXRange[by - by_last];
      }
    };
    struct BXRange {
      BYRange* pBYRange;
#ifdef DEBUG
      int bx_beg;
#endif
      int bx_last;
      inline BYRange& getBYRange(const int& bx) const {
	Assert(bx_beg <= bx && bx <= bx_last);
	return pBYRange[bx - bx_last];
      }
    };

    BXRange bxRange;
    Vector<BYRange> vBYRange;
    Vector<EXRange> vEXRange;
    Vector<EYRange> vEYRange;
    Vector<ValueT> v;

    void resize(const DPRegion::Region& region);
    const ValueT& ref(const DPRegion::Region& region,
		      const Rect& rect) const;
    ValueT& ref(const DPRegion::Region& region,
		const Rect& rect);
    const ValueT& front() const {return v.front();}
    ValueT& front() {return v.front();}
    const ValueT& back() const {return v.back();}
    ValueT& back() {return v.back();}
  };
  typedef CArray<Datum, DPRegion::RMAX> Data;
  Data data;
  int indexOf(const DPRegion::Region& region) const {
    return DPRegion::regionNum2Index(region.regionNum);
  }
  static int computeSize(const DPRegion::Region& region);
};

class 
DPDeckPSCFG::
VTBBF 
{
  typedef TBBFNode ValueT;
public:
  string toString() const;
  void resize(const DPRegion::Region& region);
  static int computeSizeInBytes(const DPRegion::Region& region);
  const  ValueT& ref(const DPRegion::Region& region, 
		     const Rect& rect) const {
    return data[indexOf(region)][region.getVSTBIndex(rect)];
  }
  ValueT& ref(const DPRegion::Region& region, const Rect& rect) {
    return data[indexOf(region)][region.getVSTBIndex(rect)];
  }
  ValueT& head(const DPRegion::Region& region) {
    return data[indexOf(region)].front();
  }
  ValueT& tail(const DPRegion::Region& region) {
    return data[indexOf(region)].back();
  }
private:
  typedef CArray<Vector<ValueT>, DPRegion::RMAX> Data;
  Data data;
  int indexOf(const DPRegion::Region& region) const {
    return DPRegion::regionNum2Index(region.regionNum);
  }
  static int computeSize(const DPRegion::Region& region);
};

class 
DPDeckPSCFG::
VSFBWB 
{
  typedef ScoreManager::SCNode ValueT;
public:
  string toString() const;
  static int computeSizeInBytes(const DPRegion::Region& region);
  void resizeI(const DPRegion::Region& region);
  void resizeO(const DPRegion::Region& region);
  ValueT& head(const DPRegion::Region& region) {
    return data[indexOf(region)].front();
  }
  ValueT& tail(const DPRegion::Region& region) {
    return data[indexOf(region)].back();
  }
  ValueT& refI(const DPRegion::Region& region,
	       const int& by, const int& ex, const int& ey) {
    return data[indexOf(region)][region.getVSFBWBIndexI(by, ex, ey)];
  }
  ValueT& refO(const DPRegion::Region& region,
	       const int& by, const int& ex, const int& ey) {
    return data[indexOf(region)][region.getVSFBWBIndexO(by, ex, ey)];
  }
private:
  typedef CArray<Vector<ValueT>, DPRegion::RMAX> Data;
  Data data;
  int indexOf(const DPRegion::Region& region) const {
    return DPRegion::regionNum2Index(region.regionNum);
  }
};

class 
DPDeckPSCFG::
VSFBWE 
{
  typedef ScoreManager::SCNode ValueT;
public:
  string toString() const;
  static int computeSizeInBytes(const DPRegion::Region& region);
  int size(const DPRegion::Region& region) const {
    return data[indexOf(region)].size();
  }
  void resizeI(const DPRegion::Region& region);
  void resizeO(const DPRegion::Region& region);
  ValueT& head(const DPRegion::Region& region) {
    return data[indexOf(region)].front();
  }
  ValueT& tail(const DPRegion::Region& region) {
    return data[indexOf(region)].back();
  }
  ValueT& refI(const DPRegion::Region& region,
	       const int& bx, const int& by, const int& ey) {
    return data[indexOf(region)][region.getVSFBWEIndexI(bx, by, ey)];
  }
  ValueT& refO(const DPRegion::Region& region,
	       const int& bx, const int& by, const int& ey) {
    return data[indexOf(region)][region.getVSFBWEIndexO(bx, by, ey)];
  }
private:
  typedef CArray<Vector<ValueT>, DPRegion::RMAX> Data;
  Data data;
  int indexOf(const DPRegion::Region& region) const {
    const int& i = DPRegion::regionNum2Index(region.regionNum);
    return i;
  }
};

class 
DPDeckPSCFG::
VFBW 
{
  typedef ScoreManager::SCNode ValueT;
public:
  string toString() const;
  void print() const;
  static int computeSizeInBytes(const DPRegion::Region& region);
  void resize(const DPRegion::Region& region);
  const ValueT& ref(const DPRegion::Region& region,
		    const Rect& rect) const  {
    const Datum& datum = data[indexOf(region)];
    return datum.ref(region, rect);
  }
  ValueT& ref(const DPRegion::Region& region, const Rect& rect)  {
    Datum& datum = data[indexOf(region)];
    return datum.ref(region, rect);
  }
  ValueT& head(const DPRegion::Region& region) {
    return data[indexOf(region)].front();
  }
  ValueT& tail(const DPRegion::Region& region) {
    return data[indexOf(region)].back();
  }
private:
  struct Datum {
    struct EYRange {
      ValueT* pVal;
      int ey_beg;
      int ey_last;
      inline ValueT& ref(const int& ey) const {
	Assert(ey_beg <= ey && ey <= ey_last);
	return pVal[ey - ey_beg];
      }
    };
    struct EXRange {
      EYRange* pEYRange;
      int ex_beg;
      int ex_last;
      inline EYRange& getEYRange(const int& ex) const {
	Assert(ex_beg <= ex && ex <= ex_last);
	return pEYRange[ex - ex_beg];
      }
    };
    struct BYRange {
      EXRange* pEXRange;
      int by_beg;
      int by_last;
      inline EXRange& getEXRange(const int& by) const {
	Assert(by_beg <= by && by <= by_last);
	return pEXRange[by - by_last];
      }
    };
    struct BXRange {
      BYRange* pBYRange;
      int bx_beg;
      int bx_last;
      inline BYRange& getBYRange(const int& bx) const {
	Assert(bx_beg <= bx && bx <= bx_last);
	return pBYRange[bx - bx_last];
      }
    };

    BXRange bxRange;
    Vector<BYRange> vBYRange;
    Vector<EXRange> vEXRange;
    Vector<EYRange> vEYRange;
    Vector<ValueT> v;

    void resize(const DPRegion::Region& region);
    const ValueT& ref(const DPRegion::Region& region,
		      const Rect& rect) const;
    ValueT& ref(const DPRegion::Region& region,
		const Rect& rect);
    const ValueT& front() const {return v.front();}
    ValueT& front() {return v.front();}
    const ValueT& back() const {return v.back();}
    ValueT& back() {return v.back();}
  };
  typedef CArray<Datum, DPRegion::RMAX> Data;
  Data data;

  int indexOf(const DPRegion::Region& region) const {
    return DPRegion::regionNum2Index(region.regionNum);
  }
};

class
DPDeckPSCFG::
EscVec
{
public:
  const ScoreManager& scoreManager;
  const VX& vx;
  const VY& vy;
  const VPairX& vpairX;
  const VPairY& vpairY;

  Rect rect;
  const ScoreManager::XCol* xColB;
  const ScoreManager::YCol* yColB;
  const ScoreManager::XCol* xColE;
  const ScoreManager::YCol* yColE;
  const ScoreManager::PairCol* xPair;
  const ScoreManager::PairCol* yPair;
  ScoreManager::MatchCol mColB;
  ScoreManager::MatchCol mColE;
  ScoreManager::ESCTable escTable;

  EscVec(const ScoreManager& scoreManager,
	 const VX& vx, const VY& vy, 
	 const VPairX& vpairX, const VPairY& vpairY)
    : scoreManager(scoreManager), 
      vx(vx),
      vy(vy),
      vpairX(vpairX),
      vpairY(vpairY),
      rect(Rect::null()),
      xColB(NULL),
      yColB(NULL),
      xColE(NULL),
      yColE(NULL),
      xPair(NULL),
      yPair(NULL),
      mColB(),
      mColE(),
      escTable(scoreManager)
  {
    escTable.reset(scoreManager);
  }
  EscVec(const EscVec& escv)
    : scoreManager(escv.scoreManager),
      vx(escv.vx),
      vy(escv.vy),
      vpairX(escv.vpairX),
      vpairY(escv.vpairY),
      rect(escv.rect),
      xColB(escv.xColB),
      yColB(escv.yColB),
      xColE(escv.xColE),
      yColE(escv.yColE),
      xPair(escv.xPair),
      yPair(escv.yPair),
      mColB(escv.mColB),
      mColE(escv.mColE),
      escTable(scoreManager) {
    escTable.reset(scoreManager);
  }
  ~EscVec() {
  }
  void reset(const Rect& rect0) {
    const ScoreManager& sm = scoreManager;
    rect = rect0;
    xColB = &vx.refB(rect.bx).xCol;
    yColB = &vy.refB(rect.by).yCol;
    xColE = &vx.refE(rect.ex - 1).xCol;
    yColE = &vy.refE(rect.ey - 1).yCol;
    xPair = &vpairX.ref(rect.bx, rect.ex - 1).pairCol;
    yPair = &vpairY.ref(rect.by, rect.ey - 1).pairCol;
    // currently vbe.ref needs region information so cannot include
    // without reference to region
    sm.getMatchCol(mColB, rect.bx, rect.by);
    sm.getMatchCol(mColE, rect.ex - 1, rect.ey - 1);
  }
  DPScoreT getVal(const State& s) const {
#if 0
#define ESC(result, s, xColB, yColB, xColE,yColE,xPair,yPair,mColB,mColE,escTable) \
  PSCFG_ENERGY_SCORE_MANAGER_ESC(result, s, xColB, yColB, xColE,yColE,xPair,yPair,mColB,mColE,escTable)
#else
#define ESC(result, s, xColB, yColB, xColE,yColE,xPair,yPair,mColB,mColE,escTable) \
  (result = scoreManager.esc(s, xColB, yColB, xColE,yColE,xPair,yPair,mColB,mColE,escTable))
#endif
    DPScoreT result;
    ESC(result, s, (*xColB), (*yColB), (*xColE), (*yColE),
	(*xPair), (*yPair), mColB, mColE, escTable);
    return result;
#undef ESC
  }
  DPScoreT getVal(const Rect& rect0, const State& s0) {
    reset(rect0);
    return getVal(s0);
  }
};
class
DPDeckPSCFG::
TscVec
{
public:
  const ScoreManager& scoreManager;
  const VX& vx;
  const VY& vy;
  const VPairX& vpairX;
  const VPairY& vpairY;

  Rect rect;
  State sp;
  const ScoreManager::XCol* xColB;
  const ScoreManager::YCol* yColB;
  const ScoreManager::XCol* xColE;
  const ScoreManager::YCol* yColE;
  const ScoreManager::PairCol* xPair;
  const ScoreManager::PairCol* yPair;
#ifdef USE_PSCFG_ENERGY
  const ScoreManager::PairCol* xPairR;
  const ScoreManager::PairCol* yPairR;
  const ScoreManager::PairCol* xPairL;
  const ScoreManager::PairCol* yPairL;
#endif
  ScoreManager::TSCTable tscTable;
  Rect rect1;

  TscVec(const ScoreManager& scoreManager,
	 const VX& vx, const VY& vy, 
	 const VPairX& vpairX, const VPairY& vpairY) 
    : scoreManager(scoreManager), 
      vx(vx),
      vy(vy),
      vpairX(vpairX),
      vpairY(vpairY),
      rect(Rect::null()),
      sp(Model::X_st),
      xColB(NULL),
      yColB(NULL),
      xColE(NULL),
      yColE(NULL),
      xPair(NULL),
      yPair(NULL),
#ifdef USE_PSCFG_ENERGY
      xPairR(NULL),
      yPairR(NULL),
      xPairL(NULL),
      yPairL(NULL),
#endif
      tscTable(scoreManager),
      rect1(Rect::null()) {
    tscTable.reset(scoreManager);
  }
  TscVec(const TscVec& tscv)
    : scoreManager(tscv.scoreManager),
      vx(tscv.vx),
      vy(tscv.vy),
      vpairX(tscv.vpairX),
      vpairY(tscv.vpairY),
      rect(tscv.rect),
      xColB(tscv.xColB),
      yColB(tscv.yColB),
      xColE(tscv.xColE),
      yColE(tscv.yColE),
      xPair(tscv.xPair),
      yPair(tscv.yPair),
#ifdef USE_PSCFG_ENERGY
      xPairR(tscv.xPairR),
      yPairR(tscv.yPairR),
      xPairL(tscv.xPairL),
      yPairL(tscv.yPairL),
#endif
      tscTable(tscv.tscTable),
      rect1(tscv.rect1) {
    tscTable.reset(scoreManager);
  }
  ~TscVec() {
  }
  void reset(const Rect& rect0) {
    rect = rect0;
    sp = Model::X_st;
    xPair = &vpairX.ref(rect.bx, rect.ex - 1).pairCol;
    yPair = &vpairY.ref(rect.by, rect.ey - 1).pairCol;
#ifdef USE_PSCFG_ENERGY
    xPairR = &(&vpairX.ref(rect.bx, rect.ex - 1) - 1)->pairCol;
    yPairR = &(&vpairY.ref(rect.by, rect.ey - 1) - 1)->pairCol;
    xPairL = &(&vpairX.ref(rect.bx + 1, rect.ex) - 1)->pairCol;
    yPairL = &(&vpairY.ref(rect.by + 1, rect.ey) - 1)->pairCol;
#endif
  }
  DPScoreT getVal(const State& sp0, const State& sc0) {
#ifndef USE_PSCFG_ENERGY
#define TSC(result,sp,sc,xColB,yColB,xColE,yColE,xPair,yPair,xPairR,yPairR,xPairL,yPairL,tscTable) \
  PSCFG_ENERGY_SCORE_MANAGER_TSC(result,sp,sc,xColB,yColB,xColE,yColE,xPair,yPair,xPairR,yPairR,xPairL,yPairL,tscTable)
#else
#define TSC(result,sp,sc,xColB,yColB,xColE,yColE,xPair,yPair,xPairR,yPairR,xPairL,yPairL,tscTable) \
  (result = scoreManager.tsc(result,sp,sc,xColB,yColB,xColE,yColE,xPair,yPair,xPairR,yPairR,xPairL,yPairL,tscTable))
#endif
    if (sp != sp0) {
      sp = sp0;
      rect1 = ScoreManager::model.addNEmit(rect, sp);
      xColB = &vx.refB(rect1.bx).xCol;
      yColB = &vy.refB(rect1.by).yCol;
      xColE = &vx.refE(rect1.ex).xCol;
      yColE = &vy.refE(rect1.ey).yCol;
    }
    DPScoreT result;
    TSC(result, sp, sc0, (*xColB), (*yColB), (*xColE), (*yColE),
	(*xPair), (*yPair), (*xPairR), (*yPairR), (*xPairL), (*yPairL),
	tscTable);
    return result;
#undef TSC
  }
  DPScoreT getVal(const Rect& rect0, const State& sp0, const State& sc0) {
    reset(rect0);
    return getVal(sp0, sc0);
  }
};


#endif
