/*
 * PSCFGEnergyScoreManager.hpp
 *
 */
#ifndef PSCFG_ENERGY_SCORE_MANAGER_HPP
#define PSCFG_ENERGY_SCORE_MANAGER_HPP

#include "PSCFG_Energy.hpp"
#include "Alpha.hpp"
#include "DPScore.hpp"
#include "ScoreParam.hpp"
#include "Profile.hpp"
#include "DPDeck.hpp"

#define PSCFG_ETSC0
//#define PSCFG_TURNER
#define PSCFG_MATRIX_SCALE
// #define PSCFG_PHYSCORE
//#define PSCFG_SMTX_SIM
//#define USE_DIRICHLET_PRIOR
//#define PSCFG_SMTX_NSEQ
//#define PSCFG_USE_STACK_MTX
//#define PSCFG_SMTX_ASYM
//#define PSCFG_USE_TURNER_STACK_SCALE_SIM
#define USE_PAIR_SMTX_MEA

#ifdef PSCFG_PHYSCORE
#include "PhylogenicScore.hpp"
#endif

#define PSCFG_NUSSINOV

class Weight;
class GuideTree;
class MSA;
class Profile;
class Rect;
class DPPath;
class AlignProbData;
class SeqIds;
class PairProbData;
class AlignProbData;

class 
PSCFG_Energy::
ScoreManager 
  : public DPScore 
{
public:
  using DPScore::DPScoreT;
  static const DPScoreT NegInfDPScore;
  static const DPScoreT ZeroDPScore;
  inline static bool IsFinite(const DPScoreT& z) {
    return DPScore::IsFiniteDPScore(z);
  }
  typedef PSCFG_Energy Model;
  typedef Model::State State;
  typedef Model::EmitModel EmitModel;
  typedef Model::EmitState EmitState;
  typedef DPDeck::ScoreMode ScoreMode;
  typedef Array<float> Vec;
  typedef NMatrix<float> Mtx;
  typedef UpTriMtx<float> PairProb;
  static const Model& model;

  //typedef Vec FVec;
  //typedef DPScore::FVecSimple FVec;
  typedef DPScore::FVecLogF FVec;
  typedef CArray<float, Alpha::REDUCED_NPBASEN> FPair;

  class FeatureCounter;
  class FeatureStats;
  class ESCTable;
  class TSCTable;
  struct XCol;
  struct YCol;
  struct PairCol;
  struct MatchCol;
  struct GapScore;
  struct SCNode;
  class Esc;
  class Tsc;
  typedef ScoreParam::Type Type;
  enum Param {
#ifdef PSCFG_ETSC0
    ESC0               = ScoreParam::PSCFG_ESC0,
    TSC0               = ScoreParam::PSCFG_TSC0,
#endif

    PAIR_PROB_CUTOFF   = ScoreParam::PSCFG_PAIR_PROB_CUTOFF   ,
    PAIR_PROB_SCALE    = ScoreParam::PSCFG_PAIR_PROB_SCALE    ,
    PAIR_PROB_COEFF    = ScoreParam::PSCFG_PAIR_PROB_COEFF    ,
    PAIR_PROB_NSEQ_COEFF = ScoreParam::PSCFG_PAIR_PROB_NSEQ_COEFF,
    LOOP_PROB_SCALE    = ScoreParam::PSCFG_LOOP_PROB_SCALE    ,
    LOOP_PROB_COEFF    = ScoreParam::PSCFG_LOOP_PROB_COEFF    ,
    ALN_PROB_CUTOFF    = ScoreParam::PSCFG_ALN_PROB_CUTOFF    ,
    ALN_PROB_SCALE     = ScoreParam::PSCFG_ALN_PROB_SCALE     ,
    ALN_PROB_COEFF     = ScoreParam::PSCFG_ALN_PROB_COEFF     ,
    ALN_PROB_STEM_SCALE     = ScoreParam::PSCFG_ALN_PROB_STEM_SCALE     ,
    ALN_PROB_STEM_COEFF     = ScoreParam::PSCFG_ALN_PROB_STEM_COEFF     ,
    GAP_PROB_SCALE     = ScoreParam::PSCFG_GAP_PROB_SCALE     ,
    GAP_PROB_COEFF     = ScoreParam::PSCFG_GAP_PROB_COEFF     ,

    ESC_LOOP           = ScoreParam::PSCFG_ESC_LOOP           ,
    ESC_STEM           = ScoreParam::PSCFG_ESC_STEM           ,
    ESC_FCOL           = ScoreParam::PSCFG_ESC_FCOL           ,
    ESC_FPAIR          = ScoreParam::PSCFG_ESC_FPAIR          ,
    ESC_LOD_SCORE      = ScoreParam::PSCFG_ESC_LOD_SCORE      ,

    SMTX_LOOP          = ScoreParam::PSCFG_SMTX_LOOP          ,
    PAIR_SMTX          = ScoreParam::PSCFG_PAIR_SMTX          ,
    SMTX_LOOP_SIM      = ScoreParam::PSCFG_SMTX_LOOP_SIM      ,
    PAIR_SMTX_SIM      = ScoreParam::PSCFG_PAIR_SMTX_SIM      ,
    SMTX_LOOP_SIM_SCALE= ScoreParam::PSCFG_SMTX_LOOP_SIM_SCALE,
    PAIR_SMTX_SIM_SCALE= ScoreParam::PSCFG_PAIR_SMTX_SIM_SCALE,
    SMTX_LOOP_NSEQ     = ScoreParam::PSCFG_SMTX_LOOP_NSEQ     ,
    PAIR_SMTX_NSEQ     = ScoreParam::PSCFG_PAIR_SMTX_NSEQ     ,
    SMTX_LOOP_ASYM     = ScoreParam::PSCFG_SMTX_LOOP_ASYM     ,
    PAIR_SMTX_ASYM     = ScoreParam::PSCFG_PAIR_SMTX_ASYM     ,
    STACK_MTX          = ScoreParam::PSCFG_STACK_MTX          ,
    TSC_MP_LOOP        = ScoreParam::PSCFG_TSC_MP_LOOP        ,
    TERM_STEM          = ScoreParam::PSCFG_TERM_STEM          ,
    MIN_BASE_PAIR      = ScoreParam::PSCFG_MIN_BASE_PAIR      ,
    MIN_CAN_FRAC       = ScoreParam::PSCFG_MIN_CAN_FRAC       ,
    SEQ_SIM_BASE       = ScoreParam::PSCFG_SEQ_SIM_BASE       ,

    GAP_EXTEND         = ScoreParam::PSCFG_GAP_EXTEND         ,
    GAP_EXTEND_OCC     = ScoreParam::PSCFG_GAP_EXTEND_OCC     ,
    GAP_OPEN           = ScoreParam::PSCFG_GAP_OPEN           ,
    GAP_OPEN_OCC       = ScoreParam::PSCFG_GAP_OPEN_OCC       ,
    GAP_OPEN_OCC_CROSS = ScoreParam::PSCFG_GAP_OPEN_OCC_CROSS ,

#ifdef PSCFG_PHYSCORE
    PI_LOOP            = ScoreParam::PSCFG_PI_LOOP            ,
    PI_STEM            = ScoreParam::PSCFG_PI_STEM            ,
    RATE_PARAM_LOOP    = ScoreParam::PSCFG_RATE_PARAM_LOOP    ,
    RATE_PARAM_STEM    = ScoreParam::PSCFG_RATE_PARAM_STEM    ,
    PHY_LOOP_SCALE     = ScoreParam::PSCFG_PHY_LOOP_SCALE     ,
    PHY_STEM_SCALE     = ScoreParam::PSCFG_PHY_STEM_SCALE     ,
#endif
#ifdef PSCFG_TURNER
    TURNER_STACK_SCALE    = ScoreParam::PSCFG_TURNER_STACK_SCALE,
    TURNER_STACK_SCALE_SIM= ScoreParam::PSCFG_TURNER_STACK_SCALE_SIM,
    TURNER_TSTACKH_SCALE  = ScoreParam::PSCFG_TURNER_TSTACKH_SCALE,
    TURNER_TSTACKI_SCALE  = ScoreParam::PSCFG_TURNER_TSTACKI_SCALE,
    TURNER_TSTACKM_SCALE  = ScoreParam::PSCFG_TURNER_TSTACKM_SCALE,
    TURNER_TERM_AU_SCALE  = ScoreParam::PSCFG_TURNER_TERM_AU_SCALE,
    TURNER_LOOP_LEN_SCALE = ScoreParam::PSCFG_TURNER_LOOP_LEN_SCALE,
    TURNER_LOOP_LEN_C0    = ScoreParam::PSCFG_TURNER_LOOP_LEN_C0,
    TURNER_LOOP_LEN_C1    = ScoreParam::PSCFG_TURNER_LOOP_LEN_C1,
#endif
#ifdef PSCFG_MATRIX_SCALE
    SMTX_LOOP_SCALE       = ScoreParam::PSCFG_SMTX_LOOP_SCALE,
    PAIR_SMTX_SCALE       = ScoreParam::PSCFG_PAIR_SMTX_SCALE,
    STACK_MTX_SCALE       = ScoreParam::PSCFG_STACK_MTX_SCALE,
    RATE_PARAM_LOOP_SCALE = ScoreParam::PSCFG_RATE_PARAM_LOOP_SCALE,
    RATE_PARAM_STEM_SCALE = ScoreParam::PSCFG_RATE_PARAM_STEM_SCALE,
#endif
    INVALID_PARAM      = (-1)
  };
  enum {
    NPARAM = (0
#ifdef PSCFG_ETSC0
	      + 2
#endif
	      + 39
#ifdef PSCFG_PHYSCORE
	      + 6
#endif
#ifdef PSCFG_TURNER
	      + 8
#endif
#ifdef PSCFG_MATRIX_SCALE
	      + 5
#endif
	      )
  };
  struct Info {
    Param param;
    Type type;
    int isFeatureParam;
    int nrow;
    int ncol;
    int size;
    char* name;
  };
  struct InfoTable {
    InfoTable() {
#define FLOAT  ScoreParam::FLOAT
#define VECTOR ScoreParam::VECTOR
#define MATRIX ScoreParam::MATRIX
#define SYMMTX ScoreParam::SYMMTX
#define ASMMTX ScoreParam::ASMMTX
#define STACK  ScoreParam::STACK
#define Item(param0, type0, isFeatureParam0, nrow0, ncol0) \
  {param0, type0, isFeatureParam0, nrow0, ncol0, \
  (((type0)==SYMMTX || (type0)==STACK) ? ((nrow0 * (nrow0 + 1)) / 2) : (((type0)==ASMMTX) ? ((nrow0 * (nrow0 - 1)) / 2) : ((nrow0)*(ncol0)))), #param0}


      enum {
	np = Alpha::REDUCED_NPBASE,
	nc = Alpha::NBASE,
	nl = Model::NLOOP
      };
      Info table0[] = {
#ifdef PSCFG_ETSC0
#ifdef USE_PSCFG_ENERGY
      Item(ESC0,               VECTOR, 1,  1, Model::getModel().getNESC0()),
#else
      Item(ESC0,               VECTOR, 0,  1, Model::getModel().getNESC0()),
#endif
      Item(TSC0,               VECTOR, 1,  1, Model::getModel().getNTSC0()),
#endif
#ifdef USE_PSCFG_ENERGY
      Item(PAIR_PROB_CUTOFF,   FLOAT,  0,  1, 1),
      Item(PAIR_PROB_SCALE,    FLOAT,  0,  1, 1),
      Item(PAIR_PROB_COEFF,    FLOAT,  0,  1, 1),
      Item(PAIR_PROB_NSEQ_COEFF,    FLOAT,  0,  1, 1),
      Item(LOOP_PROB_SCALE,    FLOAT,  0,  1, 1),
      Item(LOOP_PROB_COEFF,    FLOAT,  0,  1, 1),
      Item(ALN_PROB_CUTOFF,    FLOAT,  0,  1, 1),
      Item(ALN_PROB_SCALE,     FLOAT,  0,  1, 1), 
      Item(ALN_PROB_COEFF,     FLOAT,  0,  1, 1),
      Item(ALN_PROB_STEM_SCALE,     FLOAT,  0,  1, 1), 
      Item(ALN_PROB_STEM_COEFF,     FLOAT,  0,  1, 1),
      Item(GAP_PROB_SCALE,     FLOAT,  0,  1, 1), 
      Item(GAP_PROB_COEFF,     FLOAT,  0,  1, 1),

      Item(ESC_LOOP,           VECTOR, 1,  1, nl),     
      Item(ESC_STEM,           FLOAT,  1,  1, 1),  
      Item(ESC_FCOL,           VECTOR, 1,  1, nc), 
      Item(ESC_FPAIR,          VECTOR, 1,  1, np),
      Item(ESC_LOD_SCORE,      FLOAT,  1,  1, 1),
      Item(TSC_MP_LOOP,        VECTOR, 1,  1, nl),
      Item(TERM_STEM,          VECTOR, 1,  1, np),

      //Item(GAP_EXTEND,         VECTOR, 1,  1, nl),     
      //Item(GAP_EXTEND_OCC,     VECTOR, 1,  1, nl),
      //Item(GAP_OPEN,           VECTOR, 1,  1, nl),
      //Item(GAP_OPEN_OCC,       VECTOR, 1,  1, nl),
      //Item(GAP_OPEN_OCC_CROSS, VECTOR, 1,  1, nl),
#else
      Item(PAIR_PROB_CUTOFF,   FLOAT,  0,  1, 1),
      Item(PAIR_PROB_SCALE,    FLOAT,  0,  1, 1),
      Item(PAIR_PROB_COEFF,    FLOAT,  0,  1, 1),
      Item(PAIR_PROB_NSEQ_COEFF,    FLOAT,  0,  1, 1),
      Item(LOOP_PROB_SCALE,    FLOAT,  0,  1, 1),
      Item(LOOP_PROB_COEFF,    FLOAT,  0,  1, 1),
      Item(ALN_PROB_CUTOFF,    FLOAT,  0,  1, 1),
      Item(ALN_PROB_SCALE,     FLOAT,  0,  1, 1), 
      Item(ALN_PROB_COEFF,     FLOAT,  0,  1, 1),
      Item(ALN_PROB_STEM_SCALE,     FLOAT,  0,  1, 1), 
      Item(ALN_PROB_STEM_COEFF,    FLOAT,  0,  1, 1),
      Item(GAP_PROB_SCALE,     FLOAT,  0,  1, 1), 
      Item(GAP_PROB_COEFF,     FLOAT,  0,  1, 1),

      Item(ESC_LOOP,           VECTOR, 0,  1, nl),     
      Item(ESC_STEM,           FLOAT,  0,  1, 1),  
      Item(ESC_FCOL,           VECTOR, 0,  1, nc), 
      Item(ESC_FPAIR,          VECTOR, 0,  1, np),
      Item(ESC_LOD_SCORE,      FLOAT,  0,  1, 1),
      Item(TSC_MP_LOOP,        VECTOR, 0,  1, nl),
      Item(TERM_STEM,          VECTOR, 0,  1, np),

      //Item(GAP_EXTEND,         VECTOR, 0,  1, nl),     
      //Item(GAP_EXTEND_OCC,     VECTOR, 0,  1, nl),
      //Item(GAP_OPEN,           VECTOR, 0,  1, nl),
      //Item(GAP_OPEN_OCC,       VECTOR, 0,  1, nl),
      //Item(GAP_OPEN_OCC_CROSS, VECTOR, 0,  1, nl),
#endif

#if defined(PSCFG_USE_GAP_EXTEND)
      Item(GAP_EXTEND,         VECTOR, 1,  1, nl),   
#else
      Item(GAP_EXTEND,         VECTOR, 0,  1, nl),
#endif
      Item(GAP_EXTEND_OCC,     VECTOR, 1,  1, nl),
#if defined(PSCFG_USE_GAP_OPEN)
      Item(GAP_OPEN,           VECTOR, 1,  1, nl),
#else
      Item(GAP_OPEN,           VECTOR, 0,  1, nl),
#endif
      Item(GAP_OPEN_OCC,       VECTOR, 1,  1, nl),

#if defined(PSCFG_USE_GAP_OPTEN_OCC_CROSS)
      Item(GAP_OPEN_OCC_CROSS, VECTOR, 1,  1, nl),
#else
      Item(GAP_OPEN_OCC_CROSS, VECTOR, 0,  1, nl),
#endif
      Item(MIN_BASE_PAIR,      FLOAT,  0,  1, 1),
      Item(MIN_CAN_FRAC,       FLOAT,  0,  1, 1),
      Item(SEQ_SIM_BASE,       FLOAT,  0,  1, 1),

#ifdef PSCFG_MATRIX_SCALE
      Item(SMTX_LOOP,          SYMMTX, 0, nc, nc),      
      Item(PAIR_SMTX,          SYMMTX, 0, np, np),  
      Item(STACK_MTX,          STACK, 0, np, np),  
#else
      Item(SMTX_LOOP,          SYMMTX, 1, nc, nc),      
      Item(PAIR_SMTX,          SYMMTX, 1, np, np), 
#ifdef PSCFG_USE_STACK_MTX
      Item(STACK_MTX,          STACK, 1, np, np),
#else
      Item(STACK_MTX,          STACK, 0, np, np),
#endif
#endif
#if defined(PSCFG_SMTX_SIM)
      Item(SMTX_LOOP_SIM,       SYMMTX, 1, nc, nc),
#if defined(PSCFG_PAIR_SMTX_SIM)
      Item(PAIR_SMTX_SIM,       SYMMTX, 1, np, np),
#else
      Item(PAIR_SMTX_SIM,       SYMMTX, 0, np, np),
#endif
      Item(SMTX_LOOP_SIM_SCALE, FLOAT,  0,  1,  1),
      Item(PAIR_SMTX_SIM_SCALE, FLOAT,  0,  1,  1),
#elif defined(PSCFG_SMTX_SIM_SCALE)
      Item(SMTX_LOOP_SIM,       SYMMTX, 0, nc, nc),
      Item(PAIR_SMTX_SIM,       SYMMTX, 0, np, np),
      Item(SMTX_LOOP_SIM_SCALE, FLOAT,  1,  1,  1),
      Item(PAIR_SMTX_SIM_SCALE, FLOAT,  1,  1,  1),
#else
      Item(SMTX_LOOP_SIM,       SYMMTX, 0, nc, nc),
      Item(PAIR_SMTX_SIM,       SYMMTX, 0, np, np),
      Item(SMTX_LOOP_SIM_SCALE, FLOAT,  0,  1,  1),
      Item(PAIR_SMTX_SIM_SCALE, FLOAT,  0,  1,  1),
#endif
#if defined(PSCFG_SMTX_NSEQ)
      Item(SMTX_LOOP_NSEQ,       SYMMTX, 1, nc, nc),
      Item(PAIR_SMTX_NSEQ,       SYMMTX, 1, np, np),
#else
      Item(SMTX_LOOP_NSEQ,       SYMMTX, 0, nc, nc),
      Item(PAIR_SMTX_NSEQ,       SYMMTX, 0, np, np),
#endif
#if defined(PSCFG_SMTX_ASYM)
      Item(SMTX_LOOP_ASYM,       ASMMTX, 1, nc, nc),
      Item(PAIR_SMTX_ASYM,       ASMMTX, 1, np, np),
#else
      Item(SMTX_LOOP_ASYM,       ASMMTX, 0, nc, nc),
      Item(PAIR_SMTX_ASYM,       ASMMTX, 0, np, np),
#endif
#ifdef PSCFG_PHYSCORE
      Item(PI_LOOP,            VECTOR, 0,  1, nc),
      Item(PI_STEM,            VECTOR, 0,  1, np),  
#ifdef USE_SCFG_ENERGY
#ifdef PSCFG_MATRIX_SCALE 
      Item(RATE_PARAM_LOOP,    VECTOR, 0,  1, ((nc*(nc-1)) / 2)),
      Item(RATE_PARAM_STEM,    VECTOR, 0,  1, ((np*(np-1)) / 2)), 
#else
      Item(RATE_PARAM_LOOP,    VECTOR, 1,  1, ((nc*(nc-1)) / 2)),
      Item(RATE_PARAM_STEM,    VECTOR, 1,  1, ((np*(np-1)) / 2)),
#endif 
      Item(PHY_LOOP_SCALE,     FLOAT,  1,  1, 1),
      Item(PHY_STEM_SCALE,     FLOAT,  1,  1, 1),
#else
      Item(RATE_PARAM_LOOP,    VECTOR, 0,  1, ((nc*(nc-1)) / 2)),
      Item(RATE_PARAM_STEM,    VECTOR, 0,  1, ((np*(np-1)) / 2)), 
      Item(PHY_LOOP_SCALE,     FLOAT,  0,  1, 1),
      Item(PHY_STEM_SCALE,     FLOAT,  0,  1, 1),
#endif
#endif
#ifdef PSCFG_TURNER
      Item(TURNER_STACK_SCALE,   FLOAT,   1,  1,  1) ,
#if defined(PSCFG_USE_TURNER_STACK_SCALE_SIM)
      Item(TURNER_STACK_SCALE_SIM,   FLOAT,   1,  1,  1),
#else
      Item(TURNER_STACK_SCALE_SIM,   FLOAT,   0,  1,  1),
#endif
#ifdef USE_PSCFG_ENERGY
      Item(TURNER_TSTACKH_SCALE, FLOAT,   1,  1,  1) ,
      Item(TURNER_TSTACKI_SCALE, FLOAT,   1,  1,  1) ,
      Item(TURNER_TSTACKM_SCALE, FLOAT,   1,  1,  1) ,
      Item(TURNER_TERM_AU_SCALE, FLOAT,   1,  1,  1) ,
      Item(TURNER_LOOP_LEN_SCALE,FLOAT,   1,  1,  1) ,
      Item(TURNER_LOOP_LEN_C0,   VECTOR,  0,  1,  nl),
      Item(TURNER_LOOP_LEN_C1,   VECTOR,  0,  1,  nl),
#else
      Item(TURNER_TSTACKH_SCALE, FLOAT,   0,  1,  1) ,
      Item(TURNER_TSTACKI_SCALE, FLOAT,   0,  1,  1) ,
      Item(TURNER_TSTACKM_SCALE, FLOAT,   0,  1,  1) ,
      Item(TURNER_TERM_AU_SCALE, FLOAT,   0,  1,  1) ,
      Item(TURNER_LOOP_LEN_SCALE,FLOAT,   0,  1,  1) ,
      Item(TURNER_LOOP_LEN_C0,   VECTOR,  0,  1,  nl),
      Item(TURNER_LOOP_LEN_C1,   VECTOR,  0,  1,  nl),
#endif
#endif
#ifdef PSCFG_MATRIX_SCALE
      Item(SMTX_LOOP_SCALE      , FLOAT,   1,  1,  1),
      Item(PAIR_SMTX_SCALE      , FLOAT,   1,  1,  1),
      Item(STACK_MTX_SCALE      , FLOAT,   1,  1,  1),

#ifdef PSCFG_PHYSCORE
      Item(RATE_PARAM_LOOP_SCALE, FLOAT,   1,  1,  1),
      Item(RATE_PARAM_STEM_SCALE, FLOAT,   1,  1,  1)
#else
      Item(RATE_PARAM_LOOP_SCALE, FLOAT,   0,  1,  1),
      Item(RATE_PARAM_STEM_SCALE, FLOAT,   0,  1,  1)
#endif
#endif
#undef FLOAT
#undef VECTOR
#undef MATRIX
#undef SYMMTX
#undef ASMMTX
#undef STACK
#undef Item
      };
      int nparam = (sizeof(table0) / sizeof(Info));
      Check(nparam == NPARAM, "nparam(%d) != NPARAM(%d)", nparam, NPARAM);
      std::copy(&table0[0], &table0[0] + NPARAM, data.begin());
      setPairs();
    }
    IntPairs pairs;
    CArray<Info, NPARAM> data;
    void setPairs() {
      Array<int> v(data.size());
      for (int i = 0; i < (int)data.size(); i++) {
	v[i] = data[i].param;
      }
      pairs.assign(v.begin(), v.end());
#ifdef DEBUG
      for (int i = 0; i < (int)data.size(); i++) {
	Param param = data[i].param;
	int index = pairs.indexOf(param);
	Assert(0 <= index && index < NPARAM);
	Assert(index == i, "index(%d) != i(%d)", index, i);
      }
#endif
    }
    inline int param2index(Param param) const {return pairs.indexOf(param);}
    inline const Param& index2param(int index) const {return data[index].param;}
    inline const Info& getParamInfo(Param param) const {return data[param2index(param)];}
  };
  static const InfoTable infoTable;
  inline static const InfoTable& getInfoTable() {
    //static const InfoTable table;
    //return table;
    return infoTable;
  }
  inline static int param2index(Param param) {
    //return getInfoTable().param2index(param);
    return infoTable.param2index(param);
  }
  inline static const Param& index2param(int index) {
    //return getInfoTable().index2param(index);
    return infoTable.index2param(index);
  }
  inline static const Info& getParamInfo(Param param) {
    //return getInfoTable().getParamInfo(param);
    return infoTable.getParamInfo(param);
  }


  ///////////////////////////////////////////////////////////////////////////
  // old codes
  inline static const IntPairs& getPairs0() {
    static const struct Conv {
      Conv() : pairs() {
	typedef CArray<Param, NPARAM> Table;
	const Table table = {{
#ifdef PSCFG_ETSC0
	  ESC0                ,
	  TSC0                ,
#endif
	  PAIR_PROB_CUTOFF    ,
	  PAIR_PROB_SCALE     ,
	  ALN_PROB_CUTOFF     ,
	  ALN_PROB_SCALE      ,
	  ESC_LOOP            ,
	  ESC_STEM            ,
	  ESC_FCOL            ,
	  ESC_FPAIR           ,
	  ESC_LOD_SCORE       ,

	  SMTX_LOOP           ,
	  PAIR_SMTX           ,
	  STACK_MTX           ,
	  TSC_MP_LOOP         ,
  	  TERM_STEM           ,
	  MIN_BASE_PAIR       , 
	  MIN_CAN_FRAC        ,

	  GAP_EXTEND          ,
	  GAP_EXTEND_OCC      ,
	  GAP_OPEN            ,
	  GAP_OPEN_OCC        ,
	  GAP_OPEN_OCC_CROSS  ,

#ifdef PSCFG_PHYSCORE
	  PI_LOOP             ,
	  PI_STEM             ,
	  RATE_PARAM_LOOP       ,
	  RATE_PARAM_STEM       ,
	  PHY_LOOP_SCALE      ,
	  PHY_STEM_SCALE      ,
#endif
#ifdef PSCFG_TURNER
	  TURNER_STACK_SCALE   ,
	  TURNER_TSTACKH_SCALE ,
	  TURNER_TSTACKI_SCALE ,
	  TURNER_TSTACKM_SCALE ,
	  TURNER_TERM_AU_SCALE ,
	  TURNER_LOOP_LEN_SCALE,
	  TURNER_LOOP_LEN_C0   ,
	  TURNER_LOOP_LEN_C1   ,
#endif
#ifdef PSCFG_MATRIX_SCALE
	  SMTX_LOOP_SCALE      ,
	  PAIR_SMTX_SCALE      ,
	  STACK_MTX_SCALE      ,
	  RATE_PARAM_LOOP_SCALE,
	  RATE_PARAM_STEM_SCALE,
#endif
	}};
	pairs.assign(table.begin(), table.end());
      }
      IntPairs pairs;
    } conv;
    return conv.pairs;
  };
  inline static Param index2param0(int index) {
    const IntPairs& pairs = getPairs0();
    return Param(pairs[index]);
  }
  inline static int param2index0(Param param) {
    const IntPairs& pairs = getPairs0();
    return pairs.indexOf(param);
  }
  static const Info& getParamInfo0(const Param param) {
#define FLOAT  ScoreParam::FLOAT
#define VECTOR ScoreParam::VECTOR
#define MATRIX ScoreParam::MATRIX
#define SYMMTX ScoreParam::SYMMTX
#define STACK  ScoreParam::STACK
#define Item(param0, type0, isFeatureParam0, nrow0, ncol0) \
  {param0, type0, isFeatureParam0, nrow0, ncol0, \
  (((type0)==SYMMTX || (type0)==STACK) ? ((nrow0 * (nrow0 + 1)) / 2) : ((nrow0)*(ncol0))), #param0}
    enum {
      np = Alpha::REDUCED_NPBASE,
      nc = Alpha::NBASE,
      nl = Model::NLOOP
    };
    typedef CArray<Info, NPARAM> Table;
    static const Table table = {{
#ifdef PSCFG_ETSC0
      Item(ESC0,               VECTOR, 1,  1, Model::getModel().getNESC0()),
      Item(TSC0,               VECTOR, 1,  1, Model::getModel().getNTSC0()),
#endif
      Item(PAIR_PROB_CUTOFF,   FLOAT,  0,  1, 1),
      Item(PAIR_PROB_SCALE,    FLOAT,  1,  1, 1),
      Item(ALN_PROB_CUTOFF,    FLOAT,  0,  1, 1),
      Item(ALN_PROB_SCALE,     FLOAT,  1,  1, 1), 
      Item(ESC_LOOP,           VECTOR, 1,  1, nl),     
      Item(ESC_STEM,           FLOAT,  1,  1, 1),  
      Item(ESC_FCOL,           VECTOR, 1,  1, nc), 
      Item(ESC_FPAIR,          VECTOR, 1,  1, np),
      Item(ESC_LOD_SCORE,      FLOAT,  1,  1, 1),
#ifdef PSCFG_MATRIX_SCALE
      Item(SMTX_LOOP,          SYMMTX, 0, nc, nc),      
      Item(PAIR_SMTX,          SYMMTX, 0, np, np),  
      Item(STACK_MTX,          STACK, 0, np, np),  
#else
      Item(SMTX_LOOP,          SYMMTX, 1, nc, nc),      
      Item(PAIR_SMTX,          SYMMTX, 1, np, np), 
#ifdef PSCFG_USE_STACK_MTX
      Item(STACK_MTX,          STACK, 1, np, np),
#else
      Item(STACK_MTX,          STACK, 0, np, np),
#endif
#endif
      Item(TSC_MP_LOOP,        VECTOR, 1,  1, nl),
      Item(TERM_STEM,          VECTOR, 1,  1, np),
      Item(MIN_BASE_PAIR,      FLOAT,  0,  1, 1),
      Item(MIN_CAN_FRAC,       FLOAT,  0,  1, 1),
      Item(GAP_EXTEND,         VECTOR, 1,  1, nl),     
      Item(GAP_EXTEND_OCC,     VECTOR, 1,  1, nl),
      Item(GAP_OPEN,           VECTOR, 1,  1, nl),
      Item(GAP_OPEN_OCC,       VECTOR, 1,  1, nl),
      Item(GAP_OPEN_OCC_CROSS, VECTOR, 1,  1, nl),
#ifdef PSCFG_PHYSCORE
      Item(PI_LOOP,            VECTOR, 0,  1, nc),
      Item(PI_STEM,            VECTOR, 0,  1, np),  
#ifdef PSCFG_MATRIX_SCALE 
      Item(RATE_PARAM_LOOP,      VECTOR, 0,  1, ((nc*(nc-1)) / 2)),
      Item(RATE_PARAM_STEM,      VECTOR, 0,  1, ((np*(np-1)) / 2)), 
#else
      Item(RATE_PARAM_LOOP,      VECTOR, 1,  1, ((nc*(nc-1)) / 2)),
      Item(RATE_PARAM_STEM,      VECTOR, 1,  1, ((np*(np-1)) / 2)),
#endif 
      Item(PHY_LOOP_SCALE,     FLOAT,  1,  1, 1),
      Item(PHY_STEM_SCALE,     FLOAT,  1,  1, 1),
#endif
#ifdef PSCFG_TURNER
      Item(TURNER_STACK_SCALE,   FLOAT,   1,  1,  1) ,
      Item(TURNER_TSTACKH_SCALE, FLOAT,   1,  1,  1) ,
      Item(TURNER_TSTACKI_SCALE, FLOAT,   1,  1,  1) ,
      Item(TURNER_TSTACKM_SCALE, FLOAT,   1,  1,  1) ,
      Item(TURNER_TERM_AU_SCALE, FLOAT,   1,  1,  1) ,
      Item(TURNER_LOOP_LEN_SCALE,FLOAT,   1,  1,  1) ,
      Item(TURNER_LOOP_LEN_C0,   VECTOR,  0,  1,  nl),
      Item(TURNER_LOOP_LEN_C1,   VECTOR,  0,  1,  nl),
#endif
#ifdef PSCFG_MATRIX_SCALE
      Item(SMTX_LOOP_SCALE      , FLOAT,   1,  1,  1),
      Item(PAIR_SMTX_SCALE      , FLOAT,   1,  1,  1),
      Item(STACK_MTX_SCALE      , FLOAT,   1,  1,  1),
#ifdef PSCFG_PHYSCORE
      Item(RATE_PARAM_LOOP_SCALE, FLOAT,   1,  1,  1),
      Item(RATE_PARAM_STEM_SCALE, FLOAT,   1,  1,  1),
#else
      Item(RATE_PARAM_LOOP_SCALE, FLOAT,   0,  1,  1),
      Item(RATE_PARAM_STEM_SCALE, FLOAT,   0,  1,  1),
#endif
#endif
    }};
    const int& index = param2index(param);
    const Info& info = table[index];
    Assert(param == info.param);
    return info;
#undef FLOAT
#undef VECTOR
#undef MATRIX
#undef SYMMTX
#undef STACK
#undef Item
  }
  /////// old codes
  ///////////////////////////////////////////////////////////////////////////////////


  static const int& getOffset(const Param param);
  static const int& getParamVectorSize();

  ScoreManager();
  ScoreManager(const ScoreManager& scoreManager);
  ~ScoreManager();

  const char* offset2str(int offset0) const;
  const char* param2str(int param) const;
  void setScoreMode(ScoreMode scoreMode0) {scoreMode = scoreMode0;}
  void setDPAlongPath(bool dpAlongPath0) {dpAlongPath = dpAlongPath0;}
  void setScoreParam(ScoreParam& scoreParam);
  void setScoreParamList(Vector<ScoreParam>& scoreParamList);
  void setWeight(const Weight& weights);
  void setPairProbData(const PairProbData& pairProbData);
  void setAlignProbData(const AlignProbData& alignProbData);
  void setGuideTree(const GuideTree& guideTree);
  void resetGuideTree();
  void setDoRefine(const bool doRefine0) {doRefine = doRefine0;}
  
  void setRefinementStrategy(const string& strategy);
  void setUp();
  void setMSAs(const MSA& msaX, const MSA& msaY);
  void setMSAs(const MSA& msaX, const MSA& msaY,
	       const PairProb& pairProbX1,
	       const PairProb& pairProbY1);
  void setAlignProb(const NMatrix<float>& alignProb);// should be set after setMSAs()
  void setSequenceSimilarity(const float& z);
  void setSequenceSimilarity0(const float& z);
  void setDPPath(const DPPath& dpPath0);
  Esc getEsc() const;
  Tsc getTsc() const;
  DPScoreT esc(const Rect& rect, const State& s) const;
  DPScoreT tsc(const Rect& rect, const State& s,const State& s1) const;

  Vec desc(const Rect& rect, const State& s) const;
  Vec dtsc(const Rect& rect, 
	    const State& s,const State& s1) const;

  float getScore(const DPScoreT& dpScore) const;
  Vec getParamVector() const;
  Vec computeParamVectorFromCountVector0(const Vec& countVector) const;
  Vec computeParamVectorFromCountVector(const Vec& countVector) const;
  void setParamVector(const Vec& pv);
  FeatureCounter getFeatureCounter() const;
  FeatureStats getFeatureStats() const;
  string featureVect2String(const Vec& fv, const int& indentLevel) const;

  XCol getXCol(int x) const;// don't pass reference
  void getXCol(XCol& xCol, const int& x) const;
  YCol getYCol(int y) const;
  void getYCol(YCol& yCol, const int& y) const;
  PairCol getPairColX(int bx, int ex) const;
  void getPairColX(PairCol& pairCol, const int& bx, const int& ex) const;
  PairCol getPairColY(int by, int ey) const;
  void getPairColY(PairCol& pairCol, const int& by, const int& ey) const;
  MatchCol getMatchCol(int x, int y) const;
  void getMatchCol(MatchCol& mCol, const int& x, const int& y) const;
  MatchCol getMatchCol(const Point& pt) const;
  
  void setESCTable(ESCTable& escTable) const;
  void setTSCTable(TSCTable& tscTable) const;
  
  inline DPScoreT esc(const State& s,
		      const XCol& xColB,
		      const YCol& yColB,
		      const XCol& xColE,
		      const YCol& yColE,
		      const PairCol& xPair,
		      const PairCol& yPair,
		      const MatchCol& mColB,
		      const MatchCol& mColE,
		      const ESCTable& escTable) const;
  inline DPScoreT tsc(const State& sp,
		      const State& sc,
		      const XCol& xColB,
		      const YCol& yColB,
		      const XCol& xColE,
		      const YCol& yColE,
		      const PairCol& xPair,
		      const PairCol& yPair,
		      const PairCol& xPairR,
		      const PairCol& yPairR,
		      const PairCol& xPairL,
		      const PairCol& yPairL,
		      const TSCTable& tscTable) const;
  inline void desc(FVec& v, float w,
	    const Rect& rect, const State& sp) const;
  inline void dtsc(FVec& v, float w, const Rect& rect,
	    const State& sp, const State& sc) const;
  void testDerivatives() const;
private:
  friend class ESCTable;
  friend class TSCTable;
  friend class Esc;
  friend class Tsc;
  ScoreMode scoreMode;
  bool dpAlongPath;
  Vector<ScoreParam>* pScoreParamList;
  ScoreParam* pScoreParam;
  const Weight* pWeight;
  const GuideTree* pGuideTree;
  const PairProbData* pPairProbData;
  const AlignProbData* pAlignProbData;
  const MSA* pMSAX;
  const MSA* pMSAY;
  DPPath& dpPath;
  Profile& profileX;
  Profile& profileY;
  int lenX;
  int lenY;

#ifdef PSCFG_PHYSCORE
  typedef PhylogenicScore<Alpha::NBASE> PhyLoop;
  typedef PhylogenicScore<Alpha::REDUCED_NPBASE> PhyStem;
  SeqIds& seqIdsXY;
  PhyLoop& phyLoop;
  PhyStem& phyStem;
  GuideTree& subTree;
#endif

  PairProb pairProbX;
  PairProb pairProbY;
  NMatrix<float> alignProb;
  float sequenceSimilarity;
  string refinementStrategy;
  bool doRefine;
  float pairProbThreshold;

#ifdef PSCFG_PHYSCORE
  void resetPhylogenicScores();
  //Vec esc_PhylogenicVFPair_X(int bx, int ex) const;
  //Vec esc_PhylogenicVFPair_Y(int by, int ey) const;
  float esc_PhylogenicLoop(int x, int y) const;
  void desc_PhylogenicLoop(FVec& v, float w,int x, int y) const;
  //void desc_PhylogenicStem(FVec& v, float w,const Rect& rect) const;
#endif

  const Vec& esc_PairSMtx_X(int bx, int ex) const;
  const Vec& esc_PairSMtx_Y(int by, int ey) const;
  void desc_PairSMtx(FVec& v, float w,const Rect& rect) const;

  float esc_MatchStem(const int& x, const int& y) const;
  void desc_MatchStem(FVec& v, float w, const int& x, const int& y) const;
  float esc_MatchLoop(const int& x, const int& y,
		      const int& loopTypeIndex) const;
  void desc_MatchLoop(FVec& v, float w, const int& x, const int& y,
		      const int& loopTypeIndex) const;

  float esc_PairEScore(const UpTriMtx<float>& pairProb,
		       const Profile& profile,
		       const int& b, const int& e) const;
  void desc_PairEScore(FVec& v, float w, 
		       const UpTriMtx<float>& pairProb,
		       const Profile& profile,
		       const int& b, const int& e) const;

  float esc_PairProb(const PairProb& pairProb,
		     const int& b, const int& e, bool logVal = false) const;
  //void desc_PairProb(FVec& v, float w,
  //		     const PairProb& pairProb,
  //		     const int& b, const int& e, bool logVal = false) const;
  //float desc0_PairProb(const PairProb& pairProb,
  //		       const int& x, bool logVal = false) const;
  //float desc0_LoopProb(const PairProb& pairProb,
  //			 const int& x, bool logVal = false) const;
  float esc_LoopProb(const PairProb& pairProb,
		       const int& x, bool logVal = false) const;
  //void desc_LoopProb(FVec& v, float w,
  //		       const PairProb& pairProb,
  //		       const int& x, bool logVal = false) const;
  float esc_AlignProbLoop(int x, int y, bool logVal) const;
  //void desc_AlignProbLoop(FVec& v, float w,int x, int y, bool logVal = false) const;
  float esc_AlignProbStem(int x, int y, bool logVal) const;
  float esc_GapProbX(int x, bool logVal = false) const;
  float esc_GapProbY(int y, bool logVal = false) const;

  //float esc_EscStem() const;
  //void desc_EscStem(FVec& v, float w) const;

  float esc_FPair(const FPair& fPair) const;
  void desc_FPair(FVec& v, float w, const FPair& fPair) const;

  float esc_LODScore(const Profile::FStack& fStack) const;
  void desc_LODScore(FVec& v, float w, 
		     const Profile::FStack& fStack) const;

  float esc_LoopExtend(const int& loopTypeIndex) const;
  void desc_LoopExtend(FVec& v, float w, const int& loopTypeIndex) const;

  float esc_FCol(const Profile::FCol& fCol) const;
  void desc_FCol(FVec& v, float w, const Profile::FCol& fCol) const;

  float esc_SMtxLoop(int x, int y) const;
  void desc_SMtxLoop(FVec& v, float w,int x, int y) const;

  float tsc_Stack(const Profile& profile, int bx, int ex) const;
  void dtsc_Stack(FVec& v, float w, 
		  const Profile& profile, int b, int e) const;
  float tsc_Stack_Turner(const Profile::FStack& fStack) const;
  void dtsc_Stack_Turner(FVec& v, float w,
			 const Profile::FStack& fStack) const;

  float tsc_FromLoop(const Profile& profile,
		     const int& b, const int& e,
		     const int& loopTypeIndex) const;
  void dtsc_FromLoop(FVec& v, float w,
		     const Profile& profile,
		     const int& b, const int& e,
		     const int& loopTypeIndex) const;

  float tsc_ToLoop(const Profile& profile, const int& b, const int& e,
		   const int& loopTypeIndex) const;
  void dtsc_ToLoop(FVec& v, float w, 
		   const Profile& profile, const int& b, const int& e,
		   const int& loopTypeIndex) const;

  float tsc_FromLoop_MPLoop(const int& loopTypeIndex) const;
  float tsc_ToLoop_MPLoop(const int& loopTypeIndex) const;
  void dtsc_FromLoop_MPLoop(FVec& v, float w, 
			    const int& loopTypeIndex) const;
  void dtsc_ToLoop_MPLoop(FVec& v, float w, 
			  const int& loopTypeIndex) const;

  float tsc_FromLoop_TermStem(const FPair& fPair) const;
  void dtsc_FromLoop_TermStem(FVec& v, float w,const FPair& fPair) const;
  float tsc_ToLoop_TermStem(const FPair& fPair) const;
  void dtsc_ToLoop_TermStem(FVec& v, float w,const FPair& fPair) const;

  float tsc_ToLoop_TStackH(const Profile& profile, 
			   const int& b, const int& e, 
			   const FPair& fPair) const;
  void dtsc_ToLoop_TStackH(FVec& v, float w,
			   const Profile& profile, 
			   const int& b, const int& e, 
			   const FPair& fPair) const;

  float tsc_FromLoop_TermAU(const FPair& fPair) const;
  void dtsc_FromLoop_TermAU(FVec& v, float w, const FPair& fPair) const;
  float tsc_ToLoop_TermAU(const FPair& fPair) const;
  void dtsc_ToLoop_TermAU(FVec& v, float w, const FPair& fPair) const;


  float tsc_FromLoop_TStackI(const Profile& profile, 
			     const int& b, const int& e, 
			     const FPair& fPair) const;
  void dtsc_FromLoop_TStackI(FVec& v, float w,
			     const Profile& profile, 
			     const int& b, const int& e, 
			     const FPair& fPair) const;
  float tsc_ToLoop_TStackI(const Profile& profile, 
			   const int& b, const int& e, 
			   const FPair& fPair) const;
  void dtsc_ToLoop_TStackI(FVec& v, float w, 
			   const Profile& profile, 
			   const int& b, const int& e, 
			   const FPair& fPair) const;

  float tsc_FromLoop_TStackM(const Profile& profile, 
			     const int& b, const int& e, 
			     const FPair& fPair) const;
  void dtsc_FromLoop_TStackM(FVec& v, float w,
			     const Profile& profile, 
			     const int& b, const int& e, 
			     const FPair& fPair) const;
  float tsc_ToLoop_TStackM(const Profile& profile, 
			   const int& b, const int& e, 
			   const FPair& fPair) const;
  void dtsc_ToLoop_TStackM(FVec& v, float w, 
			   const Profile& profile, 
			   const int& b, const int& e, 
			   const FPair& fPair) const;

  void desc_Esc0(FVec& v, float w, State s) const;
  void dtsc_Tsc0(FVec& v, float w, State s, State s1) const;

  void desc_GapExtend_X(FVec& v, float w, const int& x,
			const int& loopTypeIndex) const;
  void desc_GapExtend_Y(FVec& v, float w, const int& y,
			const int& loopTypeIndex) const;

  void dtsc_GapOpen_X(FVec& v, float w, const int& x, const int& y,
		      const int& loopTypeIndex) const;
  void dtsc_GapClose_X(FVec& v, float w, const int& x, const int& y,
		       const int& loopTypeIndex) const;
  void dtsc_GapOpen_Y(FVec& v, float w, const int& x, const int& y,
		      const int& loopTypeIndex) const;
  void dtsc_GapClose_Y(FVec& v, float w, const int& x, const int& y,
		       const int& loopTypeIndex) const;
  int getParamSize(int param) const;
  Param getParamFromOffset(int offset0) const;

  float& v_ref(Vec& v, const Param& param, const int& i) const;
  const float& v_ref(const Vec& v, const Param& param, const int& i) const;
  //void v_add(Vec& v, const Param& param, const int& i, const float& z)
  //const ;
  //void v_w_add(Vec& v, float w, 
  //	         const Param& param, const int& i, const float& z) const;
  void v_w_add(FVec& v, float w, 
	       const Param& param, const int& i, const float& z) const;
  void v_w_add2(FVec& v, float w, const int& offset, const float& z) const;

  struct OffsetTable {
    typedef ScoreManager T;
    typedef CArray<int, NPARAM> Table;
    Table table;
    OffsetTable() {
      int offset = 0;
      for (int i = 0; i < T::NPARAM; i++) {
	const T::Info& info = T::getParamInfo(T::index2param(i));
	if (info.isFeatureParam) {
	  table[i] = offset;
	  offset += info.size;
	} else {
	  table[i] = (-1);
	}
      }
    }
    inline const int& getVal(const Param param) const {
      const int& i = T::param2index(param);
      Assert(table[i] >= 0);
      return table[i];
    }
  };
  static const OffsetTable offsetTable;
};

#ifdef PSCFG_ETSC0
#ifdef USE_PSCFG_ENERGY
struct
PSCFG_Energy::ScoreManager::
ESCTable 
{
  typedef Array<DPScoreT> Table;
  Table table;

  const ScoreManager* pScoreManager;
  ESCTable() : table(ScoreManager::model.getNESC0()), pScoreManager(NULL) {table.fill(NegInfDPScore);}
  ESCTable(const ScoreManager& scoreManager)
    : pScoreManager(&scoreManager) {reset();}
  void setScoreManager(const ScoreManager& scoreManager) {
    pScoreManager = &scoreManager;
    reset();
  }
  void reset() {pScoreManager->setESCTable(*this);}
  void reset(const ScoreManager& scoreManager) {
    pScoreManager = &scoreManager;
    reset();
  }
  int size() const {return table.size();}
  void clear() {table.fill(ZeroDPScore);}
  inline const DPScoreT& getVal(const State& s) const {
    return table[ScoreManager::model.getESC0Index(s)];
  }
};
#else
struct 
PSCFG_Energy::ScoreManager::
ESCTable 
{
  typedef Array<DPScoreT> Table;
  Table table;
  ScoreMode scoreMode;
  const ScoreManager* pScoreManager;
  ESCTable() : table() , scoreMode(DPDeck::SCORE_MODE_GUES_BY_SIM), pScoreManager(NULL){}
  ESCTable(const ScoreManager& scoreManager) {
    pScoreManager = &scoreManager;
    reset();
  }
  void setScoreManager(const ScoreManager& scoreManager) {
    pScoreManager = &scoreManager;
    reset();
  }
  void reset() {pScoreManager->setESCTable(*this);}
  void reset(const ScoreManager& scoreManager) {
    pScoreManager = &scoreManager;
    reset();
  }
  int size() const {return 0;}
  void clear() {}
  inline const DPScoreT& getVal(const State& s) const {
    static const DPScoreT esc0 = ZeroDPScore;
    return esc0;
  }
};
#endif
struct 
PSCFG_Energy::ScoreManager::
TSCTable 
{
  enum {SIZE = Model::NTSC0};
  typedef CArray<DPScoreT,SIZE> Table;
  Table table;

  const ScoreManager* pScoreManager;
  TSCTable() : table(), pScoreManager(NULL) {table.fill(NegInfDPScore);}
  TSCTable(const ScoreManager& scoreManager)
    : pScoreManager(&scoreManager) {reset();}
  void setScoreManager(const ScoreManager& scoreManager) {
    pScoreManager = &scoreManager;
  }
  void reset(const ScoreManager& scoreManager) {
    pScoreManager = &scoreManager;
    reset();
  }
  void reset() {pScoreManager->setTSCTable(*this);}
  int size() const {return SIZE;}
  void clear() {table.fill(ZeroDPScore);}
  inline const DPScoreT& getVal(const State& sp, const State& sc) const {
    return table[ScoreManager::model.getTSC0Index(sp, sc)];
  }
};
#else
struct 
PSCFG_Energy::ScoreManager::
ESCTable 
{
  enum {SIZE = 0};
  typedef CArray<DPScoreT,SIZE> Table;
  Table table;
  ScoreMode scoreMode;
  const ScoreManager* pScoreManager;
  ESCTable() : table(), scoreMode(DPDeck::SCORE_MODE_GUES_BY_SIM), pScoreManager(NULL)  {}
  ESCTable(const ScoreManager& scoreManager) {
    pScoreManager = &scoreManager;
    reset();
  }
  void setScoreManager(const ScoreManager& scoreManager) {
    pScoreManager = &scoreManager;
    reset();
  }
  void reset() {pScoreManager->setESCTable(*this);}
  void reset(const ScoreManager& scoreManager) {
    pScoreManager = &scoreManager;
    reset();
  }
  int size() const {return SIZE;}
  void clear() {}
  inline const DPScoreT& getVal(const State& s) const {
    static const DPScoreT esc0 = ZeroDPScore;
    return esc0;
  }
};
struct 
PSCFG_Energy::ScoreManager::
TSCTable 
{
  TSCTable() {}
  TSCTable(const ScoreManager& scoreManager){}
  void setScoreManager(const ScoreManager& scoreManager) {}
  void reset() {}
  void reset(const ScoreManager& scoreManager) {}

  int size() const {return 0;}
  void clear() {}
  inline const DPScoreT& getVal(const State& sp, const State& sc) const {
    static const DPScoreT tsc0 = ZeroDPScore;
    return tsc0;
  }
};
#endif

struct 
PSCFG_Energy::ScoreManager::
GapScore 
{
  DPScoreT gapOpen;
  DPScoreT gapOpenCross;
  DPScoreT gapClose;
  DPScoreT gapCloseCross;
  DPScoreT gapExtend;
  static const GapScore& null() {
    static const GapScore nullNode = {
      NegInfDPScore,
      NegInfDPScore,
      NegInfDPScore,
      NegInfDPScore,
      NegInfDPScore
    };
    return nullNode;
  };
  static const GapScore& zero() {
    static const GapScore zeroNode = {
      ZeroDPScore,
      ZeroDPScore,
      ZeroDPScore,
      ZeroDPScore,
      ZeroDPScore
    };
    return zeroNode;
  };
};

struct 
PSCFG_Energy::ScoreManager::
XCol 
{
  enum {SIZE = Model::NLOOP};
  CArray<GapScore,SIZE> gaps;
  void setNull() {gaps.fill(GapScore::zero());}
  inline static const XCol& null() {
    static const struct NullNode {
      XCol  nd;
      NullNode() {nd.setNull();}
    } nn;
    return nn.nd;
  }
  inline const DPScoreT& scGapOpen(const State& s) const {
    return gaps[ScoreManager::model.getLoopTypeIndex(s)].gapOpen;
  }
  inline const DPScoreT& scGapOpenCross(const State& s) const {
    return gaps[ScoreManager::model.getLoopTypeIndex(s)].gapOpenCross;
  }
  inline const DPScoreT& scGapClose(const State& s) const {
    return gaps[ScoreManager::model.getLoopTypeIndex(s)].gapClose;
  }
  inline const DPScoreT& scGapCloseCross(const State& s) const {
    return gaps[ScoreManager::model.getLoopTypeIndex(s)].gapCloseCross;
  }
  inline const DPScoreT& scGapExtend(const State& s) const {
    return gaps[ScoreManager::model.getLoopTypeIndex(s)].gapExtend;
  }
};
struct 
PSCFG_Energy::ScoreManager::
YCol 
{
  enum {SIZE = Model::NLOOP};
  typedef CArray<GapScore,SIZE> Gaps;
  Gaps gaps;
  void setNull() {gaps.fill(GapScore::zero());}
  inline static const YCol& null() {
    static const struct NullNode {
      YCol  nd;
      NullNode() {nd.setNull();}
    } nn;
    return nn.nd;
  }
  inline const DPScoreT& scGapOpen(const State& s) const {
    return gaps[ScoreManager::model.getLoopTypeIndex(s)].gapOpen;
  }
  inline const DPScoreT& scGapOpenCross(const State& s) const {
    return gaps[ScoreManager::model.getLoopTypeIndex(s)].gapOpenCross;
  }
  inline const DPScoreT& scGapClose(const State& s) const {
    return gaps[ScoreManager::model.getLoopTypeIndex(s)].gapClose;
  }
  inline const DPScoreT& scGapCloseCross(const State& s) const {
    return gaps[ScoreManager::model.getLoopTypeIndex(s)].gapCloseCross;
  }
  inline const DPScoreT& scGapExtend(const State& s) const {
    return gaps[ScoreManager::model.getLoopTypeIndex(s)].gapExtend;
  }
};
struct 
PSCFG_Energy::ScoreManager::
PairCol 
{
  enum {FPAIRSIZE = Alpha::REDUCED_NPBASE};
  enum {FPAIRSIZE_1 = Alpha::REDUCED_NPBASE};
  enum {TSCSIZE = Model::NLOOP};
  typedef CArray<DPScoreT,FPAIRSIZE> FPair;

#ifdef PSCFG_PHYSCORE
  //typedef CArray<DPScoreT,FPAIRSIZE_1> PhySC;
#endif

  typedef CArray<DPScoreT,TSCSIZE> TSCFromLoop;
  typedef CArray<DPScoreT,TSCSIZE> TSCToLoop;
  FPair fPair_;

#ifdef PSCFG_PHYSCORE
  //PhySC physc_;
#endif
  TSCFromLoop tscFromLoop_;
  TSCToLoop tscToLoop_;
  DPScoreT pairEScore_;
  DPScoreT stackScore_;
  inline void setNull() {
    tscFromLoop_.fill(NegInfDPScore);
    tscToLoop_.fill(NegInfDPScore);
    pairEScore_ = NegInfDPScore;
    stackScore_ = NegInfDPScore;
  }
  inline static const PairCol& null() {
    static const struct NullNode {
      PairCol  nd;
      NullNode() {nd.setNull();}
    } nn;
    return nn.nd;
  }
  inline DPScoreT escMP(const PairCol& pn
			//, const ScoreManager& scoreManager
			) const { 
    //return (pairEScore_ + pn.pairEScore_
#ifdef USE_DIRICHLET_PRIOR
	    // assuming dpScoreScale == 1.0
    //	    + log(dot_prod<FPAIRSIZE>(&fPair_[0], &(pn.fPair_[0])))
#else
    //	    + dot_prod<FPAIRSIZE>(&fPair_[0], &(pn.fPair_[0]))
#endif
#ifdef PSCFG_PHYSCORE
    //	    + computePhyScore(pn.physc_, scoreManager)
#endif
    //	    );
    return (pairEScore_ + pn.pairEScore_);
  }
#ifdef PSCFG_PHYSCORE
  //inline DPScoreT 
  //computePhyScore(const PhySC& physc1,
  //		  const ScoreManager& scoreManager) const {
  //  DPScoreT z = NegInfDPScore;
  //  int n = physc_.size();
  //  const DPScoreT* p = &physc_[0];
  //  const DPScoreT* p1 = &physc1[0];
  //  while (n-- > 0) {
  //    scoreManager.logAdd(z, (*p++) + (*p1++));
  //  }
  //  return z;
  //}
#endif

  inline const DPScoreT& tscMP() const {return stackScore_;}
  inline const DPScoreT& tscFromLoop(const State& s) const {
    return tscFromLoop_[ScoreManager::model.getLoopTypeIndex(s)];
  }
  inline const DPScoreT& tscToLoop(const State& s) const {
    return tscToLoop_[ScoreManager::model.getLoopTypeIndex(s)];
  }
};
struct 
PSCFG_Energy::ScoreManager::
MatchCol 
{
  enum {SIZE = Model::NLOOP};
  CArray<DPScoreT,SIZE> escLoop_;
  DPScoreT escStem_;

  inline void setNull() {
    escLoop_.fill(ZeroDPScore);
    escStem_ = ZeroDPScore;
  }
  inline static const MatchCol& null() {
    static const struct NullNode {
      MatchCol  nd;
      NullNode() {nd.setNull();}
    } nn;
    return nn.nd;
  }
  inline const DPScoreT& escLoop(const State& s) const {
    return escLoop_[ScoreManager::model.getLoopTypeIndex(s)];
  }
  inline const DPScoreT& escStem() const {
    return escStem_;
  }
};
struct 
PSCFG_Energy::ScoreManager::
SCNode 
{
  enum {SIZE = Model::NSTATE};
  typedef CArray<DPScoreT,SIZE> Data;
  Data data;

  inline void setNull() {data.fill(NegInfDPScore);}
  static const SCNode& null() {
    static const struct NullNode {
      SCNode scn;
      NullNode() {scn.setNull();}
    } nn;
    return nn.scn;
  };
  bool operator==(const SCNode& sc) const {
    return equal(data.begin(), data.end(), sc.data.begin());
  }
  bool isNull() const {return (*this == null());}

  string toString() const {
    bool hasOne = false;
    ostringstream oss;
    oss << "[ ";
    for (int i = 0; i < SIZE; i++) {
      if (data[i] != null()[i]) {
	if (hasOne) {
	  oss << ", ";
	} else {
	  hasOne = true;
	}
	oss << ScoreManager::model.state2str(i) << ": " << data[i];
      }
    }
    oss << " ]";
    return oss.str();
  }
  void print() const {
    cout << toString() << flush;
  }
  bool equals(const SCNode& sc) const;
  DPScoreT& operator[](int i) {return data[i];}
  const DPScoreT& operator[](int i) const {return data[i];}
};
inline PSCFG_Energy::ScoreManager::DPScoreT
PSCFG_Energy::ScoreManager::
esc(const State& s,
    const XCol& xColB,
    const YCol& yColB,
    const XCol& xColE,
    const YCol& yColE,
    const PairCol& xPair,
    const PairCol& yPair,
    const MatchCol& mColB,
    const MatchCol& mColE,
    const ESCTable& escTable) const
{
#if 0
#define Exp(x) PROB_CONS::EXP(x)
#else
#define Exp(x) exp(x)
#endif

  const DPScoreT& esc0 = escTable.getVal(s);
  const EmitState& s_es = model.toPSCFGBaseState(s);
  switch (s_es) {
  case EmitModel::ST_st: break;
  case EmitModel::MP_st:
    //if (escTable.scoreMode == DPDeck::SCORE_MODE_MEA) {
#if defined(PSCFG_NUSSINOV)
      return (
	      exp(esc0 
		  + xPair.escMP(yPair /* , *this */) 
		  + mColB.escStem() + mColE.escStem()
		  )
	      /* + exp(mColB.escStem()) + exp(mColE.escStem()) */
	      /* + mColB.escLoop(s) + mColE.escLoop(s) */
	      );
      //return (xPair.escMP(yPair) * mColB.escStem() * mColE.escStem());
      //} else {
#else
      return (esc0 
	      + xPair.escMP(yPair /* , *this */) 
#ifdef USE_PSCFG_ENERGY
	      //+ mColB.escStem() + mColE.escStem()
#endif
	      );
      //}
#endif
  case EmitModel::MR_st: return (esc0 
				 + mColE.escLoop(s) //substitution score
				 );
#if !defined(PSCFG_NUSSINOV)
  case EmitModel::IR_st: return (esc0 + yColE.scGapExtend(s));
  case EmitModel::DR_st: return (esc0 + xColE.scGapExtend(s));
#endif
  case EmitModel::ML_st: return (esc0 
				 + mColB.escLoop(s) //substitution score
				 );
#if !defined(PSCFG_NUSSINOV)
  case EmitModel::IL_st: return (esc0 + yColB.scGapExtend(s));
  case EmitModel::DL_st: return (esc0 + xColB.scGapExtend(s));
#endif
  case EmitModel::BF_st: break; 
  default: break; 
  }
  return esc0;
}
#if !defined(PSCFG_NUSSINOV)
#define PSCFG_ENERGY_SCORE_MANAGER_ESC(result,s,xColB,yColB,xColE,yColE,xPair,yPair,mColB,mColE,escTable) \
do {                                                                \
  result = escTable.getVal(s);				            \
  const EmitState& s_es = ScoreManager::model.toPSCFGBaseState(s);  \
  switch (s_es) {			                            \
  case EmitModel::MP_st:                                            \
   if (escTable.scoreMode == DPDeck::SCORE_MODE_MEA) {              \
     result = (                                                     \
                exp(result + xPair.escMP(yPair) + mColB.escStem() + mColE.escStem()) \
                /* + exp(mColB.escStem()) + exp(mColE.escStem()) */ \
                /* + mColB.escLoop(s) + mColE.escLoop(s)  */        \
              );                                                    \
   } else {                                                         \
     result += xPair.escMP(yPair);                                  \
   }                                                                \
   break;                                                           \
  case EmitModel::MR_st: result += (mColE.escLoop(s));     break;   \
  case EmitModel::IR_st: result += (yColE.scGapExtend(s)); break;   \
  case EmitModel::DR_st: result += (xColE.scGapExtend(s)); break;   \
  case EmitModel::ML_st: result += (mColB.escLoop(s));     break;   \
  case EmitModel::IL_st: result += (yColB.scGapExtend(s)); break;   \
  case EmitModel::DL_st: result += (xColB.scGapExtend(s)); break;   \
  default: break;				                    \
  }                                                                 \
} while (0)
#endif
 void
PSCFG_Energy::ScoreManager::
desc(FVec& v, float w, const Rect& rect, const State& sp) const
{// should be consistent with esc()
  desc_Esc0(v, w, sp);

  const EmitState& sp_es = model.toPSCFGBaseState(sp);
  switch (sp_es) {
  case EmitModel::MP_st:
    desc_PairSMtx(v, w, rect);

    //desc_PairProb_X(v, w, rect.bx, rect.ex - 1);
    //desc_PairProb_Y(v, w, rect.by, rect.ey - 1);
    //desc_EscStem();
    desc_PairEScore(v, w, pairProbX, profileX, rect.bx, rect.ex - 1);
    desc_PairEScore(v, w, pairProbY, profileY, rect.by, rect.ey - 1);

    //desc_AlnProb(v, w, rect.bx, rect.by);
    //desc_AlnProb(v, w, rect.ex - 1, rect.ey - 1);
    desc_MatchStem(v, w, rect.bx, rect.by);
    desc_MatchStem(v, w, rect.ex - 1, rect.ey - 1);
    // no derivative formula is written

#ifdef PSCFG_PHYSCORE
    //desc_PhylogenicStem(v, w, rect);
#endif
    break;
  case EmitModel::MR_st:
    //desc_SMtxLoop(v, w, rect.ex - 1, rect.ey - 1);
    //desc_EscLoop(v, w, sp);
    //desc_AlnProb(v, w, rect.ex-1,rect.ey-1);
#ifdef PSCFG_PHYSCORE
    //desc_RateMtxLoop(v, w, rect.ex-1,rect.ey-1);
#endif
    desc_MatchLoop(v, w, rect.ex - 1, rect.ey - 1,
		   model.getLoopTypeIndex(sp));
    break;
  case EmitModel::IR_st:
    //desc_GapExtend(v, w, sp);
    //desc_GapExtendOcc_Y(v, w, rect.ey-1, sp);
    desc_GapExtend_Y(v, w, rect.ey-1, model.getLoopTypeIndex(sp));
    break;
  case EmitModel::DR_st:
    //desc_GapExtend(v, w, sp);
    //desc_GapExtendOcc_X(v, w, rect.ex-1, sp);
    desc_GapExtend_X(v, w, rect.ex-1, model.getLoopTypeIndex(sp));
    break;
  case EmitModel::ML_st:
    //desc_EscLoop(v, w, sp);
    //desc_SMtxLoop(v, w, rect.bx, rect.by);
    //desc_AlnProb(v, w, rect.bx,rect.by);
#ifdef PSCFG_PHYSCORE
    //desc_RateMtxLoop(v, w, rect.bx,rect.by);
#endif
    desc_MatchLoop(v, w, rect.bx, rect.by, 
		   model.getLoopTypeIndex(sp));
    break;
  case EmitModel::IL_st:
    //desc_GapExtend(v, w, sp);
    //desc_GapExtendOcc_Y(v, w, rect.by, sp);
    desc_GapExtend_Y(v, w, rect.by, model.getLoopTypeIndex(sp));
    break;
  case EmitModel::DL_st:
    //desc_GapExtend(v, w, sp);
    //desc_GapExtendOcc_X(v, w, rect.bx, sp);
    desc_GapExtend_X(v, w, rect.bx, model.getLoopTypeIndex(sp));
    break;
  default:
    break;
  }
}

inline PSCFG_Energy::ScoreManager::DPScoreT
PSCFG_Energy::ScoreManager::
tsc(const State& sp,
    const State& sc,
    const XCol& xColB,
    const YCol& yColB,
    const XCol& xColE,
    const YCol& yColE,
    const PairCol& xPair,
    const PairCol& yPair,
    const PairCol& xPairR,
    const PairCol& yPairR,
    const PairCol& xPairL,
    const PairCol& yPairL,
    const TSCTable& tscTable) const
{
#if !defined(PSCFG_NUSSINOV)
  const DPScoreT& tsc0 = tscTable.getVal(sp, sc);

  const EmitState& sp_es = model.toPSCFGBaseState(sp);
  const EmitState& sc_es = model.toPSCFGBaseState(sc);
  switch (sp_es) {
  case EmitModel::ST_st:
    {
      switch (sc_es) {
      case EmitModel::ST_st:
	break;
      case EmitModel::MP_st:
	break;
      case EmitModel::MR_st:
	break;
      case EmitModel::IR_st:
	return (tsc0 + yColE.scGapClose(sc) + xColE.scGapCloseCross(sc));
      case EmitModel::DR_st:
	return (tsc0 + xColE.scGapClose(sc) + yColE.scGapCloseCross(sc));
      case EmitModel::ML_st:
	break;
      case EmitModel::IL_st:
	return (tsc0 + yColB.scGapOpen(sc) + xColB.scGapOpenCross(sc));
      case EmitModel::DL_st:
	return (tsc0 + xColB.scGapOpen(sc) + yColB.scGapOpenCross(sc));
      case EmitModel::BF_st:
	break;
      default:
	break;
      }
    }
    break;
  case EmitModel::MP_st:
    {
      switch (sc_es) {
      case EmitModel::ST_st:
	break;
      case EmitModel::MP_st:
	return (tsc0 + xPair.tscMP() + yPair.tscMP());
      case EmitModel::MR_st:
	return (tsc0 
#ifdef USE_PSCFG_ENERGY
		+ xPair.tscToLoop(sc) + yPair.tscToLoop(sc)
#endif
		);
      case EmitModel::IR_st:
	return (tsc0 
#ifdef USE_PSCFG_ENERGY
		+ yPair.tscToLoop(sc) + xPair.tscToLoop(sc)
#endif
		+ yColE.scGapClose(sc) + xColE.scGapCloseCross(sc));
      case EmitModel::DR_st:
	return (tsc0 
#ifdef USE_PSCFG_ENERGY
		+ xPair.tscToLoop(sc) + yPair.tscToLoop(sc)
#endif
		+ xColE.scGapClose(sc) + yColE.scGapCloseCross(sc));
      case EmitModel::ML_st:
	return (tsc0 
#ifdef USE_PSCFG_ENERGY
		+ xPair.tscToLoop(sc) + yPair.tscToLoop(sc)
#endif
		);
      case EmitModel::IL_st:
	return (tsc0 
#ifdef USE_PSCFG_ENERGY
		+ yPair.tscToLoop(sc) + xPair.tscToLoop(sc)
#endif
		+ yColB.scGapOpen(sc) + xColB.scGapOpenCross(sc));
      case EmitModel::DL_st:
	return (tsc0 
#ifdef USE_PSCFG_ENERGY
		+ xPair.tscToLoop(sc) + yPair.tscToLoop(sc)
#endif
		+ xColB.scGapOpen(sc) + yColB.scGapOpenCross(sc));
      case EmitModel::BF_st:
	break;
      default:
	break;
      }
    }
    break;
  case EmitModel::MR_st:
    {
      switch (sc_es) {
      case EmitModel::ST_st:
	break;
      case EmitModel::MP_st:
	return (tsc0 
#ifdef USE_PSCFG_ENERGY
		+ xPairR.tscFromLoop(sp) + yPairR.tscFromLoop(sp)
#endif
		);
      case EmitModel::MR_st:
	break;
      case EmitModel::IR_st:
	return (tsc0 + yColE.scGapClose(sc) + xColE.scGapCloseCross(sc));
      case EmitModel::DR_st:
	return (tsc0 + xColE.scGapClose(sc) + yColE.scGapCloseCross(sc));
      case EmitModel::ML_st:
	break;
      case EmitModel::IL_st:
	return (tsc0 + yColB.scGapOpen(sc) + xColB.scGapOpenCross(sc));
      case EmitModel::DL_st:
	return (tsc0 + xColB.scGapOpen(sc) + yColB.scGapOpenCross(sc));
      default: break;
      } 
      break;
    case EmitModel::IR_st:
      {
	const DPScoreT& go = 
	  (yColE.scGapOpen(sp) + xColE.scGapOpenCross(sp));
	switch (sc_es) {

	case EmitModel::ST_st:
	  break;
	case EmitModel::MP_st:
	  return (tsc0 + go 
#ifdef USE_PSCFG_ENERGY
		  + yPairR.tscFromLoop(sp) 
		  + xPair.tscFromLoop(sp)
#endif
		  );
	case EmitModel::MR_st:
	  return (tsc0 + go);
	case EmitModel::IR_st:
	  break;
	case EmitModel::DR_st:
	  break;
	case EmitModel::ML_st:
	  return (tsc0 + go);
	case EmitModel::IL_st:
	  return (tsc0 + go
		  + yColB.scGapOpen(sc) + xColB.scGapOpenCross(sc));
	case EmitModel::DL_st:
	  return (tsc0 + go
		  + xColB.scGapOpen(sc) + yColB.scGapOpenCross(sc));
	case EmitModel::BF_st:
	  return (tsc0 + go);
	default: break;
	}
      }
      break;
    case EmitModel::DR_st:
      {
	const DPScoreT& go = 
	  (xColE.scGapOpen(sp) + yColE.scGapOpenCross(sp));
	switch (sc_es) {

	case EmitModel::ST_st:
	  break;
	case EmitModel::MP_st:
	  return (tsc0 + go 
#ifdef USE_PSCFG_ENERGY
		  + xPairR.tscFromLoop(sp) 
		  + yPair.tscFromLoop(sp)
#endif
		  );
	case EmitModel::MR_st:
	  return (tsc0 + go);
	case EmitModel::IR_st:
	  break;
	case EmitModel::DR_st:
	  break;
	case EmitModel::ML_st:
	  return (tsc0 + go);
	case EmitModel::IL_st:
	  return (tsc0 + go
		  + yColB.scGapOpen(sc) + xColB.scGapOpenCross(sc));
	case EmitModel::DL_st:
	  return (tsc0 + go 
		  + xColB.scGapOpen(sc) + yColB.scGapOpenCross(sc));
	case EmitModel::BF_st:
	  return (tsc0 + go);
	default: break;
	}
      }
      break;
    case EmitModel::ML_st:
      {
	switch (sc_es) {
	case EmitModel::ST_st:
	  break;
	case EmitModel::MP_st:
	  return (tsc0 
#ifdef USE_PSCFG_ENERGY
		  + xPairL.tscFromLoop(sp)    //vpair.refX(bx+1,ex)
		  + yPairL.tscFromLoop(sp)    //vpair.refY(by+1,ey)
#endif
		  );
	case EmitModel::MR_st:
	  break;
	case EmitModel::IR_st:
	  break;
	case EmitModel::DR_st:
	  break;
	case EmitModel::ML_st:
	  break;
	case EmitModel::IL_st:
	  return (tsc0 + yColB.scGapOpen(sc) + xColB.scGapOpenCross(sc));
	case EmitModel::DL_st:
	  return (tsc0 + xColB.scGapOpen(sc) + yColB.scGapOpenCross(sc));
	case EmitModel::BF_st:
	  break;
	default: break;
	}
      }
      break;
    case EmitModel::IL_st:
      {
	const DPScoreT& go = 
	  (yColB.scGapClose(sp) + xColB.scGapCloseCross(sp));
	switch (sc_es) {
	case EmitModel::ST_st:
	  break;
	case EmitModel::MP_st:
	  return (tsc0 + go 
#ifdef USE_PSCFG_ENERGY
		  + yPairL.tscFromLoop(sp) 
		  + xPair.tscFromLoop(sp)
#endif
		  );
	case EmitModel::MR_st:
	  break;
	case EmitModel::IR_st:
	  break;
	case EmitModel::DR_st:
	  break;
	case EmitModel::ML_st:
	  return (tsc0 + go);
	case EmitModel::IL_st:
	  break;
	case EmitModel::DL_st:
	  break;
	case EmitModel::BF_st:
	  return (tsc0 + go);
	default: break;
	}
      }
      break;
    case EmitModel::DL_st:
      {
	const DPScoreT& go = 
	  (xColB.scGapClose(sp) + yColB.scGapCloseCross(sp));
	switch (sc_es) {
	case EmitModel::ST_st:
	  break;
	case EmitModel::MP_st:
	  return (tsc0 + go
#ifdef USE_PSCFG_ENERGY 
		  + xPairL.tscFromLoop(sp)
		  + yPair.tscFromLoop(sp)
#endif
		  );
	case EmitModel::MR_st:
	  break;
	case EmitModel::IR_st:
	  break;
	case EmitModel::DR_st:
	  break;
	case EmitModel::ML_st:
	  return (tsc0 + go);
	case EmitModel::IL_st:
	  break;
	case EmitModel::DL_st:
	  break;
	case EmitModel::BF_st:
	  return tsc0 + go;
	default: break;
	}
      }
      break;
    case EmitModel::BF_st:
      {
	switch (sc_es) {
	case EmitModel::ST_st:
	  break;
	case EmitModel::MP_st:
	  break;
	case EmitModel::MR_st:
	  break;
	case EmitModel::IR_st:
	  break;
	case EmitModel::DR_st:
	  break;
	case EmitModel::ML_st:
	  break;
	case EmitModel::IL_st:
	  break;
	case EmitModel::DL_st:
	  break;
	case EmitModel::BF_st:
	  break;
	default:
	  break;
	}
      }
      break;
    default:
      break;
    }
  }
#else
  const DPScoreT tsc0 = ZeroDPScore;
#endif
  return tsc0;
}
#if defined(PSCFG_NUSSINOV)
#define PSCFG_ENERGY_SCORE_MANAGER_TSC(result,sp,sc,xColB,yColB,xColE,yColE,xPair,yPair,xPairR,yPairR,xPairL,yPairL,tscTable) do {result = ZeroDPScore;} while (0)
#else
#ifndef USE_PSCFG_ENERGY
#define PSCFG_ENERGY_SCORE_MANAGER_TSC(result,sp,sc,xColB,yColB,xColE,yColE,xPair,yPair,xPairR,yPairR,xPairL,yPairL,tscTable) \
do {                                                                                              \
  result = tscTable.getVal(sp, sc);								  \
  const EmitState& sp_es = ScoreManager::model.toPSCFGBaseState(sp);			      	  \
  const EmitState& sc_es = ScoreManager::model.toPSCFGBaseState(sc);			       	  \
  switch (sp_es) {										  \
  case EmitModel::ST_st:									  \
    switch (sc_es) {										  \
    case EmitModel::IR_st:									  \
      (result += yColE.scGapClose(sc) + xColE.scGapCloseCross(sc));				  \
      break;											  \
    case EmitModel::DR_st:									  \
      (result += xColE.scGapClose(sc) + yColE.scGapCloseCross(sc));				  \
      break;											  \
    case EmitModel::IL_st:									  \
      (result += yColB.scGapOpen(sc) + xColB.scGapOpenCross(sc));				  \
      break;											  \
    case EmitModel::DL_st:									  \
      (result += xColB.scGapOpen(sc) + yColB.scGapOpenCross(sc));				  \
      break;											  \
    default:											  \
      break;											  \
    }												  \
    break;											  \
  case EmitModel::MP_st:									  \
    switch (sc_es) {										  \
    case EmitModel::MP_st:									  \
      (result += xPair.tscMP() + yPair.tscMP());						  \
      break;											  \
    case EmitModel::IR_st:									  \
      (result += yColE.scGapClose(sc) + xColE.scGapCloseCross(sc));				  \
      break;											  \
    case EmitModel::DR_st:									  \
      (result += xColE.scGapClose(sc) + yColE.scGapCloseCross(sc));				  \
      break;											  \
    case EmitModel::IL_st:									  \
      (result += yColB.scGapOpen(sc) + xColB.scGapOpenCross(sc));				  \
      break;											  \
    case EmitModel::DL_st:									  \
      (result += xColB.scGapOpen(sc) + yColB.scGapOpenCross(sc));				  \
      break;											  \
    default:											  \
      break;											  \
    }												  \
    break;											  \
  case EmitModel::MR_st:									  \
    switch (sc_es) {										  \
    case EmitModel::IR_st:									  \
      (result += yColE.scGapClose(sc) + xColE.scGapCloseCross(sc));				  \
      break;											  \
    case EmitModel::DR_st:									  \
      (result += xColE.scGapClose(sc) + yColE.scGapCloseCross(sc));				  \
      break;											  \
    case EmitModel::IL_st:									  \
      (result += yColB.scGapOpen(sc) + xColB.scGapOpenCross(sc));				  \
      break;											  \
    case EmitModel::DL_st:									  \
      (result += xColB.scGapOpen(sc) + yColB.scGapOpenCross(sc));				  \
      break;											  \
    default: 											  \
      break;											  \
    } 												  \
    break;											  \
  case EmitModel::IR_st:									  \
    {												  \
      const DPScoreT& go = 									  \
	(yColE.scGapOpen(sp) + xColE.scGapOpenCross(sp));					  \
      switch (sc_es) {										  \
      case EmitModel::MP_st:									  \
	(result += go);										  \
	break;											  \
      case EmitModel::MR_st:									  \
	(result += go);										  \
	break;											  \
      case EmitModel::ML_st:									  \
	(result += go);										  \
	break;											  \
      case EmitModel::IL_st:									  \
	(result += go + yColB.scGapOpen(sc) + xColB.scGapOpenCross(sc));			  \
	break;											  \
      case EmitModel::DL_st:									  \
	(result += go + xColB.scGapOpen(sc) + yColB.scGapOpenCross(sc));			  \
	break;											  \
      case EmitModel::BF_st:									  \
	(result += go);										  \
	break;											  \
      default:											  \
	break;											  \
      }												  \
    }												  \
    break;											  \
  case EmitModel::DR_st:									  \
    {												  \
      const DPScoreT& go = 									  \
	(xColE.scGapOpen(sp) + yColE.scGapOpenCross(sp));					  \
      switch (sc_es) {										  \
      case EmitModel::MP_st:									  \
	(result += go);										  \
	break;											  \
      case EmitModel::MR_st:									  \
	(result += go);										  \
	break;											  \
      case EmitModel::ML_st:									  \
	(result += go);										  \
	break;											  \
      case EmitModel::IL_st:									  \
	(result += go + yColB.scGapOpen(sc) + xColB.scGapOpenCross(sc));			  \
	break;											  \
      case EmitModel::DL_st:									  \
	(result += go + xColB.scGapOpen(sc) + yColB.scGapOpenCross(sc));			  \
	break;											  \
      case EmitModel::BF_st:									  \
	(result += go);										  \
	break;											  \
      default: 											  \
	break;											  \
      }												  \
    }												  \
    break;											  \
  case EmitModel::ML_st:									  \
    switch (sc_es) {										  \
    case EmitModel::IL_st:									  \
      (result += yColB.scGapOpen(sc) + xColB.scGapOpenCross(sc));				  \
      break;											  \
    case EmitModel::DL_st:									  \
      (result += xColB.scGapOpen(sc) + yColB.scGapOpenCross(sc));				  \
      break;											  \
    default: 											  \
      break;											  \
    }												  \
    break;											  \
  case EmitModel::IL_st:									  \
    {												  \
      const DPScoreT& go = 									  \
	(yColB.scGapClose(sp) + xColB.scGapCloseCross(sp));					  \
      switch (sc_es) {										  \
      case EmitModel::MP_st:									  \
	(result += go);										  \
	break;											  \
      case EmitModel::ML_st:									  \
	(result += go);										  \
	break;											  \
      case EmitModel::BF_st:									  \
	(result += go);										  \
	break;											  \
      default: 											  \
	break;											  \
      }												  \
    }												  \
    break;											  \
  case EmitModel::DL_st:									  \
    {												  \
      const DPScoreT& go = 									  \
	(xColB.scGapClose(sp) + yColB.scGapCloseCross(sp));					  \
      switch (sc_es) {										  \
      case EmitModel::MP_st:									  \
	(result += go);										  \
	break;											  \
      case EmitModel::ML_st:									  \
	(result += go);										  \
	break;											  \
      case EmitModel::BF_st:									  \
	(result += go);										  \
	break;											  \
      default: 											  \
	break;											  \
      }												  \
    }												  \
    break;											  \
  default:											  \
    break;											  \
  }                                                                                               \
} while (0)
#endif
#endif
inline void
PSCFG_Energy::ScoreManager::
dtsc(FVec& v, float w, const Rect& rect,
     const State& sp, const State& sc) const
{ // should be consistent with tsc()
  // extract function for each tsc() term
  //const DPScoreT& tsc0 = tscTable.getVal(sp, sc);
  dtsc_Tsc0(v, w, sp, sc);

  const EmitState& sp_es = model.toPSCFGBaseState(sp);
  const EmitState& sc_es = model.toPSCFGBaseState(sc);
  const Rect r = EmitModel::addNEmit(rect, sp_es);
  Assert((rect.b == rect.e) || r.isNormal());

  switch (sp_es) {
  case EmitModel::ST_st:
    switch (sc_es) {
    case EmitModel::ST_st:
      break;
    case EmitModel::MP_st:
      // add tscToLoop ? context of multi stem
      break;
    case EmitModel::MR_st:
      break;
    case EmitModel::IR_st:
      //return (yColE.scGapClose(sc) + xColE.scGapCloseCross(sc));

      //dtsc_GapOpen_CloseY(v, w, r.ey, sc);
      //dtsc_GapOpenOcc_CloseY(v, w, r.ey, sc);
      //dtsc_GapOpenOccCross_CloseX(v, w, r.ex, sc);
      dtsc_GapClose_Y(v, w, r.ex, r.ey, model.getLoopTypeIndex(sc));
      break;
    case EmitModel::DR_st:
      //return (xColE.scGapClose(sc) + yColE.scGapCloseCross(sc));

      //dtsc_GapOpen_CloseX(v, w, r.ex, sc);
      //dtsc_GapOpenOcc_CloseX(v, w, r.ex, sc);
      //dtsc_GapOpenOccCross_CloseY(v, w, r.ey, sc);
      dtsc_GapClose_X(v, w, r.ex, r.ey, model.getLoopTypeIndex(sc));
      break;
    case EmitModel::ML_st:
      break;
    case EmitModel::IL_st:
      //return (yColB.scGapOpen(sc) + xColB.scGapOpenCross(sc));

      //dtsc_GapOpen_OpenY(v, w, r.by, sc);
      //dtsc_GapOpenOcc_OpenY(v, w, r.by, sc);
      //dtsc_GapOpenOccCross_OpenX(v, w, r.bx, sc);
      dtsc_GapOpen_Y(v, w, r.bx, r.by, model.getLoopTypeIndex(sc));
      break;
    case EmitModel::DL_st:
      //return (xColB.scGapOpen(sc) + yColB.scGapOpenCross(sc));
      //dtsc_GapOpen_OpenX(v, w, r.bx, sc);
      //dtsc_GapOpenOcc_OpenX(v, w, r.bx, sc);
      //dtsc_GapOpenOccCross_OpenY(v, w, r.by, sc);
      dtsc_GapOpen_X(v, w, r.bx, r.by, model.getLoopTypeIndex(sc));
      break;
    case EmitModel::BF_st:
      break;
    default:
      break;
    }
    break;
    case EmitModel::MP_st:
      switch (sc_es) {
      case EmitModel::ST_st:
	// add tscToLoop ? context of multi stem
	break;
      case EmitModel::MP_st:
	//return (xPair.tscMP() + yPair.tscMP());
	dtsc_Stack(v, w, profileX, rect.bx, rect.ex - 1);
	dtsc_Stack(v, w, profileY, rect.by, rect.ey - 1);
	break;
      case EmitModel::MR_st:
	//return (xPair.tscToLoop(sc) + yPair.tscToLoop(sc));
#ifdef USE_PSCFG_ENERGY 
	//dtsc_TscMPLoop(v, w, sc);
	//dtsc_TermStem_X(v, w, rect.bx, rect.ex - 1);
	//dtsc_TermStem_Y(v, w, rect.by, rect.ey - 1);
	dtsc_ToLoop(v, w, profileX, rect.bx, rect.ex - 1,
		       model.getLoopTypeIndex(sc));
	dtsc_ToLoop(v, w, profileY, rect.by, rect.ey - 1,
		       model.getLoopTypeIndex(sc));
#endif
	break;
      case EmitModel::IR_st:
	//return (yPair.tscToLoop(sc) + xPair.tscToLoop(sc)
	//	+ yColE.scGapClose(sc) + xColE.scGapCloseCross(sc));
#ifdef USE_PSCFG_ENERGY 
	//dtsc_TscMPLoop(v, w, sc);
	//dtsc_TermStem_X(v, w, rect.bx, rect.ex - 1);
	//dtsc_TermStem_Y(v, w, rect.by, rect.ey - 1);
	dtsc_ToLoop(v, w, profileX, rect.bx, rect.ex - 1,
		       model.getLoopTypeIndex(sc));
	dtsc_ToLoop(v, w, profileY, rect.by, rect.ey - 1,
		       model.getLoopTypeIndex(sc));
#endif

	//dtsc_GapOpen_CloseY(v, w, r.ey, sc);
	//dtsc_GapOpenOcc_CloseY(v, w, r.ey, sc);
	//dtsc_GapOpenOccCross_CloseX(v, w, r.ex, sc);
	dtsc_GapClose_Y(v, w, r.ex, r.ey, model.getLoopTypeIndex(sc));
	break;
      case EmitModel::DR_st:
	//return (xPair.tscToLoop(sc) + yPair.tscToLoop(sc)
	//	+ xColE.scGapClose(sc) + yColE.scGapCloseCross(sc));
#ifdef USE_PSCFG_ENERGY 
	//dtsc_TscMPLoop(v, w, sc);
	//dtsc_TermStem_X(v, w, rect.bx, rect.ex - 1);
	//dtsc_TermStem_Y(v, w, rect.by, rect.ey - 1);
	dtsc_ToLoop(v, w, profileX, rect.bx, rect.ex - 1,
		       model.getLoopTypeIndex(sc));
	dtsc_ToLoop(v, w, profileY, rect.by, rect.ey - 1,
		       model.getLoopTypeIndex(sc));
#endif

	//dtsc_GapOpen_CloseX(v, w, r.ex, sc);
	//dtsc_GapOpenOcc_CloseX(v, w, r.ex, sc);
	//dtsc_GapOpenOccCross_CloseY(v, w, r.ey, sc);
	dtsc_GapClose_X(v, w, r.ex, r.ey, model.getLoopTypeIndex(sc));
	break;
      case EmitModel::ML_st:
	//return (xPair.tscToLoop(sc) + yPair.tscToLoop(sc));
#ifdef USE_PSCFG_ENERGY 
	//dtsc_TscMPLoop(v, w, sc);
	//dtsc_TermStem_X(v, w, rect.bx, rect.ex - 1);
	//dtsc_TermStem_Y(v, w, rect.by, rect.ey - 1);
	dtsc_ToLoop(v, w, profileX, rect.bx, rect.ex - 1,
		       model.getLoopTypeIndex(sc));
	dtsc_ToLoop(v, w, profileY, rect.by, rect.ey - 1,
		       model.getLoopTypeIndex(sc));
#endif
	break;
      case EmitModel::IL_st:
	//return (yPair.tscToLoop(sc) + xPair.tscToLoop(sc)
	//	+ yColB.scGapOpen(sc) + xColB.scGapOpenCross(sc));
#ifdef USE_PSCFG_ENERGY 
	//dtsc_TscMPLoop(v, w, sc);
	//dtsc_TermStem_X(v, w, rect.bx, rect.ex - 1);
	//dtsc_TermStem_Y(v, w, rect.by, rect.ey - 1);
	dtsc_ToLoop(v, w, profileX, rect.bx, rect.ex - 1,
		       model.getLoopTypeIndex(sc));
	dtsc_ToLoop(v, w, profileY, rect.by, rect.ey - 1,
		       model.getLoopTypeIndex(sc));
#endif

	//dtsc_GapOpen_OpenY(v, w, r.by, sc);
	//dtsc_GapOpenOcc_OpenY(v, w, r.by, sc);
	//dtsc_GapOpenOccCross_OpenX(v, w, r.bx, sc);
	dtsc_GapOpen_Y(v, w, r.bx, r.by, model.getLoopTypeIndex(sc));
	break;
      case EmitModel::DL_st:
	//return (xPair.tscToLoop(sc) + yPair.tscToLoop(sc)
	//	+ xColB.scGapOpen(sc) + yColB.scGapOpenCross(sc));
#ifdef USE_PSCFG_ENERGY 
	//dtsc_TscMPLoop(v, w, sc);
	//dtsc_TermStem_X(v, w, rect.bx, rect.ex-1);
	//dtsc_TermStem_Y(v, w, rect.by, rect.ey-1);
	dtsc_ToLoop(v, w, profileX, rect.bx, rect.ex - 1,
		       model.getLoopTypeIndex(sc));
	dtsc_ToLoop(v, w, profileY, rect.by, rect.ey - 1,
		       model.getLoopTypeIndex(sc));
#endif

	//dtsc_GapOpen_OpenX(v, w, r.bx, sc);
	//dtsc_GapOpenOcc_OpenX(v, w, r.bx, sc);
	//dtsc_GapOpenOccCross_OpenY(v, w, r.by, sc);
	dtsc_GapOpen_X(v, w, r.bx, r.by, model.getLoopTypeIndex(sc));
	break;
      case EmitModel::BF_st:
	// add tscToLoop ? context of multi loop
	break;
      default:
	break;
      }
      break;
    case EmitModel::MR_st:
      switch (sc_es) {
      case EmitModel::ST_st:
	break;
      case EmitModel::MP_st:
	//return (xPairR.tscFromLoop(sp) + yPairR.tscFromLoop(sp));
#ifdef USE_PSCFG_ENERGY 
	//dtsc_TscMPLoop(v, w, sp);
	//dtsc_TermStem_X(v, w, rect.bx, rect.ex-2);
	//dtsc_TermStem_Y(v, w, rect.by, rect.ey-2);
	dtsc_FromLoop(v, w, profileX, rect.bx, rect.ex - 2,
			 model.getLoopTypeIndex(sp));
	dtsc_FromLoop(v, w, profileY, rect.by, rect.ey - 2,
			 model.getLoopTypeIndex(sp));
#endif
	break;
      case EmitModel::MR_st:
	break;
      case EmitModel::IR_st:
	//return (yColE.scGapClose(sc) + xColE.scGapCloseCross(sc));
	//dtsc_GapOpen_CloseY(v, w, r.ey, sc);
	//dtsc_GapOpenOcc_CloseY(v, w, r.ey, sc);
	//dtsc_GapOpenOccCross_CloseX(v, w, r.ex, sc);
	dtsc_GapClose_Y(v, w, r.ex, r.ey, model.getLoopTypeIndex(sc));
	break;
      case EmitModel::DR_st:
	//return (xColE.scGapClose(sc) + yColE.scGapCloseCross(sc));
	//dtsc_GapOpen_CloseX(v, w, r.ex, sc);
	//dtsc_GapOpenOcc_CloseX(v, w, r.ex, sc);
	//dtsc_GapOpenOccCross_CloseY(v, w, r.ey, sc);
	dtsc_GapClose_X(v, w, r.ex, r.ey, model.getLoopTypeIndex(sc));
	break;
      case EmitModel::ML_st:
	break;
      case EmitModel::IL_st:
	//return (yColB.scGapOpen(sc) + xColB.scGapOpenCross(sc));
	//dtsc_GapOpen_OpenY(v, w, r.by, sc);
	//dtsc_GapOpenOcc_OpenY(v, w, r.by, sc);
	//dtsc_GapOpenOccCross_OpenX(v, w, r.bx, sc);
	dtsc_GapOpen_Y(v, w, r.bx, r.by, model.getLoopTypeIndex(sc));
	break;
      case EmitModel::DL_st:
	//return (xColB.scGapOpen(sc) + yColB.scGapOpenCross(sc));
	//dtsc_GapOpen_OpenX(v, w, r.bx, sc);
	//dtsc_GapOpenOcc_OpenX(v, w, r.bx, sc);
	//dtsc_GapOpenOccCross_OpenY(v, w, r.by, sc);
	dtsc_GapOpen_X(v, w, r.bx, r.by, model.getLoopTypeIndex(sc));
	break;
      default:
	break;
      } 
      break;
    case EmitModel::IR_st:
      switch (sc_es) {
	//const DPScoreT& go = 
	//  (yColE.scGapOpen(sp) + xColE.scGapOpenCross(sp));
      case EmitModel::ST_st:
	break;
      case EmitModel::MP_st:
	//return (go + yPairR.tscFromLoop(sp) + xPair.tscFromLoop(sp);
	//dtsc_GapOpen_OpenY(v, w, r.ey, sp);
	//dtsc_GapOpenOcc_OpenY(v, w, r.ey, sp);
	//dtsc_GapOpenOccCross_OpenX(v, w, r.ex, sp);
	dtsc_GapOpen_Y(v, w, r.ex, r.ey, model.getLoopTypeIndex(sp));
	
#ifdef USE_PSCFG_ENERGY 
	//dtsc_TscMPLoop(v, w, sp);
	//dtsc_TermStem_X(v, w, rect.bx, rect.ex-1);
	//dtsc_TermStem_Y(v, w, rect.by, rect.ey-2);
	dtsc_FromLoop(v, w, profileX, rect.bx, rect.ex - 1,
			 model.getLoopTypeIndex(sp));
	dtsc_FromLoop(v, w, profileY, rect.by, rect.ey - 2,
			 model.getLoopTypeIndex(sp));
#endif
	break;
      case EmitModel::MR_st:
	//dtsc_GapOpen_OpenY(v, w, r.ey, sp);
	//dtsc_GapOpenOcc_OpenY(v, w, r.ey, sp);
	//dtsc_GapOpenOccCross_OpenX(v, w, r.ex, sp);
	dtsc_GapOpen_Y(v, w, r.ex, r.ey, model.getLoopTypeIndex(sp));
	break;
      case EmitModel::IR_st:
	break;
      case EmitModel::DR_st:
	break;
      case EmitModel::ML_st:
	//dtsc_GapOpen_OpenY(v, w, r.ey, sp);
	//dtsc_GapOpenOcc_OpenY(v, w, r.ey, sp);
	//dtsc_GapOpenOccCross_OpenX(v, w, r.ex, sp);
	dtsc_GapOpen_Y(v, w, r.ex, r.ey, model.getLoopTypeIndex(sp));
	break;
      case EmitModel::IL_st:
	//return (go + yColB.scGapOpen(sc) + xColB.scGapOpenCross(sc));
	//dtsc_GapOpen_OpenY(v, w, r.ey, sp);
	//dtsc_GapOpenOcc_OpenY(v, w, r.ey, sp);
	//dtsc_GapOpenOccCross_OpenX(v, w, r.ex, sp);
	dtsc_GapOpen_Y(v, w, r.ex, r.ey, model.getLoopTypeIndex(sp));

	//dtsc_GapOpen_OpenY(v, w, r.by, sc);
	//dtsc_GapOpenOcc_OpenY(v, w, r.by, sc);
	//dtsc_GapOpenOccCross_OpenX(v, w, r.bx, sc);
	dtsc_GapOpen_Y(v, w, r.bx, r.by, model.getLoopTypeIndex(sc));
	break;
      case EmitModel::DL_st:
	//return (go + xColB.scGapOpen(sc) + yColB.scGapOpenCross(sc));
	//dtsc_GapOpen_OpenY(v, w, r.ey, sp);
	//dtsc_GapOpenOcc_OpenY(v, w, r.ey, sp);
	//dtsc_GapOpenOccCross_OpenX(v, w, r.ex, sp);
	dtsc_GapOpen_Y(v, w, r.ex, r.ey, model.getLoopTypeIndex(sp));

	//dtsc_GapOpen_OpenX(v, w, r.bx, sc);
	//dtsc_GapOpenOcc_OpenX(v, w, r.bx, sc);
	//dtsc_GapOpenOccCross_OpenY(v, w, r.by, sc);
	dtsc_GapOpen_X(v, w, r.bx, r.by, model.getLoopTypeIndex(sc));
	break;
      case EmitModel::BF_st:
	//dtsc_GapOpen_OpenY(v, w, r.ey, sp);
	//dtsc_GapOpenOcc_OpenY(v, w, r.ey, sp);
	//dtsc_GapOpenOccCross_OpenX(v, w, r.ex, sp);
	dtsc_GapOpen_Y(v, w, r.ex, r.ey, model.getLoopTypeIndex(sp));
	break;
      default:
	break;
      }
      break;
    case EmitModel::DR_st:
      switch (sc_es) {
	//const DPScoreT& go = 
	//  (xColE.scGapOpen(sp) + yColE.scGapOpenCross(sp));
      case EmitModel::ST_st:
	break;
      case EmitModel::MP_st:
	//return (go + xPairR.tscFromLoop(sp) + yPair.tscFromLoop(sp));
	//dtsc_GapOpen_OpenX(v, w, r.ex, sp);
	//dtsc_GapOpenOcc_OpenX(v, w, r.ex, sp);
	//dtsc_GapOpenOccCross_OpenY(v, w, r.ey, sp);
	dtsc_GapOpen_X(v, w, r.ex, r.ey, model.getLoopTypeIndex(sp));
#ifdef USE_PSCFG_ENERGY 
	//dtsc_TscMPLoop(v, w, sp);
	//dtsc_TermStem_X(v, w, rect.bx, rect.ex-2);
	//dtsc_TermStem_Y(v, w, rect.by, rect.ey-1);
	dtsc_FromLoop(v, w, profileX, rect.bx, rect.ex - 2,
			 model.getLoopTypeIndex(sp));
	dtsc_FromLoop(v, w, profileY, rect.by, rect.ey - 1,
			 model.getLoopTypeIndex(sp));
#endif
	break;
      case EmitModel::MR_st:
	//dtsc_GapOpen_OpenX(v, w, r.ex, sp);
	//dtsc_GapOpenOcc_OpenX(v, w, r.ex, sp);
	//dtsc_GapOpenOccCross_OpenY(v, w, r.ey, sp);
	dtsc_GapOpen_X(v, w, r.ex, r.ey, model.getLoopTypeIndex(sp));
	break;
      case EmitModel::IR_st:
	break;
      case EmitModel::DR_st:
	break;
      case EmitModel::ML_st:
	//dtsc_GapOpen_OpenX(v, w, r.ex, sp);
	//dtsc_GapOpenOcc_OpenX(v, w, r.ex, sp);
	//dtsc_GapOpenOccCross_OpenY(v, w, r.ey, sp);
	dtsc_GapOpen_X(v, w, r.ex, r.ey, model.getLoopTypeIndex(sp));
	break;
      case EmitModel::IL_st:
	//return (go + yColB.scGapOpen(sc) + xColB.scGapOpenCross(sc);
	//dtsc_GapOpen_OpenX(v, w, r.ex, sp);
	//dtsc_GapOpenOcc_OpenX(v, w, r.ex, sp);
	//dtsc_GapOpenOccCross_OpenY(v, w, r.ey, sp);
	dtsc_GapOpen_X(v, w, r.ex, r.ey, model.getLoopTypeIndex(sp));

	//dtsc_GapOpen_OpenY(v, w, r.by, sc);
	//dtsc_GapOpenOcc_OpenY(v, w, r.by, sc);
	//dtsc_GapOpenOccCross_OpenX(v, w, r.bx, sc);
	dtsc_GapOpen_Y(v, w, r.bx, r.by, model.getLoopTypeIndex(sc));
	break;
      case EmitModel::DL_st:
	//return (go + xColB.scGapOpen(sc) + yColB.scGapOpenCross(sc);
	//dtsc_GapOpen_OpenX(v, w, r.ex, sp);
	//dtsc_GapOpenOcc_OpenX(v, w, r.ex, sp);
	//dtsc_GapOpenOccCross_OpenY(v, w, r.ey, sp);
	dtsc_GapOpen_X(v, w, r.ex, r.ey, model.getLoopTypeIndex(sp));

	//dtsc_GapOpen_OpenX(v, w, r.bx, sc);
	//dtsc_GapOpenOcc_OpenX(v, w, r.bx, sc);
	//dtsc_GapOpenOccCross_OpenY(v, w, r.by, sc);
	dtsc_GapOpen_X(v, w, r.bx, r.by, model.getLoopTypeIndex(sc));
	break;
      case EmitModel::BF_st:
	//return go;
	//dtsc_GapOpen_OpenX(v, w, r.ex, sp);
	//dtsc_GapOpenOcc_OpenX(v, w, r.ex, sp);
	//dtsc_GapOpenOccCross_OpenY(v, w, r.ey, sp);
	dtsc_GapOpen_X(v, w, r.ex, r.ey, model.getLoopTypeIndex(sp));
	break;
      default:
	break;
      }
      break;
    case EmitModel::ML_st:
      switch (sc_es) {
      case EmitModel::ST_st:
       break;
      case EmitModel::MP_st:
	//return (xPairL.tscFromLoop(sp)    //vpair.refX(bx+1,ex)
	//	+ yPairL.tscFromLoop(sp);//vpair.refY(by+1,ey)
#ifdef USE_PSCFG_ENERGY 
	//dtsc_TscMPLoop(v, w, sp);
	//dtsc_TermStem_X(v, w, rect.bx+1, rect.ex-1);
	//dtsc_TermStem_Y(v, w, rect.by+1, rect.ey-1);
	dtsc_FromLoop(v, w, profileX, rect.bx + 1, rect.ex - 1,
			 model.getLoopTypeIndex(sp));
	dtsc_FromLoop(v, w, profileY, rect.by + 1, rect.ey - 1,
			 model.getLoopTypeIndex(sp));
#endif
	break;
      case EmitModel::MR_st:
	break;
      case EmitModel::IR_st:
	break;
      case EmitModel::DR_st:
	break;
      case EmitModel::ML_st:
	break;
      case EmitModel::IL_st:
	//return (yColB.scGapOpen(sc) + xColB.scGapOpenCross(sc);
	//dtsc_GapOpen_OpenY(v, w, r.by, sc);
	//dtsc_GapOpenOcc_OpenY(v, w, r.by, sc);
	//dtsc_GapOpenOccCross_OpenX(v, w, r.bx, sc);
	dtsc_GapOpen_Y(v, w, r.bx, r.by, model.getLoopTypeIndex(sc));
	break;
      case EmitModel::DL_st:
	//return (xColB.scGapOpen(sc) + yColB.scGapOpenCross(sc);
	//dtsc_GapOpen_OpenX(v, w, r.bx, sc);
	//dtsc_GapOpenOcc_OpenX(v, w, r.bx, sc);
	//dtsc_GapOpenOccCross_OpenY(v, w, r.by, sc);
	dtsc_GapOpen_X(v, w, r.bx, r.by, model.getLoopTypeIndex(sc));
	break;
      case EmitModel::BF_st:
	break;
      default:
	break;
      }
      break;
    case EmitModel::IL_st:
      switch (sc_es) {
	//const DPScoreT& go = 
	//  (yColB.scGapClose(sp) + xColB.scGapCloseCross(sp);
      case EmitModel::ST_st:
	break;
      case EmitModel::MP_st:
	//return (go + yPairL.tscFromLoop(sp) + xPair.tscFromLoop(sp);
	//dtsc_GapOpen_CloseY(v, w, r.by, sp);
	//dtsc_GapOpenOcc_CloseY(v, w, r.by, sp);
	//dtsc_GapOpenOccCross_CloseX(v, w, r.bx, sp);
	dtsc_GapClose_Y(v, w, r.bx, r.by, model.getLoopTypeIndex(sp));
#ifdef USE_PSCFG_ENERGY 
	//dtsc_TscMPLoop(v, w, sp);
	//dtsc_TermStem_X(v, w, rect.bx, rect.ex-1);
	//dtsc_TermStem_Y(v, w, rect.by+1, rect.ey-1);
	dtsc_FromLoop(v, w, profileX, rect.bx, rect.ex - 1,
			 model.getLoopTypeIndex(sp));
	dtsc_FromLoop(v, w, profileY, rect.by + 1, rect.ey - 1,
			 model.getLoopTypeIndex(sp));
#endif
	break;
      case EmitModel::MR_st:
	break;
      case EmitModel::IR_st:
	break;
      case EmitModel::DR_st:
	break;
      case EmitModel::ML_st:
	//dtsc_GapOpen_CloseY(v, w, r.by, sp);
	//dtsc_GapOpenOcc_CloseY(v, w, r.by, sp);
	//dtsc_GapOpenOccCross_CloseX(v, w, r.bx, sp);
	dtsc_GapClose_Y(v, w, r.bx, r.by, model.getLoopTypeIndex(sp));
	break;
      case EmitModel::IL_st:
	break;
      case EmitModel::DL_st:
	break;
      case EmitModel::BF_st:
	//dtsc_GapOpen_CloseY(v, w, r.by, sp);
	//dtsc_GapOpenOcc_CloseY(v, w, r.by, sp);
	//dtsc_GapOpenOccCross_CloseX(v, w, r.bx, sp);
	dtsc_GapClose_Y(v, w, r.bx, r.by, model.getLoopTypeIndex(sp));
	break;
      default: 
	break;
      }
      break;
    case EmitModel::DL_st:
      switch (sc_es) {
	//const DPScoreT& go = 
	//  (xColB.scGapClose(sp) + yColB.scGapCloseCross(sp);
      case EmitModel::ST_st:
	break;
      case EmitModel::MP_st:
	//return (go + xPairL.tscFromLoop(sp) + yPair.tscFromLoop(sp);
	//dtsc_GapOpen_CloseX(v, w, r.bx, sp);
	//dtsc_GapOpenOcc_CloseX(v, w, r.bx, sp);
	//dtsc_GapOpenOccCross_CloseY(v, w, r.by, sp);
	dtsc_GapClose_X(v, w, r.bx, r.by, model.getLoopTypeIndex(sp));
#ifdef USE_PSCFG_ENERGY 
	//dtsc_TscMPLoop(v, w, sp);
	//dtsc_TermStem_X(v, w, rect.bx+1, rect.ex-1);
	//dtsc_TermStem_Y(v, w, rect.by, rect.ey-1);
	dtsc_FromLoop(v, w, profileX, rect.bx + 1, rect.ex - 1,
			 model.getLoopTypeIndex(sp));
	dtsc_FromLoop(v, w, profileY, rect.by, rect.ey - 1,
			 model.getLoopTypeIndex(sp));
#endif
	break;
      case EmitModel::MR_st:
	break;
      case EmitModel::IR_st:
	break;
      case EmitModel::DR_st:
	break;
      case EmitModel::ML_st:
	//return go;
	//dtsc_GapOpen_CloseX(v, w, r.bx, sp);
	//dtsc_GapOpenOcc_CloseX(v, w, r.bx, sp);
	//dtsc_GapOpenOccCross_CloseY(v, w, r.by, sp);
	dtsc_GapClose_X(v, w, r.bx, r.by, model.getLoopTypeIndex(sp));
	break;
      case EmitModel::IL_st:
	break;
      case EmitModel::DL_st:
	break;
      case EmitModel::BF_st:
	//return go;
	//dtsc_GapOpen_CloseX(v, w, r.bx, sp);
	//dtsc_GapOpenOcc_CloseX(v, w, r.bx, sp);
	//dtsc_GapOpenOccCross_CloseY(v, w, r.by, sp);
	dtsc_GapClose_X(v, w, r.bx, r.by, model.getLoopTypeIndex(sp));
	break;
      default:
	break;
      }
      break;
    case EmitModel::BF_st:
      switch (sc_es) {
      case EmitModel::ST_st:
	break;
      case EmitModel::MP_st:
	// add tscToLoop ? context of multi stem
	break;
      case EmitModel::MR_st:
	break;
      case EmitModel::IR_st:
	break;
      case EmitModel::DR_st:
	break;
      case EmitModel::ML_st:
	break;
      case EmitModel::IL_st:
	break;
      case EmitModel::DL_st:
	break;
      case EmitModel::BF_st:
	break;
      default:
	break;
      }
      break;
  default:
    break;
  }
}


class
PSCFG_Energy::ScoreManager::
Esc
{
public:
  const ScoreManager* pScoreManager;
  Rect rect;
  XCol xColB;
  YCol yColB;
  XCol xColE;
  YCol yColE;
  PairCol xPair;
  PairCol yPair;
  MatchCol mColB;
  MatchCol mColE;
  ESCTable escTable;

  Esc()
    : pScoreManager(NULL), rect(Rect::null()) {
  }
  Esc(const ScoreManager& scoreManager)
    : pScoreManager(NULL), rect(Rect::null()) {
    reset(scoreManager);
  }
  Esc(const Esc& esc)
    : pScoreManager(esc.pScoreManager),
      rect(esc.rect),
      xColB(esc.xColB),
      yColB(esc.yColB),
      xColE(esc.xColE),
      yColE(esc.yColE),
      xPair(esc.xPair),
      yPair(esc.yPair),
      mColB(esc.mColB),
      mColE(esc.mColE),
      escTable(esc.escTable) {
  }
  ~Esc() {
  }
  void reset(const ScoreManager& scoreManager, const Rect& rect0) {
    reset(scoreManager);
    reset(rect0);
  }
  void reset(const ScoreManager& scoreManager) {
    pScoreManager = &scoreManager;
    escTable.reset(*pScoreManager);
  }
  void reset(const Rect& rect0) {
    const ScoreManager& sm = *pScoreManager;
    rect = rect0;
    sm.getXCol(xColB, rect.bx);
    sm.getYCol(yColB, rect.by);
    sm.getXCol(xColE, rect.ex - 1);
    sm.getYCol(yColE, rect.ey - 1);
    sm.getPairColX(xPair, rect.bx, rect.ex - 1);
    sm.getPairColY(yPair, rect.by, rect.ey - 1);
    sm.getMatchCol(mColB, rect.bx, rect.by);
    sm.getMatchCol(mColE, rect.ex - 1, rect.ey - 1);
  }
  DPScoreT getVal(const State& s) const {
    const ScoreManager& sm = *pScoreManager;
    return sm.esc(s, xColB, yColB, xColE, yColE,
		  xPair, yPair, mColB, mColE, escTable);
  }
  DPScoreT getVal(const Rect& rect0, const State& s0) {
    reset(rect0);
    return getVal(s0);
  }
};

class
PSCFG_Energy::ScoreManager::
Tsc
{
public:
  const ScoreManager* pScoreManager;
  Rect rect;
  State sp;
  XCol xColB;
  YCol yColB;
  XCol xColE;
  YCol yColE;
  PairCol xPair;
  PairCol yPair;
  PairCol xPairR;
  PairCol yPairR;
  PairCol xPairL;
  PairCol yPairL;
  TSCTable tscTable;
  Rect rect1;

  Tsc() 
    : pScoreManager(NULL), rect(Rect::null()) {
  }
  Tsc(const ScoreManager& scoreManager)
    : pScoreManager(NULL), rect(Rect::null()) {
    reset(scoreManager);
  }
  Tsc(const Tsc& tsc)
    : pScoreManager(tsc.pScoreManager),
      rect(tsc.rect),
      xColB(tsc.xColB),
      yColB(tsc.yColB),
      xColE(tsc.xColE),
      yColE(tsc.yColE),
      xPair(tsc.xPair),
      yPair(tsc.yPair),
      xPairR(tsc.xPairR),
      yPairR(tsc.yPairR),
      xPairL(tsc.xPairL),
      yPairL(tsc.yPairL),
      tscTable(tsc.tscTable),
      rect1(tsc.rect1) {
  }
  ~Tsc() {
  }
  void reset(const ScoreManager& scoreManager, const Rect& rect0) {
    reset(scoreManager);
    reset(rect0);
  }
  void reset(const ScoreManager& scoreManager) {
    pScoreManager = &scoreManager;
    tscTable.reset(*pScoreManager);
  }
  void reset(const Rect& rect0) {
    const ScoreManager& sm = *pScoreManager;
    rect = rect0;
    sp = Model::X_st;
    sm.getPairColX(xPair, rect.bx, rect.ex - 1);
    sm.getPairColY(yPair, rect.by, rect.ey - 1);
    sm.getPairColX(xPairR, rect.bx, rect.ex - 2);
    sm.getPairColY(yPairR, rect.by, rect.ey - 2);
    sm.getPairColX(xPairL, rect.bx + 1, rect.ex - 1);
    sm.getPairColY(yPairL, rect.by + 1, rect.ey - 1);
  }
  DPScoreT getVal(const State& sp0, const State& sc0) {
    const ScoreManager& sm = *pScoreManager;
    if (sp != sp0) {
      sp = sp0;
      rect1 = Model::getModel().addNEmit(rect, sp);
      sm.getXCol(xColB, rect1.bx);
      sm.getYCol(yColB, rect1.by);
      sm.getXCol(xColE, rect1.ex);
      sm.getYCol(yColE, rect1.ey);
    }
    return sm.tsc(sp, sc0, xColB, yColB, xColE, yColE,
		  xPair, yPair, xPairR, yPairR, xPairL, yPairL, tscTable);
  }
  DPScoreT getVal(const Rect& rect0, const State& sp0, const State& sc0) {
    reset(rect0);
    return getVal(sp0, sc0);
  }
};

// may be defined in a separate file 
// as rather complex functions are needed for addTCCount()
class 
PSCFG_Energy::ScoreManager::
FeatureCounter 
{
public:
  FeatureCounter(const FeatureCounter& fc);
  ~FeatureCounter();
  void addECount(const DPScoreT& e, const Rect& rect,
		 const State& sp);
  void addTCount(const DPScoreT& t, const Rect& rect,
		 const State& so, const State& sc);
  Vec getFeatureVector() const;
private:
  friend class ScoreManager;
  const ScoreManager& scoreManager;
  FVec featureVector;

  FeatureCounter(const ScoreManager& scoreManager);
  FeatureCounter();
  FeatureCounter& operator=(const FeatureCounter& fc);
};

class 
PSCFG_Energy::ScoreManager::
FeatureStats 
{
public:
  FeatureStats(const FeatureStats& fc);
  ~FeatureStats();
  void addECount(const DPScoreT& e, const Rect& rect,
		 const State& sp);
  void addTCount(const DPScoreT& t, const Rect& rect,
		 const State& so, const State& sc);
  Vec getFeatureVector() const;
  Vec getCountVector() const;
private:
  friend class ScoreManager;
  const ScoreManager& scoreManager;
  FVec featureVector;
  FVec countVector;

  FeatureStats(const ScoreManager& scoreManager);
  FeatureStats();
  FeatureStats& operator=(const FeatureStats& fc);
};

#endif
