/*
 * SkipSet.cpp
 *
 */
#include "SkipSet.hpp"
#include "DPPath.hpp"

// how to check these codes ? 
// write at least toString() method
const Point
SkipSet::
pt0 = {0, 0};

SkipSet::
SkipSet()
  : skipSize(1)
{
}
SkipSet::
SkipSet(const SkipSet& skipSet)
  : skipSize(skipSet.skipSize)
{
  Assert(skipSize > 0);
}
SkipSet::
~SkipSet()
{
}
SkipSet&
SkipSet::
operator=(const SkipSet& skipSet)
{
  skipSize = skipSet.skipSize;
  Assert(skipSize > 0);
  return *this;
}

SkipSet::DivT
SkipSet::
divmodX(const int& x) const
{
  return std::div((x - pt0.x), skipSize);
}
void
SkipSet::
setSkipSize(const int& skipSize0)
{
  skipSize = skipSize0;
  Assert(skipSize > 0);
}
void
SkipSet::
setYOffsets(const DPPath& dpPath)
{
  const Rect& rect0 = dpPath.getRect0();
  Check(rect0.isNormal());
  const int& nx = getNSkipRow(rect0.bx, rect0.ex);
  Assert(rect0.bx == pt0.x);// assuming dpPath starts from pt0.x
  y0.resize(nx);
  y0.fill(0);

  DPPath::Iterator it(dpPath);
  while (it.getNext()) {
    if (it.emitsX() && isSkipRow(it.x)) {
      const int& mx = ((it.x - pt0.x) / skipSize);
      const int& ry = (it.y % skipSize);
      const int& k = (ry == 0 ? 0 : (skipSize - ry));
      y0[mx] = (-k);
      Assert((it.y - y0[mx]) % skipSize == 0);
    }
  }
  if (isSkipRow(rect0.ex)) {
    const int& mx = ((rect0.ex - pt0.x) / skipSize);
    const int& ry = (rect0.ey & skipSize);
    const int& k = (ry == 0 ? 0 : (skipSize - ry));
    y0[mx] = (-k);
  }
#ifdef DEBUG
  it.reset();
  Array<bool> flag(rect0.ex + 1);
  flag.fill(false);
  int k = 0;
  while (it.getNext()) {
    if (it.emitsX() && isSkipPoint(it.pt)) {
      if (!flag[it.x]) {
	  k++;
	  flag[it.x] = true;
      }
    }
    Assert(k <= nx);
  }
  if (isSkipPoint(rect0.e)) {
    k++;
  }
  Assert(k == nx);
#endif
}
int
SkipSet::
getSkipSize() const
{
  Assert(skipSize > 0);
  return skipSize;
}
int
SkipSet::
getSkipRowPeriod() const
{
  return skipSize;
}
int
SkipSet::
getHorizPeriod() const
{
  return skipSize;
}
//int
//SkipSet::
//getVertPeriod() const
//{
//  return (skipSize * skipSize);
//}
int
SkipSet::
getDiagPeriod() const
{
  return (skipSize * skipSize);
}
bool
SkipSet::
isSkipRow(const int& x) const
{
  const int& nx = (x - pt0.x);
  Assert(nx >= 0);
  const int& rx = nx % skipSize;
  return (rx == 0);
}
bool
SkipSet::
isSkipPoint(const Point& pt) const
{
  const DivT& dvx = divmodX(pt.x);
  if (dvx.rem) return false;

  const int& ny = (pt.y - y0[dvx.quot]);
  const int& k = (ny % skipSize);
  return (k == 0);
}

bool
SkipSet::
isSkipPoint0(int x, int y) const
{
  int nx = (x - pt0.x);
  int rx = nx % skipSize;
  int mx = nx / skipSize;
  int ny = (y - y0[mx]);
  Assert(nx >= 0);
  Assert(ny >= 0);
  int ry = ny % skipSize;
  if (rx > 0) return false;
  if (ry > 0) return false;

  return true;
}
bool
SkipSet::
isSkipRect(const Rect& rect) const
{
  return (isSkipPoint(rect.b) && isSkipPoint(rect.e));
}
int
SkipSet::
findFirstSkipRow(const int& x) const
{
  const int& nx = (x - pt0.x);
  Assert(nx >= 0);
  const int& rx = nx % skipSize;
  if (rx == 0) return x;

  const int& dx = (skipSize - rx);
  const int& retVal = (x + dx);
  Assert(retVal == findFirstSkipRow0(x));
  return retVal;
}
int
SkipSet::
findFirstSkipRow0(int x) const
{
  int x1 = x;
  while (!isSkipRow(x1)) {
    x1++;
  }
  return x1;
}
int
SkipSet::
findLastSkipRow(const int& x) const
{
  const int& nx = (x - pt0.x);
  Assert(nx >= 0);
  const int& rx = nx % skipSize;
  if (rx == 0) return x;

  const int& dx = (-rx);
  const int& x1 = (x + dx);
  if (x1 < 0) return (-1);

  return x1;
}
int
SkipSet::
findFirstSkipRowXRange(const int& x_beg, const int& x_last) const
{
  const int& x = findFirstSkipRow(x_beg);
  if (x_last < x) return (-1);

  return x;
}
int
SkipSet::
findLastSkipRowXRange(const int& x_beg, const int& x_last) const
{
  const int& x = findLastSkipRow(x_last);
  if (x < x_beg) return (-1);

  return x;
}
int
SkipSet::
getNSkipRow(const int& x_beg, const int& x_last) const
{
  const int& x0 = findFirstSkipRow(x_beg);
  if (x_last < x0) return 0;

  const int& lx = (x_last - x0 + 1);
  const int& period = getSkipRowPeriod();
  const int& n = (lx + period - 1) / period;
  return n;
}
int
SkipSet::
findFirstSkipY(const Point& pt) const
{
  const DivT& dvx = divmodX(pt.x);
  if (dvx.rem) return (-1);

  const int& ny = (pt.y - y0[dvx.quot]);
  const int& k = (ny % skipSize);
  if (k == 0) return pt.y;

  const int& dy = (skipSize - k);// 0 < dy < skipSize
  return (pt.y + dy);
}
int
SkipSet::
findLastSkipY(const Point& pt) const
{
  const DivT& dvx = divmodX(pt.x);
  if (dvx.rem) return (-1);

  const int& ny = (pt.y - y0[dvx.quot]);
  const int& k = (ny % skipSize);
  const int& dy = (-k);
  const int& y1 = (pt.y + dy);
  if (y1 < 0) return (-1);

  return y1;
}
int
SkipSet::
findFirstSkipYRange(const int& x,
		    const int& y_beg, const int& y_last) const
{
  const Point pt = {x, y_beg};
  const int& y = findFirstSkipY(pt);
  if (y_last < y) return (-1);

  return y;
}
int
SkipSet::
findLastSkipYRange(const int& x,
		   const int& y_beg, const int& y_last) const
{
  const Point pt = {x, y_last};
  const int& y = findLastSkipY(pt);
  if (y < y_beg) return (-1);

  return y;
}
int
SkipSet::
findFirstSkipX(const Point& pt) const
{
  const DivT& dvx = divmodX(pt.x);
  int x1 = pt.x;
  int mx1 = dvx.quot;
  if (dvx.rem > 0) {
    x1 += (skipSize - dvx.rem);
    mx1++;
  }
  const int& ny = (pt.y - y0[mx1]);
  const int& k = (ny % skipSize);
  if (k == 0) return x1;

  const int& dmx1 = (skipSize - k);
  return (x1 + (dmx1 * skipSize));
}
int
SkipSet::
findLastSkipX(const Point& pt) const
{
  const DivT& dvx = divmodX(pt.x);
  const int& x1 = pt.x + (-dvx.rem);
  if (x1 < 0) return (-1);

  const int& ny = (pt.y - y0[dvx.quot]);
  const int& k = (ny % skipSize);
  const int& dmx1 = (-k);
  const int& x2 = (x1 + (dmx1 * skipSize));
  if (x2 < 0) return (-1);

  return x2;
}
Point
SkipSet::
findFirstSkipDiag(const Point& pt) const
{
  const DivT& dvx = divmodX(pt.x);
  Point pt1 = pt;
  int mx1 = dvx.quot;
  if (dvx.rem > 0) {
    const int& d = (skipSize - dvx.rem);
    pt1.x += d;
    pt1.y += d;
    mx1++;
  }
  const int& nx1 = (pt1.y - y0[mx1]);
  const int& k = (nx1 % skipSize);
  if (k == 0) return pt1;

  const int& dmx1 = (skipSize - k);
  const int& d = (dmx1 * skipSize);
  pt1.x += d;
  pt1.y += d;
  return pt1;
}
Point
SkipSet::
findLastSkipDiag(const Point& pt) const
{
  const DivT& dvx = divmodX(pt.x);
  const int& d = (-dvx.rem);
  Point pt1 = {pt.x + d, pt.y + d};
  if (pt1.x < 0 || pt1.y < 0) return Point::null();

  const int& ny = (pt1.y - y0[dvx.quot]);
  const int& k = (ny % skipSize);
  const int& dmx1 = (-k);
  const int& d1 = (dmx1 * skipSize);
  pt1.x += d1;
  pt1.y += d1;
  if (pt1.x < 0 || pt1.y < 0) return Point::null();

  return pt1;
}
int
SkipSet::
getNSkipPointY(const int& x, 
	       const int& y_beg, const int& y_last) const
{
  const Point pt = {x, y_beg};
  const int& byk = findFirstSkipY(pt);
  if (byk < 0 || y_last < byk) return 0;

  const int& ly = (y_last - byk + 1);
  const int& period = getHorizPeriod();
  const int& n = (ly + period - 1) / period;
  return n;
}
void
SkipSet::
getSkipYRange(const int& x, 
	      int& y_beg, int& y_last, int& ly) const
{// assuming x is a skipRow
  Assert(isSkipRow(x));
  const int& mx = (x - pt0.x) / skipSize;
  const int& nyb = (y_beg - y0[mx]);
  const int& kb = (nyb % skipSize);
  if (kb > 0) {
    y_beg += (skipSize - kb);
  }
  if (y_last < y_beg) {
    y_beg = -1;
    y_last = -2;
    ly = 0;
    return;
  }
  ly = ((y_last - y_beg + 1) + (skipSize - 1)) / skipSize;
  y_last = y_beg + (ly - 1) * skipSize;
}
int
SkipSet::
getNSkipPointX(const int& x_beg, const int& x_last,
	       const int& y) const
{
  const int& bxk = findFirstSkipRow(x_beg);
  if (bxk < 0) return 0;

  const int& period = getSkipRowPeriod();
  Point pt = {bxk, y};
  int n = 0;
  while (pt.x <= x_last) {
    if (isSkipPoint(pt)) {
      n++;
    }
    pt.x += period;
  }
  return n;
}
string
SkipSet::
toString() const
{
  ostringstream oss;
  oss << "--- !skip_set\n";
  oss << "pt0: " << pt0.toString() << "\n";
  oss << "skip_size: " << skipSize << "\n";
  oss << "y0: \n";
  for (int i = 0; i < (int)y0.size(); i++) {
    oss << "- " << y0[i] << "\n";
  }
  return oss.str();
}
void
SkipSet::
print() const
{
  cout << toString() << flush;
}
