/*
 * StripRegionSkip.hpp
 *
 */
#ifndef STRIP_REGION_SKIP_HPP
#define STRIP_REGION_SKIP_HPP
#include "StripRegion.hpp"

class SkipSet;
class StripRegionSkip 
{
public:
  StripRegionSkip(const StripRegion& sr, const SkipSet& skipSet);
  StripRegionSkip(const StripRegionSkip& srk);
  ~StripRegionSkip();
  StripRegionSkip& operator=(const StripRegionSkip& srk);

  string toString() const;
  void print() const;
  bool empty() const;
  int getNSkipRow() const;
  int getSkipSize() const;
  int getArea() const;
  int getAreaRight(const int& y) const;
  int getAreaBefore(const Point& pt) const;
  int getAreaTopLeft(const Point& pt) const;// accept all point
  int getAreaBottomRight(const Point& pt) const;
  Point findFirstSkipPoint(const Point& pt) const;
  Point findLastSkipPoint(const Point& pt) const;
  Rect getRect0() const {return sr.rect0;}
  Rect getRectK() const;
  Rect getRectK(const Rect& rect) const;
  Point advance(const Point& pt, const int& i) const;
  // accept all rect
  StripRegionSkip getBoundedRegion(const Rect& rect) const; 

  int getFirstSkipRow() const;
  int getLastSkipRow() const;
  int getFirstSkipRow(const int& y) const;
  int getLastSkipRow(const int& y) const;
  int getNSkipRow(const int& y) const;
    
  template <bool fwd> class RowIterator;
  template <bool fwd> class Iterator;
  template <bool fwd> class Iterator0;

private:
  StripRegion sr;
  const SkipSet& skipSet;

  StripRegionSkip();
};

template <bool fwd>
class 
StripRegionSkip::
RowIterator 
{
public:
  int x;
  int ly;
  int y_beg;
  int y_last;
  int dy;

  RowIterator(const StripRegionSkip& sr);
  ~RowIterator();
  void reset(const Rect& rect);
  void reset();
  bool getNext();

#ifndef DEBUG
private:
#endif

  const StripRegion& sr;
  const SkipSet& skipSet;
  const int skipSize;
  Rect bbox;
  int x_lim;
  int next_x;
  int idx_lim;
  int idx;

  RowIterator();
  RowIterator& operator=(const RowIterator& it);
};

template <bool fwd>
class 
StripRegionSkip::
Iterator 
{
public:
  union {
    Point pt;
    struct {int x, y;};
  };
  RowIterator<fwd> rit;

  Iterator(const StripRegionSkip& sr);
  Iterator(const StripRegionSkip& sr, const Rect& rect0);
  ~Iterator();
  bool getNext();
  void reset(const Rect& rect0);
  void reset();
private:
  int y_lim;
  int next_y;
  void reset0();
};

template <bool fwd>
class 
StripRegionSkip::
Iterator0 
{
public:
  union {
    Point pt;
    struct {int x, y;};
  };
  StripRegion::Iterator<fwd> it;
  const SkipSet& skipSet;

  Iterator0(const StripRegionSkip& sr);
  Iterator0(const StripRegionSkip& sr, const Rect& rect0);
  ~Iterator0();
  bool getNext();
  void reset(const Rect& rect0);
  void reset();
};

#include "StripRegionSkipIterators.cpp"
#endif
