/*
 * StripRegionSkipIterators.cpp
 *
 */
#ifndef STRIP_REGION_SKIP_ITERATORS_CPP
#define STRIP_REGION_SKIP_ITERATORS_CPP

#include "StripRegionSkip.hpp"
#include "SkipSet.hpp"

template <bool fwd>
StripRegionSkip::RowIterator<fwd>::
RowIterator(const StripRegionSkip& srk)
  : x(-1), ly(0), y_beg(-1), y_last(-2), dy(srk.skipSet.getSkipSize()),
    sr(srk.sr), skipSet(srk.skipSet), skipSize(srk.skipSet.getSkipSize()),
    bbox(srk.getRect0()), x_lim(-1), next_x(-1), idx_lim(-1), idx(-1)
{
  reset();
}
template <bool fwd>
StripRegionSkip::RowIterator<fwd>::
~RowIterator()
{
}
template <bool fwd>
void
StripRegionSkip::RowIterator<fwd>::
reset(const Rect& rect)
{
  bbox = rect;
  reset();
}
template <bool fwd>
void
StripRegionSkip::RowIterator<fwd>::
reset()
{
  x = -1;
  ly = 0;
  y_beg = -1;
  y_last = -2;

  if (!bbox.isNull()) {
    const int& x_beg = skipSet.findFirstSkipRow(max(sr.bx0, bbox.bx));
    const int& x_last = skipSet.findLastSkipRow(min(sr.ex0, bbox.ex));
    if (0 <= x_beg && x_beg <= x_last) {
      x_lim = (fwd ? (x_last + skipSize) : (x_beg - skipSize));
      next_x = (fwd ? x_beg : x_last);
      idx_lim = sr.getXTIndex(fwd ? x_last : x_beg);
      idx = sr.getXTIndex(next_x);
      return;
    }
  }
  x_lim = -1;
  next_x = -1;
  idx_lim = -1;
  idx = -1;
}
template <bool fwd>
bool
StripRegionSkip::RowIterator<fwd>::
getNext()
{
  if (next_x == x_lim) return false;

  x = next_x;
  //const int& y_beg0 = max(sr[idx].getBY(x), bbox.by);
  //const int& y_last0 = min(sr[idx].getEY(x), bbox.ey);
  //ly = skipSet.getNSkipPointY(x, y_beg0, y_last0);
  //if (ly == 0) {
  //  y_beg = -1;
  //  y_last = -2;
  //} else {
  //  const Point pt = {x, y_beg0};
  //  y_beg = skipSet.findFirstSkipY(pt);
  //  y_last = y_beg + (ly - 1) * skipSize;
  //}
  y_beg = max(sr[idx].getBY(x), bbox.by);
  y_last = min(sr[idx].getEY(x), bbox.ey);
  skipSet.getSkipYRange(x, y_beg, y_last, ly);
  Assert(ly == skipSet.getNSkipPointY(x, 
				      max(sr[idx].getBY(x), bbox.by),
				      min(sr[idx].getEY(x), bbox.ey)));
  next_x += (fwd ? skipSize : (-skipSize));
  while (fwd ? (sr[idx].ex < next_x) : (next_x < sr[idx].bx)) {
    if (idx == idx_lim) break;

    (fwd ? idx++ : idx--);
  }
  return true;
}

template <bool fwd>
StripRegionSkip::Iterator<fwd>::
Iterator(const StripRegionSkip& sr)
  : pt(Point::null()), rit(sr), y_lim(-1), next_y(-1)
{
  reset0();
}
template <bool fwd>
StripRegionSkip::Iterator<fwd>::
~Iterator()
{
}
template <bool fwd>
void
StripRegionSkip::Iterator<fwd>::
reset(const Rect& rect)
{
  rit.reset(rect);
  reset0();
}
template <bool fwd>
void
StripRegionSkip::Iterator<fwd>::
reset()
{
  rit.reset();
  reset0();
}
template <bool fwd>
void
StripRegionSkip::Iterator<fwd>::
reset0()
{
  pt.setNull();
  rit.reset();
  while (rit.getNext()) {
    if (rit.ly == 0) continue;

    y_lim = (fwd ? (rit.y_last + rit.dy) : (rit.y_beg - rit.dy));
    next_y = (fwd ? rit.y_beg : rit.y_last);
    x = rit.x;
    return;
  }
  y_lim = -1;
  next_y = -1;
  x = -1;
}
template <bool fwd>
bool
StripRegionSkip::Iterator<fwd>::
getNext()
{
  while (next_y == y_lim) {
    if (!rit.getNext()) return false;
    if (rit.ly == 0) continue;

    y_lim = (fwd ? (rit.y_last + rit.dy) : (rit.y_beg - rit.dy));
    next_y = (fwd ? rit.y_beg : rit.y_last);
    x = rit.x;
  }
  y = next_y;
  next_y += (fwd ? rit.dy : (-rit.dy));

#ifdef DEBUG
  Assert(rit.skipSet.isSkipPoint0(x, y));
#endif

  return true;
}

template <bool fwd>
StripRegionSkip::Iterator0<fwd>::
Iterator0(const StripRegionSkip& sr)
      : pt(Point::null()), it(sr.sr), skipSet(sr.skipSet)
{
}
template <bool fwd>
StripRegionSkip::Iterator0<fwd>::
Iterator0(const StripRegionSkip& sr, const Rect& rect0)
  : pt(Point::null()), it(sr.sr), skipSet(sr.skipSet)
{
  reset(rect0);
}
template <bool fwd>
StripRegionSkip::Iterator0<fwd>::
~Iterator0()
{
}
template <bool fwd>
void
StripRegionSkip::Iterator0<fwd>::
reset(const Rect& rect0)
{
  it.reset(rect0);
}
template <bool fwd>
void
StripRegionSkip::Iterator0<fwd>::
reset()
{
  it.reset();
}
template <bool fwd>
bool
StripRegionSkip::Iterator0<fwd>::
getNext() 
{
  while (it.getNext()) {
    if (skipSet.isSkipPoint0(it.pt.x, it.pt.y)) {
      pt = it.pt;
      return true;
    }
  }
  return false;
}

#endif
